#include <AVR_LCD.h>
#include <gc.h>
#include <javatypes.h>
#include <avr/io.h>
#include <compat/ina90.h>
#include "ClockDevice.h"
#include <AVR_Kernel.h>
#include "done_ClockInput.h"

// Directional buttons
#define BTN_NONE	0
#define BTN_LEFT	1
#define BTN_RIGHT	2
#define BTN_UP		3
#define BTN_DOWN	4

// Clock input mode
#define MODE_NONE	0
#define MODE_TSHOW	1
#define MODE_ASET	2
#define MODE_TSET	3
#define MODE_ATOGGLE	4

// These need to correspond to the constants given in ClockInput.java
#define IN_TSHOW	0
#define IN_ASET		1
#define IN_TSET		2

static uint8_t btnstate = 0xFF;
static int32_t alarmtime = 0;
static uint8_t clockmode = MODE_NONE;
static volatile int32_t clocktime = 0;// ClockOutput
static volatile int32_t inputval = 0;// ClockInput
static volatile uint8_t inmode = IN_TSHOW;// ClockInput
static volatile uint8_t alarmflag = 0;// ClockInput
static volatile uint8_t alarmpulse = 0;
static volatile uint8_t cursor = 0;// 54:32:10
static Mutex *clockmutex;
static char lcdcache[] = {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '};
static char lcdcache2[] = {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '}; 


uint8_t GetClockMode();
uint8_t GetDirection();

void ClockDeviceThread()
{
  void *eFlag = 0;
  void **__eFlag__ = &eFlag;

  uint8_t prevstate;
  uint8_t prevmode;

  clockmutex = allocMutex(0);

  PrintClockLCD();

  while (1) {
    // Read push button states
    prevstate = btnstate;
    btnstate = input(PINF);

    // Buttons pressed/released?
    if (btnstate != prevstate) {
      prevmode = clockmode;
      clockmode = GetClockMode();

      // Any button changed?
      if (clockmode == MODE_TSET) {
        int status = disableInts();
	TimeSet(prevmode);
        enableInts(status);
      } else if (clockmode == MODE_ASET) {
        int status = disableInts();
	AlarmSet(prevmode);
        enableInts(status);
      } else {
	cursor = 0;
      }

      // Mode changed?
      if (clockmode != prevmode) {
	if (clockmode == MODE_TSHOW) {
	  takeMutex(0, clockmutex);
	  inmode = IN_TSHOW;
	  giveMutex(0, clockmutex);
	} else if (clockmode == MODE_ATOGGLE) {
	  takeMutex(0, clockmutex);
	  if (alarmflag) {
	    alarmflag = 0;
	  } else {
	    alarmflag = 1;
	  }
	  giveMutex(0, clockmutex);
	}
      }

      PrintClockLCD();
      // Signal ClockInput
      GC_PROC_CALL(done_ClockInput_giveInput);
    }

    sleep((long)120);
  }
}

uint8_t GetClockMode()
{
	switch (btnstate & 0x11) {
		case 0x11:
			return MODE_TSHOW;
		case 0x01:
			return MODE_TSET;
		case 0x10:
			return MODE_ASET;
		case 0x00:
			return MODE_ATOGGLE;
		default:
			return MODE_NONE;
	}
}

uint8_t GetDirection()
{
	switch (btnstate & 0x2E) {
		case 0x2C:
			return BTN_LEFT;
		case 0x2A:
			return BTN_DOWN;
		case 0x26:
			return BTN_RIGHT;
		case 0x0E:
			return BTN_UP;
		default:
			return BTN_NONE;
	}
}

// ClockOutput
void SetClockTime(int32_t hhmmss)
{
	takeMutex(0, clockmutex);
	if (clockmode != MODE_TSET) {
		clocktime = hhmmss;
	}
	giveMutex(0, clockmutex);
}

void SetAlarmPulse(uint8_t onoff)
{
	takeMutex(0, clockmutex);
	alarmpulse = onoff;
	giveMutex(0, clockmutex);
}

// ClockInput
uint8_t GetInputMode()
{
	uint8_t result;
	takeMutex(0, clockmutex);
	result = inmode;
	giveMutex(0, clockmutex);
	return result;
}

// ClockInput
int32_t GetInputValue()
{
	int32_t result;
	takeMutex(0, clockmutex);
	result = inputval;
	giveMutex(0, clockmutex);
	return result;
}

// ClockInput
uint8_t GetAlarmFlag()
{
	uint8_t result;
	takeMutex(0, clockmutex);
	result = alarmflag;
	giveMutex(0, clockmutex);
	return result;
}

// Print clock or input time on the LCD
void PrintClockLCD()
{
	// "54:32:10"
	//  01234567
	char lcdout[8];
	int32_t hhmmss;
	int i;
	int status;

	takeMutex(0, clockmutex);
	status = disableInts();

	hhmmss = clocktime;
	lcdout[7] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[6] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[5] = ':';
	lcdout[4] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[3] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[2] = ':';
	lcdout[1] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[0] = (char)(hhmmss % 10 + '0');

	for (i = 0; i < 8; ++i) {
		if (lcdout[i] != lcdcache[i]) {
			lcdcache[i] = lcdout[i];
			LCD_moveCursor(i);
			LCD_printChar(lcdout[i]);
		}
	}

	hhmmss = alarmtime;
	lcdout[7] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[6] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[4] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[3] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[1] = (char)(hhmmss % 10 + '0');
	hhmmss /= 10;
	lcdout[0] = (char)(hhmmss % 10 + '0');

	if (alarmpulse) {
		lcdout[5] = ' ';
		lcdout[2] = ' ';
	} else {
		if (alarmflag) {
			lcdout[5] = ':';
			lcdout[2] = ':';
		} else {
			lcdout[5] = '_';
			lcdout[2] = '_';
		}
	}

	for (i = 0; i < 8; ++i) {
		if (lcdout[i] != lcdcache2[i]) {
			lcdcache2[i] = lcdout[i];
			LCD_moveCursor(40+i);
			LCD_printChar(lcdout[i]);
		}
	}

	if (clockmode == MODE_TSET || clockmode == MODE_ASET) {
		uint8_t cur = 5-cursor;
		if (cursor < 4) {
			cur ++;
			if (cursor < 2) {
				cur ++;
			}
		}
		if (clockmode == MODE_ASET) {
			cur += 40;
		}
		LCD_moveCursor(cur);
		LCD_enableCursor();
	} else {
		LCD_disableCursor();
	}

	enableInts(status);
	giveMutex(0, clockmutex);
}

uint8_t GetDigit(int32_t number, uint8_t pos)
{
	uint8_t i;
	for (i = 0; i < pos; ++i) number /= 10;
	return (uint8_t) (number % 10);
}

int32_t IncDigit(int32_t number, uint8_t pos)
{
	int32_t term = 1;
	uint8_t i;
	for (i = 0; i < pos; ++i) term *= 10;
	return number + term;
}

int32_t DecDigit(int32_t number, uint8_t pos)
{
	int32_t term = 1;
	uint8_t i;
	for (i = 0; i < pos; ++i) term *= 10;
	return number - term;
}

int32_t IncTime(int32_t time, uint8_t pos)
{
	int32_t newtime = time;
	if ((pos == 5 && GetDigit(time, 5) < 2 && GetDigit(time, 4) < 4) ||
			(pos == 5 && GetDigit(time, 5) < 1) || 
			(pos == 4 && GetDigit(time, 4) < 3 && GetDigit(time, 5) <= 2) ||
			(pos == 4 && GetDigit(time, 4) < 9 && GetDigit(time, 5) <= 1) ||
			(pos == 3 && GetDigit(time, 3) < 5) ||
			(pos == 1 && GetDigit(time, 1) < 5) ||
			((pos == 0 || pos == 2) && GetDigit(time, pos) < 9)) {
		newtime = IncDigit(time, pos);
	}
	return newtime;
}

void TimeSet(uint8_t prevmode)
{
	takeMutex(0, clockmutex);
	uint8_t dir = GetDirection();

	if (prevmode != MODE_TSET) {
		inputval = clocktime;
	}

	if (dir == BTN_LEFT && cursor < 5) {
		cursor++;
	} else if (dir == BTN_RIGHT && cursor > 0) {
		cursor--;
	} else if (dir == BTN_UP) {
		inputval = IncTime(inputval, cursor);
	} else if (dir == BTN_DOWN) {
		if (GetDigit(inputval, cursor) > 0) {
			inputval = DecDigit(inputval, cursor);
		}
	}

	clocktime = inputval;
	inmode = IN_TSET;
	giveMutex(0, clockmutex);
}

void AlarmSet(uint8_t prevmode)
{
	takeMutex(0, clockmutex);
	uint8_t dir = GetDirection();

	if (prevmode != MODE_ASET) {
		inputval = alarmtime;
	}

	if (dir == BTN_LEFT && cursor < 5) {
		cursor++;
	} else if (dir == BTN_RIGHT && cursor > 0) {
		cursor--;
	} else if (dir == BTN_UP) {
		inputval = IncTime(inputval, cursor);
	} else if (dir == BTN_DOWN) {
		if (GetDigit(inputval, cursor) > 0) {
			inputval = DecDigit(inputval, cursor);
		}
	}

	alarmtime = inputval;
	inmode = IN_ASET;
	giveMutex(0, clockmutex);
}

