#ifndef EXCEPTION_H
#define EXCEPTION_H

/* exception.h - a simple implementation of java-like exceptions 
 * (c) 2002 Sven Gestegrd Robertz <sven@cs.lth.se>
 *     2009 Anders Nilsson <andersn@cs.lth.se>
 *
 * usage: the program
 *
 *	void foo() throws SomeException, AnotherException {
 *		...
 *		throw(new SomeException());
 *		...
 *		throw(new AnotherException());
 *	}
 *	...
 * 	try {
 *		foo();
 *	} catch (SomeException e) {
 *		bar(e);
 *	} catch (AnotherException e) {
 *		snafu(e);
 *	} finally {
 *		something();
 *	}
 *
 * should be translated to
 *
 *	void foo() {
 *	EXCEPTION_THROWS;	// NB! this is a declaration 
 *		...
 *		SomeException e = new SomeException();
 *		EXCEPTION_STORE_OBJ(e);
 *		EXCEPTION_THROW(&SomeException_Class);
 *		...
 *		SomeException e = new AnotherException();
 *		EXCEPTION_STORE_OBJ(e);
 *		EXCEPTION_THROW(&AnotherException_Class);
 *	}
 *	...
 *	EXCEPTION_TRY{
 *		foo();
 *	} 
 *  	EXCEPTION_CATCH(&SomeException_Class) {
 *		SomeException e;
 *		EXCEPTION_GET_OBJ(e);
 *		bar(e);
 *	}
 *	EXCEPTION_CATCH_MORE(&AnotherException_Class) {
 *		AnotherException e;
 *		EXCEPTION_GET_OBJ(e);
 *		snafu(e);
 *	}
 *	EXCEPTION_FINALLY {
 *		something();
 *	}
 *	EXCEPTION_AFTER_CATCH
 *
 * Comments:
 *
 * The handling of uncaught exceptions is a bit crude and might be improved.
 *
 * Finally is not fully implemented: the finally block is not run if the
 * method does a /return/ within the try-block. The compiler should check
 * if there is a return in the try-block. If so, it should /goto finally_n/
 * instead of goto /gc_leave/. Also, since *all* finally blocks should be
 * run, the finally block should /goto finally_(n-1)/ (where block n-1 is the
 * static parent of block n).  *** NB *** We also need to change the 
 * implementation of the GC__IMPL_RETURN_* macros.
 *
 * Addendum 2003-11-26:
 * Fran's hack was a workaround for a logical error in my original 
 * implementation, but had the drawback that it put a greater responsibility 
 * on the code generator (and -- worse -- the implementor of native methods). 
 * The original idea of the GCI and friends was keeping as much of the 
 * dirty work hidden from the public, and this version is an attempt to stick 
 * to that idea...
 *
 * The current implementation moves RESTORE_ROOT_STACK to the catch clause
 * where it belongs. I.e., when an interrupt is caught, the root stack is
 * restored to the state it was in when the try block was entered.
 *
 * A small caveat: the root stack is only restored if an exception is
 * actually caught. The rationale behind this is that if the exception
 * is not caught at this level, it will be on an enclosing level, and
 * we really only need to do one RESTORE_ROOT_STACK per exception.
 *
 * The gain is that we may save a few RESTORE_ROOT_STACKs if we have
 * deeply nested try blocks and, more importantly, that the code is
 * a bit simpler.
 * 
 * The drawback is that the root stack is unneccesarily deep in the
 * finally block (which might be a problem if the finally block allocates
 * much memory and/or takes a long time to execute)
 *
 * The EXCEPTION_THROWS macro is now empty, but is kept for future use.
 */

#include <javatypes.h>
#include <java_lang_Object.h>

// Possible exception configurations.
#define NONE   0
#define SETJMP 1
#define GOTO   2

#if HAS_REFLECTION > 0
#include <reflection.h>
#define classpointer_to_index(p) \
  (((struct java_lang_Object_ClassStruct*) p)->classIndex)
/*   (*((int*)(p+3*sizeof(int*)))) */

#define classpointer_to_name(p) \
	 (clazzes[classpointer_to_index(p)].name)
#else
#define classpointer_to_index(p) 0
#define classpointer_to_name(p) "[no reflection]"
#endif

/********************************
 *
 * Exceptions using setjmp
 *
 ********************************/

#if HAS_EXCEPTIONS==SETJMP

#include <gc.h>
#include <gc_roots.h>
#include <setjmp.h>


#ifndef HOST_RTAI_KERNEL
#define UNCAUGHT_EXCEPTION(__nbr) \
 	printf("Uncaught exception: (%d,%s)\n", classpointer_to_index(__nbr), classpointer_to_name(__nbr));  
#else
#define UNCAUGHT_EXCEPTION(__nbr) \
	printk("Uncaught exception: (%d,%s)\n", classpointer_to_index(__nbr), classpointer_to_name(__nbr)); 
#endif

#define EXCEPTION_THROWS 

#define CHECK(__catch_lbl) 

#define EXCEPTION_THROW(unused_lbl,__nbr)	\
{ \
	ex_env_t *__tmp_env; \
	EXCEPTION_POP(__tmp_env); \
	if(__tmp_env) { \
		longjmp(__tmp_env->buf, __nbr);	\
	} else { \
		UNCAUGHT_EXCEPTION(__nbr); \
	} \
}

#ifdef GC_LAZY_GET_THREAD_ROOT
// HERE BE DRAGONS! This is a work-around for the lazy implementation
// of SAVE_ROOT_STACK. This is a good idea for functions, but doesn't
// matter as much for try-blocks. For now, the context "func" is hard-
// coded in the lazy version of the PUSH_ROOT macros, and in order to
// keep that reasonably clean, we use the following hack to actually
// get it to save the root stack directly at the start of each try-block.
#define __SAVE_ROOTSTACK(context) \
	 SAVE_ROOT_STACK(context); \
	 SAVE_ROOT_STACK_LAZY(context)
#else
#define __SAVE_ROOTSTACK(context)  SAVE_ROOT_STACK(context)
#endif
 
 
#define EXCEPTION_TRY \
{	__SAVE_ROOTSTACK(exception); \
	{\
		ex_env_t __env; \
		volatile int __ex_nbr; \
		volatile int __ex_throw = 1;\
		EXCEPTION_PUSH(__env); \
		if ( (__ex_nbr = setjmp(__env.buf) ) == 0) {

#define EXCEPTION_CATCH(unused_lbl,__catch_nbr)	\
			__ex_throw = 0; \
			EXCEPTION_POP_DISCARD; \
                        } else if (isCompatibleClass(__ex_nbr, __catch_nbr)) { \
			RESTORE_ROOT_STACK(exception); \
			__ex_throw = 0;

#define EXCEPTION_NO_CATCH(__catch_lbl)


#define EXCEPTION_CATCH_MORE(__catch_nbr)	\
		} else if (__ex_nbr == __catch_nbr) { \
			RESTORE_ROOT_STACK(exception); \
			__ex_throw = 0; \

#define EXCEPTION_FINALLY } {

#define EXCEPTION_END(unused_lbl)\
		}\
                if(__ex_throw) {\
				  EXCEPTION_THROW(1,__ex_nbr)};	\
	}\
}

#define EXCEPTION_PUSH(__env) \
	__env.next = gc_thread_get_current()->env; \
	gc_thread_get_current()->env=&__env 

#define EXCEPTION_POP(__env_out) \
	__env_out = gc_thread_get_current()->env;\
	if(__env_out) \
		gc_thread_get_current()->env=gc_thread_get_current()->env->next

#define EXCEPTION_POP_DISCARD \
	gc_thread_get_current()->env;\
	gc_thread_get_current()->env= gc_thread_get_current()->env->next 

#define EXCEPTION_STORE_OBJ(obj) \
  GC_ASSIGN(gc_thread_get_current()->eObj,obj)
/*   gc_thread_get_current()->eObj =  obj */

#define EXCEPTION_GET_OBJ(obj) \
  GC_ASSIGN(obj,gc_thread_get_current()->eObj)
/*   obj = gc_thread_get_current()->eObj */


/********************************
 *
 * Exceptions using goto
 *
 ********************************/
#elif HAS_EXCEPTIONS==GOTO

#include <gc.h>
/* #include <gc_roots.h> */
/* #include <setjmp.h> */


#ifndef HOST_RTAI_KERNEL
#define UNCAUGHT_EXCEPTION(__nbr) \
	printf("Uncaught exception: (%d,%s)\n", classpointer_to_index(__nbr), classpointer_to_name(__nbr)); 
#else
#define UNCAUGHT_EXCEPTION(__nbr) \
	printk("Uncaught exception: (%d,%s)\n", classpointer_to_index(__nbr), classpointer_to_name(__nbr)); 
#endif

#define EXCEPTION_THROWS 

#define CHECK(__catch_lbl) \
  if (*__eFlag__) goto catch_##__catch_lbl

#define EXCEPTION_THROW(__catch_lbl, __nbr)	\
  *__eFlag__ = __nbr; goto catch_##__catch_lbl;

//  printf("Throw %d %d\n",__catch_lbl, __nbr);			\

#ifdef GC_LAZY_GET_THREAD_ROOT
// HERE BE DRAGONS! This is a work-around for the lazy implementation
// of SAVE_ROOT_STACK. This is a good idea for functions, but doesn't
// matter as much for try-blocks. For now, the context "func" is hard-
// coded in the lazy version of the PUSH_ROOT macros, and in order to
// keep that reasonably clean, we use the following hack to actually
// get it to save the root stack directly at the start of each try-block.
#define __SAVE_ROOTSTACK(context) \
	 SAVE_ROOT_STACK(context); \
	 SAVE_ROOT_STACK_LAZY(context)
#else
#define __SAVE_ROOTSTACK(context)  SAVE_ROOT_STACK(context)
#endif
 
 
#define EXCEPTION_TRY \
{	int __try_exit = 0;\
    __SAVE_ROOTSTACK(exception); 

#define EXCEPTION_CATCH(__catch_lbl,__catch_nbr)		\
  catch_##__catch_lbl: \
if (*__eFlag__) { \
  if (isCompatibleClass(*__eFlag__, __catch_nbr)) {	\
    RESTORE_ROOT_STACK(exception);			\
    *__eFlag__ = 0;

//  printf("Catch %d %d\n",__catch_lbl,__catch_nbr);			\

#define EXCEPTION_NO_CATCH(__catch_lbl) \
  catch_##__catch_lbl: \
  {{

#define EXCEPTION_CATCH_MORE(__catch_nbr)	\
  }		\
 else if (isCompatibleClass(*__eFlag__, __catch_nbr)) { \
    RESTORE_ROOT_STACK(exception);			  \
    *__eFlag__ = 0;

//   printf("Catch %d\n",__catch_nbr);

/* #define EXCEPTION_AFTER_CATCH(__catch_lbl)	\ */
/* #define EXCEPTION_FINALLY } { */
#define EXCEPTION_FINALLY	\
  }}{	
/*   if(*__eFlag__) {  */

#define EXCEPTION_END(__catch_lbl) \
  }}
/* 	EXCEPTION_THROW(__catch_lbl,*__eFlag__)}; \ */
/*   } */

//	printf("No catch for %d\n",__catch_lbl); \

#define EXCEPTION_PUSH(__env) \
	__env.next = gc_thread_get_current()->env; \
	gc_thread_get_current()->env=&__env 

#define EXCEPTION_POP(__env_out) \
	__env_out = gc_thread_get_current()->env;\
	if(__env_out) \
		gc_thread_get_current()->env=gc_thread_get_current()->env->next

#define EXCEPTION_POP_DISCARD \
	gc_thread_get_current()->env;\
	gc_thread_get_current()->env= gc_thread_get_current()->env->next 

#define EXCEPTION_STORE_OBJ(obj) \
  GC_ASSIGN(gc_thread_get_current()->eObj,obj)

#define EXCEPTION_GET_OBJ(obj) \
  GC_ASSIGN(obj,gc_thread_get_current()->eObj)


/********************************
 *
 * No exceptions
 *
 ********************************/
#else // HAS_EXCEPTIONS

// HERE BE DRAGONS
// Define the empty equivalents of the exception macros to allow code
// that catches exceptions to work with exceptions turned off



#define CHECK(__catch_lbl)
#define EXCEPTION_THROWS
#define EXCEPTION_THROW(label,a)					\
  printf("WARNING: EXCEPTION_THROW(%d,%s): exceptions not enabled\n", classpointer_to_index(a), classpointer_to_name(a)); 
#define EXCEPTION_TRY
#define EXCEPTION_CATCH(a,b) if(0) {
#define EXCEPTION_CATCH_MORE(a) } if(0) {
#define EXCEPTION_FINALLY } {
#define EXCEPTION_AFTER_CATCH(__catch_lbl)	}
#define EXCEPTION_END(a) }
#define EXCEPTION_STORE_OBJ(obj) 
#define EXCEPTION_GET_OBJ(obj) 

#endif // HAS_EXCEPTIONS


#endif // EXCEPTION_H
