#ifndef __GC_H__
#define __GC_H__
//
// The GC macros are defined in layers, whith the actual GC calls at
// the bottom layer and the interface to be used in programs is in the
// topmost layer.
//
//  +-----------------------------------+
//  | User layer (GC_*)                 |
//  +-----------------------------------+
//  | Thread layer (GC__THREAD_*)       | @@@@ Atomic?
//  +-----------------------------------+
//  | Debug layer (GC__DEBUG_*)         |
//  +-----------------------------------+
//  | Implementation layer (GC__IMPL_*) |
//  +-----------------------------------+
//
//
// The macros to be used in programs
//
// A. One time
//
//    GC_GLOBALS		Declare all needed global GC variables
//    GC_INIT(heapsize)		Called once in the main thread before
//				any othe GC calls are done. 
//    GC_EXIT                   Is this ever needed?
//
// B. Object layout declaration 
//
//    GC_STRUCT_BEGIN(name)	Beginning of  object
//    GC_STRUCT_REF(type, name)	Reference variable
//    GC_STRUCT_VAR(type, name)	Ordinary variable (char, ...)
//    GC_STRUCT_END(name)	End of object
//
//    GC_NEW_LAYOUT(layout,object,#descriptors, descriptor1, ...)
//    GC_LAYOUT_FIELD(object, field)
//    GC_LAYOUT_ARRAY(array_length, array_of_offsets)
//
// C. Reference declaration
//
//    GC_REF(type, name)	Reference variable   
//    GC_PUSH_ROOT(var)	        Make variable known to GC
//    GC_POP_ROOT(var)		Make GC forget about variable
//    GC_PUSH_STATIC_ROOT(var)  Make global variable known to GC
//    GC_POP_STATIC_ROOT(var)   Make GC forget about global variable
//
// D. Object allocation
//
//    GC_NEW(var, layout)
//
// E. Reference access
//
//    GC_ASSIGN(destination, source)
//    GC_ASSIGN_NULL(destination)
//    GC_IS_NULL(ref)
//    GC_EQUAL(ref1, ref2)  Check if the pointers refer to the same object
//
// F. Field access
//
//    GC_GET(var, object, field)
//    GC_SET(object, field, value)
//    GC_GET_REF(var, object, field)
//    GC_SET_REF(object, field, value)
//    GC_SET_NULL(object, field)
//
// G. Function declaration 
//    GC_REF_FUNC_BEGIN(type, func, ...)
//    GC_VAR_FUNC_BEGIN(type, func, ...)
//    GC_PROC_BEGIN(func, ...)
//    GC_PARAM(type, name)      Declaration reference parameter 
//    GC_PARAM_REF(type, name)  Body reference parameter
//    GC_PUSH_PARAM(name)	Make parameter known to GC
//    GC_FUNC_ENTER
//    GC_RETURN_REF(result)
//    GC_RETURN_VAR(result)
//    GC_RETURN_NULL
//    GC_RETURN
//    GC_FUNC_LEAVE
//    GC_POP_PARAM(name)	Make GC forget about parameter
//    GC_REF_FUNC_END(type, func)
//    GC_VAR_FUNC_END(type, func)
//    GC_PROC_END(func)
//
// H. Function call
//    
//    GC_REF_FUNC_CALL(result, func_name, ...)
//    GC_VAR_FUNC_CALL(result, func_name, ...)
//    GC_PROC_CALL(func_name, ...)
//    GC_PASS(name)
//
//
// I. Magic
//
//    GC_SET_FINALIZER(object,finalizer)
//
// @@@@ What have we forgotten
//
// 1. A "real" equals method should be added besides the equivalence operator.
//    eg. equivaleve operator:      ref1 == ref2
//        deep pointer comparision: GC_EQUAL(ref1, ref2)
//
// Examples:
//
//   B. Object layout
//
//     struct Object {
//       int anInt;
//       struct Object *anObject;
//     };
// 
//   is translated to
//
//     GC_STRUCT_BEGIN(Object)
//       GC_STRUCT_VAR(int, anInt);
//       GC_STRUCT_REF(Object, anObject);
//     GC_STRUCT_END(Object);

/*******************************************************/
/*** Define constants to be used in selection macros ***/
/*******************************************************/

#define GC_THREAD_PTHREAD 1
#define GC_THREAD_STORK 2
#define GC_THREAD_TOBBE 3
#define GC_THREAD_RTAIKERNEL 4
#define GC_THREAD_XENOMAI 5
#define GC_THREAD_T2 6

#define GC_ALGORITHM_BATCH_COPY 1
#define GC_ALGORITHM_INCREMENTAL_MS 2
#define GC_ALGORITHM_INCREMENTAL_MS_BENGTSSON 3
#define GC_ALGORITHM_INCREMENTAL_MS_NONMOVING 4
#define GC_ALGORITHM_INCREMENTAL_COPY 5

/*******************************************************/
/*** start configuration                             ***/
/*******************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>							
#endif

#ifndef GC_ALGORITHM
#error No GC algorithm defined
#endif

#if GC_ALGORITHM == GC_ALGORITHM_INCREMENTAL_MS_NONMOVING
#define GC_IS_NONMOVING
#else
#undef GC_IS_NONMOVING
#endif

#undef UPDATE_MACRO
#ifdef UPDATE_MACRO
#define gc_update(a, c) \
        *GC___PTR(a) = GC___PTR(c); \
        gc_shade(GC___PTR(c)) 
#endif

#ifdef GC_HAS_THREADS
#ifdef GC_NO_GC_LOCKING
#define gc_lock()
#define gc_unlock()
#else
#undef GC_LOCK_AS_MACROS
#ifdef GC_LOCK_AS_MACROS
#define gc_lock    rt_global_cli
#define gc_unlock  rt_global_sti
#else
//These are defined in gci_thread.h, and that is enough...
//void gc_lock(void);
//void gc_unlock(void);
#endif
#endif

#ifndef GC_THREAD
#error GC_HAS_THREADS defined but no thread model selected?
#endif
#else
/* HERE BE DRAGONS!!!! FIXME!!! */
#define gc_lock()
#define gc_unlock()
#endif

#if GC_ALGORITHM == GC_ALGORITHM_BATCH_COPY
#include "gc_impl_batch_copy.h"
#elif GC_ALGORITHM == GC_ALGORITHM_INCREMENTAL_MS
#include "gc_impl_incremental_ms.h"
#elif GC_ALGORITHM == GC_ALGORITHM_INCREMENTAL_MS_NONMOVING
#include "gc_impl_nonmoving.h"
#elif GC_ALGORITHM == GC_ALGORITHM_INCREMENTAL_COPY
#include "gc_impl_incremental_copy.h"
#else
#error "Non-existent GC_ALGORITHM specified"
#endif

/*******************************************************/
/*** start implementation                            ***/
/*******************************************************/

#include <stddef.h>
#if GC_THREAD == GC_THREAD_RTAIKERNEL
#include <linux/types.h>     // for size_t, needed by offsetof(.)
#include <rtai/sem.h>
#include <rtai/task.h>
#elif GC_THREAD == GC_THREAD_XENOMAI
#include <xenomai/native/sem.h>
#include <xenomai/native/task.h>
#elif GC_THREAD == GC_THREAD_PTHREAD
#include <pthread.h>
#include <semaphore.h>
#endif

int gc_layout(int descriptors, ...);


// Common implementation macros

#define OBJECTHEAD_HEAD \
  GCFields gc;\
  GC___REF(struct LayoutHead) tmplate;

#if GC_THREAD == GC_THREAD_RTAIKERNEL
#define OBJECTHEAD \
  OBJECTHEAD_HEAD \
  struct gc_thread_t* waitQ_first;\
  struct gc_thread_t* waitQ_last;\
  RT_SEM* lock;
#elif GC_THREAD == GC_THREAD_XENOMAI
#define OBJECTHEAD \
  OBJECTHEAD_HEAD \
  struct gc_thread_t* waitQ_first;\
  struct gc_thread_t* waitQ_last;\
  RT_SEM* lock;
#elif GC_THREAD == GC_THREAD_PTHREAD
#define OBJECTHEAD \
  OBJECTHEAD_HEAD \
  struct ljrt_monitor_t *lock;
#else
#define OBJECTHEAD \
  OBJECTHEAD_HEAD 
#endif

typedef unsigned short     uint16 ;  /* char  - unsigned 8-bits */

#ifdef GC_HAS_FINALIZERS
#define LAYOUTHEAD \
  OBJECTHEAD_HEAD;\
  GC___REF(struct Uint8Array) gcInfo; \
  uint16 objectSize;\
  void (*finalizer)(void *)
#else
#define LAYOUTHEAD \
  OBJECTHEAD_HEAD;\
  GC___REF(struct Uint8Array) gcInfo; \
  uint16 objectSize
#endif

/* This struct is only for internal use (currently, calculating the size
 * of the thread (etc) dependent fields, and should not be used for any
 * instantiation.
 */
typedef struct ObjectHead_Head {
  OBJECTHEAD_HEAD;
} ObjectHead_Head;

/* These are the "real" structs to be used */
typedef struct ObjectHead {
  OBJECTHEAD;
} ObjectHead;

typedef struct LayoutHead {
  LAYOUTHEAD;
} LayoutHead;


typedef struct gc_root {
  GC___REF(ObjectHead) ref;
  struct gc_root *next_root;
  struct gc_root *top_root;
  struct gc_root *next_thread;
  struct gc_root *prev_thread;
  int local_result;
} gc_root;

void *gc_new_2(GC___REF(LayoutHead) layout);

/* local_result in the gc_thread struct is used as a temporary variable for
   making GC_IS_NULL and GC_EQUAL expressions. (see e.g. GC__IMPL_EQUAL)*/

// HERE BE DRAGONS! temporarily moved local_result to gc_root

#if HAS_EXCEPTIONS > 0
#include <setjmp.h>
typedef struct ex_env_t{
	jmp_buf buf;
	struct ex_env_t *next;
	int type;
} ex_env_t;
#endif

#define GC__IMPL_REF_DECL(type, name) \
  struct name##_struct{ GC___REF(struct type) ref; gc_root *next_root; }

/* Moved for declaration of gc_thread_t */
#define GC__IMPL_REF(type, name) \
  GC__IMPL_REF_DECL(type,name) name##_real;\
  struct name##_struct * name




typedef struct gc_thread_t{
  gc_root root;
#if HAS_EXCEPTIONS > 0
  ex_env_t *env;
  GC__IMPL_REF(ObjectHead,eObj);
#endif
#if GC_THREAD == GC_THREAD_STORK || GC_THREAD == GC_THREAD_TOBBE || GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI || GC_THREAD == GC_THREAD_T2
  long pid;
  struct gc_thread_t *next;
#endif
 //    int local_result;
  GC___REF(ObjectHead) this;
  GC__IMPL_REF(ObjectHead,javaThis);
  void *semObj;
#if GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI
  struct gc_thread_t *waitQ_next;
  struct gc_thread_t *waitQ_previous;
#endif
} gc_thread_t;

gc_root *gc_thread_get_root(void);
//extern gc_root *gc_thread_roots;
extern gc_root *gc_static_roots;
//extern void ****gc_roottop;
//extern int gc_marking;
//extern struct objecthead *markstk;

void gc_incremental_ms_init(int heapsize);


#define GC__IMPL_KIND(object, field) \
  gc_field_kind(GC___PTR(object.ref), &(GC___PTR(object.ref)->field), \
		sizeof(GC___PTR(object.ref)->field))

GC___REF(LayoutHead) gc_new_layout(int size, int descriptors, ...);

#ifdef GC_GET_THREAD_ROOT_IN_SCHEDULER
  /* The scheduler (or context switch callback) sets this variable
   * to the root stack of the current thread at each context switch
   */
  extern gc_root *thread_root;
#endif

//
// One time
//
#define GC__IMPL_GLOBALS

#define GC__IMPL_INIT(heapsize) \
    gci_common_init(heapsize); \
    GC__IMPL_THREAD_INIT \
    gci_thread_start();\
    {

#ifdef GC_HAS_THREADS
#define GC__IMPL_EXIT \
   GC__IMPL_THREAD_EXIT \
   gc_exit(); \
  }
#else
#define GC__IMPL_EXIT \
   GC__IMPL_THREAD_EXIT \
  }
#endif

//
// For each thread
//

/* HERE BE DRAGONS!
 * I don't know if we need to do a proper IMPL_PUSH_ROOT on javaThis...
 */
#if HAS_EXCEPTIONS > 0
#define __PUSH_EXCEPTION_ROOT \
  GC__IMPL_PUSH_ROOT(gc_thread.eObj)
#define __POP_EXCEPTION_ROOT \
  GC__IMPL_POP_ROOT(gc_thread.eObj)
#else
#define __PUSH_EXCEPTION_ROOT do{}while(0)
#define __POP_EXCEPTION_ROOT do{}while(0)
#endif

#define GC__IMPL_THREAD_INIT \
{ \
  gc_thread_t gc_thread; \
  gc_thread.javaThis = &gc_thread.javaThis_real; \
  gc_thread_init(&gc_thread); \
  GC__GET_THREAD_ROOT; \
  __PUSH_EXCEPTION_ROOT;\
  {

#define GC__IMPL_THREAD_EXIT \
    } \
  __POP_EXCEPTION_ROOT;\
  gc_thread_exit(&gc_thread); \
  }
//
// Object layout declaration
//
#define GC__IMPL_STRUCT_BEGIN(name) struct name { OBJECTHEAD;
#define GC__IMPL_STRUCT_REF(type, name) GC___REF(struct type) name
#define GC__IMPL_STRUCT_VAR(type, name) type name
#define GC__IMPL_STRUCT_END(name) }
#define GC__IMPL_NEW_LAYOUT(layout, object, ...) \
  layout->ref = gc_new_layout(sizeof(struct object),__VA_ARGS__)
#define GC__IMPL_LAYOUT_FIELD(object, field) offsetof(struct object, field)
#ifdef GC_PRINT_WARNINGS
#define GC__IMPL_LAYOUT_ARRAY(length, offsets) \
fprintf(stderr, "GC_LAYOUT_ARRAY(%s, %s) not supported (yet)\n", \
          #length, #offsets)
#else
#define GC__IMPL_LAYOUT_ARRAY(length, offsets)
#endif


//
// Reference declaration
//

// HERE BE DRAGONS!
// We only perform the empty root stack test if we have GC_PRINT_WARNINGS 
// enabled. If the macros, the compiler  and native implementations are 
// correct, it can never happen, and thus the run-time overhead of the check 
// is unneccessary.
//
// This applies also to GC_POP_PARAM (which also uses ...DO_POP_ROOT)
// *****************************
// *H*E*R*E**B*E**D*R*A*G*O*N*S*
// *****************************
static int __push_count;  // Hack to make native code compile... is not used...
static void *__saved_root_func; // ditto...
// *H*E*R*E**B*E**D*R*A*G*O*N*S*
#ifdef GC_LAZY_GET_THREAD_ROOT

#define GC__DO_GET_THREAD_ROOT \
	thread_root = gc_thread_get_root(); 

#define GC__GET_THREAD_ROOT_LAZY \
	if(!thread_root) { \
		GC__DO_GET_THREAD_ROOT \
	}
#else
#define GC__GET_THREAD_ROOT_LAZY 
#endif

#ifdef GC_PRINT_WARNINGS
#define POP_ERROR(obj, type) printf("Error: Trying to pop "#type" ["#obj"] from empty root stack\n")
#define GC__IMPL__DO_POP_ROOT(obj, type) \
  { \
  if(thread_root->top_root && thread_root->top_root->next_root) { \
  thread_root->top_root = thread_root->top_root->next_root; \
  FIX_ROOT_STACK(thread_root->top_root); \
  } else { POP_ERROR(obj, type); } \
  }
#else
#define GC__IMPL__DO_POP_ROOT(obj, type) \
  { \
  __push_count--;\
  thread_root->top_root = thread_root->top_root->next_root; \
  FIX_ROOT_STACK(thread_root->top_root); \
  }
#endif

#ifdef GC_PROFILING
void push_root_dummy(void);
#define GC_PROFILING_PUSH push_root_dummy();
#else
#define GC_PROFILING_PUSH 
#endif


//#define GC__IMPL_PUSH_ALIAS(var) GC__IMPL_PUSH_ROOT(var)
//#define GC__IMPL_POP_ALIAS(var) GC__IMPL_POP_ROOT(var)

#define GC__IMPL_PUSH_ALIAS(var) \
  var = & var##_real

#define GC__IMPL_POP_ALIAS(var) 

/* TODO! For copying GC w/ fp, the ALIAS macros should be
 * equal to the ROOT macros
 */

#ifdef GC_LAZY_GET_THREAD_ROOT
#define __PUSH_SAVE_LAZY SAVE_ROOT_STACK_LAZY(func);
#else
#define __PUSH_SAVE_LAZY
#endif

#define GC__IMPL_PUSH_ROOT(var) \
  var = & var##_real;\
  var->ref = 0; \
  GC_PROFILING_PUSH \
  GC__GET_THREAD_ROOT_LAZY;\
  __PUSH_SAVE_LAZY;\
  __push_count++;\
  var->next_root = (thread_root->top_root); \
  (thread_root->top_root) = (gc_root*)var;


#define GC__IMPL_POP_ROOT(var) \
  var->ref = 0; \
  GC__IMPL__DO_POP_ROOT(var, root);

#define GC__IMPL_PUSH_STATIC_ROOT(var) \
  var = & var##_real;\
  var->ref = 0; \
  var->next_root = gc_static_roots->top_root; \
  gc_static_roots->top_root = (gc_root*)var
#define GC__IMPL_POP_STATIC_ROOT(var) \
  var->ref = 0; \
  gc_static_roots->top_root = gc_static_roots->top_root->next_root; \
  FIX_ROOT_STACK(gc_static_roots->top_root)

//
// Object allocation
//
#ifdef IAR
#define GC__IMPL_NEW(obj,layout) \
  GC___ASSIGN(obj->ref, (void *)gc_new_2(layout->ref));
#else
#define GC__IMPL_NEW(obj,layout) \
  GC___ASSIGN(obj->ref, gc_new_2(layout->ref));
#endif

//
// Reference access
//
#define GC__IMPL_ASSIGN(target, source) \
  GC___ASSIGN(target->ref, source->ref) 
#define GC__IMPL_ASSIGN_NULL(target) target->ref = 0
#define GC__IMPL_IS_NULL(var) (var->ref == 0)
#define GC__IMPL_EQUAL(ref1, ref2) ((void *)(ref1->ref)==(void *)(ref2->ref))

//
// Field access
//

#include <class_indices.h>
// TODO: FIXME: Fix macro dependency chain caused by including
// java_lang_Object.h via exception.h below. When this works,
// remove the && 0 at the end of the following expression!
// Jesper 2010-02-02
#if defined JAVA_2_C && HAS_EXCEPTIONS > 0 && defined JAVA_LANG_NULLPOINTEREXCEPTION_INDEX && 0
#include <exception.h>
#include <gci_common.h>
// HERE BE FILTHY BUGS!
// Neither label 0 or JAVA_LANG_NULLPOINTEREXCEPTION_INDEX are 
// apropriate with existing exception implementation
// Should be FIXED!!  060407, AndersN
#define CHECK_NULL(a) if(!a->ref) {\
    gc_unlock(); EXCEPTION_THROW(0,JAVA_LANG_NULLPOINTEREXCEPTION_INDEX); \
}
#elif defined GC_PRINT_WARNINGS
#define CHECK_NULL(a) if(!GC___PTR(a->ref)) {printf("Error! dereferencing NULL reference ["#a"]\n");}
#else
#define CHECK_NULL(a)
#endif

#define GC__IMPL_GET(var, object, field) CHECK_NULL(object); var = GC___PTR(object->ref)->field 
//#define GC__IMPL_CGET(var, type, object, field) CHECK_NULL(object); var = GC___PTR((GC___REF(type)) object->ref)->field 
#define GC__IMPL_CGET(var, type, object, field) CHECK_NULL(object); var = GC___CPTR(type, object->ref)->field 
#define GC__IMPL_CGET_REF(var, type, object, field) \
CHECK_NULL(object);   GC___ASSIGN(var->ref, GC___CPTR(type, object->ref)->field);

#define GC__IMPL_GET_PTR(var, type, object, field) CHECK_NULL(object); var = (type) GC___PTR(object->ref)->field 
#define GC__IMPL_SET(object, field, value) CHECK_NULL(object); GC___PTR(object->ref)->field = value 
#ifdef IAR
#define GC__IMPL_GET_REF(var, object, field) \
CHECK_NULL(object);   GC___ASSIGN(var->ref, (void*)GC___PTR(object->ref)->field);
#else
#define GC__IMPL_GET_REF(var, object, field) \
CHECK_NULL(object);   GC___ASSIGN(var->ref, GC___PTR(object->ref)->field);
#endif
#define GC__IMPL_SET_REF(object, field, value) \
CHECK_NULL(object);   GC___ASSIGN((GC___PTR(object->ref)->field), value->ref);
#define GC__IMPL_SET_NULL(object, field) GC___PTR(object->ref)->field = 0

//
// Function declaration
//
#ifndef GC_LAZY_GET_THREAD_ROOT
#ifdef GC_GET_THREAD_ROOT_IN_SCHEDULER
#define GC__GET_THREAD_ROOT 
#else
#define GC__GET_THREAD_ROOT \
  gc_root *thread_root = gc_thread_get_root()
#endif // IN_SCHEDULER
#else
#define GC__GET_THREAD_ROOT \
  gc_root *thread_root = 0; 
#endif

#define GC_REF_FUNC_PROTO(type,name,...) \
  GC___REF(type) name(void **__eFlag__ , ##__VA_ARGS__)


#define GC__IMPL_REF_FUNC_BEGIN(type, name, ...) \
  GC_REF_FUNC_PROTO(type,name,##__VA_ARGS__) {			  \
  GC__GET_THREAD_ROOT; { GC__IMPL_REF(type, gc_result);	  \
 GC__IMPL_PUSH_ROOT(gc_result); {
#define GC__IMPL_VAR_FUNC_BEGIN(type, name, ...) \
  type name(void **__eFlag__ , ##__VA_ARGS__) { \
    GC__GET_THREAD_ROOT; { type gc_result; {
#define GC__IMPL_PROC_BEGIN(name , ...) \
  void name(void **__eFlag__ , ##__VA_ARGS__) {	\
    GC__GET_THREAD_ROOT; { {

#ifdef GC__REF_CALL_PARAMS
#define GC__IMPL_PARAM(type, name) void *gc_param_##name
#define GC__IMPL_PARAM_REF(type, name) \
  struct { GC___REF(struct type) ref; gc_root *next_root; } *name = gc_param_##name

#define GC__IMPL_PUSH_PARAM(local,passed) 
#define GC__IMPL_POP_PARAM(local, passed) 
#else
#define GC__IMPL_PARAM(type, name) GC___REF(struct type) gc_param_##name
#define GC__IMPL_PARAM_REF(type, name) GC__IMPL_REF(type, name)
#define GC__IMPL_PUSH_PARAM(local,passed) \
  { \
  GC__GET_THREAD_ROOT_LAZY;\
  local = & local##_real;\
  GC___ASSIGN(local->ref, gc_param_##passed); \
  local->next_root = (thread_root->top_root); \
  thread_root->top_root = (gc_root*)local; \
  }
#define GC__IMPL_POP_PARAM(local,passed) \
  local->ref = 0; \
  GC__IMPL__DO_POP_ROOT(local, param);
#endif //REF_CALL_PARAMS

#define GC__IMPL_FUNC_ENTER { SAVE_ROOT_STACK(func); {
#define GC__IMPL_RETURN_REF(result) \
  GC__IMPL_ASSIGN(gc_result, result);  GC__IMPL_RETURN
#define GC__IMPL_RETURN_VAR(result) \
  gc_result = result; GC__IMPL_RETURN
#define GC__IMPL_RETURN_NULL \
  gc_result->ref = 0; \
  GC__IMPL_RETURN
#define GC__IMPL_RETURN RESTORE_ROOT_STACK(func); goto gc_leave
#define GC__IMPL_FUNC_LEAVE } } catch_0: gc_leave: 
#define GC__IMPL_REF_FUNC_END(type, name) \
  }{ void *kludge = gc_result->ref; GC__IMPL_POP_ROOT(gc_result); return kludge; }}}
#define GC__IMPL_VAR_FUNC_END(type, name) \
  } return gc_result; }}
#define GC__IMPL_PROC_END(name) \
  } }}

//
// Function call
// 
#include <gci_thread.h>

#define GC__IMPL_REF_FUNC_CALL(result, func, ...) \
  GC___ASSIGN(result->ref, func(__eFlag__ , ##__VA_ARGS__)); 
// Test below to evaluate new exception desing. AndersN 060123
/*   if (func == 101) {goto gc_leave;} */
/*   if (gc_thread_get_current()->pid == 101) {goto gc_leave;} */
#define GC__IMPL_VAR_FUNC_CALL(result, func, ...) \
  result = func(__eFlag__ , ##__VA_ARGS__); 
/*   if (func == 101) {goto gc_leave;} */
/*   if (gc_thread_get_current()->pid == 101) {goto gc_leave;} */
#define GC__IMPL_PROC_CALL(proc, ...) \
  proc(__eFlag__ , ##__VA_ARGS__); 
/*   if (proc == 101) {goto gc_leave;} */
/*   if (gc_thread_get_current()->pid == 101) {goto gc_leave;} */

#ifdef GC__REF_CALL_PARAMS
#define GC__IMPL_PASS(obj) obj
#else
#ifdef IAR
#define GC__IMPL_PASS(obj) (void*)GC___ADDR_OF_PTR(obj->ref)
#else
#define GC__IMPL_PASS(obj) GC___ADDR_OF_PTR(obj->ref)
#endif
#endif


//
// Array macros
// 
#define GC__IMPL_GET_VAR_ARRAY(var, arrayRef, index) var = (GC___PTR(arrayRef->ref)->data[index].bytes)
#define GC__IMPL_SET_VAR_ARRAY(arrayRef, index, value) (GC___PTR(arrayRef->ref)->data[index].bytes = value)
#define GC__IMPL_GET_REF_ARRAY(var, arrayRef, index) GC___ASSIGN(var->ref, (GC___PTR(arrayRef->ref)->data[index].bytes))
#define GC__IMPL_SET_REF_ARRAY(arrayRef, index, value) GC___ASSIGN((GC___PTR(arrayRef->ref)->data[index].bytes), value->ref)


//
// Debugging macros
//
#ifdef GC_NODEBUG

//
// One time
//
#define GC__DEBUG_GLOBALS \
  GC__IMPL_GLOBALS
#define GC__DEBUG_INIT(heapsize) \
  GC__IMPL_INIT(heapsize)
#define GC__DEBUG_EXIT \
  GC__IMPL_EXIT

//
// For each thread
//

#define GC__DEBUG_THREAD_INIT GC__IMPL_THREAD_INIT
#define GC__DEBUG_THREAD_EXIT GC__IMPL_THREAD_EXIT

//
// Object layout declaration 
//
#define GC__DEBUG_STRUCT_BEGIN(name) GC__IMPL_STRUCT_BEGIN(name)
#define GC__DEBUG_STRUCT_REF(type, name) GC__IMPL_STRUCT_REF(type, name)
#define GC__DEBUG_STRUCT_VAR(type, name) GC__IMPL_STRUCT_VAR(type, name)
#define GC__DEBUG_STRUCT_END(name) GC__IMPL_STRUCT_END(name)
#define GC__DEBUG_NEW_LAYOUT(layout, object, ...) \
  GC__IMPL_NEW_LAYOUT(layout, object, __VA_ARGS__)
#define GC__DEBUG_LAYOUT_FIELD(object, field) \
  GC__IMPL_LAYOUT_FIELD(object, field)
#define GC__DEBUG_LAYOUT_ARRAY(length, array) \
  GC__IMPL_LAYOUT_ARRAY(length, array)

//
// Reference declaration
//
#define GC__DEBUG_REF(type, name) GC__IMPL_REF(type, name)
#define GC__DEBUG_PUSH_ROOT(var) GC__IMPL_PUSH_ROOT(var)
#define GC__DEBUG_PUSH_ALIAS(var) GC__IMPL_PUSH_ALIAS(var)
#define GC__DEBUG_POP_ALIAS(var) 
#define GC__DEBUG_POP_ROOT(var) GC__IMPL_POP_ROOT(var)
#define GC__DEBUG_PUSH_STATIC_ROOT(var) GC__IMPL_PUSH_STATIC_ROOT(var) 
#define GC__DEBUG_POP_STATIC_ROOT(var) GC__IMPL_POP_STATIC_ROOT(var)

//
// Object allocation
//
#define GC__DEBUG_NEW(var,layout) GC__IMPL_NEW(var,layout)

//
// Reference access
//
#define GC__DEBUG_ASSIGN(target, source) GC__IMPL_ASSIGN(target, source)
#define GC__DEBUG_ASSIGN_NULL(target) GC__IMPL_ASSIGN_NULL(target)
#define GC__DEBUG_IS_NULL(var) GC__IMPL_IS_NULL(var)
#define GC__DEBUG_EQUAL(ref1, ref2) GC__IMPL_EQUAL(ref1, ref2)

//
// Field access
//
#define GC__DEBUG_GET(var, object, field) GC__IMPL_GET(var, object, field)
#define GC__DEBUG_GET_PTR(var, type, object, field) GC__IMPL_GET_PTR(var, type, object, field)
#define GC__DEBUG_SET(object, field, value) GC__IMPL_SET(object, field, value)
#define GC__DEBUG_GET_REF(var, object, field) \
  GC__IMPL_GET_REF(var, object, field)
#define GC__DEBUG_SET_REF(object, field, value) \
  GC__IMPL_SET_REF(object, field, value)
#define GC__DEBUG_SET_NULL(object, field) \
  GC__IMPL_SET_NULL(object, field)

//
// Function declaration
//
#define GC__DEBUG_REF_FUNC_BEGIN(type, name, ...) \
  GC__IMPL_REF_FUNC_BEGIN(type, name , ##__VA_ARGS__) 
#define GC__DEBUG_VAR_FUNC_BEGIN(type, name, ...) \
  GC__IMPL_VAR_FUNC_BEGIN(type, name , ##__VA_ARGS__) 
#define GC__DEBUG_PROC_BEGIN(name, ...) \
  GC__IMPL_PROC_BEGIN(name , ##__VA_ARGS__) 
#define GC__DEBUG_PARAM(type, name) GC__IMPL_PARAM(type, name)
#define GC__DEBUG_PARAM_REF(type, name)  \
    GC__IMPL_PARAM_REF(type, name)
#define GC__DEBUG_PUSH_PARAM(local,passed) \
  GC__IMPL_PUSH_PARAM(local,passed)
#define GC__DEBUG_FUNC_ENTER \
  GC__IMPL_FUNC_ENTER
#define GC__DEBUG_RETURN_REF(result) GC__IMPL_RETURN_REF(result)
#define GC__DEBUG_RETURN_VAR(result) GC__IMPL_RETURN_VAR(result)
#define GC__DEBUG_RETURN_NULL GC__IMPL_RETURN_NULL
#define GC__DEBUG_RETURN GC__IMPL_RETURN
#define GC__DEBUG_FUNC_LEAVE \
  GC__IMPL_FUNC_LEAVE 
#define GC__DEBUG_POP_PARAM(local,passed)  \
  GC__IMPL_POP_PARAM(local,passed)
#define GC__DEBUG_REF_FUNC_END(type, name) GC__IMPL_REF_FUNC_END(type, name)
#define GC__DEBUG_VAR_FUNC_END(type, name) \
  GC__IMPL_VAR_FUNC_END(type, name)
#define GC__DEBUG_PROC_END(name) \
  GC__IMPL_PROC_END(name)

//
// Function call
//
#define GC__DEBUG_REF_FUNC_CALL(result, func, ...) \
  GC__IMPL_REF_FUNC_CALL(result, func , ##__VA_ARGS__)
#define GC__DEBUG_VAR_FUNC_CALL(result, func, ...) \
  GC__IMPL_VAR_FUNC_CALL(result, func , ##__VA_ARGS__); 
#define GC__DEBUG_PROC_CALL(name, ...) \
  GC__IMPL_PROC_CALL(name , ##__VA_ARGS__) 
#define GC__DEBUG_PASS(obj) \
    GC__IMPL_PASS(obj)

//
// Array macros
// 
#define GC__DEBUG_GET_VAR_ARRAY(var, arrayRef, index) \
  GC__IMPL_GET_VAR_ARRAY(var, arrayRef, index)
#define GC__DEBUG_SET_VAR_ARRAY(arrayRef, index, value) \
  GC__IMPL_SET_VAR_ARRAY(arrayRef, index, value)
#define GC__DEBUG_GET_REF_ARRAY(var, arrayRef, index) \
  GC__IMPL_GET_REF_ARRAY(var, arrayRef, index)
#define GC__DEBUG_SET_REF_ARRAY(arrayRef, index, value) \
  GC__IMPL_SET_REF_ARRAY(arrayRef, index, value)


#else

//
// Debug declarations and internal macros
//
#ifdef GC_PRINT_WARNINGS
#include <stdio.h>
#endif

typedef struct gc__debug_ref_t {
  struct gc__debug_ref_t *next; 
  char *name; 
} gc__debug_ref_t;

extern gc__debug_ref_t *gc_paranoia;
typedef enum { 
  gc_state_boot, gc_state_globals, gc_state_run, gc_state_call, 
  gc_state_func_begin, gc_state_leave, gc_state_return, gc_state_exit 
} gc_debug_state_t;
extern gc_debug_state_t gc_debug_state;

#define GC__DEBUG_STATE_TEXT(x) \
  (x==gc_state_boot)?"gc_boot": \
  (x==gc_state_globals)?"gc_globals":\
  (x==gc_state_run)?"gc_run":\
  (x==gc_state_call)?"gc_call":\
  (x==gc_state_func_begin)?"gc_func_begin":\
  (x==gc_state_leave)?"gc_leave":\
  (x==gc_state_return)?"gc_return":\
  "gc_unknown"
#define GC__DEBUG_ASSERT(cond, ...) \
  (!(cond))?( \
    fprintf(stderr, "%s:%d: ", __FILE__, __LINE__), \
    fprintf(stderr, __VA_ARGS__) \
  ): \
    0
#define GC__DEBUG_ASSERT_REF(object, field, ...) \
  (!(GC__IMPL_KIND(object.ref, field.ref) == 1)?( \
    fprintf(stderr, __VA_ARGS__), \
    fprintf(stderr, " invalid reference (%s:%d)\n",__FILE__, __LINE__) \
    ):0)

#define GC__DEBUG_ASSERT_STATE(state,...) \
  (!(gc_debug_state == state)?( \
    fprintf(stderr, __VA_ARGS__), \
    fprintf(stderr, " state is %s, but expected %s (%s:%d)\n", \
      GC__DEBUG_STATE_TEXT(gc_debug_state), GC__DEBUG_STATE_TEXT(state), \
      __FILE__, __LINE__) \
    ):0)

//
// One time
//
#define GC__DEBUG_GLOBALS \
  gc__debug_ref_t *gc_paranoia = 0; \
  gc_debug_state_t gc_debug_state = gc_state_globals; \
  GC__IMPL_GLOBALS
#define GC__DEBUG_INIT(heapsize) \
  GC__DEBUG_ASSERT_STATE(gc_state_globals, "GC_INIT(%d)", heapsize); \
  gc_debug_state = gc_state_run; \
  GC__IMPL_INIT(heapsize)
#define GC__DEBUG_EXIT \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_EXIT"); \
  gc_debug_state = gc_state_exit; \
  GC__IMPL_EXIT

//
// For each thread
//

#define GC__DEBUG_THREAD_INIT GC__IMPL_THREAD_INIT
#define GC__DEBUG_THREAD_EXIT GC__IMPL_THREAD_EXIT

//
// Object layout declaration 
//
#define GC__DEBUG_STRUCT_BEGIN(name) GC__IMPL_STRUCT_BEGIN(name)
#define GC__DEBUG_STRUCT_REF(type, name) \
  struct{ GC__IMPL_STRUCT_REF(type, ref);} name
#define GC__DEBUG_STRUCT_VAR(type, name) GC__IMPL_STRUCT_VAR(type, name)
#define GC__DEBUG_STRUCT_END(name) GC__IMPL_STRUCT_END(name)
#define GC__DEBUG_NEW_LAYOUT(layout, object, ...) \
  GC__IMPL_NEW_LAYOUT(layout.ref, object, __VA_ARGS__)
#define GC__DEBUG_LAYOUT_FIELD(object, field) \
  GC__IMPL_LAYOUT_FIELD(object, field.ref)
#define GC__DEBUG_LAYOUT_ARRAY(length, array) \
  GC__IMPL_LAYOUT_ARRAY(length, array)

//
// Reference declaration
//
#define GC__DEBUG_REF(type, name) \
  struct { \
    gc__debug_ref_t debug; \
    GC__IMPL_REF(type, ref); \
  } name
#define GC__DEBUG_PUSH_ROOT(var) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_PUSH_ROOT(%s)", #var); \
  var.debug.next = gc_paranoia; \
  var.debug.name = #var; \
  gc_paranoia = &var.debug; \
  GC_GET_GLOBAL(paranoia) = &var.debug; \
/*  printf("%s %x\n", #var, GC_GET_GLOBAL(paranoia)); */ \
  GC__IMPL_PUSH_ROOT(var.ref); 
#define GC__DEBUG_POP_ROOT(var) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_POP_ROOT(%s)", #var); \
  GC__DEBUG_ASSERT(gc_paranoia == &var.debug, \
		   "GC_POP_ROOT(%s): exepected '%s'\n", \
		   var.debug.name, \
		   gc_paranoia->name); \
  gc_paranoia = gc_paranoia->next; \
  GC__IMPL_POP_ROOT(var.ref)
#define GC__DEBUG_PUSH_STATIC_ROOT(var) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_PUSH_STATIC_ROOT(%s)", #var); \
  var.debug.next = gc_paranoia; \
  var.debug.name = #var; \
  gc_paranoia = &var.debug; \
  GC__IMPL_PUSH_STATIC_ROOT(var.ref); 
#define GC__DEBUG_POP_STATIC_ROOT(var) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_POP_STATIC_ROOT(%s)", #var); \
  GC__DEBUG_ASSERT(gc_paranoia == &var.debug, \
		   "GC_POP_STATIC_ROOT(%s): exepected '%s'\n", \
		   var.debug.name, \
		   gc_paranoia->name); \
  gc_paranoia = gc_paranoia->next; \
  GC__IMPL_POP_STATIC_ROOT(var.ref)

//
// Object allocation
//
#define GC__DEBUG_NEW(var,layout) GC__IMPL_NEW(var.ref,layout.ref)

//
// Reference access
//
#define GC__DEBUG_ASSIGN(target, source) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_ASSIGN(%s,%s)", #target, #source); \
  GC__IMPL_ASSIGN(target.ref, source.ref)
#define GC__DEBUG_ASSIGN_NULL(target) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_ASSIGN_NULL(%s)", #target); \
  GC__IMPL_ASSIGN_NULL(target.ref)
#define GC__DEBUG_IS_NULL(var) GC__IMPL_IS_NULL(var.ref)
#define GC__DEBUG_EQUAL(ref1, ref2) GC__IMPL_EQUAL(ref1.ref, ref2.ref)

//
// Field access
//
#define GC__DEBUG_GET(var, object, field) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_GET(%s,%s,%s)",#var,#object,#field); \
  GC__IMPL_GET(var, object.ref, field)
#define GC__DEBUG_SET(object, field, value) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_SET(%s,%s,%s)",#object,#field,#value); \
  GC__IMPL_SET(object.ref, field, value)
#define GC__DEBUG_GET_REF(var, object, field) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_GET_REF(%s,%s,%s)", \
                         #var, #object, #field); \
  GC__DEBUG_ASSERT_REF(object, field, "GC_GET_REF(%s, %s, %s)", \
                       #var, #object, #field); \
  GC__IMPL_GET_REF(var.ref, object.ref, field.ref)
#define GC__DEBUG_SET_REF(object, field, value) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_SET_REF(%s,%s,%s)", \
                         #object, #field, #value); \
  GC__DEBUG_ASSERT_REF(object, field, "GC_SET_REF(%s, %s, %s)", \
                       #object, #field, #value); \
  GC__IMPL_SET_REF(object.ref, field.ref, value.ref)
#define GC__DEBUG_SET_NULL(object, field) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_SET_NULL(%s,%s)", \
                         #object, #field); \
  GC__DEBUG_ASSERT_REF(object, field, "GC_SET_NULL(%s, %s)", \
                       #object, #field); \
  GC__IMPL_SET_NULL(object.ref, field.ref)

//
// Function declaration
//
#define GC__DEBUG_REF_FUNC_BEGIN(type, name, ...) \
  GC__IMPL_REF_FUNC_BEGIN(type, name , ##__VA_ARGS__) \
  GC__DEBUG_ASSERT_STATE(gc_state_call, "GC_REF_FUNC_BEGIN(%s,%s)", #type, #name); \
  gc_debug_state = gc_state_func_begin; {
#define GC__DEBUG_VAR_FUNC_BEGIN(type, name, ...) \
  GC__IMPL_VAR_FUNC_BEGIN(type, name , ##__VA_ARGS__) \
  GC__DEBUG_ASSERT_STATE(gc_state_call, "GC_VAR_FUNC_BEGIN(%s,%s)", #type, #name); \
  gc_debug_state = gc_state_func_begin; {
#define GC__DEBUG_PROC_BEGIN(name, ...) \
  GC__IMPL_PROC_BEGIN(name , ##__VA_ARGS__) \
  GC__DEBUG_ASSERT_STATE(gc_state_call, "GC_PROC_BEGIN(%s,%s)", #name); \
  gc_debug_state = gc_state_func_begin; {
#define GC__DEBUG_PARAM(type, name) GC__IMPL_PARAM(type, name)
#define GC__DEBUG_PARAM_REF(type, name)  \
  struct { \
    gc__debug_ref_t debug; \
    GC__IMPL_PARAM_REF(type, ref); \
  } name
#define GC__DEBUG_PUSH_PARAM(local,passed) \
  GC__DEBUG_ASSERT_STATE(gc_state_func_begin, "GC_PUSH_PARAM(%s)", #local); \
  local.debug.next = gc_paranoia; \
  local.debug.name = #local; \
  gc_paranoia = &local.debug; \
  GC__IMPL_PUSH_PARAM(local.ref,passed)
#define GC__DEBUG_FUNC_ENTER \
  GC__DEBUG_ASSERT_STATE(gc_state_func_begin, "GC_FUNC_ENTER"); \
  gc_debug_state = gc_state_run; \
  GC__IMPL_FUNC_ENTER
#define GC__DEBUG_RETURN_REF(result) GC__IMPL_RETURN_REF(result.ref)
#define GC__DEBUG_RETURN_VAR(result) GC__IMPL_RETURN_VAR(result)
#define GC__DEBUG_RETURN_NULL GC__IMPL_RETURN_NULL
#define GC__DEBUG_RETURN GC__IMPL_RETURN
#define GC__DEBUG_FUNC_LEAVE \
  GC__IMPL_FUNC_LEAVE \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "GC_FUNC_LEAVE"); \
  gc_debug_state = gc_state_leave; 
#define GC__DEBUG_POP_PARAM(local,passed)  \
  GC__DEBUG_ASSERT_STATE(gc_state_leave, "GC_POP_PARAM(%s)", #local); \
  GC__DEBUG_ASSERT(gc_paranoia == &local.debug, \
	 	   "GC_POP_ROOT(%s): exepected '%s'\n", \
		   local.debug.name, \
		   gc_paranoia->name); \
  gc_paranoia = gc_paranoia->next; \
  GC__IMPL_POP_PARAM(local.ref,passed)
#define GC__DEBUG_REF_FUNC_END(type, name) \
  } \
  GC__DEBUG_ASSERT_STATE(gc_state_leave, "GC_REF_FUNC_END(%s,%s)", #type, #name); \
  gc_debug_state = gc_state_return; \
  GC__IMPL_REF_FUNC_END(type, name)
#define GC__DEBUG_VAR_FUNC_END(type, name) \
  } \
  GC__DEBUG_ASSERT_STATE(gc_state_leave, "GC_VAR_FUNC_END(%s,%s)", #type, #name); \
  gc_debug_state = gc_state_return; \
  GC__IMPL_VAR_FUNC_END(type, name)
#define GC__DEBUG_PROC_END(name) \
  } \
  GC__DEBUG_ASSERT_STATE(gc_state_leave, "GC_PROC_END(%s,%s)", #name); \
  gc_debug_state = gc_state_return; \
  GC__IMPL_PROC_END(name)

//
// Function call
//
#define GC__DEBUG_REF_FUNC_CALL(result, func, ...) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "before GC_REF_FUNC_CALL(%s,%s)", \
                         #result, #func); \
  gc_debug_state = gc_state_call; \
  GC__IMPL_REF_FUNC_CALL(result.ref, func , ##__VA_ARGS__); \
  GC__DEBUG_ASSERT_STATE(gc_state_return, "after GC_REF_FUNC_CALL(%s,%s)", \
                         #result, #func); \
  gc_debug_state = gc_state_run;
#define GC__DEBUG_VAR_FUNC_CALL(result, func, ...) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "before GC_VAR_FUNC_CALL(%s,%s)", \
                         #result, #func); \
  gc_debug_state = gc_state_call; \
  GC__IMPL_VAR_FUNC_CALL(result, func , ##__VA_ARGS__); \
  GC__DEBUG_ASSERT_STATE(gc_state_return, "after GC_VAR_FUNC_CALL(%s,%s)", \
                         #result, #func); \
  gc_debug_state = gc_state_run;
#define GC__DEBUG_PROC_CALL(name, ...) \
  GC__DEBUG_ASSERT_STATE(gc_state_run, "before GC_PROC_CALL(%s)", #name); \
  gc_debug_state = gc_state_call; \
  GC__IMPL_PROC_CALL(name , ##__VA_ARGS__); \
  gc_debug_state = gc_state_run
#define GC__DEBUG_PASS(obj) \
  ( GC__DEBUG_ASSERT_STATE(gc_state_call, "GC_PASS(%s)", #obj), \
    GC__IMPL_PASS(obj.ref) )

//
// Array macros
// 

// NB!!! Same code as for GC_NODEBUG!!!!!
// Should be changed.
#define GC__DEBUG_GET_VAR_ARRAY(var, arrayRef, index) \
  GC__IMPL_GET_VAR_ARRAY(var, arrayRef, index)
#define GC__DEBUG_SET_VAR_ARRAY(arrayRef, index, value) \
  GC__IMPL_SET_VAR_ARRAY(arrayRef, index, value)
#define GC__DEBUG_GET_REF_ARRAY(var, arrayRef, index) \
  GC__IMPL_GET_REF_ARRAY(var, arrayRef, index)
#define GC__DEBUG_SET_REF_ARRAY(arrayRef, index, value) \
  GC__IMPL_SET_REF_ARRAY(arrayRef, index, value)
#endif


//
// Thread securing macros
//
#ifdef GC_HAS_THREADS

// GC algorithm (etc) dependent optimizations

/* For non-moving GC algorithms, the read-barrier is empty, and therefore
 * the heap need not be locked on read acesses
 */
#if defined GC_IS_NONMOVING || defined GC_NO_GC_LOCKING
#define GC__READ_LOCK 
#define GC__READ_UNLOCK 
#define HACK_COMMA
#else
#define GC__READ_LOCK gc_lock()
#define GC__READ_UNLOCK gc_unlock()
#define HACK_COMMA ,
#endif

/* If the mark roots phase is done atomically, the root operations does
 * not require any (additional) synchronization with the GC.
 * (Thus, this is only worthwile if the mark roots phase can be made atomic
 * by some mechanism that is cheaper than gc_lock/unlock (e.g., turning off
 * interrupts)). 
 */

#ifdef GC_MARK_ROOTS_ATOMIC
#define GC__ROOT_LOCK 
#define GC__ROOT_UNLOCK 
#else
#define GC__ROOT_LOCK gc_lock()
#define GC__ROOT_UNLOCK gc_unlock()
#endif

/* If an argument to a function can be the only live reference to an object,
 * the argument passing at function calls must be protected or the GC may
 * falsely identify objects as garbage. If a moving GC is used, the params
 * must be protected if they are pointers to heap objects. When the references
 * are reference called, this is not necessary.
 */

#ifndef GC_IS_NONMOVING
#ifdef GC__REF_CALL_PARAMS
#undef GC_ARGS_MUST_BE_PROTECTED
#else
#define GC_ARGS_MUST_BE_PROTECTED
#endif
#else
#undef GC_ARGS_MUST_BE_PROTECTED
#endif

#ifdef GC_ARGS_MUST_BE_PROTECTED
#define GC__ARG_LOCK gc_lock()
#define GC__ARG_UNLOCK gc_unlock()
#else
#define GC__ARG_LOCK
#define GC__ARG_UNLOCK
#endif

//
// One time 
//
#define GC__THREAD_GLOBALS GC__DEBUG_GLOBALS
#define GC__THREAD_INIT(heapsize) GC__DEBUG_INIT(heapsize)
#define GC__THREAD_EXIT GC__DEBUG_EXIT

//
// For each thread
//

#define GC__THREAD_THREAD_INIT gc_lock(); GC__DEBUG_THREAD_INIT gc_unlock(); {
#define GC__THREAD_THREAD_EXIT GC__DEBUG_THREAD_EXIT }

//
// Object layout declaration
//
#define GC__THREAD_STRUCT_BEGIN(name) GC__DEBUG_STRUCT_BEGIN(name)
#define GC__THREAD_STRUCT_REF(type, name) GC__DEBUG_STRUCT_REF(type, name)
#define GC__THREAD_STRUCT_VAR(type, name) GC__DEBUG_STRUCT_VAR(type, name)
#define GC__THREAD_STRUCT_END(name) GC__DEBUG_STRUCT_END(name)
#define GC__THREAD_NEW_LAYOUT(layout, object, ...) \
  gc_lock(); GC__DEBUG_NEW_LAYOUT(layout, object,__VA_ARGS__); gc_unlock();
#define GC__THREAD_LAYOUT_FIELD(object, field) \
  GC__DEBUG_LAYOUT_FIELD(object, field)
#define GC__THREAD_LAYOUT_ARRAY(length, array) \
  GC__DEBUG_LAYOUT_ARRAY(length, array); 

//
// Reference declaration
//
#define GC__THREAD_REF(type, name) GC__DEBUG_REF(type, name)
#define GC__THREAD_PUSH_ROOT(var) \
  GC__ROOT_LOCK; GC__DEBUG_PUSH_ROOT(var); GC__ROOT_UNLOCK
#define GC__THREAD_PUSH_ALIAS(var) \
  GC__DEBUG_PUSH_ALIAS(var); 
#define GC__THREAD_POP_ALIAS(var) \
  GC__DEBUG_POP_ALIAS(var); 
#define GC__THREAD_POP_ROOT(var) \
  GC__ROOT_LOCK; GC__DEBUG_POP_ROOT(var); GC__ROOT_UNLOCK
#define GC__THREAD_PUSH_STATIC_ROOT(var) \
  GC__ROOT_LOCK; GC__DEBUG_PUSH_STATIC_ROOT(var); GC__ROOT_UNLOCK
#define GC__THREAD_POP_STATIC_ROOT(var) \
  GC__ROOT_LOCK; GC__DEBUG_POP_STATIC_ROOT(var); GC__ROOT_UNLOCK

//
// Object allocation
//
#define GC__THREAD_NEW(ref,layout) \
  gc_lock(); GC__DEBUG_NEW(ref,layout); gc_unlock()

//
// Reference access
//

#define GC__THREAD_ASSIGN(target, source) \
  gc_lock(); GC__DEBUG_ASSIGN(target, source); gc_unlock()
#define GC__THREAD_ASSIGN_NULL(target) \
  gc_lock(); GC__DEBUG_ASSIGN_NULL(target); gc_unlock()

/* HERE BE DRAGONS!
 * The (.,.,.)-hack with local result is not needed if we don't have
 * to lock the heap for the comparison. This should be cleared up
 */

#ifdef GC_IS_NONMOVING
#define GC__THREAD_IS_NULL(var) GC__DEBUG_IS_NULL(var)
#define GC__THREAD_EQUAL(ref1, ref2) GC__DEBUG_EQUAL(ref1, ref2)
#else
#if defined GC_GET_THREAD_ROOT_IN_SCHEDULER
#define GC__THREAD_IS_NULL(var) \
( GC__READ_LOCK HACK_COMMA \
  thread_root->local_result=(GC__DEBUG_IS_NULL(var)), \
  GC__READ_UNLOCK HACK_COMMA \
  thread_root->local_result)

#define GC__THREAD_EQUAL(ref1, ref2) \
( GC__READ_LOCK HACK_COMMA  \
  thread_root->local_result=(GC__DEBUG_EQUAL(ref1, ref2)), \
  GC__READ_UNLOCK HACK_COMMA \
  thread_root->local_result)
#else
#define GC__THREAD_IS_NULL(var) \
( GC__READ_LOCK HACK_COMMA \
  gc_thread_get_root()->local_result=(GC__DEBUG_IS_NULL(var)), \
  GC__READ_UNLOCK HACK_COMMA \
  gc_thread_get_root()->local_result)

#define GC__THREAD_EQUAL(ref1, ref2) \
( GC__READ_LOCK HACK_COMMA  \
  gc_thread_get_root()->local_result=(GC__DEBUG_EQUAL(ref1, ref2)), \
  GC__READ_UNLOCK HACK_COMMA \
  gc_thread_get_root()->local_result)
#endif
#endif
//
// Field access
//
#define GC__THREAD_GET(var, object, field) \
  GC__READ_LOCK; GC__DEBUG_GET(var, object, field); GC__READ_UNLOCK
#define GC__THREAD_CGET(var, type, object, field) \
  GC__READ_LOCK; GC__IMPL_CGET(var, type, object, field); GC__READ_UNLOCK
#define GC__THREAD_CGET_REF(var, type, object, field) \
  GC__READ_LOCK; GC__IMPL_CGET_REF(var, type, object, field); GC__READ_UNLOCK
#define GC__THREAD_GET_PTR(var, type, object, field) \
  GC__READ_LOCK; GC__DEBUG_GET_PTR(var, type, object, field); GC__READ_UNLOCK
#define GC__THREAD_SET(object, field, value) \
  GC__READ_LOCK; GC__DEBUG_SET(object, field, value); GC__READ_UNLOCK
#define GC__THREAD_GET_REF(var, object, field) \
  GC__READ_LOCK; GC__DEBUG_GET_REF(var, object, field); GC__READ_UNLOCK
#define GC__THREAD_SET_REF(object, field, value) \
  gc_lock(); GC__DEBUG_SET_REF(object, field, value); gc_unlock()
#define GC__THREAD_SET_NULL(object, field) \
  gc_lock(); GC__DEBUG_SET_NULL(object, field); gc_unlock()

//
// Function declaration
//
/* HERE BE DRAGONS!
 * Hack: we only need to protect the return value if we are
 * returning a reference. We rely on the compiler back-end
 * to do constant propagation and remove the if-statement.
 * Other solutions include introducing a special GC_REF_FUNC_LEAVE
 * macro or some macro redefinition hack.
 */

#define GC__RETURN_LOCK if(__gc_is_ref_func) gc_lock(); 
#define GC__RETURN_UNLOCK gc_unlock();

#define GC__THREAD_REF_FUNC_BEGIN(type, name, ...) \
  GC__DEBUG_REF_FUNC_BEGIN(type, name , ##__VA_ARGS__) \
  const int __gc_is_ref_func = 1;\
  GC_REF(java_lang_Object, __gc_monitor_obj);\
  GC_PUSH_ALIAS(__gc_monitor_obj);
#define GC__THREAD_VAR_FUNC_BEGIN(type, name, ...) \
  GC__DEBUG_VAR_FUNC_BEGIN(type, name , ##__VA_ARGS__) \
  const int __gc_is_ref_func = 0;\
  GC_REF(java_lang_Object, __gc_monitor_obj);\
  GC_PUSH_ALIAS(__gc_monitor_obj);
#define GC__THREAD_PROC_BEGIN(name, ...) \
  GC__DEBUG_PROC_BEGIN(name , ##__VA_ARGS__) \
  const int __gc_is_ref_func = 0;\
  GC_REF(java_lang_Object, __gc_monitor_obj);\
  GC_PUSH_ALIAS(__gc_monitor_obj);
#define GC__THREAD_PARAM(type, name) GC__DEBUG_PARAM(type, name)
#define GC__THREAD_PARAM_REF(type, name) GC__DEBUG_PARAM_REF(type, name)
#define GC__THREAD_PUSH_PARAM(local,passed) GC__DEBUG_PUSH_PARAM(local,passed);
#define GC__THREAD_FUNC_ENTER GC__ARG_UNLOCK;GC__DEBUG_FUNC_ENTER 
#define GC__THREAD_RETURN_REF(result) GC__DEBUG_RETURN_REF(result)
#define GC__THREAD_RETURN_VAR(result) GC__DEBUG_RETURN_VAR(result)
#define GC__THREAD_RETURN_NULL GC__DEBUG_RETURN_NULL
#define GC__THREAD_RETURN GC__DEBUG_RETURN
#define GC__THREAD_FUNC_LEAVE GC__DEBUG_FUNC_LEAVE; GC__RETURN_LOCK
#define GC__THREAD_POP_PARAM(local,passed) GC__DEBUG_POP_PARAM(local,passed)
#define GC__THREAD_REF_FUNC_END(type, name) GC__DEBUG_REF_FUNC_END(type, name)
#define GC__THREAD_VAR_FUNC_END(type, name) GC__DEBUG_VAR_FUNC_END(type, name)
#define GC__THREAD_PROC_END(name) GC__DEBUG_PROC_END(name)

//
// Function call
//
#define GC__THREAD_REF_FUNC_CALL(result, func, ...) \
  GC__ARG_LOCK; GC__DEBUG_REF_FUNC_CALL(result, func , ##__VA_ARGS__); GC__RETURN_UNLOCK
#define GC__THREAD_VAR_FUNC_CALL(result, func, ...) \
  GC__ARG_LOCK; GC__DEBUG_VAR_FUNC_CALL(result, func , ##__VA_ARGS__); 
#define GC__THREAD_PROC_CALL(name, ...) \
  GC__ARG_LOCK; GC__DEBUG_PROC_CALL(name , ##__VA_ARGS__); 
#define GC__THREAD_PASS(obj) GC__DEBUG_PASS(obj)

//
// Array macros
// 
#define GC__THREAD_GET_VAR_ARRAY(var, arrayRef, index) \
  GC__READ_LOCK; GC__DEBUG_GET_VAR_ARRAY(var, arrayRef, index); GC__READ_UNLOCK;
#define GC__THREAD_SET_VAR_ARRAY(arrayRef, index, value) \
  GC__READ_LOCK; GC__DEBUG_SET_VAR_ARRAY(arrayRef, index, value); GC__READ_UNLOCK;
#define GC__THREAD_GET_REF_ARRAY(var, arrayRef, index) \
  GC__READ_LOCK; GC__DEBUG_GET_REF_ARRAY(var, arrayRef, index); GC__READ_UNLOCK;
#define GC__THREAD_SET_REF_ARRAY(arrayRef, index, value) \
  gc_lock(); GC__DEBUG_SET_REF_ARRAY(arrayRef, index, value); gc_unlock();


#else

//
// One time 
//
#define GC__THREAD_GLOBALS GC__DEBUG_GLOBALS
#define GC__THREAD_INIT(heapsize) GC__DEBUG_INIT(heapsize)
#define GC__THREAD_EXIT GC__DEBUG_EXIT

//
// For each thread
//

#define GC__THREAD_THREAD_INIT GC__DEBUG_THREAD_INIT
#define GC__THREAD_THREAD_EXIT GC__DEBUG_THREAD_EXIT

//
// Object layout declaration
//
#define GC__THREAD_STRUCT_BEGIN(name) GC__DEBUG_STRUCT_BEGIN(name)
#define GC__THREAD_STRUCT_REF(type, name) GC__DEBUG_STRUCT_REF(type, name)
#define GC__THREAD_STRUCT_VAR(type, name) GC__DEBUG_STRUCT_VAR(type, name)
#define GC__THREAD_STRUCT_END(name) GC__DEBUG_STRUCT_END(name)
#define GC__THREAD_NEW_LAYOUT(layout, object, ...) \
  GC__DEBUG_NEW_LAYOUT(layout, object,__VA_ARGS__)
#define GC__THREAD_LAYOUT_FIELD(object, field) \
  GC__DEBUG_LAYOUT_FIELD(object, field)
#define GC__THREAD_LAYOUT_ARRAY(length, array) \
  GC__DEBUG_LAYOUT_ARRAY(length, array)

//
// Reference declaration
//
#define GC__THREAD_REF(type, name) GC__DEBUG_REF(type, name)
#define GC__THREAD_PUSH_ROOT(var) GC__DEBUG_PUSH_ROOT(var)
#define GC__THREAD_PUSH_ALIAS(var) GC__DEBUG_PUSH_ALIAS(var)
#define GC__THREAD_POP_ALIAS(var) GC__DEBUG_PUSH_ALIAS(var)
#define GC__THREAD_POP_ROOT(var) GC__DEBUG_POP_ROOT(var)
#define GC__THREAD_PUSH_STATIC_ROOT(var) GC__DEBUG_PUSH_STATIC_ROOT(var)
#define GC__THREAD_POP_STATIC_ROOT(var) GC__DEBUG_POP_STATIC_ROOT(var)

//
// Object allocation
//
#define GC__THREAD_NEW(ref,layout) GC__DEBUG_NEW(ref,layout)

//
// Reference access
//
#define GC__THREAD_ASSIGN(target, source) GC__DEBUG_ASSIGN(target, source)
#define GC__THREAD_ASSIGN_NULL(target) GC__DEBUG_ASSIGN_NULL(target)
#define GC__THREAD_IS_NULL(var) GC__DEBUG_IS_NULL(var)
#define GC__THREAD_EQUAL(ref1, ref2) GC__DEBUG_EQUAL(ref1, ref2)

//
// Field access
//
#define GC__THREAD_GET(var, object, field) \
  GC__DEBUG_GET(var, object, field)
#define GC__THREAD_GET_PTR(var, type, object, field) \
  GC__DEBUG_GET_PTR(var, type, object, field)
#define GC__THREAD_SET(object, field, value) \
  GC__DEBUG_SET(object, field, value)
#define GC__THREAD_GET_REF(var, object, field) \
  GC__DEBUG_GET_REF(var, object, field)
#define GC__THREAD_SET_REF(object, field, value) \
  GC__DEBUG_SET_REF(object, field, value)
#define GC__THREAD_SET_NULL(object, field) \
  GC__DEBUG_SET_NULL(object, field)

//
// Function declaration
//
#define GC__THREAD_REF_FUNC_BEGIN(type, name, ...) \
  GC__DEBUG_REF_FUNC_BEGIN(type, name , ##__VA_ARGS__)
#define GC__THREAD_VAR_FUNC_BEGIN(type, name, ...) \
  GC__DEBUG_VAR_FUNC_BEGIN(type, name , ##__VA_ARGS__)
#define GC__THREAD_PROC_BEGIN(name, ...) \
  GC__DEBUG_PROC_BEGIN(name , ##__VA_ARGS__)
#define GC__THREAD_PARAM(type, name) GC__DEBUG_PARAM(type, name)
#define GC__THREAD_PARAM_REF(type, name) GC__DEBUG_PARAM_REF(type, name)
#define GC__THREAD_PUSH_PARAM(local,passed) GC__DEBUG_PUSH_PARAM(local,passed)
#define GC__THREAD_FUNC_ENTER GC__DEBUG_FUNC_ENTER
#define GC__THREAD_RETURN_REF(result) GC__DEBUG_RETURN_REF(result)
#define GC__THREAD_RETURN_VAR(result) GC__DEBUG_RETURN_VAR(result)
#define GC__THREAD_RETURN_NULL GC__DEBUG_RETURN_NULL
#define GC__THREAD_RETURN GC__DEBUG_RETURN
#define GC__THREAD_FUNC_LEAVE GC__DEBUG_FUNC_LEAVE
#define GC__THREAD_POP_PARAM(local,passed) GC__DEBUG_POP_PARAM(local,passed)
#define GC__THREAD_REF_FUNC_END(type, name) GC__DEBUG_REF_FUNC_END(type, name)
#define GC__THREAD_VAR_FUNC_END(type, name) GC__DEBUG_VAR_FUNC_END(type, name)
#define GC__THREAD_PROC_END(name) GC__DEBUG_PROC_END(name)

//
// Function call
//
#define GC__THREAD_REF_FUNC_CALL(result, func, ...) \
  GC__DEBUG_REF_FUNC_CALL(result, func , ##__VA_ARGS__)
#define GC__THREAD_VAR_FUNC_CALL(result, func, ...) \
  GC__DEBUG_VAR_FUNC_CALL(result, func , ##__VA_ARGS__)
#define GC__THREAD_PROC_CALL(name, ...) \
  GC__DEBUG_PROC_CALL(name , ##__VA_ARGS__)
#define GC__THREAD_PASS(obj) GC__DEBUG_PASS(obj)

//
// Array macros
// 
#define GC__THREAD_GET_VAR_ARRAY(var, arrayRef, index) \
  GC__DEBUG_GET_VAR_ARRAY(var, arrayRef, index)
#define GC__THREAD_SET_VAR_ARRAY(arrayRef, index, value) \
  GC__IMPL_SET_VAR_ARRAY(arrayRef, index, value)
#define GC__THREAD_GET_REF_ARRAY(var, arrayRef, index) \
  GC__DEBUG_GET_REF_ARRAY(var, arrayRef, index)
#define GC__THREAD_SET_REF_ARRAY(arrayRef, index, value) \
  GC__IMPL_SET_REF_ARRAY(arrayRef, index, value)

#endif

/* OLD thread macros kept, just in case... /s
#ifdef GC_THREAD

#if GC_THREAD == GC_THREAD_PTHREAD

#define GC__THREAD_THREAD_LOCAL \
  gc_roots roots;
#error PTHREAD macros not finished yet



#endif
#error Thread securing macros not finished yet

#define GC__THREAD_GLOBALS GC__DEBUG_GLOBALS
#define GC__THREAD_INIT(heapsize) \
  gc_lock(); GC__DEBUG_INIT(heapsize); gc_unlock()
#define GC__THREAD_EXIT \
  gc_lock(); GC__DEBUG_EXIT; gc_unlock()
#define GC__THREAD_REF(type, name) GC__DEBUG_REF(type, name)
#define GC__THREAD_PUSH_ROOT(var) \
  gc_lock(); GC__DEBUG_PUSH_ROOT(var); gc_unlock()
#define GC__THREAD_POP_ROOT(var) \
  gc_lock(); GC__DEBUG_POP_ROOT(var); gc_unlock()
#define GC__THREAD_PUSH_STATIC_ROOT(var) \
  gc_lock(); GC__DEBUG_PUSH_ROOT(var); gc_unlock()
#define GC__THREAD_POP_STATIC_ROOT(var) \
  gc_lock(); GC__DEBUG_POP_ROOT(var); gc_unlock()
#define GC__THREAD_NEW(ref,layout) \
  gc_lock(); GC__DEBUG_NEW(ref,layout); gc_unlock()
#define GC__THREAD_NEW_LAYOUT(layout, object, ...) \
  gc_lock(); GC__DEBUG_NEW_LAYOUT(layout, object,__VA_ARGS__); gc_unlock()
#define GC__THREAD_LAYOUT_FIELD(object, field) \
  GC__DEBUG_LAYOUT_FIELD(object, field)
#define GC__THREAD_SET(object, field, value) \
  gc_lock(); GC__DEBUG_SET(object, field, value); gc_unlock()
#define GC__THREAD_GET(var, object, field) 
  gc_lock(); GC__DEBUG_GET(var, object, field); gc_unlock()
#define GC__THREAD_REF_PASS(name) GC__DEBUG_REF_PASS(name)
#define GC__THREAD_CALL_PROC(name, ...) \
  gc_lock(); GC__DEBUG_CALL_PROC(name, __VA_ARGS__); gc_unlock()

#else

//
// One time 
//
#define GC__THREAD_GLOBALS GC__DEBUG_GLOBALS
#define GC__THREAD_INIT(heapsize) GC__DEBUG_INIT(heapsize)
#define GC__THREAD_EXIT GC__DEBUG_EXIT

//
// Object layout declaration
//
#define GC__THREAD_STRUCT_BEGIN(name) GC__DEBUG_STRUCT_BEGIN(name)
#define GC__THREAD_STRUCT_REF(type, name) GC__DEBUG_STRUCT_REF(type, name)
#define GC__THREAD_STRUCT_VAR(type, name) GC__DEBUG_STRUCT_VAR(type, name)
#define GC__THREAD_STRUCT_END(name) GC__DEBUG_STRUCT_END(name)
#define GC__THREAD_NEW_LAYOUT(layout, object, ...) \
  GC__DEBUG_NEW_LAYOUT(layout, object,__VA_ARGS__)
#define GC__THREAD_LAYOUT_FIELD(object, field) \
  GC__DEBUG_LAYOUT_FIELD(object, field)
#define GC__THREAD_LAYOUT_ARRAY(length, array) \
  GC__DEBUG_LAYOUT_ARRAY(length, array)

//
// Reference declaration
//
#define GC__THREAD_REF(type, name) GC__DEBUG_REF(type, name)
#define GC__THREAD_PUSH_ROOT(var) GC__DEBUG_PUSH_ROOT(var)
#define GC__THREAD_POP_ROOT(var) GC__DEBUG_POP_ROOT(var)
#define GC__THREAD_PUSH_STATIC_ROOT(var) GC__DEBUG_PUSH_STATIC_ROOT(var)
#define GC__THREAD_POP_STATIC_ROOT(var) GC__DEBUG_POP_STATIC_ROOT(var)

//
// Object allocation
//
#define GC__THREAD_NEW(ref,layout) GC__DEBUG_NEW(ref,layout)

//
// Reference access
//
#define GC__THREAD_ASSIGN(target, source) GC__DEBUG_ASSIGN(target, source)
#define GC__THREAD_ASSIGN_NULL(target) GC__DEBUG_ASSIGN_NULL(target)
#define GC__THREAD_IS_NULL(var) GC__DEBUG_IS_NULL(var)

//
// Field access
//
#define GC__THREAD_GET(var, object, field) GC__DEBUG_GET(var, object, field)
#define GC__THREAD_SET(object, field, value) \
  GC__DEBUG_SET(object, field, value)
#define GC__THREAD_GET_REF(var, object, field) \
  GC__DEBUG_GET_REF(var, object, field)
#define GC__THREAD_SET_REF(object, field, value) \
  GC__DEBUG_SET_REF(object, field, value)
#define GC__THREAD_SET_NULL(object, field) \
  GC__DEBUG_SET_NULL(object, field)

//
// Function declaration
//
#define GC__THREAD_FUNC_BEGIN(type, name, ...) \
  GC__DEBUG_FUNC_BEGIN(type, name,__VA_ARGS__)
#define GC__THREAD_PARAM(type, name) GC__DEBUG_PARAM(type, name)
#define GC__THREAD_PARAM_REF(type, name) GC__DEBUG_PARAM_REF(type, name)
#define GC__THREAD_PUSH_PARAM(local,passed) GC__DEBUG_PUSH_PARAM(local,passed)
#define GC__THREAD_FUNC_ENTER GC__DEBUG_FUNC_ENTER
#define GC__THREAD_RETURN(result) GC__DEBUG_RETURN(result)
#define GC__THREAD_FUNC_LEAVE GC__DEBUG_FUNC_LEAVE
#define GC__THREAD_POP_PARAM(local,passed) GC__DEBUG_POP_PARAM(local,passed)
#define GC__THREAD_FUNC_END(type, name) GC__DEBUG_FUNC_END(type, name)

//
// Function call
//
#define GC__THREAD_FUNC_CALL(result, func, ...) \
  GC__DEBUG_FUNC_CALL(result, func, __VA_ARGS__)
#define GC__THREAD_PROC_CALL(name, ...) \
  GC__DEBUG_PROC_CALL(name, __VA_ARGS__)
#define GC__THREAD_REF_PASS(name) GC__DEBUG_REF_PASS(name)
#define GC__THREAD_PASS(obj) GC__DEBUG_PASS(obj)

#endif
*/

//
// User macros
//

//
// One time 
//
#define GC_GLOBALS GC__THREAD_GLOBALS \
  struct gc_globals gc_global;
#define GC_INIT(heapsize) GC__THREAD_INIT(heapsize)
#define GC_EXIT GC__THREAD_EXIT

//
// For each thread
//

#define GC_THREAD_INIT GC__THREAD_THREAD_INIT
#define GC_THREAD_EXIT GC__THREAD_THREAD_EXIT

// 
// Object layout declaration
// 
#define GC_STRUCT_BEGIN(name) GC__THREAD_STRUCT_BEGIN(name)
#define GC_STRUCT_REF(type, name) GC__THREAD_STRUCT_REF(type, name)
#define GC_STRUCT_VAR(type, name) GC__THREAD_STRUCT_VAR(type, name)
#define GC_STRUCT_END(name) GC__THREAD_STRUCT_END(name)
#define GC_NEW_LAYOUT(layout, object, ...) \
  GC__THREAD_NEW_LAYOUT(layout, object,__VA_ARGS__)
#define GC_LAYOUT_FIELD(object, field) GC__THREAD_LAYOUT_FIELD(object, field)
#define GC_LAYOUT_ARRAY(length, array) GC__THREAD_LAYOUT_ARRAY(length, array)

//
// Reference declaration
//
#define GC_REF(type, name) GC__THREAD_REF(type,name)
#define GC_PUSH_ROOT(var) GC__THREAD_PUSH_ROOT(var)
#define GC_PUSH_ALIAS(var) GC__THREAD_PUSH_ALIAS(var)
#define GC_POP_ALIAS(var) GC__THREAD_PUSH_ALIAS(var)
#define GC_POP_ROOT(var) GC__THREAD_POP_ROOT(var)
#define GC_PUSH_STATIC_ROOT(var) GC__THREAD_PUSH_STATIC_ROOT(var)
#define GC_POP_STATIC_ROOT(var) GC__THREAD_POP_STATIC_ROOT(var)

//
// Object allocation
//
#define GC_NEW(ref,layout) GC__THREAD_NEW(ref,layout); 

//
// Reference acess
//
#define GC_ASSIGN(target, source) GC__THREAD_ASSIGN(target, source)
#define GC_ASSIGN_NULL(target) GC__THREAD_ASSIGN_NULL(target)
#define GC_IS_NULL(var) GC__THREAD_IS_NULL(var)
#define GC_EQUAL(ref1, ref2) GC__THREAD_EQUAL(ref1, ref2)

//
// Field access
//
#define GC_GET(var, object, field) GC__THREAD_GET(var, object, field)
#define GC_CGET(var, type, object, field) GC__THREAD_CGET(var, type, object, field)
#define GC_CGET_REF(var, type, object, field) GC__THREAD_CGET_REF(var, type, object, field)
#define GC_GET_PTR(var, type, object, field) GC__THREAD_GET_PTR(var, type, object, field)
#define GC_SET(object, field, value) GC__THREAD_SET(object, field, value)
#define GC_GET_REF(var, object, field) \
  GC__THREAD_GET_REF(var, object, field)
#define GC_SET_REF(object, field, value) \
  GC__THREAD_SET_REF(object, field, value)
#define GC_SET_NULL(object, field) GC__THREAD_SET_NULL(object, field)




//
// Function declaration
//
#define GC_REF_FUNC_BEGIN(type, name, ...) \
  GC__THREAD_REF_FUNC_BEGIN(type, name , ##__VA_ARGS__)
#define GC_VAR_FUNC_BEGIN(type, name, ...) \
  GC__THREAD_VAR_FUNC_BEGIN(type, name , ##__VA_ARGS__)
#define GC_PROC_BEGIN(name, ...) \
  GC__THREAD_PROC_BEGIN(name , ##__VA_ARGS__)
#define GC_PARAM(type, name) GC__THREAD_PARAM(type, name)
#define GC_PARAM_REF(type, name) GC__THREAD_PARAM_REF(type, name)
#define GC_PUSH_PARAM(obj) GC__THREAD_PUSH_PARAM(obj,obj)
#define GC_FUNC_ENTER GC__THREAD_FUNC_ENTER
#define GC_RETURN_REF(result) GC__THREAD_RETURN_REF(result)
#define GC_RETURN_VAR(result) GC__THREAD_RETURN_VAR(result)
#define GC_RETURN_NULL GC__THREAD_RETURN_NULL
#define GC_RETURN GC__THREAD_RETURN
#define GC_FUNC_LEAVE GC__THREAD_FUNC_LEAVE 
#define GC_POP_PARAM(obj) GC__THREAD_POP_PARAM(obj,obj)
#define GC_REF_FUNC_END(type, name) GC__THREAD_REF_FUNC_END(type, name)
#define GC_VAR_FUNC_END(type, name) GC__THREAD_VAR_FUNC_END(type, name)
#define GC_PROC_END(name) GC__THREAD_PROC_END(name)

//
// Function call
//
#define GC_REF_FUNC_CALL(result, func, ...) \
  GC__THREAD_REF_FUNC_CALL(result, func , ##__VA_ARGS__)
#define GC_VAR_FUNC_CALL(result, func, ...) \
  GC__THREAD_VAR_FUNC_CALL(result, func , ##__VA_ARGS__)
#define GC_PROC_CALL(name, ...) \
  GC__THREAD_PROC_CALL(name , ##__VA_ARGS__)
#define GC_PASS(obj) GC__THREAD_PASS(obj)

//
// Array macros
//
#define GC_GET_VAR_ARRAY(var, arrayRef, index) \
  GC__THREAD_GET_VAR_ARRAY(var, arrayRef, index)
#define GC_SET_VAR_ARRAY(arrayRef, index, value) \
  GC__THREAD_SET_VAR_ARRAY(arrayRef, index, value)
#define GC_GET_REF_ARRAY(var, arrayRef, index) \
  GC__THREAD_GET_REF_ARRAY(var, arrayRef, index)
#define GC_SET_REF_ARRAY(arrayRef, index, value) \
  GC__THREAD_SET_REF_ARRAY(arrayRef, index, value)




//
// Men at work, beware...
//

#define GC__IMPL_GLOBAL_STRUCT \
  void *impl;

// HERE BE DRAGONS
// I didn't dare moving any code so there are a few ifdef GC_NODEBUG here...
// /sven
#ifdef GC_NODEBUG
#define GC__DEBUG_GLOBAL_STRUCT GC__IMPL_GLOBAL_STRUCT 
#else
#define GC__DEBUG_GLOBAL_STRUCT GC__IMPL_GLOBAL_STRUCT \
  gc__debug_ref_t *paranoia;
#endif  

#define GC__THREAD_GLOBAL_STRUCT GC__DEBUG_GLOBAL_STRUCT \
  void *thread;
#define GC_GLOBAL_STRUCT GC__THREAD_GLOBAL_STRUCT

extern struct gc_globals {
  GC_GLOBAL_STRUCT
} gc_global;

#define GC_GET_GLOBAL(name) gc_global.name

#ifdef GC_NODEBUG
#define GC_PUSH_STATICROOT(r) \
  GC__IMPL_PUSH_ROOT(r)
#define GC_POP_STATICROOT(r) \
  GC__IMPL_POP_ROOT(r)
#else  
#ifdef GC_PRINT_WARNINGS
#define GC_PUSH_STATICROOT(r) \
  fprintf(stderr, " GC_PUSH_STATICROOT(%s)\n", #r); \
  GC__IMPL_PUSH_ROOT(r)
#else
#define GC_PUSH_STATICROOT(r) \
  GC__IMPL_PUSH_ROOT(r)
#endif
#ifdef GC_PRINT_WARNINGS
#define GC_POP_STATICROOT(r) \
  fprintf(stderr, " GC_POP_STATICROOT(%s)\n", #r); \
  GC__IMPL_POP_ROOT(r)
#else
#define GC_POP_STATICROOT(r) \
  GC__IMPL_POP_ROOT(r)
#endif
#endif

#ifdef GC_NODEBUG
#define GC_NEW_ARRAY(ref,layout,arrSize, objSize, length) (void*)(ref) = gc_new_array(layout, arrSize, objSize, length)
#else
#define GC_NEW_ARRAY(ref,layout,arrSize, objSize, length) (void*)(ref.ref) = gc_new_array2(layout, arrSize, objSize, length)
#endif

#define GC__IMPL_BEGIN_ROOT(type, name) \
  { GC_REF(type, name); GC_PUSH_ROOT(name)
#define GC__DEBUG_BEGIN_ROOT(type, name) GC__IMPL_BEGIN_ROOT(type, name)
#define GC__THREAD_BEGIN_ROOT(type, name) GC__DEBUG_BEGIN_ROOT(type, name)
#define GC_BEGIN_ROOT(type, name) GC__THREAD_BEGIN_ROOT(type, name)

#define GC__IMPL_END_ROOT(name) GC_POP_ROOT(name); }
#define GC__DEBUG_END_ROOT(name) GC__IMPL_END_ROOT(name)
#define GC__THREAD_END_ROOT(name) GC__DEBUG_END_ROOT(name)
#define GC_END_ROOT(name) GC__THREAD_END_ROOT(name)


#ifdef GC_HAS_FINALIZERS

/* Macro to set a finalizer on an object.
**   NB! The finalizer pointer is located in the tmplate-object. I.e., 
**   calling GC_SET_FINALIZER() overrides any previous finalizers on
**   that *class* of objects 
*/

#define GC__IMPL_SET_FINALIZER(tmplate, finalize_ptr) \
    GC___PTR(tmplate)->finalizer = finalize_ptr

#define GC_SET_FINALIZER(object, finalize_ptr) \
  do { GC___REF(LayoutHead) tmplate; \
    GC_GET(tmplate, object, tmplate); \
    GC__IMPL_SET_FINALIZER(tmplate, finalize_ptr); \
  } while(0)
#else
// an empty implementation:
#ifdef GC_PRINT_WARNINGS
#define FINALIZER_WARNING printf("Warning: GC_SET_FINALIZER() called but GC_HAS_FINALIZERS is not set\n");
#else
#define FINALIZER_WARNING
#endif
#define GC_SET_FINALIZER(a,b) do{FINALIZER_WARNING}while(0)
#endif

#include <gc_roots.h>
#endif
