/*
  gc_incremental_ms/gc.c,
  The implementation of an inremental garbage collector for the IVM.

  Copyright (C) 2000 Roger Henriksson <roger@cs.lth.se>,
                     Sven Gestegrd Robertz <sven@cs.lth.se>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 

static int gc_started = 0;
#include <stdarg.h>

#ifdef GC_PRINT_WARNINGS
#if !(defined HOST_AVR || defined HOST_RTAI_KERNEL)
#include <stdio.h>
#endif
#endif

#include <string.h>
#include "gc.h"
#include "gc_roots.h"
#include "gc_incremental_ms.h"
#include "gci_sched.h"


#define DO_ALIGN
#define ALIGNMENT 4
//#define GC_PRINT_DEBUG_ALIGN

#ifdef CPU_UTILIZATION_LOGGING
#include "cpu_utilization_logger.h"
#endif

#ifdef CPU_UTILIZATION_LOGGING
#include "cpu_utilization_logger.h"
#endif

#ifdef LPMEM
extern unsigned int HP_reserved;
#endif 

#ifdef TIME_METRIC
#ifdef STORK_PPC
#include "Times.h"
void GC_SetDeadline (Times_Time deadline);
void GC_SetDeadlineAndWaitUntil (Times_Time deadline, Times_Time time);
#endif
#endif

#ifdef LOGGING
#include "gc_logging.h"
#endif

#undef VIRTUAL_FLIP
// Virtual flip: do not make the freed memory available until at the start
// of the next GC cycle. This is needed to keep the scheduler's GC cycles
// in sync with the variations in free memory.

#ifndef VIRTUAL_FLIP
#define alloctop otabletop
#endif
//#include <assert.h>

ObjectHead *markstk;
void **otabletop;
#ifdef VIRTUAL_FLIP
void **alloctop;
#endif
static void *freeoptrs;
static uint8 *allocpos;

int gc_marking,gc_working;
static uint32 gc_state;

uint8 gc_heap[GC_HEAPSIZE];

void gc(void);
void gc_cycle_time_adapt();

static ObjectHead *curobj,*curdest;

static uint32 workcnt;
static float32 workfctr;
static uint32 noretained,allocated;

#ifdef STORK_PPC
#define printf cprintf
#endif

#ifdef TIME_METRIC
#define MIN_OBJSIZE 20
#define MARK_FACTOR 0.3
#define GC_CYCLE_LENGTH 1900
#define GC_INIT_HOLDOFF 0	// Wait this time to give the system a chance 
				// to reach a stable state before starting GC.
				// May not be necessary.
uint32 gc_cycle_length;  // effective GC cycle length
uint32 gc_cycle_olength; // nominal GC cycle length
#if 0
#ifdef STORK_PPC
#ifdef GC_ONLY_TIME
Times_Time gc_cycle_oldstart;
#endif
Times_Time gc_cycle_start;
Times_Time gc_cycle_end;
#else
unsigned long gc_cycle_start;   // start time for current cycle
unsigned long gc_cycle_end;     // start_time + cycle_length
#endif
#endif // if 0... I think this has moved to gci_sched
#endif // TIME_METRIC

#ifdef GC_CYCLE_TIME_ADAPTIVE  
#ifdef STORK_PPC
static Times_Time lastSample;
#else
long lastSample;
#endif
static uint32 lastAlloced;
double allocrate;
#endif

///////// HEAP STATE
#ifdef GC_MEASURE_HEAPSTATE
static unsigned int alloccount,livecount, livebytes, deadcount, deadbytes;
static unsigned int oldlivecount, oldlivebytes;
static unsigned int olddeadcount, olddeadbytes;
int deltaDc, deltaDb;
int deltaDcp, deltaDbp;
#endif

#ifdef GC_PREDICT_C
#include <rtai/timer.h>
extern RTIME gc_cycle_start;
extern RTIME gc_cycle_end;
#endif

#ifdef DO_ALIGN
static unsigned int roundup( unsigned int size )
{
	unsigned int newsize = (size + ALIGNMENT - 1) & -ALIGNMENT;
#ifdef GC_PRINT_DEBUG_ALIGN
	if(newsize != size) printf("roundup: %d -> %d\n", size, newsize);
#endif
	return newsize;
}
#endif

int gc_marking_roots(void)
{
	return (gc_state==1) ;
}

void gc_mark_object(GC___REF(ObjectHead) *o) {
	if(*o && GC___PTR(*o) && GC___PTR(*o)->gc.mark==0) {
	//	printf("marking 0x%x (0x%x)\n", o, *o);
		GC___PTR(*o)->gc.mark = markstk;
		markstk = *GC___PTR(o);
#ifdef GC_MEASURE_HEAPSTATE
		{
		int foosize = getsize(GC___PTR(*o));
#ifdef DO_ALIGN
		foosize = roundup(foosize);
#endif
		livecount ++;
		livebytes += foosize;
		}
#endif
	 //      printf("*** pushed new object: 0x%x (0x%x)\n", markstk, *markstk);
	} else {
#ifdef GC_PRINT_DEBUG_MARK
		printf("gc_mark_object: o   = %x\n", o);
		printf("              : *o  = %x\n", *o);
#endif
	}
}

int gc_init(int heapsize)
{
  int result = 0;
  static int called = 0;
  if (!called) {
    called = 1;
    result = 1;

    markstk = IVM_NULL;
    otabletop = (void **) (void *) (gc_heap+sizeof(gc_heap));
#ifdef VIRTUAL_FLIP    
    alloctop = otabletop;
#endif
    freeoptrs = IVM_NULL;
    allocpos = gc_heap;
    curdest = (ObjectHead *) (void *) gc_heap;

    gc_marking = 0;
    gc_working = 0;
    gc_state = 0;
    noretained = 0;
    allocated = 0;
#ifdef GC_MEASURE_HEAPSTATE
    oldlivecount = 0;
    oldlivebytes = 0;
    olddeadcount = 0;
    olddeadbytes = 0;
    livecount = 0;
    livebytes = 0;
    alloccount= 0;
#endif
    gci_sched_init();
  }
  gc_roots_gc_init(&gc_mark_object);
  gc_started = 1;
  return result;
}

int gc_sleeping = 0;

static void* debug_ptr;

static inline void traverse_inc(void)
{
	if ((uint8 *)curobj<allocpos) {
		uint32 osize;

		osize = getsize(curobj);
#ifdef DO_ALIGN
		osize = roundup(osize);
#endif
		if(osize == 0 || osize == 1) {
#ifndef HOST_AVR
			printf("osize == %d curobj == %x\n", osize,curobj);
			printf("debug_ptr == %x\n", debug_ptr);
			printf("gc_heap == %x\n", gc_heap);
#endif
		} else {
			debug_ptr = curobj;
		}
		if (curobj->gc.mark) {
			if (curdest!=curobj) {
				memmove(curdest,curobj,osize);
				*(((void **) curdest->gc.fp)) = curdest;
			}
			curdest->gc.mark = 0;
			curdest = (ObjectHead *) (void *) (((uint8 *) curdest)+osize);
			noretained++;
		} else {
#ifdef GC_HAS_FINALIZERS
			if(GC___PTR(curobj->tmplate)->finalizer) {
#ifdef GC_PRINT_DEBUG
				printf("finalizing object\n");
#endif
				GC___PTR(curobj->tmplate)->finalizer(curobj); 
#ifdef GC_PRINT_DEBUG
				printf("done\n");
#endif
			}
#endif
			// Better destroy monitor after java finalizer
			// has been run (in case it is synchronized)
		        ljrt_monitor_free(curobj);
			*(((void **) curobj->gc.fp)) = freeoptrs;
			freeoptrs = curobj->gc.fp;
		}
		curobj = (ObjectHead *) (void *) (((uint8 *) curobj)+osize);
		workcnt += osize;
	} else {

		//printf("GC-cycle END: freemem = %d\n", gc_freemem());
#ifdef VIRTUAL_FLIP	
		alloctop=(void *)((uint8*)otabletop-(uint8*)allocpos+(uint8*)curdest);
#endif	
		allocpos = (uint8 *) curdest;
#ifdef GC_PRINT_DEBUG
#ifndef HOST_AVR
		printf("GC-cycle END: freemem = %d \n", gc_freemem());
#endif
#endif
#ifdef LPMEM
		// make sure that this GC cycle's NC allocation don't "steal"
		// memory from the next cycle.
		// This should probably be a function call
		HP_reserved += (unsigned int) curobj- (unsigned int) curdest;
#endif
#ifdef LOGGING
		log_gc_cycle_end();
#endif
		gci_sched_cycle_end();
		gc_working = 0;
		gc_state = 0;
	}
}

void initiate_gc_cycle(void);

void gc(void)
{
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
	cpu_utilization_logger_setGCRunning();
#endif
#endif
	switch (gc_state) {
#ifdef GC_ONLY_TIME
	        case 0:
		  initiate_gc_cycle();
#endif
		case 1: /* Mark roots */
			{
				if (gc_roots_mark_one()) {
					workcnt++;
				} else {
					gc_state=2;
			}
			break;
		}
		case 2: /* Mark recursively */
		{
			if (markstk!=(ObjectHead *) -1) {
			ObjectHead *obj;

			obj = markstk;
			markstk = obj->gc.mark;
			gc_parse_object(obj, gc_mark_object);
			workcnt++;
			} else {
				gc_marking = 0;
				curobj = (ObjectHead *) (void *) gc_heap;
				workfctr = (((uint8*)otabletop)-gc_heap)/
				  (float32)(((uint8*)otabletop)-allocpos);
				noretained = 0;
#ifdef GC_CYCLE_TIME_ADAPTIVE	
				// Hack to avoid anomalies around cycle bundaries.
				lastAlloced -= allocated;
				allocated = 0;
#endif	
#ifdef GC_MEASURE_HEAPSTATE
				{
				int deadcount, deadbytes;
				deadcount = oldlivecount-livecount;
				deadbytes = oldlivebytes-livebytes;
				deltaDc = deadcount - olddeadcount;
				deltaDb = deadbytes - olddeadbytes;
				deltaDcp = (100*deltaDc)/deadcount;
				deltaDbp = (100*deltaDb)/deadbytes;
				olddeadcount = deadcount;
				olddeadbytes = deadbytes;
//				printf("Sh: [%d, %d]\n", livecount, livebytes);
				printf("Sh: [%d, %d, %d, %d], freemem=%d, sum=%d\n", 
						livecount+alloccount, 
						livebytes+allocated, 
						deadcount, 
						deadbytes,
						gc_freemem(), 
						livebytes+allocated+deadbytes+gc_freemem()+ 
						((unsigned int) (gc_heap+sizeof(gc_heap)))-(unsigned int)otabletop);
				printf("    DeltaD= %d, %d\n", deltaDc, deltaDb);
#ifdef GC_PREDICT_C
				{
#if 0
				long a1 = 19;
				long a2 = 482;
				long a3 = 77;
				long a4 = 423;
#else
//				long a1 = 106;
//				long a2 = 53;
//				long a3 = -7000;
//				long a4 = 180;
				long a1 = 141;
				long a2 = 52;
				long a3 = -2372;
				long a4 = 116;
				
#endif
				int acpred = 0;
				int abpred = 0;
				RTIME W, Wp;
#if 0
				RTIME now = rt_timer_read();
				RTIME el = now - gc_cycle_start;
				float achat = alloccount/(el/1000000);
				float abhat = allocated/(el/1000000);
				acpred = 1000000*achat * (gc_cycle_end-now);
				abpred = 1000000*abhat * (gc_cycle_end-now);

				printf("pred: %d, %d, el=%llu, ac=%d, ab=%d,achat=%f, abhat=%f\n", acpred, abpred,el, alloccount, allocated, achat, abhat);
#else
				//abpred = gc_freemem()/2;
				//acpred = abpred/64;
#endif
				W = (a1 * (livecount+alloccount+acpred))/1000 +
				    (a2 * (livebytes+allocated+abpred))/1000 +
				    (a3 * (deadcount))/1000 +
				    (a4 * (deadbytes))/1000;
				
				Wp= (a1 * (livecount+alloccount+acpred-deltaDc))/1000 +
				    (a2 * (livebytes+allocated+abpred-deltaDb))/1000 +
				    (a3 * (deadcount+deltaDc))/1000 +
				    (a4 * (deadbytes+deltaDb))/1000;
				printf("What = %llu, next = %llu\n", W,Wp);

				}
#endif
				}
#endif
				workcnt = 0;
				gc_state = 3;
			}
			break;
		}
		case 3: /* Traverse A */
		{
			traverse_inc();
			break;
		}
	}
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
	cpu_utilization_logger_setGCnotRunning();
#endif
#endif
}

void initiate_gc_cycle(void) 
{
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
    cpu_utilization_logger_setGCRunning();
#endif
#endif
#ifdef GC_PRINT_DEBUG
        printf("gc.c: initiate_gc_cycle\n");
#endif 
#ifdef CPU_UTILIZATION_LOGGING
#if GC_CYCLE > 0
  cpu_utilization_logger_gc_cycle_start();
#endif
#endif
#ifdef LOGGING
  log_gc_cycle_start();
#endif
  gc_working = 1;
  gc_marking = 1;
  gc_state = 1;
  markstk = (ObjectHead *) -1;
  gci_sched_init_cycle();
  gc_roots_initiate_gc_cycle();

  curobj = (ObjectHead *) (void *) gc_heap;
  curdest = (ObjectHead *) (void *) gc_heap;

  workcnt = 0;
  workfctr = (GC_MAX_ROOTS+noretained)/
             (float32)(GC_MARKRATIO(gc_heap+sizeof(gc_heap)-allocpos));
#ifdef GC_MEASURE_HEAPSTATE
    oldlivecount = livecount + alloccount;
    oldlivebytes = livebytes + allocated;
    livecount = 0;
    livebytes = 0;
    alloccount= 0;
#endif
  allocated = 0;
#ifdef TIME_METRIC
#ifdef VIRTUAL_FLIP  
  alloctop = otabletop;
#endif  
#ifdef LPMEM
  gc_lpmem_init_cycle();
#endif //LPMEM
#endif // TIME_METRIC
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
    cpu_utilization_logger_setGCnotRunning();
#endif
#endif

}

void conclude_gc_cycle(void) 
{
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
    cpu_utilization_logger_setGCRunning();
#endif
cpu_utilization_logger_gc_log_conclude();
#endif
#ifdef LPMEM
    gc_lpmem_panic();
#endif
#ifdef LOGGING
    log_gc_panic();
#endif
#ifdef LOGGING
  log_gc_inc_start();
#endif
  while (gc_working)
    gc();
#ifdef LOGGING
  log_gc_inc_end();
#endif

#ifdef TIME_METRIC
#ifdef STORK_PPC
	gc_cycle_end = Times_Now();
#endif
#ifdef VIRTUAL_FLIP	
	alloctop = otabletop;
#endif	
#endif
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
    cpu_utilization_logger_setGCnotRunning();
#endif
#endif
}

void gc_full_cycle(void)
{
	if (!gc_is_working()) initiate_gc_cycle();
	conclude_gc_cycle();
}


/* aux functions factored out of gc_get_handle
 * to make it readable
 */
static inline ObjectHead ** try_freeoptr(void)
{
	ObjectHead **r;

	r = (ObjectHead **) freeoptrs;
	if(r) {
		freeoptrs = *((void **) freeoptrs);
	}
	return r;
}

static inline ObjectHead ** try_alloc_new_handle(void)
{
	ObjectHead **r;
	if ((uint8 *)(otabletop-1)>=allocpos) {
		r = (ObjectHead **) --otabletop;
#ifdef VIRTUAL_FLIP
		alloctop--;
#endif	      
		allocated += sizeof(void *);
	} else {
		r = 0;
	}

	return r;
}

static inline ObjectHead ** try_get_handle(void)
{
	ObjectHead ** r;
	if ( (r  = try_freeoptr() ) == 0) {
		r  = try_alloc_new_handle();
	}
	return r;
}

static ObjectHead **gc_get_handle()
/* aux funciton, was inlined in gc_new before but moved here for readability */
{
	ObjectHead **new_handle;

	/* Obtain a handle */
	if ( (new_handle = try_get_handle() ) == 0) {
#ifndef HOST_AVR
		printf("Trying to allocate object handle\n"); 
#endif
		gc_full_cycle();
		//if (!gc_is_working()) initiate_gc_cycle();
		//conclude_gc_cycle();
		if ( (new_handle = try_get_handle() ) == 0) {
			initiate_gc_cycle();
			conclude_gc_cycle();
			if ( (new_handle = try_get_handle() == 0) ) {
#ifdef GC_PRINT_WARNINGS
				printf("gc.c: out of memory when allocating handle\n"); 
#endif
#if HAS_EXCEPTIONS == 0 && !defined HOST_RTAI_KERNEL
				EXIT_IVM;
#endif
				return 0;
			}
		}
	}
	return new_handle;
}

static inline int check_free_mem_or_do_gc(uint32 size)
/* Returns 1 if enough memory for the allocation (possibly after panic GC)
 * or 0 if out of memory
 */
{
	if (allocpos+size>(uint8*)alloctop) {
#ifdef GC_IS_INCREMENTAL
#ifdef GC_PRINT_WARNINGS
#ifndef HOST_AVR
		printf("gc.c: *** performing full GC cycle ***\n");
#endif
#endif
		if (!gc_working)
			initiate_gc_cycle();
	} 
	if (allocpos+size>(uint8*)alloctop) {
		conclude_gc_cycle();
#endif //GC_IS_INCREMENTAL
		if (allocpos+size>(uint8*)alloctop) {
			initiate_gc_cycle();
		}
		if (allocpos+size>(uint8*)alloctop) {
#ifdef GC_IS_INCREMENTAL
#ifdef GC_PRINT_WARNINGS
#ifndef HOST_AVR
			printf("gc.c: *** performing another full GC cycle ***\n");
#endif
#endif
#endif
			conclude_gc_cycle();
			if (allocpos+size>(uint8*)alloctop) {
#ifdef GC_PRINT_WARNINGS
				 printf("gc.c: out of memory when allocating object\n"); 
#endif
#if HAS_EXCEPTIONS == 0 && !defined HOST_RTAI_KERNEL
				EXIT_IVM;
#endif
				return 0;
			}
		}
	}

	return 1;
}

void *gc_new(uint32 size,void *tmplate) 
{
	ObjectHead *new_object;
	ObjectHead **new_handle;
#ifdef GC_ONLY_TIME
#ifdef GC_CYCLE_TIME_ADAPTIVE  
	gc_cycle_time_adapt();
#endif  
#endif  

#ifdef GC_PRINT_WARNINGS
	if (size == 0) {
#ifndef HOST_AVR
		printf(" Trying to allocate object with size==0!\n");
		printf("tmplate == %x\n", tmplate);
#endif
	}
#endif

	new_handle = gc_get_handle();
#if HAS_EXCEPTIONS > 0
	// HERE BE DRAGONS! we return null so that the layer above
	// can throw an OutOfMemoryError.
	// If exceptions are off, gc_get_handle will do EXIT_IVM
	// instead of returning null.
	if(new_handle == 0) return 0;
#endif
#ifdef DO_ALIGN
	size = roundup(size);
#endif

	allocated += size;
#ifdef GC_MEASURE_HEAPSTATE
	alloccount++;
#endif
#ifdef LPMEM
	gc_lpmem_alloc(size);
#endif

#ifdef GC_IS_INCREMENTAL
	if (!gc_is_working()) {
		noretained++;
		if(gc_start_new_cycle()) { 
			initiate_gc_cycle();
		}
	}
#ifndef GC_ONLY_TIME
	trigger_gc();
#endif
#endif
	/* Check for free mem and possibly do panic GC */
	if (check_free_mem_or_do_gc(size) == 0) {
		return 0;
	} else {
	/* Allocate the new object */

		new_object = (ObjectHead *) (void *) allocpos;
		*new_handle = new_object;
		allocpos += size;

		memset(new_object,0,size);
		new_object->gc.mark = gc_working ? (void *) -1 : 0;
		new_object->gc.fp = (void *) new_handle;
		new_object->tmplate = (GC___REF(LayoutHead)) tmplate;
#ifdef LOGGING
		log_malloc();
#endif
		return new_handle;
	}
}

void fastDealloc(GC___REF(void) objectStart, int objectSize) 
{
/* HERE BE DRAGONS! this should be put back in... */
  if (((GC___CPTR(uint8, objectStart) + objectSize) == allocpos)
      && ((uint32) (GC___CPTR(ObjectHead, objectStart)->gc.mark) == 0xffffffff)) {
    /* NOTE: It is very unlikey that white objects could be popped. */
    /*    || GC___CPTR(ObjectHead, objectStart)->gc.mark == 0 ) */
    allocpos = GC___CPTR(uint8, objectStart);
    *(((void **) (((ObjectHead*)(void *)allocpos)->gc.fp))) = freeoptrs;
    freeoptrs = ((ObjectHead*)(void *)allocpos)->gc.fp;
  }
}

unsigned int gc_freemem(void)
{
  /* return (gc_heap + sizeof(gc_heap)) - allocpos; */
        return ((uint8*)alloctop) - allocpos;
 // return ((uint8*)otabletop) - allocpos;
}


/************ HACK... temporary implementations... */

inline int gc_is_working(void) {return gc_working;}
inline int gc_workcnt(void) {return workcnt;}
inline float gc_workfctr(void) {return workfctr;}
inline unsigned int gc_allocated(void) {return allocated;}
inline int gc_heapsize(void) {return ((uint8*)otabletop)-gc_heap;}

void gc_finalize_all(void)
/* HERE BE DRAGONS... scan heap and run all finalizers */
{
	ObjectHead *fobj = (ObjectHead *) (void *) gc_heap;
#ifdef GC_PRINT_WARNINGS
	printf("****Running finalizers for ALL objects on the heap...\n");
#endif
	while ((uint8 *)fobj<allocpos) {
		uint32 osize;

		osize = getsize(fobj);
#ifdef DO_ALIGN
		osize = roundup(osize);
#endif

#ifdef GC_HAS_FINALIZERS
		if(GC___PTR(fobj->tmplate)->finalizer) {
#ifdef GC_PRINT_DEBUG_FINALIZERS
			printf("*** finalizing object\n");
#endif
			GC___PTR(fobj->tmplate)->finalizer(fobj); 
#ifdef GC_PRINT_DEBUG_FINALIZERS
			printf("done\n");
#endif
			// Better destroy monitor after java finalizer
			// has been run (in case it is synchronized)
		        ljrt_monitor_free(fobj);
		}
#endif
		fobj = (ObjectHead *) (void *) (((uint8 *) fobj)+osize);
		workcnt += osize;
	}
}

