/*
  All roots for a thread are kept in a singly linked list. Each of these
  lists are kept in a doubly linked list. All list elements are described 
  by the structure:

    struct gc_root {
      ObjectHead **ref;
      struct gc_root *next_root;
      struct gc_root *top_root;
      struct gc_root *next_thread;
      struct gc_root *prev_thread;
    } gc_root;

  The fields 'top_root', 'next_thread' and 'previous_thread' only 
  exists in the thread elements (indicated by next_root == 0). The
  ref field is currently unused in the thread elements (and could be 
  overlaid with the top_root field). 
 
  A non-incremental traversal of this structure can be written as: 
    
    gc_mark_root = gc_roots.next_thread->top_root;
    while (gc_mark_root != 0) {
      next_root = gc_mark_root->next_root;
      if (next == 0) {
        // No more roots in this thread
        next = gc_mark_root->next_thread->top_root;
      } else {
        // mark the root
      }
      gc_mark_root = next_root;
    }

                               +--------+        +--------+
                               |        |        |        |            
                               v        |        v        |
                            +------+    |     +------+    |
                            | ref  |    |     | ref  |    |
                            +------+    |     +------+    |
                            | root |-+  |     | root |-+  |
                            +------+ |  |     +------+ |  |       
                               +-----+  |        +-----+  |
                               |        |        |        |            
                               v        |        v        |
                            +------+    |     +------+    | 
                            | ref  |    |     | ref  |    |
                            +------+    |     +------+    |
                            | root |-+  |     | root |-+  |
                            +------+ |  |     +------+ |  |
                               +-----+  |        +-----+  |       +--------+
                               |        |        |        |       |        |
                               v        |        v        |       v        |
           +------+         +------+    |     +------+    |    +------+    |  
           | ref  |->0      | ref  |    |     | ref  |    |    | ref  |    | 
           +------+         +------+    |     +------+    |    +------+    | 
           | root |->0      | root |->0 |     | root |->0 |    | root |->0 | 
           +------+         +------+    |     +------+    |    +------+    | 
           | top  |->0      | top  |----+     | top  |----+    | top  |----+  
           +------+         +------+          +------+         +------+     
       +-->| next |-------->| next |--------->| next |-  ... ->| next |----+
       |   +------+         +------+          +------+         +------+    |
       | +-| prev |<--------| prev |<---------| prev |<- ...  -| prev |<-+ |
       | | +------+         +------+          +------+         +------+  | |
       | | gc_roots         gc_static         thread_1   ...   thread_N  | |
       | |                                                               | |
       | |                                                               | |
       | |                                                               | |
       | +---------------------------------------------------------------+ |
       |                                                                   |
       +-------------------------------------------------------------------+

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gc_roots.h"

gc_root *gc_thread_roots = 0;
gc_root *gc_static_roots = 0;
volatile gc_root *gc_mark_root;
gc_root gc_roots;
gc_root gc_static;
static void (*p_gc_mark_object)(GC___REF(ObjectHead));

void gc_roots_fix_thread_root_stack(gc_root *newStackTop)
/*  If we are in the mark_roots phase and the root we are about to
 *  mark is popped from the root stack, we must advance the 
 *  gc_mark_root pointer to the next live root 
 *
 *  newStackTop should point to the stack top *after* the root(s) have
 *  been popped.
 *
 *  A typical use of this function is to call it after popping many roots
 *  by restoring a previously saved stack state:
 *
 *      gc_thread_get_root()->top_root = saved_top_root;
 *      gc _roots_fix_thread_root_stack(saved_top_root);
 */
{
#ifdef GC_IS_INCREMENTAL
/* Moved this test to the FIX_ROOTS macro, in order to avoid unneccessary
 * function calls 
 */
//	if( gc_marking_roots() ) {	
#ifdef GC_PRINT_DEBUG_FIX
		printf("gc_fix_thread_root_stack:\n");
#endif
		gc_root *foo;
		int found = 0;
		foo = gc_mark_root;
// HERE BE DRAGONS!
// is the null-check for foo really necessary in "production" version?
// I.e., can gc_mark_root be NULL?
// If not, move the check to #IFDEF GC_PRINT_WARNINGS or DEBUG
		while(!found && foo && foo->next_root) {
			found = (foo == newStackTop);	
			foo = foo->next_root;
		}
		if(found) {
#ifdef GC_PRINT_DEBUG
			printf("fixing gc_mark_root. new stack top = %x, ref = %x, next_root = %x \n",newStackTop,(newStackTop->ref), newStackTop->next_root);
#endif
			gc_mark_root = newStackTop;
		} 
#ifdef GC_PRINT_DEBUG_FIX
		else {
			printf("                        : nothing to do\n");
		}
#endif
//	}	
#endif // GC_IS_INCREMENTAL
}

int gc_roots_mark_one(void) 
/* returns 1 if it marked a root and 0 if no more roots to mark */
{
//	gc_roots_debug_print_stack();
#ifdef GC_PRINT_DEBUG_MARK_ONE
	printf("gc_roots_mark_one\n");
#endif
	if (gc_mark_root) {
		gc_root *next_root = gc_mark_root->next_root;
		if (next_root == 0) {
			 /*
			 printf("%8x %8x %8x %8x\n", gc_mark_root, 
			 gc_mark_root->next_thread->top_root,
			 gc_mark_root->next_thread,
			 &gc_roots);
			 */
		  next_root = gc_mark_root->next_thread->top_root;
		} else {
#ifdef GC_PRINT_DEBUG_MARK_ONE
			printf("gc_roots_mark_one 4 (%x)\n",gc_mark_root);
			printf("gc_roots_mark_one 41(%x)\n",gc_mark_root->ref);
#endif
			(*p_gc_mark_object)(&gc_mark_root->ref);
#ifdef GC_PRINT_DEBUG_MARK_ONE
			printf("gc_roots_mark_one 5\n");
#endif
		}
#ifdef GC_PRINT_DEBUG		
		if((((unsigned int)next_root) & 3) != 0) {
			printf("********** next_root = %x\n", next_root);
		}
#endif		
		gc_mark_root = next_root;
		return 1;	
	} else {
		return 0;
	}
}

void gc_roots_initiate_gc_cycle(void) 
{
  gc_mark_root = gc_roots.next_thread->top_root;
}

void gc_roots_insert_thread_root(gc_root *thread_roots)
{
  // Setup and insert thread_roots in list...
  thread_roots->ref = 0;
  thread_roots->next_root = 0;
  thread_roots->top_root = thread_roots;
  thread_roots->next_thread = &gc_roots;
  thread_roots->prev_thread = gc_roots.prev_thread;
  gc_roots.prev_thread->next_thread = thread_roots;
  gc_roots.prev_thread = thread_roots;

#ifndef GC_HAS_THREADS
  gc_thread_roots = thread_roots; // @@@@ FIX this
#endif  
}

void gc_roots_remove_thread_root(gc_root *thread_root)
{
  thread_root->next_thread->prev_thread = thread_root->prev_thread;
  thread_root->prev_thread->next_thread = thread_root->next_thread;
}

void gc_roots_gc_init(void (*p_mark)(GC___REF(ObjectHead) )) 
{
  static int called = 0;
  if (!called) {
    called = 1;

    gc_roots.ref = 0;
    gc_roots.next_root = 0;
    gc_roots.top_root = 0;
    gc_roots.next_thread = &gc_static;
    gc_roots.prev_thread = &gc_static;
    
    gc_static.ref = 0;
    gc_static.next_root = 0;
    gc_static.top_root = &gc_static;
    gc_static.next_thread = &gc_roots;
    gc_static.prev_thread = &gc_roots;
    
    gc_static_roots = &gc_static;
  }
  p_gc_mark_object = p_mark;
}

void gc_roots_debug_print_stack(void)
{
#ifdef GC_PRINT_DEBUG
    printf("\n***********ROOT STACK***********\n");
    {
	    gc_root *gc_mark_root = gc_roots.next_thread->top_root;
	    while (gc_mark_root != 0) {
	      gc_root *next_root = gc_mark_root->next_root;
	      if (next_root == 0) {
		// No more roots in this thread
		printf("\n");
		next_root = gc_mark_root->next_thread;
		printf("\n:");
		next_root = next_root->top_root;
		printf("\n:");
	      } else {
		// mark the root
		printf("[ 0x%x ]", next_root);
	      }
	      gc_mark_root = next_root;
	    }
    }	
    printf("\n***********  END     ***********\n");
#endif
}
