#ifndef GC_ROOTS
#define GC_ROOTS

/*
  All roots for a thread are kept in a singly linked list. Each of these
  lists are kept in a doubly linked list. All list elements are described 
  by the structure:

    struct gc_root {
      ObjectHead **ref;
      struct gc_root *next_root;
      struct gc_root *top_root;
      struct gc_root *next_thread;
      struct gc_root *prev_thread;
    } gc_root;

  The fields 'top_root', 'next_thread' and 'previous_thread' only 
  exists in the thread elements (indicated by next_root == 0). The
  ref field is currently unused in the thread elements (and could be 
  overlaid with the top_root field). 
 
  A non-incremental traversal of this structure can be written as: 
    
    gc_mark_root = gc_roots.next_thread->top_root;
    while (gc_mark_root != 0) {
      next_root = gc_mark_root->next_root;
      if (next == 0) {
        // No more roots in this thread
        next = gc_mark_root->next_thread->top_root;
      } else {
        // mark the root
      }
      gc_mark_root = next_root;
    }

                               +--------+        +--------+
                               |        |        |        |            
                               v        |        v        |
                            +------+    |     +------+    |
                            | ref  |    |     | ref  |    |
                            +------+    |     +------+    |
                            | root |-+  |     | root |-+  |
                            +------+ |  |     +------+ |  |       
                               +-----+  |        +-----+  |
                               |        |        |        |            
                               v        |        v        |
                            +------+    |     +------+    | 
                            | ref  |    |     | ref  |    |
                            +------+    |     +------+    |
                            | root |-+  |     | root |-+  |
                            +------+ |  |     +------+ |  |
                               +-----+  |        +-----+  |       +--------+
                               |        |        |        |       |        |
                               v        |        v        |       v        |
           +------+         +------+    |     +------+    |    +------+    |  
           | ref  |->0      | ref  |    |     | ref  |    |    | ref  |    | 
           +------+         +------+    |     +------+    |    +------+    | 
           | root |->0      | root |->0 |     | root |->0 |    | root |->0 | 
           +------+         +------+    |     +------+    |    +------+    | 
           | top  |->0      | top  |----+     | top  |----+    | top  |----+  
           +------+         +------+          +------+         +------+     
       +-->| next |-------->| next |--------->| next |-  ... ->| next |----+
       |   +------+         +------+          +------+         +------+    |
       | +-| prev |<--------| prev |<---------| prev |<- ...  -| prev |<-+ |
       | | +------+         +------+          +------+         +------+  | |
       | | gc_roots         gc_static         thread_1   ...   thread_N  | |
       | |                                                               | |
       | |                                                               | |
       | |                                                               | |
       | +---------------------------------------------------------------+ |
       |                                                                   |
       +-------------------------------------------------------------------+

*/
#ifdef HAVE_CONFIG_H 
#include <config.h>                      
#endif
#include "gc.h"

/*
typedef struct gc_root {
	ObjectHead **ref;
	struct gc_root *next_root;
	struct gc_root *top_root;
	struct gc_root *next_thread;
	struct gc_root *prev_thread;
} gc_root;
*/

int gc_roots_mark_one(void);
/* returns 1 if it marked a root and 0 if no more roots to mark */

int gc_marking_roots(void);

void gc_roots_initiate_gc_cycle(void);
void gc_roots_insert_thread_root(gc_root *thread_roots);
void gc_roots_remove_thread_root(gc_root *thread_roots);
void gc_roots_gc_init(void (*mark_object)(GC___REF(ObjectHead)));

#ifdef GC_PROFILING
void save_root_dummy(void);
#define GC_PROFILING_SAVE_ROOT save_root_dummy();
#else
#define GC_PROFILING_SAVE_ROOT
#endif

#ifndef GC_LAZY_GET_THREAD_ROOT
// HERE BE DRAGONS!
#define SAVE_ROOT_STACK(context)  \
        gc_root *__saved_root_##context; \
	GC_PROFILING_SAVE_ROOT \
	gc_lock(); \
        __saved_root_##context = thread_root->top_root; \
	gc_unlock()

#define RESTORE_ROOT_STACK(context) \
	gc_lock();\
        thread_root->top_root = __saved_root_##context; \
	FIX_ROOT_STACK(__saved_root_##context);\
	gc_unlock()
#else
#define SAVE_ROOT_STACK_LAZY(context) \
	if(!__saved_root_##context) { \
		GC_PROFILING_SAVE_ROOT; \
		GC__GET_THREAD_ROOT_LAZY; \
        	__saved_root_##context = thread_root->top_root; \
	}
#define SAVE_ROOT_STACK(context)  \
        gc_root *__saved_root_##context = 0

//#define SAVE_ROOT_STACK(context)  
//	 int __push_count = 0;

#define DO_RESTORE_ROOT_STACK(context) \
do { \
	gc_lock(); \
        thread_root->top_root = __saved_root_##context; \
	FIX_ROOT_STACK(__saved_root_##context);\
	GC__DO_GET_THREAD_ROOT;\
	gc_unlock();\
} while(0)

#define RESTORE_ROOT_STACK(context) \
	if(__saved_root_##context) {DO_RESTORE_ROOT_STACK(context);} \

//#define RESTORE_ROOT_STACK(context) 
//  while(__push_count-- > 0) GC__IMPL__DO_POP_ROOT(0,0);
#endif

#ifdef GC_IS_INCREMENTAL
#define FIX_ROOT_STACK(new_top) \
        if(gc_marking_roots()) gc_roots_fix_thread_root_stack(new_top) 
#else
#define FIX_ROOT_STACK(new_top) 
#endif

void gc_roots_fix_thread_root_stack(gc_root *restoredStackTop);
#endif
