#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "mem.h"
#include "gc-ms.h"
#if !defined HOST_AVR && !defined HOST_RTAI_KERNEL
#include <stdio.h>
#endif

#ifndef NULL
#define NULL 0
#endif

#define false 0
#define true 1
 
// Simple coalescing allocator

// The free list is stored on a doubly linked list, each member of
// which is stored in the body of the free memory.  The head of the
// list has the same structure but its size field is zero.  This
// resides in the memory pool structure.  Always having at least one
// item on the list simplifies the alloc and free code.

struct block theBlock;

void do_memset(void *s, int c, size_t n) {
#ifdef DEBUG_ALLOC_MEMSET
	printf("do_memset %x, %d\n",s, n);
#endif
	memset(s,c,n);
//	printf("done\n");
}

pointer roundup_obj( unsigned int size )
{
#ifdef ALLOC_COALESCE
    size += sizeof(struct memtail);
#endif
    size = (size + theBlock.alignment - 1) & -theBlock.alignment;
    return size;
}

pointer roundup( pointer size )
{
    //cprintf("roundup: objsize = %x\n", size);
    size += sizeof(struct memdq);
#ifdef ALLOC_COALESCE
    size += sizeof(struct memtail);
#endif
	pointer p1 = size + theBlock.alignment - 1;
	pointer p2 = -((pointer) theBlock.alignment);
	size = p1 & p2;
    //size = (size + theBlock.alignment - 1) &   -theBlock.alignment;
    //cprintf("roundup: size = %x\n", size);
    return size;
}

void mempool_create(void *base, unsigned int size)
{
    int align = 4;
    struct memdq *hdq;
    struct memdq *dq;

    theBlock.obase=base;
    theBlock.osize=size;
    theBlock.oalign=align;

    theBlock.allocCount = 0;

    theBlock.alignment = align;
#ifdef ALLOC_COALESCE
    while(theBlock.alignment <= (unsigned int)sizeof(struct memdq)+(unsigned int) sizeof(struct memtail))
#ifdef POWER_OF_TWO_ALIGN
        theBlock.alignment += theBlock.alignment;
#else
        theBlock.alignment += HEAP_BLOCK_ALIGN_STEP;
#endif
#else
    while(theBlock.alignment < (unsigned int)sizeof(struct memdq))
#ifdef POWER_OF_TWO_ALIGN
        theBlock.alignment += theBlock.alignment;
#else
        theBlock.alignment += HEAP_BLOCK_ALIGN_STEP;
#endif
#endif

#ifdef GC_PRINT_DEBUG
#ifndef HOST_AVR
printf("alignment = %d\n", theBlock.alignment);
#endif
#endif
    theBlock.bottom = (void *)roundup((pointer)base);
    theBlock.top = (void *)((pointer)(base+size) & -(pointer)theBlock.alignment);
#ifdef GC_PRINT_DEBUG
#ifndef HOST_AVR
printf("bottom = %d\n", theBlock.bottom);
#endif
#endif
/*
    CYG_ASSERT( top > bottom , "heap too small" );
    CYG_ASSERT( ((cyg_int32)bottom & alignment-1)==0, "bottom badly aligned" );
    CYG_ASSERT( ((cyg_int32)top & alignment-1)==0, "top badly aligned" );
*/

    hdq = &(theBlock.head); 
    dq = (struct memdq *)theBlock.bottom;
    
    hdq->prev = hdq->next = dq;
    hdq->size = 0;
    dq->prev = dq->next = hdq;

#ifndef ALLOC_COALESCE
    theBlock.freemem = dq->size = theBlock.top-theBlock.bottom;
#else
    theBlock.freemem = dq->size = theBlock.top-theBlock.bottom - sizeof(struct memtail);
    ((struct memtail *) ( (pointer) theBlock.top - sizeof(struct memtail) ))->size = dq->size | 0x01;
#endif
    theBlock.totalmem = theBlock.freemem;
}

void debug_print_freelist(void)
{
#ifndef HOST_AVR
	struct memdq *tdq = &theBlock.head;
	printf("\nfreelist: ");
	do {
#ifdef FREELIST_PRINT_NEXT
		printf("(%x:%d [%x])", tdq,tdq->size, tdq->next);
#else
		printf("(%x:%d)", tdq,tdq->size);
#endif
		//printf("(%d)", tdq->size);
		if(tdq->next == MAGIC_ALLOCED) {
			printf("[***ERROR! allocated block!!!***]\n");
			return;
		 }
		tdq = tdq->next;
	} while(tdq->size > 0);
	printf("\n");
#endif
}

// allocation is simple
// First we look down the free list for a large enough block
// If we find a block the right size, we unlink the block from
//    the free list and return a pointer to it.
// If we find a larger block, we chop a piece off the end
//    and return that
// Otherwise we will eventually get back to the head of the list
//    and return NULL
void *alloc_obj( unsigned int size )
{
// This should be the same as alloc -- the only difference is that it should
// return the pointer to the memdq instead of to the data. For now, we return
// 1 to break any program that uses this function (shouldn't be any)...

	printf("mem.c: WARNING! alloc_obj is not implemented\n");
	return 1;
}
void *alloc( unsigned int objsize )
{
	struct memdq *dq = &theBlock.head;
	void *alloced, *result;
	unsigned int size;

	size = roundup(objsize);

	do {
#ifdef SANITY_CHECKS
		/* CYG_ASSERT( dq->next->prev==dq, "Bad link in dq"); */
		if(dq->next->prev != dq) printf("mem.c: Warning! bad link in freelist (dq = %x)\n",dq);
#endif
		//printf("mem.c: alloc: searching freelist, dq = %x, next = %x\n", dq, dq->next);
		dq = dq->next;
		if(0 == dq->size) {
#ifdef SANITY_CHECKS
			/* CYG_ASSERT(dq == &head, "bad free block"); */
			if(dq != &theBlock.head) printf("mem.c: WARNING! bad free block (dq = %x)!\n", dq);
#endif
			return NULL;
		}
	} while(dq->size < size);
	//printf("mem.c: alloc: found large enough block, dq = %x, next = %x\n", dq, dq->next);

#ifdef SANITY_CHECKS
	if(dq->next == MAGIC_ALLOCED) {
		printf("mem.c: ERROR! Allocated block in freelist???\n");
		exit(1);
	}
#endif

#ifdef ALLOC_COALESCE
	if( size == dq->size || dq->size - size <= sizeof(struct memdq) + sizeof(struct memtail)) 
#else
	if( size == dq->size || dq->size - size <= sizeof(struct memdq)) 
#endif
	{
	// exact fit -- unlink from free list
		//printf("   exact fit\n");
		// HERE BE DRAGONS! is this correct?
		size = dq->size;
		dq->prev->next = dq->next;
		dq->next->prev = dq->prev;
		alloced = (void *)dq;
	} else {

		/* CYG_ASSERT( dq->size > size, "block found is too small"); */

		// allocate portion of memory from end of block
        
		dq->size -=size;
#ifdef ALLOC_COALESCE
		((struct memtail *)
		((pointer)dq+dq->size-sizeof(struct memtail)))->size=dq->size|0x01;
#endif

		// The portion left over has to be large enough to store a
		// struct memdq.  This is guaranteed because the alignment is
		// larger than the size of this structure.

		/* CYG_ASSERT( (cyg_int32)sizeof(struct memdq)<=dq->size ,
			"not enough space for list item" ); */

		alloced = (void *)dq + dq->size;

		// Set size on allocated block
		((memdq *)alloced)->size = size;
	}

	/* CYG_ASSERT( bottom<=alloced && alloced<=top, "alloced outside pool" ); */


	dq = (struct memdq *)alloced;
	dq->next = dq->prev = (struct memdq *) MAGIC_ALLOCED; // magic number

#ifdef ALLOC_COALESCE
	((struct memtail *)((pointer)dq+size-sizeof(struct memtail)))->size=size;
#endif
	//printf("ALLOC: size=%d (%d -- %d). Block at %x ends at %x\n", size, dq->size, ((struct memtail *)((int)dq+size-sizeof(struct memtail)))->size, dq,(int)dq+dq->size);
	theBlock.freemem -=size;
	theBlock.allocCount++;
	result = (void *) ((pointer)alloced + sizeof(struct memdq));
#ifdef INIT_ON_ALLOC
	do_memset(result, 0, objsize);
#endif
	//debug_print_freelist();
	return result;
}

void freelist_unlink(struct memdq* dq)
{
//	printf("freelist_unlink %x\n", dq);
	dq->prev->next = dq->next;
	dq->next->prev = dq->prev;
	dq->next = MAGIC_ALLOCED;
	dq->prev = MAGIC_ALLOCED;
	theBlock.freemem -= dq->size;
}

// If no coalescing is done, free is simply a matter of using the
// freed memory as an element of the free list linking it in at the
// start.
    
int hfree(void *p, unsigned int size )
{
  hfree_obj(p-sizeof(struct memdq), size); 
}

void * hfree_obj(struct memdq *dq, unsigned int size )
{
	struct memdq *hdq;
	// struct memdq *dq;

#ifdef ALLOC_COALESCE
	struct memdq *codq, *nextdq;
	struct memtail *prevtail;
	int cosize;
	char co_prev = 0;
	char co_next = 0;
#endif
#ifdef INIT_ON_FREE
	int *tmp, tmpsize, j;
#endif
//	printf("mem.c: hfree_obj: dq = %x, size = %d, dq->size = %d\n", dq, size,dq->size);

	if(!(theBlock.bottom<=(void *)dq&&(void *)dq<=theBlock.top)){
#if !(defined HOST_AVR || HOST_RTAI_KERNEL)
		fprintf(stderr, "hfree: pointer outside block!\n");
#endif
		return false;
    	}
	hdq=&theBlock.head;
	// dq=(struct memdq *)(p-sizeof(struct memdq));

	// check magic number in block to be freed
	if(dq->next != dq->prev || dq->next != (struct memdq *)MAGIC_ALLOCED) {
#if !(defined HOST_AVR || HOST_RTAI_KERNEL)
		fprintf(stderr, "hfree: trying to free free object!\n");
#endif
		return false;
	}

	if(0==size) {
		size = dq->size;
	} 
// HERE BE DRAGONS! change the ifdef below when CONTIGOUS_HACK (w/ better
// name, is in config.h!
#if 0
//#ifndef GC_FREE_CONTIGOUS_HACK
        else {
		size = roundup(size);
	}
printf("mem.c: hfree_obj: using size %d\n", size);

	if(dq->size != size) {
		printf("WARNING! mem.c: hfree_obj: dq->size != size!\n");
		return false;
	}
#endif

	/* CYG_ASSERT( (cyg_int32)sizeof(struct memdq)<=size ,
		"not enough space for list item" ); */

#ifdef ALLOC_COALESCE
//	printf("ALLOC_COALESCE: size=%d\n", size);
	cosize = size;
	codq = dq;
	nextdq = (struct memdq *) ( ((pointer) dq) + size);
	if(nextdq <= (struct memdq*) theBlock.top && 
           nextdq->next != (struct memdq *)MAGIC_ALLOCED && 
           nextdq->prev != (struct memdq *)MAGIC_ALLOCED && 
           nextdq->size > 0) {
		cosize += nextdq->size;
		nextdq->prev->next = nextdq->next;
		nextdq->next->prev = nextdq->prev;

		co_next = 1;
#ifdef DEBUG_ALLOC_COALESCE
		printf("ALLOC_COALESCE next:\n");
#endif

// HERE BE DRAGONS! 
// When coalescing with the next block, update this block's size
// to avoid having the garbage collecter sweeping nextdq.
// (assuming that the sweeper does sweep = sweep + sweep->size...)

//	*** free returns the pointer to the possibly coalesced block,
//	so the GC should do sweep = free(sweep).
//		dq->size+=nextdq->size;
	}
	
	prevtail = (struct memtail*) ( ((pointer) dq) - sizeof(struct memtail));

	if(prevtail >= (struct memtail*) theBlock.bottom && 
           prevtail->size & 0x01) {
		int prevsize = prevtail->size & ~1;
		struct memdq *prevdq = (struct memdq *) ((pointer) codq - prevsize);
 #ifdef SANITY_CHECKS
		if(prevdq->next != (struct memdq *)MAGIC_ALLOCED && 
		   prevdq->prev != (struct memdq *)MAGIC_ALLOCED) {
 #endif
			cosize += prevsize;
			codq = (struct memdq *) ((pointer) dq - prevsize);
			prevdq->prev->next = prevdq->next;
			prevdq->next->prev = prevdq->prev;
#ifdef DEBUG_ALLOC_COALESCE
			printf("ALLOC_COALESCE prev\n");
#endif
			co_prev = 1;
#ifdef SANITY_CHECKS
		} else {
			fprintf(stdout,"WARNING! block with odd tail size not free! prevdq = %x, prevsize = %d\n", prevdq, prevsize);
		}
 #endif
	}

    ( (struct memtail *) ((pointer)codq+cosize-sizeof(struct memtail) ))->size = cosize | 0x01;

//    dq = codq;
#endif

#ifdef INIT_ON_FREE
#ifdef ALLOC_COALESCE
    		// naive version: always init whole object
#if 0
		tmpsize = cosize-sizeof(struct memdq)-sizeof(struct memtail);
#else
		if(co_prev && co_next) {
#ifdef DEBUG_ALLOC_COALESCE
			printf("prev+next\n");
#endif
			// init prevtail + this incl head and tail + nexthead	
			tmp = (int *) ((pointer) dq - sizeof(struct memtail));
			tmpsize = size + sizeof(struct memtail) + sizeof(struct memdq);
		} else if(co_prev) {
#ifdef DEBUG_ALLOC_COALESCE
			printf("prev\n");
#endif
			// init prevtail + this incl head
			tmp = (int *) ((pointer) dq - sizeof(struct memtail));
			tmpsize = size;// + sizeof(memtail)-sizeof(memtail);
		} else if(co_next) {
#ifdef DEBUG_ALLOC_COALESCE
			printf("next\n");
#endif
			// init nexthead + this incl tail
			tmp = (int *) ((pointer) dq + sizeof(struct memdq));
			tmpsize = size;// + sizeof(memtail)-sizeof(memtail);
		} else {
#ifdef DEBUG_ALLOC_COALESCE
			printf("this\n");
#endif
			// init this excl head and tail
			tmp = (int *) ((pointer) dq + sizeof(struct memdq));
			tmpsize = size-sizeof(struct memdq)-sizeof(struct memtail);
		}
		//tmpsize = size-sizeof(struct memdq)-sizeof(struct memtail);
#endif
#else
                tmp = (int *) ((pointer)dq + sizeof(struct memdq));
		tmpsize = size-sizeof(struct memdq);
#endif
                /* initialise object */ 
//		printf("dq = %x, size = %d\n", dq, size);
                // size is in bytes, but memory is (at least) word
                // aligned, so we can write whole ints instead.
//                for(j=0; j<tmpsize/sizeof(int); j++) {
//                        *tmp = 0;
//                        tmp++;
//                }
// better to use do_memset... 
		do_memset(tmp, 0, tmpsize);
#endif

#ifdef ALLOC_COALESCE
    dq = codq;
#endif

    dq->prev = hdq;
    dq->next = hdq->next;
#ifdef ALLOC_COALESCE
    dq->size = cosize;
#else
    dq->size = size;
#endif
    hdq->next = dq;
    dq->next->prev=dq;

    theBlock.freemem +=size;
    theBlock.allocCount--;
    //debug_print_freelist();
//    return true;
    return dq;
}    

unsigned int get_freemem(void) 
{
	return theBlock.freemem;
}
