/*
 * fmtio.h
 *
 * Formatted IO; Provide Console printf in STORK (and other) targets.
 */

#ifndef _FMTIO_H_
#define _FMTIO_H_

/*
 * The same as printf, but returns a negative value in case of internal
 * buffer overflow. The absolute value of the returned value is the
 * number of characters printed on the console.
 */
int cprintf(const char *, ...);


/* 
 * Floating point with precition: 
 */
#define cprintff(expr,prec) cprintf(#expr " = %*g\n", expr, prec)

/* 
 * General floating point (default %g precition): 
 */
#define cprintfg(expr) cprintf(#expr " = %g\n", expr)

/* 
 * Integer output: 
 */
#define cprintfi(expr) cprintf(#expr " = %i\n", expr)

/* 
 * Unsigned integer: 
 */
#define cprintfu(expr) cprintf(#expr " = %u\n", expr)

/* 
 * Hex (%X) output of expression: 
 */
#define cprintfx(expr) cprintf(#expr " = 0x%X\n", expr)

/* 
 * Pointer value, in hex: 
 */
#define cprintfp(expr) cprintf(#expr " = %p\n", expr)


/* Line-break settings for array display below: */
#ifndef FMTIO_HEXCOLS
#define FMTIO_HEXCOLS 16
#endif
#ifndef FMTIO_FLTCOLS
#define FMTIO_FLTCOLS 6
#endif

/* 
 * Warning: Printing arrays is problematic since for an array in C, 
 * the size is not known within the called function. Hence, you need
 * to put an array in a struct to make it dumpable/printable using 
 * the following macros:
 */

/* 
 * Memory dump of data area, in hex: 
 */
#define cprintfm(data) { int i; \
  cprintf(#data "[0..0x%X] = ", sizeof(data)-1); \
  for (i=0; i<sizeof(data)/sizeof(long*); i++) { \
    if (!(i%FMTIO_HEXCOLS)) cprintf("\n"); \
    cprintf("%X ", &((long*)&(data))[i]); \
  } \
  cprintf("\n"); \
}

/* 
 * Single precision (float) array dump displayed in Matlab format: 
 */
#define cprintfs(data) { int i; \
  cprintf(#data "(1:%u) = [", sizeof(data)/sizeof(float)); \
  for (i=0; i<sizeof(data)/sizeof(float); i++) { \
    if (!(i%FMTIO_FLTCOLS)) cprintf(" ...\n"); \
    cprintf("%g ", ((float*)&(data))[i]); \
  } \
  cprintf("];\n"); \
}

/* 
 * Double precision (double) array dump displayed in Matlab format: 
 */
#define cprintfd(data) { int i; \
  cprintf(#data "(1:%u) = [", sizeof(data)/sizeof(double)); \
  for (i=0; i<sizeof(data)/sizeof(double); i++) { \
    if (!(i%FMTIO_FLTCOLS)) cprintf(" ...\n"); \
    cprintf("%g ", ((double*)&(data))[i]); \
  } \
  cprintf("];\n"); \
}

/* 
 * Single precision (float) matrix dump displayed in Matlab format: 
 */
#define cprintfs2(data,ncols) { int i; \
  cprintf(#data "(1:%u,1:%u) = [", sizeof(data)/sizeof(float)/(ncols), ncols);\
  for (i=0; i<sizeof(data)/sizeof(float); i++) { \
    if (!(i%(ncols))) if (i) cprintf(" ; ...\n"); else cprintf(" ...\n"); \
    cprintf("%g ", ((float*)&(data))[i]); \
  } \
  cprintf("];\n"); \
}

/* 
 * Double precision (double) matrix dump displayed in Matlab format: 
 */
#define cprintfd2(data,ncols) { int i; \
  cprintf(#data "(1:%u,1:%u) = [", sizeof(data)/sizeof(double)/(ncols),ncols);\
  for (i=0; i<sizeof(data)/sizeof(double); i++) { \
    if (!(i%(ncols))) if (i) cprintf(" ; ...\n"); else cprintf(" ...\n");\
    cprintf("%g ", ((double*)&(data))[i]); \
  } \
  cprintf("];\n"); \
}


/* Only target (STORK or VxWorks) support; no hosts yet: */
// Define this when building
// #ifndef VW
///#define STORK_PPC
// #endif

/* Compatibilty with printf. */
#ifdef STORK_PPC
#define printf cprintf
#endif

#endif
/*EOF*/
