/* $Id: ActivityImpl.java,v 1.6 2005/08/19 09:37:16 d02en Exp $ 
 * 
 * Part of LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * Let a <code>java.lang.Thread</code> object be treated as an <code>Activity</code>, 
 * which it (almost) is even if that interface (in <code>se.lth.cs.realtime</code>)is 
 * not implemented (in <code>java.lang</code>). Also threads that already implements 
 * <code>Activity</code> (instances of <code>JThread</code> and <code>RTThread</code>)
 * can be wrapped, thereby ensuring that the constructor always returns a valid activity.
 * 
 * @see License
 */

public final class ActivityImpl implements Activity {

  /**
   * Encapsulated Java thread.
   */
  protected final Thread jActivity;

  /**
   * Encapsulated real-time thread.
   */
  protected final RTThread rtActivity;
  
  /**
   * Not part of the <code>Activity</code> interface, obtain the represented 
   * <code>Thread</source> or <code>RTThread</source>. The intended use of this 
   * class is only for <code>ActivityEvent</code>s and configuration of activities. 
   * For example, to support obtaining the actual thread in <code>RTEvent</code>,
   * to have the source and owner attributes referring to the actual thread objects,
   * the <code>RTEvent</code> base class will use this method. Similar need might 
   * appear outside this package, hence the protected visibility.
   *
   * @return the (real-time or Java) tread object.
   */
  public Object getThread(){
    return (jActivity != null ? (Object)jActivity : (Object)rtActivity);
  }

  /**
   * Create an activity representing the currently executing thread.
   */
  public ActivityImpl() {
    rtActivity = RTThread.currentRTThread();
    if (rtActivity == null) {
      jActivity = Thread.currentThread();
    } else {
      jActivity = null;
    }
  }

  /**
   * Make an activity out of a Java thread.
   * 
   * @param thread the Java thread to represent.
   */
  public ActivityImpl(Thread thread) {
    jActivity = thread;
    rtActivity = null;
  }

  /**
   * Create an activity object, that we already have, for completeness.
   * 
   * @param thread the RTThread to be wrapped.
   */
  public ActivityImpl(RTThread thread) {
    jActivity = null;
    rtActivity = thread;
  }

  /**
   * @see se.lth.cs.realtime.Activity#getEnv()
   */
  public Environment getEnv() {
    return null;
  }

  /**
   * @see se.lth.cs.realtime.Activity#getTimebase()
   */
  public Timebase getTimebase() {
    return null;
  }

  /**
   * Calls <code>isAlive</code> for the represented thread.
   * 
   * @see se.lth.cs.realtime.Activity#isAlive()
   * @see java.lang.Thread#isAlive()
   */
  public boolean isAlive() {
    return (jActivity != null ? jActivity.isAlive() : rtActivity.isAlive());
  }

  /**
   * Calls <code>isAlive</code> for the represented thread.
   * 
   * @see se.lth.cs.realtime.Activity#isAlive()
   * @see java.lang.Thread#isAlive()
   */
  public void join() throws InterruptedException {
    if (jActivity != null)
      jActivity.join();
    else
      rtActivity.join();
  }

  /**
   * Calls <code>isAlive</code> for the represented thread.
   * 
   * @see se.lth.cs.realtime.JThread#isAlive()
   * @see java.lang.Thread#isAlive()
   */
  public void start() {
    if (jActivity != null)
      jActivity.start();
    else
      rtActivity.start();
  }

  /**
   * Interrupt and join the represented thread.
   * 
   * @see se.lth.cs.realtime.Activity#terminate()
   * @see se.lth.cs.realtime.JThread#terminate()
   */
  public void terminate() {
    if (jActivity != null) {
      if (jActivity.isAlive() && !jActivity.isInterrupted()) {
        jActivity.interrupt();
        try {
          jActivity.join();
        }
        catch (InterruptedException exc) {
          throw new RTInterrupted(
            "Termination (join) of " + jActivity.getName() + " interrupted");
        }
      };
    }
    else {
      if (rtActivity.isAlive() && !rtActivity.isInterrupted()) {
        rtActivity.interrupt();
        try {
          rtActivity.join();
        }
        catch (InterruptedException exc) {
          throw new RTInterrupted(
            "Termination (join) of " + rtActivity.getName() + " interrupted");
        }
      };
    }
  }
}
