/* $Id: RTThread.java,v 1.8 2005/08/19 09:37:16 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

import se.lth.cs.realtime.event.*;

/**
 * This class defines a thread of execution in a real-time environment.
 * <p>
 * A normal <b><code>java.lang.Thread</code> </b> is not suitable to
 * real-time problems since it does not provide primitives like
 * <code>sleepUntil()</code>, necessary to implement periodic real-time
 * processes properly, while, conversely, it provides means to asynchronously
 * change the priority of a thread which sometimes is not acceptable. Since some
 * methods of <code>java.lang.Thread</code> are declared <code>final</code>,
 * subclassing <b>java.lang.Thread </b> is not a solution.
 * <p>
 * Every real-time thread has a priority (if <code>FixedPriority</code> is
 * implemented, it can be modified only before starting the thread); each thread
 * may or may not also be marked as a daemon. When code run by some thread
 * creates a new <code>RTThread</code> object, the new thread has its priority
 * initially set equal to the priority of the creating thread, and it is set to
 * be a daemon thread if and only if the creating thread is a daemon.
 * <p>
 * Every <code>RTThread</code> belongs to a <code>RTThreadGroup</code>.
 * <p>
 * Every thread has a name for identification purposes. More than one thread may
 * have the same name. If a name is not specified when a thread is created, a
 * new name is generated for it.
 * 
 * The user may of course call <code>Thread.currentThread</code> to obtain the
 * <code>java.lang.Thread</code> which is encapsulated by the
 * <code>RTThread</code> object, and, for instance, calling
 * <code>interrupt</code> on that object will work as if the
 * <code>interrupt</code> of <code>RTThread</code> would have been called,
 * the user is supposed to retrieve the <code>RTThread</code> object to gain
 * full support for embedded applications.
 * 
 * @see License
 * @see java.lang.Thread
 * @see se.lth.cs.realtime.FixedPriority
 * @see se.lth.cs.realtime.RTThreadGroup
 */

public abstract class RTThread implements Activity, RTEventListener,
		RTEventSource {

	private volatile RTThreadGroup group;
	private volatile String name;
	private volatile int id;
	private volatile int priority;

	// Moved interruptedFlag to java.lang.Object instead so that we
	// may have the same monitor implementation for both Thread and
	// RTThread. AndersN
// 	private volatile boolean interruptedFlag;


	private volatile boolean isAlive;
	private volatile boolean started;

   /* Runnable object */
	private Runnable targetRun;
	// Identifier of this RTThread object 
	private long target;
	/* Pointer to the TSD - Oooops */
	private long TSD;

	/** The event input buffer. */
	protected volatile RTEventBuffer mailbox;

	private volatile int maxTimeLag;

	/**
	 * Default size of the event buffer, used by constructor possibly after
	 * being changed by some subclass constructor.
	 */
	protected volatile int defaultBufferSize = RTSystem.defaultEventBufferSize;

	/** Whether or not the thread is a daemon thread. */
	//private volatile boolean daemon;

	/** Private link to contained java.lang.Thread object. */
	//
// 	private WrappedThread langThread;
	// 

	/** For auto-numbering unnamed threads and determination of thread id number. */
	private static int threadInitNumber = 0;

	/*
	 * Get and increment thread number, to be called in <code>init</code>
	 * for named  and unnamed threads. The returned number should be assigned 
	 * to <code>id</code> for thread identification.
	 */
	private static synchronized int nextThreadId() {
		return threadInitNumber++;
	}

	/**
	 * The minimum priority that a thread can have.
	 */
	public final static int MIN_PRIORITY = Thread.MIN_PRIORITY;

	/**
	 * The maximum priority that a thread can have.
	 */
	public final static int MAX_PRIORITY = Thread.MAX_PRIORITY;

	/**
	 * The maximum priority that a thread can have.
	 */
	public final static int NORM_PRIORITY = Thread.NORM_PRIORITY;

	public void setMaxTimeLag(int maxResponseTime) {
		maxTimeLag = maxResponseTime;
	}

	public int getMaxTimeLag() {
		return maxTimeLag;
	}

	/**
	 * Obtain the unique <code>RTThread</code> number assigned to this thread
	 * object. When a new thread object is created
	 * 
	 * @return Returns the number of this object
	 */
	public int getRTThreadId() {
		return id;
	}

	/**
	 * Put the time-stamped event in the input buffer of this thread. If the
	 * buffer is full, the caller is blocked.
	 * 
	 * @return null, but subclasses are free to return an event if desired to
	 *         inform about blocking or results.
	 */
	public RTEvent putEvent(RTEvent ev) {
		mailbox.doPost(ev);
		return null;
	}

	/**
	 * The list of the listeners of this thread.
	 * 
	 * @see se.lth.cs.realtime.event.RTEventDispatcher
	 */
	protected RTEventDispatcher dispatcher = new RTEventDispatcher(this);

	/**
	 * Obtain the object that internally stores the subscriptions of events,
	 * often referred to as the listener list. Opposed to listener lists in GUI
	 * packages (such as AWT and Swing), where the listener typically gets its
	 * processing method called and fully performed in the context of the GUI
	 * thread, the subscribers/listeners of <code>RTEvent</code> s gets a call
	 * of <code>putEvent</code> which typically only stores/buffers the event
	 * for later processing by the receiving thread (object). Thus, the
	 * <code>EventSource</code> simply dispatches the event distribution,
	 * hence the name dispatcher of the lister list storage object.
	 */
	public RTEventDispatcher getDispatcher() {
		return dispatcher;
	}

	public void addListener(RTEventListener subscriber) {
		RTThread rtt = currentRTThread();
		if (rtt != null) {
			dispatcher.addListener(subscriber, currentRTThread().getPriority(),
					null);
		} else {
			dispatcher.addListener(subscriber, Thread.currentThread()
					.getPriority(), null);
		}
	}

	/**
	 * Returns a reference to the currently executing <code>RTThread</code>.
	 * If the currently executing thread is not an <code>RTThread</code>,
	 * <code>null</code> is returned.
	 * 
	 * @return the currently executing real-time thread.
	 */
// 	public static RTThread currentRTThread() {
	
// 		Object self = Thread.currentThread();
// 		//return self.getRTThread();
// 		//
	
// 		if (self instanceof WrappedThread) {
// 			return ((WrappedThread) self).getRTThread();
// 		} else {
// 			return null;
// 		}
		
// 	}
	
	public static native RTThread currentRTThread();
	

	/**
	 * Returns a reference to the currently executing thread object, which can
	 * be either an <code>RTThread</code> or a <code>java.lang.Thread</code>.
	 * Since those two types of threads do not share a common thread base class,
	 * the return type is <code>Object</code>.
	 * 
	 * @return the currently executing thread.
	 */
// 	public static Object currentThread() {
	
// 	        Object obj = currentRTThread();
// 		if (obj == null)
// 			return Thread.currentThread();
// 		else
// 			return obj;
// 	}
	public static Object currentThread() {
		return currentRTThread();
	}

	/**
	 * Causes the currently executing thread object to temporarily pause and
	 * allow other threads to execute. If the currently executing thread is a
	 * <code>java.lang.Thread</code>,<code>yield</code> of that class is
	 * called.
	 */
	public static final void yield() {
		RTThread self = currentRTThread();
		if (self == null)
			Thread.yield();
		else
			WrappedThread.yield();
	}
	//public static final native void yield();

	/**
	 * Causes the currently executing thread to sleep (temporarily cease
	 * execution) until the specified moment in time. Time is measured in
	 * milliseconds, starting from 00am, January 1st 1970, UTC. The thread does
	 * not lose ownership of any monitors.	 * 
	 * @param wakeUpTime
	 *            the time to wake up.
	 * @throws RTInterrupted
	 *             if another thread has interrupted this thread.
	 * @throws RTDelayed
	 *             if execution continues more than <code>maxTimeLag</code>
	 *             too late.
	 */
// 	public static void sleepUntil(long wakeUpTime) throws RTInterrupted,
// 			RTDelayed {
// 		RTSystem.sleepUntil(wakeUpTime);
// 	}

	public static native void sleepUntil(long wakeUpTime) throws RTInterrupted,
																 RTDelayed;
/**
	 * Sleep for (at least) millis ms, handled by <code>RTSystem</code>.
	 * 
	 * @see se.lth.cs.realtime.RTSystem#sleep(long)
	 */
	public static native void sleep(long millis) throws RTInterrupted, RTDelayed;
// 	public static void sleep(long millis) throws RTInterrupted, RTDelayed {
// 		RTSystem.sleep(millis);
// 	}

	/**
	 * Sleep for (at least) millis ms plus nanos ns, handled by
	 * <code>RTSystem</code>.
	 * 
	 * @see se.lth.cs.realtime.RTSystem#sleep(long,int)
	 */
	public static native void sleep(long millis, int nanos) throws RTInterrupted,
																   RTDelayed;
// 	public static void sleep(long millis, int nanos) throws RTInterrupted,
// 			RTDelayed {
// 		RTSystem.sleep(millis, nanos);
// 	}

	/**
	 * Native method coupled to the constructor of the class java.lang.Thread
	 * @returns A long representing the ID of the thread instance
	 **/ 
    
    private final native long createThread();

	/**
	 * Initialize a RTThread.
	 * 
	 * @param group
	 *            the thread group of this thread.
	 * @param name
	 *            the name of the new RTThread.
	 */
	private void init(RTThreadGroup group, String name) {
		if (false) {
			RTInterrupted e = new RTInterrupted();
		}
		//Thread father = Thread.currentThread();
// 		langThread = new WrappedThread();
		//if (group == null && father instanceof RTThread) {
		//	group = father.getRTThreadGroup();
		//}
		//this.group = group; // FIXME check access and/or register...
		//this.daemon = father.isDaemon();
		this.priority = NORM_PRIORITY; //father.getPriority();
// 		langThread.setPriority(this.priority);
		
		//setPriority(NORM_PRIORITY);
		
		long auxLong;
		int  auxInt;
 		this.interruptedFlag = false;
 		auxLong = this.createThread();
		this.target = auxLong; 
// 		this.targetRun = null;

		if (name == null) {
			name = "RTThread-" + nextThreadId();
		}
		this.name = name;
		
		if (mailbox == null && defaultBufferSize > 0) {
			mailbox = new Buffer(defaultBufferSize);
		}
	}

	/**
	 * Allocates a new <code>RTThread</code> object with a newly generated
	 * name in its father's <code>RTThreadGroup</code>. Automatically
	 * generated names are of the form <code>"RTThread-"+</code> <i>n </i>,
	 * where <i>n </i> is an integer.
	 * <p>
	 */
	public RTThread() {
		init(null, null);
	}

	/**
	 * Allocates a new <code>RTThread</code> object with the specified name in
	 * its father's <code>RTThreadGroup</code>.
	 * 
	 * @param name
	 *            the name of the new thread.
	 */
	public RTThread(String name) {
		init(null, name);
	}

	/**
	 * Allocates a new <code>RTThread</code> object with a newly generated
	 * name in the specified <code>RTThreadGroup</code>.
	 * <p>
	 * Automatically generated names are of the form <code>"RTThread-"+</code>
	 * <i>n </i>, where <i>n </i> is an integer.
	 * 
	 * @param group
	 *            the <code>RTThreadGroup</code> of the new thread.
	 */
	public RTThread(RTThreadGroup group) {
		init(group, null);
	}

	/**
	 * Allocates a new <code>RTThread</code> object with the specified name in
	 * the specified <code>RTThreadGroup</code>.
	 * 
	 * @param name
	 *            the name of the new thread.
	 * @param group
	 *            the <code>RTThreadGroup</code> of the new thread.
	 */
	public RTThread(RTThreadGroup group, String name) {
		init(group, name);
	}

	/**
	 * Causes this thread to begin execution.
	 * <p>
	 * The result is that two threads are running concurrently: the current
	 * thread (which returns from the call to the <code>start</code> method)
	 * and the other thread (which executes its <code>run</code> method).
	 * 
	 * @exception IllegalThreadStateException
	 *                if the thread was already started.
	 * @see RTThread#run()
	 */
	
// 	public synchronized void start() {
// 		if (started)
// 			throw new IllegalThreadStateException();
// 		langThread.start();
// 		started = true;
// 	}
	
	public synchronized native void start();

	/**
	 * Default non-empty <code>run</code> method calling <code>perform</code>
	 * until interrupt or termination, corresponding to <code>JThread.run</code>
	 * which in turn corresponds to <code>java.lang.Thread.run</code> (having
	 * the same semantics). Compared to <code>java.lang.Thread.run</code>,
	 * this method defines the real-time execution of the real-time thread, and
	 * not only concurrency, if an RTOS is used.
	 * 
	 * @see java.lang.Thread#run()
	 * @see se.lth.cs.realtime.JThread#run()
	 * @see RTThread#start()
	 */
	protected void run() {
		while (!isInterrupted()) {
			perform();
		}
	}

	/**
	 * Empty perform method, to be replaced in subclasses. There is no
	 * additional semantics in terms of scheduling properties compared to
	 * implementing <code>run</code> with a loop calling this method directly;
	 * the default implementation of <code>run</code> calling
	 * <code>perform</code> is for programming convenience only. If real-time
	 * scheduling should be able to utilize cyclic or periodic behavior, inherit
	 * from <code>CyclicThread</code> or <code>PeriodicThread</code>
	 * instead.
	 * 
	 * @see se.lth.cs.realtime.CyclicThread
	 * @see se.lth.cs.realtime.PeriodicThread
	 */
	protected void perform() {
	}

	/**
	 * This method is called by the system to give a Thread a chance to clean up
	 * before it actually exits.
	 */
	private void exit() {
	}

	/**
	 * Interrupts this thread. Thus an "interruption-flag" is set, which may
	 * raise an InterruptedException in the interrupted thread; otherwise the
	 * flag can be checked via <code>interrupted</code> or
	 * <code>isInterrupted</code>.
	 * 
	 * @see RTThread#interrupted()
	 * @see RTThread#isInterrupted()
	 */
// 	public void interrupt() {
// 		/*
// 		 * Note that this method is not synchronized. Three reasons for this: 1)
// 		 * It changes the API. 2) It's another place where the system could
// 		 * hang. 3) All we're doing is turning on a one-way bit. It doesn't
// 		 * matter exactly when it's done WRT probes via the interrupted()
// 		 * method.
// 		 */
// 		langThread.interrupt();
// 	}
	public native void interrupt();

	/**
	 * Tests if the current thread has been interrupted. The "interruption-flag"
	 * is reset.
	 * <p>
	 * Note that <code>interrupted</code> is a static method, while
	 * <code>isInterrupted</code> is called on this <code>Thread</code>
	 * instance.
	 * 
	 * @return <code>true</code> if the current thread has been interrupted;
	 *         <code>false</code> otherwise.
	 * @see RTThread#interrupt()
	 * @see RTThread#isInterrupted()
	 */
	public boolean interrupted() {
		return isInterrupted(true);
		//return currentRTThread().isInterrupted(true);
	}

	/**
	 * Tests if this thread has been interrupted without modifying the
	 * "interruption-flag".
	 * <p>
	 * Note that <code>isInterrupted</code> is called on this
	 * <code>RTThread</code> instance; by contrast, <code>interrupted</code>
	 * is a static method.
	 * 
	 * @return <code>true</code> if this thread has been interrupted;
	 *         <code>false</code> otherwise.
	 * @see RTThread#interrupt()
	 * @see RTThread#interrupted()
	 */
	public boolean isInterrupted() {
		return isInterrupted(false);
	}

	/**
	 * Tests if this thread has been interrupted. The interrupted state is reset
	 * or not based on the value of clearInterrupted that is passed.
	 */
	private boolean isInterrupted(boolean clearInterrupted) {
		boolean tmp = interruptedFlag;
		if (clearInterrupted) {
			interruptedFlag = false;
		} 
		return tmp;
	} 

	/**
	 * Tests if this thread is alive. A thread is alive if it has been started
	 * and has not yet died.
	 * 
	 * @return <code>true</code> if this thread is alive; <code>false</code>
	 *         otherwise.
	 */
	public final boolean isAlive() {
		return isAlive;
		/*
		return langThread.isAlive();
		*/
	}

	/**
	 * Changes the priority of this thread.
	 * 
	 * @exception FixedPriorityError
	 *                If this thread is alive and implements the
	 *                <code>FixedPriority</code> interface.
	 * @exception IllegalArgumentException
	 *                If the priority is not in the range
	 *                <code>MIN_PRIORITY</code> to <code>MAX_PRIORITY</code>.
	 * 
	 * @see RTThread#getPriority()
	 * @see se.lth.cs.realtime.FixedPriority
	 */
	
	public final synchronized void setPriority(int newPriority)
			throws FixedPriorityError {
		if (newPriority > MAX_PRIORITY || newPriority < MIN_PRIORITY) {
			throw new IllegalArgumentException("priority cannot be "
					+ newPriority + ".");
		}
		if ((this instanceof FixedPriority) && isAlive()) {
			throw new FixedPriorityError();
		}
		priority = newPriority;
		langThread.setPriority(newPriority);
	}
	
	//public final synchronized native void setPriority(int newPriority);

	/**
	 * Obtain the priority of this thread.
	 * 
	 * @return this thread's priority.
	 * @see RTThread#setPriority(int)
	 */
	public final int getPriority() {
		return priority;
	}

	/**
	 * Changes the name of this thread to be equal to the argument
	 * <code>name</code>.
	 * <p>
	 * 
	 * @param name
	 *            the new name for this thread.
	 * @see RTThread#getName()
	 */
	public final void setName(String name) {
		this.name = name;
	}

	/**
	 * Returns this thread's name.
	 * 
	 * @return this thread's name.
	 * @see RTThread#setName(java.lang.String)
	 */
	public final String getName() {
		return name;
	}

	/**
	 * Returns this thread's <code>RTThreadGroup</code>.
	 * 
	 * @return this thread's <code>RTThreadGroup</code>.
	 */
	public final RTThreadGroup getThreadGroup() {
		return group;
	}

	/**
	 * Waits at most <code>millis</code> milliseconds for this thread to die.
	 * A timeout of <code>0</code> means to wait forever. A CyclicThread is
	 * not allowed to join another thread. (TO FIX)
	 * 
	 * @param millis
	 *            the time to wait in milliseconds.
	 * @exception InterruptedException
	 *                if another thread has interrupted the current thread.
	 * @exception IllegalStateException
	 *                if the calling thread is an instance of <b>CyclicRTThread
	 *                </b>.
	 * @see se.lth.cs.realtime.CyclicThread
	 */
// 	public final synchronized void join(long millis)
// 			throws InterruptedException {
// 		/*
// 		 * The RTThread which wants to join this one is put in the waiting queue
// 		 * of this RTThread (as in java.lang.Thread); this could be dangerous if
// 		 * some other RTThread is in the queue for different purposes (the
// 		 * latter may not be notified at the proper time); we could use a
// 		 * different object to wait on.
// 		 */
// 		/*
// 		 * long base = System.currentTimeMillis(); long now = 0;
// 		 * 
// 		 * //if (currentRTThread() instanceof CyclicThread) { // throw new
// 		 * IllegalStateException("cyclic real-time threads can not join other
// 		 * threads."); //}
// 		 * 
// 		 * if (millis < 0) { throw new IllegalArgumentException("timeout value
// 		 * is negative"); }
// 		 * 
// 		 * if (millis == 0) { while (isAlive()) { wait(0); } } else { while
// 		 * (isAlive()) { long delay = millis - now; if (delay <= 0) { break; }
// 		 * wait(delay); now = System.currentTimeMillis() - base; } }
// 		 */
		
// 		if (currentRTThread() instanceof CyclicThread) {
// 			throw new UnsupportedOperationException("Cyclic real-time threads" + 
// 					" are not allowed to join other threads.");
// 		}
		
// 		langThread.join(millis);
// 	}
	
	//public final synchronized native void join(long millis);

	/**
	 * Waits at most <code>millis</code> milliseconds plus <code>nanos</code>
	 * nanoseconds for this thread to die.
	 * 
	 * @param millis
	 *            the time to wait in milliseconds.
	 * @param nanos
	 *            0-999999 additional nanoseconds to wait.
	 * @exception IllegalArgumentException
	 *                if the value of millis is negative the value of nanos is
	 *                not in the range 0-999999.
	 * @exception InterruptedException
	 *                if another thread has interrupted the current thread.
	 */
// 	public final synchronized void join(long millis, int nanos)
// 			throws InterruptedException {

// 		if (millis < 0) {
// 			throw new IllegalArgumentException("timeout value is negative");
// 		}

// 		if (nanos < 0 || nanos > 999999) {
// 			throw new IllegalArgumentException(
// 					"nanosecond timeout value out of range");
// 		}

// 		if (nanos >= 500000 || (nanos != 0 && millis == 0)) {
// 			millis++;
// 		}

// 		join(millis);
// 	}

	/**
	 * Waits for this thread to die.
	 * 
	 * @exception InterruptedException
	 *                if another thread has interrupted the current thread.
	 */
// 	public final void join() throws InterruptedException {
// 		join(0);
// 	}

	/**
	 * Interrupt and join.
	 */
	public void terminate() {
		if (isAlive() && !isInterrupted()) {
			interrupt();
			try {
				join();
			} catch (InterruptedException exc) {
				throw new RTInterrupted("Termination (join) of " + name
						+ " interrupted");
			}
		}
	}

	/**
	 * Prints a stack trace of the current thread. This method is used only for
	 * debugging.
	 * 
	 * @see java.lang.Throwable#printStackTrace()
	 */
	public static void dumpStack() {
		new Exception("Stack trace").printStackTrace();
	}

	/**
	 * Marks this thread as either a daemon thread or a user thread.
	 * <p>
	 * This method must be called before the thread is started.
	 * 
	 * @param on
	 *            if <code>true</code>, marks this thread as a daemon thread.
	 * @exception IllegalThreadStateException
	 *                if this thread is active.
	 * @see RTThread#isDaemon()
	 */
	/*
	public final void setDaemon(boolean on) {
		if (isAlive()) {
			throw new IllegalThreadStateException();
		}
		daemon = on;
	}
	*/

	/**
	 * Tests if this thread is a daemon thread.
	 * 
	 * @return <code>true</code> if this thread is a daemon thread;
	 *         <code>false</code> otherwise.
	 * @see RTThread#setDaemon(boolean)
	 */
	/*
	public final boolean isDaemon() {
		return daemon;
	}
	*/

	/**
	 * Time bases not implemented yet.
	 * 
	 * @return null.
	 */
	public Timebase getTimebase() {
		return null;
	}

	/**
	 * Environments not implemented yet.
	 * 
	 * @return null.
	 */
	public Environment getEnv() {
		return null;
	}

	/**
	 * Returns a string representation of this thread, including the thread's
	 * name and priority.
	 * 
	 * @return a string representation of this thread.
	 */
	public String toString() {
		return "RTThread[" + getName() + ", " + getPriority() + "]";
	}

	//-------------------------------------------------------------------
	// The encapsulation of the ordinary Thread. Note that this class has
	// to be internal to RTThread and a subclass to Thread to provide the
	// mapping both ways between these types of thread objects.
// 	private class WrappedThread extends java.lang.Thread {

// 		public void run() {
// 			System.out.println(RTThread.this.toString());
// 			RTThread.this.run();
// 		}

// 		RTThread getRTThread() {
// 			return RTThread.this;
// 		}

// 	}
}
