/* $Id: RTThreadGroup.java,v 1.6 2005/08/19 09:37:16 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * A thread group represents a set of threads. In addition, a thread
 * group can also include other thread groups. The thread groups form
 * a tree in which every thread group except the initial thread group
 * has a parent.
 * <p>
 * This class is mainly aimed at giving a more convenient way to issue
 * the same function call to several threads, which is often needed.
 * It poses also the basis for an optional, future implementation of
 * some kind of scheduling or priority settings for a group of threads.
 * 
 * @see License
 */

public class RTThreadGroup {

  private RTThreadGroup parent;
  private String name;
  private int maxPriority;
  private int minPriority;
  private boolean destroyed;
  private boolean daemon;

  private int nthreads;
  private RTThread threads[];

  private int ngroups;
  private RTThreadGroup groups[];

  /**
   * Creates an empty Thread group that is not in any Thread group.
   * This method is used to create the system Thread group.
   */
  RTThreadGroup() {
    this.name = "system";
    this.maxPriority = RTThread.MAX_PRIORITY;
    this.minPriority = RTThread.MIN_PRIORITY;
  }

  /**
   * Constructs a new thread group. The parent of this new group is
   * the thread group of the currently running real-time thread.
   *
   * @param   name   the name of the new thread group.
   */
  public RTThreadGroup(String name) {
    RTThread t = (RTThread) RTThread.currentThread();
    if (t instanceof RTThread) {
      this.name = name;
      this.maxPriority = parent.maxPriority;
      this.minPriority = parent.minPriority;
      this.daemon = parent.daemon;
      parent.add(this);
    }
    else {
//      throw new UnsupportedOperationException("Not an RT thread.");
	System.out.println("Not an RT thread.");
    }
  }

  /**
   * Creates a new thread group. The parent of this new group is the
   * specified thread group.
   *
   * @param     parent   the parent thread group.
   * @param     name     the name of the new thread group.
   * @exception  NullPointerException  if the thread group argument is
   *               <code>null</code>.
   */
  public RTThreadGroup(RTThreadGroup parent, String name) {
    if (parent == null) {
      //throw new NullPointerException();
      System.out.println("NullPointerException");
      System.exit(2);
    }
    this.name = name;
    this.maxPriority = parent.maxPriority;
    this.minPriority = parent.minPriority;
    this.daemon = parent.daemon;
    this.parent = parent;
    parent.add(this);
  }

  /**
   * Returns the name of this thread group.
   *
   * @return  the name of this thread group.
   */
  public final String getName() {
    return name;
  }

  /**
   * Returns the parent of this thread group.
   *
   * @return  the parent of this thread group. The top-level thread group
   *          is the only thread group whose parent is <code>null</code>.
   */
  public final RTThreadGroup getParent() {
    return parent;
  }

  /**
   * Returns the maximum priority of this thread group. Threads that are
   * part of this group cannot have a higher priority than the maximum
   * priority.
   *
   * @return  the maximum priority that a thread in this thread group
   *          can have.
   */
  public final int getMaxPriority() {
    return maxPriority;
  }

  /**
   * Returns the minimum priority of this thread group. Threads that are
   * part of this group cannot have a lower priority than the minimum
   * priority.
   *
   * @return  the minimum priority that a thread in this thread group
   *          can have.
   */
  public final int getMinPriority() {
    return minPriority;
  }

  /**
   * Tests if this thread group is a daemon thread group. A
   * daemon thread group is automatically destroyed when its last
   * thread is stopped or its last thread group is destroyed.
   *
   * @return  <code>true</code> if this thread group is a daemon thread group;
   *          <code>false</code> otherwise.
   */
  public final boolean isDaemon() {
    return daemon;
  }

  /**
   * Tests if this thread group has been destroyed.
   *
   */
  public synchronized boolean isDestroyed() {
    return destroyed;
  }

  /**
   * Changes the daemon status of this thread group.
   * <p>
   * A daemon thread group is automatically destroyed when its last
   * thread is stopped or its last thread group is destroyed.
   *
   * @param      daemon   if <code>true</code>, marks this thread group as
   *                      a daemon thread group; otherwise, marks this
   *                      thread group as normal.
   */
  public final void setDaemon(boolean daemon) {
    this.daemon = daemon;
  }

  /**
   * Sets the maximum priority of the group.
   * <p>
   * Threads in the thread group that already have a higher priority
   * are not affected.
   *
   * @param      pri   the new priority of the thread group.
   */
  public final void setMaxPriority(int pri) {
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      if (pri < RTThread.MIN_PRIORITY) {
        maxPriority = RTThread.MIN_PRIORITY;
      }
      else if (pri < maxPriority) {
        maxPriority = pri;
      }
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
    }
    for (int i = 0; i < ngroupsSnapshot; i++) {
      groupsSnapshot[i].setMaxPriority(pri);
    }
  }

  /**
   * Sets the minimum priority of the group.
   * <p>
   * Threads in the thread group that already have a lower priority
   * are not affected.
   *
   * @param      pri   the new priority of the thread group.
   */
  public final void setMinPriority(int pri) {
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      if (pri > RTThread.MAX_PRIORITY) {
        minPriority = RTThread.MAX_PRIORITY;
      }
      else if (pri > minPriority) {
        minPriority = pri;
      }
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
    }
    for (int i = 0; i < ngroupsSnapshot; i++) {
      groupsSnapshot[i].setMinPriority(pri);
    }
  }
  /**
   * Tests if this thread group is either the thread group
   * argument or one of its ancestor thread groups.
   *
   * @param   g   a thread group.
   * @return  <code>true</code> if this thread group is the thread group
   *          argument or one of its ancestor thread groups;
   *          <code>false</code> otherwise.
   */
  public final boolean parentOf(RTThreadGroup g) {
    for (; g != null; g = g.parent) {
      if (g == this) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns an estimate of the number of active threads in this
   * thread group.
   *
   * @return  the number of active threads in this thread group and in any
   *          other thread group that has this thread group as an ancestor.
   */
  public int activeCount() {
    int result;
    // Snapshot sub-group data so we don't hold this lock
    // while our children are computing.
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      if (destroyed) {
        return 0;
      }
      result = nthreads;
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
    }
    for (int i = 0; i < ngroupsSnapshot; i++) {
      result += groupsSnapshot[i].activeCount();
    }
    return result;
  }

  /**
   * Copies into the specified array every active thread in this
   * thread group and its subgroups.
   * <p>
   * An application should use the <code>activeCount</code> method to
   * get an estimate of how big the array should be. If the array is
   * too short to hold all the threads, the extra threads are silently
   * ignored.
   *
   * @param   list   an array into which to place the list of threads.
   * @return  the number of threads put into the array.
   * @see     java.lang.ThreadGroup#activeCount()
   */
  public int enumerate(RTThread list[]) {
    return enumerate(list, 0, true);
  }

  /**
   * Copies into the specified array every active thread in this
   * thread group. If the <code>recurse</code> flag is
   * <code>true</code>, references to every active thread in this
   * thread's subgroups are also included. If the array is too short to
   * hold all the threads, the extra threads are silently ignored.
   * <p>
   * An application should use the <code>activeCount</code> method to
   * get an estimate of how big the array should be.
   *
   * @param   list      an array into which to place the list of threads.
   * @param   recurse   a flag indicating whether also to include threads
   *                    in thread groups that are subgroups of this
   *                    thread group.
   * @return  the number of threads placed into the array.
   * @see     java.lang.ThreadGroup#activeCount()
   */
  public int enumerate(/**/
  RTThread list[], boolean recurse) {
    return enumerate(list, 0, recurse);
  }

  private int enumerate(/**/
  RTThread list[], int n, boolean recurse) {
    int ngroupsSnapshot = 0;
    RTThreadGroup[] groupsSnapshot = null;
    synchronized (this) {
      if (destroyed) {
        return 0;
      }
      int nt = nthreads;
      if (nt > list.length - n) {
        nt = list.length - n;
      }
      for (int i = 0; i < nt; i++) {
        if (threads[i].isAlive()) {
          list[n++] = threads[i];
        }
      }
      if (recurse) {
        ngroupsSnapshot = ngroups;
        if (groups != null) {
          groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
          System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
        }
        else {
          groupsSnapshot = null;
        }
      }
    }
    if (recurse) {
      for (int i = 0; i < ngroupsSnapshot; i++) {
        n = groupsSnapshot[i].enumerate(list, n, true);
      }
    }
    return n;
  }

  /**
   * Returns an estimate of the number of active groups in this
   * thread group.
   *
   * @return  the number of active thread groups with this thread group as
   *          an ancestor.
   */
  public int activeGroupCount() {
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      if (destroyed) {
        return 0;
      }
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
    }
    int n = ngroupsSnapshot;
    for (int i = 0; i < ngroupsSnapshot; i++) {
      n += groupsSnapshot[i].activeGroupCount();
    }
    return n;
  }

  /**
   * Copies into the specified array references to every active
   * subgroup in this thread group.
   * <p>
   * An application should use the <code>activeGroupCount</code>
   * method to get an estimate of how big the array should be. If the
   * array is too short to hold all the thread groups, the extra thread
   * groups are silently ignored.
   *
   * @param   list   an array into which to place the list of thread groups.
   * @return  the number of thread groups put into the array.
   * @see     java.lang.ThreadGroup#activeGroupCount()
   */
  public int enumerate(RTThreadGroup list[]) {
    return enumerate(list, 0, true);
  }

  /**
   * Copies into the specified array references to every active
   * subgroup in this thread group. If the <code>recurse</code> flag is
   * <code>true</code>, references to all active subgroups of the
   * subgroups and so forth are also included.
   * <p>
   * An application should use the <code>activeGroupCount</code>
   * method to get an estimate of how big the array should be.
   *
   * @param   list      an array into which to place the list of threads.
   * @param   recurse   a flag indicating whether to recursively enumerate
   *                    all included thread groups.
   * @return  the number of thread groups put into the array.
   * @see     java.lang.ThreadGroup#activeGroupCount()
   */
  public int enumerate(RTThreadGroup list[], boolean recurse) {
    return enumerate(list, 0, recurse);
  }

  private int enumerate(RTThreadGroup list[], int n, boolean recurse) {
    int ngroupsSnapshot = 0;
    RTThreadGroup[] groupsSnapshot = null;
    synchronized (this) {
      if (destroyed) {
        return 0;
      }
      int ng = ngroups;
      if (ng > list.length - n) {
        ng = list.length - n;
      }
      if (ng > 0) {
        System.arraycopy(groups, 0, list, n, ng);
        n += ng;
      }
      if (recurse) {
        ngroupsSnapshot = ngroups;
        if (groups != null) {
          groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
          System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
        }
        else {
          groupsSnapshot = null;
        }
      }
    }
    if (recurse) {
      for (int i = 0; i < ngroupsSnapshot; i++) {
        n = groupsSnapshot[i].enumerate(list, n, true);
      }
    }
    return n;
  }

  /**
   * Interrupts all threads in this thread group.
   * <p>
   * This method calls the <code>interrupt</code> method on all the
   * threads in this thread group and in all of its subgroups.
   *
   * @see        RTThread#interrupt()
   */
  public final void interrupt() {
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      for (int i = 0; i < nthreads; i++) {
        threads[i].interrupt();
      }
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
    }
    for (int i = 0; i < ngroupsSnapshot; i++) {
      groupsSnapshot[i].interrupt();
    }
  }

  /**
   * Destroys this thread group and all of its subgroups. This thread
   * group must be empty, indicating that all threads that had been in
   * this thread group have since stopped.
   *
   * @exception  IllegalThreadStateException  if the thread group is not
   *               empty or if the thread group has already been destroyed.
   */
  public final void destroy() {
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      if (destroyed || (nthreads > 0)) {
        throw new IllegalThreadStateException();
      }
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
      if (parent != null) {
        destroyed = true;
        ngroups = 0;
        groups = null;
        nthreads = 0;
        threads = null;
      }
    }
    for (int i = 0; i < ngroupsSnapshot; i += 1) {
      groupsSnapshot[i].destroy();
    }
    if (parent != null) {
      parent.remove(this);
    }
  }

  /**
   * Adds the specified Thread group to this group.
   * @param g the specified Thread group to be added
   * @exception IllegalThreadStateException If the Thread group has been destroyed.
   */
  private final void add(RTThreadGroup g) {
    synchronized (this) {
      if (destroyed) {
        throw new IllegalThreadStateException();
      }
      if (groups == null) {
        groups = new RTThreadGroup[8];
      }
      else if (ngroups == groups.length) {
        RTThreadGroup newgroups[] = new RTThreadGroup[ngroups * 2];
        System.arraycopy(groups, 0, newgroups, 0, ngroups);
        groups = newgroups;
      }
      groups[ngroups] = g;
      ngroups++;
    }
  }

  /**
   * Removes the specified Thread group from this group.
   * @param g the Thread group to be removed
   * @return if this Thread has already been destroyed.
   */
  private void remove(RTThreadGroup g) {
    synchronized (this) {
      if (destroyed) {
        return;
      }
      for (int i = 0; i < ngroups; i++) {
        if (groups[i] == g) {
          ngroups -= 1;
          System.arraycopy(groups, i + 1, groups, i, ngroups - i);
          groups[ngroups] = null;
          break;
        }
      }
      if (nthreads == 0) {
        notifyAll();
      }
      if (daemon && (nthreads == 0) && (ngroups == 0)) {
        destroy();
      }
    }
  }

  /**
   * Adds the specified Thread to this group.
   * @param t the Thread to be added
   * @exception IllegalThreadStateException If the Thread group has been destroyed.
   */
  void add(RTThread t) {
    synchronized (this) {
      if (destroyed) {
        throw new IllegalThreadStateException();
      }
      if (threads == null) {
        threads = new RTThread[8];
      }
      else if (nthreads == threads.length) {
        RTThread newthreads[] = new RTThread[nthreads * 2];
        System.arraycopy(threads, 0, newthreads, 0, nthreads);
        threads = newthreads;
      }
      threads[nthreads] = t;
      nthreads++;
    }
  }

  /**
   * Removes the specified Thread from this group.
   * @param t the Thread to be removed
   * @return if the Thread has already been destroyed.
   */
  void remove(RTThread t) {
    synchronized (this) {
      if (destroyed) {
        return;
      }
      for (int i = 0; i < nthreads; i++) {
        if (threads[i] == t) {
          System.arraycopy(threads, i + 1, threads, i, --nthreads - i);
          threads[nthreads] = null;
          break;
        }
      }
      if (nthreads == 0) {
        notifyAll();
      }
      if (daemon && (nthreads == 0) && (ngroups == 0)) {
        destroy();
      }
    }
  }

  /**
   * Prints information about this thread group to the standard
   * output. This method is useful only for debugging.
   */
  public void list() {
    // list(System.out, 0);
  }
  void list(java.io.PrintStream out, int indent) {
    int ngroupsSnapshot;
    RTThreadGroup[] groupsSnapshot;
    synchronized (this) {
      for (int j = 0; j < indent; j++) {
        out.print(" ");
      }
      out.println(this);
      indent += 4;
      for (int i = 0; i < nthreads; i++) {
        for (int j = 0; j < indent; j++) {
          out.print(" ");
        }
        out.println(threads[i]);
      }
      ngroupsSnapshot = ngroups;
      if (groups != null) {
        groupsSnapshot = new RTThreadGroup[ngroupsSnapshot];
        System.arraycopy(groups, 0, groupsSnapshot, 0, ngroupsSnapshot);
      }
      else {
        groupsSnapshot = null;
      }
    }
    for (int i = 0; i < ngroupsSnapshot; i++) {
      groupsSnapshot[i].list(out, indent);
    }
  }

  /**
   * Returns a string representation of this Thread group.
   *
   * @return  a string representation of this thread group.
   */
  public String toString() {
    return //getClass().getName()
	    "RTThreadGroup "
      + "[name="
      + getName()
      + ",maxpri="
      + maxPriority
      + "]";
  }

}
