/* $Id: SporadicThread.java,v 1.2 2005/08/22 08:45:17 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

import se.lth.cs.realtime.semaphore.*;

/**
 * This class models a real-time thread that executes the same function
 * every time it is triggered.
 * <p>
 * The execution frequency must be upper-bound specifying a shortest period
 * for this thread(if <code>FixedPeriod</code> is implemented, the shortest
 * period can be modified only before this thread is started).
 * <p>
 * Being a <code>CyclicThread</code>,
 * a <code>SporadicThread</code> may be suspended and resumed
 * by other threads. By default, the "triggering events"
 * that occur during the suspension time are served after the
 * resumption.
 * 
 * @see License
 */

// TODO: Have the default trig sending a trigg-event.

public class SporadicThread extends CyclicThread {

  private long shortestPeriod;

  private long lastTriggerTime = 0;

  protected volatile int failedTriggerCount = 0;
  
  private volatile RTThread triggerThread = null;
  
  private BinarySem sem;

  /**
   * Allocates a new <code>SporadicThread</code> object.
   * <p>
   * Equivalent to <code>SporadicThread(null, null, true, shortestPeriod)</code>.
   *
   * @param	shortestPeriod	    the shortest period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the shortest period is not positive.
   *
   * @see SporadicThread#SporadicThread(RTThreadGroup, java.lang.String, boolean, long)
   */
  public SporadicThread(long shortestPeriod) {
    this(null, null, shortestPeriod);
  }

  /**
   * Allocates a new <code>SporadicThread</code> object.
   * <p>
   * Equivalent to <code>SporadicThread(null, name, true, shortestPeriod)</code>.
   *
   * @param   name	    the name of the new thread.
   * @param	shortestPeriod	    the shortest period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the shortest period is not positive.
   *
   * @see SporadicThread#SporadicThread(RTThreadGroup, java.lang.String, boolean, long)
   */
  public SporadicThread(String name, long shortestPeriod) {
    this(null, name, shortestPeriod);
  }

  /**
   * Allocates a new <code>SporadicThread</code> object.
   * <p>
   * Equivalent to <code>SporadicThread(group, null, true, shortestPeriod)</code>.
   *
   * @param	group		    this thread's <code>RTThreadGroup</code>..
   * @param	shortestPeriod	    the shortest period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the shortest period is not positive.
   *
   * @see SporadicThread#SporadicThread(RTThreadGroup, java.lang.String, long)
   */
  public SporadicThread(RTThreadGroup group, long shortestPeriod) {
    this(group, null, shortestPeriod);
  }

  /**
   * Allocates a new <code>SporadicThread</code> object.
   * <p>
   * Equivalent to <code>SporadicThread(group, name, true, shortestPeriod)</code>.
   *
   * @param   name	    the name of the new thread.
   * @param	group	    this thread's <code>RTThreadGroup</code>..
   * @param	shortestPeriod	    the shortest period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the shortest period is not positive.
   *
   * @see SporadicThread#SporadicThread(RTThreadGroup, java.lang.String, long)
   */
  public SporadicThread(
    RTThreadGroup group,
    String name,
    long shortestPeriod) {
    super(group, name);
    
    sem = new BinarySem();
    
    try {
      setShortestPeriod(shortestPeriod);
    }
    catch (FixedPeriodError e) {/* impossible */}
  }

  /**
   * Causes the execution of the <code>perform</code> method.
   * 
   * Each SporadicThread can have at most one triggering thread.
   * The first RTThread which tries to trigger this thread will  
   * be the triggering thread.
   *
   * @return -2 - This thread isn't alive to be triggered. 
   * @return -1 - The calling thread hasn't the right to trigger this thread.
   * @return  0 - All OK.
   * @return >0 - Trigger failed, the number of failed calls is returned.
   */
  public int trigger() {
	  
	  if (!isAlive()) {
		  return -2;
	  }
	  if (currentRTThread() != null) {
		  if (triggerThread == null) {
			  triggerThread = currentRTThread();
		  } 
				
		  if (triggerThread == currentRTThread()) {
			  long now = System.currentTimeMillis();
			  if (lastTriggerTime + shortestPeriod > now) {
				  failedTriggerCount++;
				  return failedTriggerCount;
			  }
			  lastTriggerTime = now;
			  sem.give();
			  return 0;
		  }
	  }
			
	  return -1;
  }

  /**
   * Does nothing.
   * This empty implementation allows the use of a <b>SporadicThread</b>
   * as a <code>TriggerEvent</code> generator only.
   *
   * @see se.lth.cs.realtime.event.TriggerEvent
   *
   */
  protected void perform() {}

  /**
   * Lets this thread wait for a call to the function <code>trigger</code>.
   *
   * @exception InterruptedException if another thread has interrupted this one.
   */
  protected void waitToPerform() throws InterruptedException {
	  sem.take();
  }

  /**
   * Returns the shortest period of this thread, in milliseconds.
   *
   * @return the shortest period in ms.
   */
  public synchronized long getShortestPeriod() {
    return shortestPeriod;
  }

  /**
   * Sets the shortest period of this thread, in milliseconds.
   * If this thread implements <code>FixedPeriod</code>and is
   * alive , an exception is thrown.
   *
   * @param shortestPeriod the new shortest period of this thread.
   *
   * @exception IllegalArgumentException if the argument is not positive.
   * @exception FixedPeriodError if the thread is alive and implements
   *		    <code>FixedPeriod</code>.
   */
  public synchronized void setShortestPeriod(long shortestPeriod)
    throws FixedPeriodError {
    if (shortestPeriod <= 0) {
      throw new IllegalArgumentException("Shortest period value must be positive.");
    }
    if (this instanceof FixedPeriod && isAlive()) {
      throw new FixedPeriodError();
    }
    this.shortestPeriod = shortestPeriod;
  }

  /**
  * Returns a string representation of this thread, including the
  * thread's name, priority and shortest period in milliseconds.
  *
  * @return  a string representation of this thread.
  */
  public String toString() {
    return "SporadicThread["
      + getName()
      + ", "
      + getPriority()
      + ", "
      + getShortestPeriod()
      + "]";
  }
}
