package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;

/**
 * A thread safe basic queue/buffer for <code>RTEvent</code> objects.
 * Essentially a protected bounded buffer with both blocking and non-blocking
 * methods for posting and fetching messages in terms of references to objects.
 * <p>
 * For the functionality provided by this basic class, the buffer could have
 * been typed to contain <code>EventObject</code>s (the base class of RTEvent),
 * but to permit subclasses of <code>RTEventBuffer</code> to utilize the
 * properties of the RTEvents that class is used here. For instance, one might
 * need a buffer that keeps track of the total delay of data by referring to
 * the time-stamps of the newest and oldest RTEvent.
 * <p>
 * The methods checking or awaiting buffer full/empty are useful for operation
 * supervision, capturing events until a buffer is filled, mode changes, etc.
 * Do not misuse them for polled (concurrently bad) or busy-wait solutions.
 * <p>
 * To meet the efficiency demands of some applications, most methods
 * are final.  The class as such is, however, not final since
 * subclassing is expected for adding more features, hence the
 * protected visibility of fields. Additionally, the methods post and
 * fetch are virtual (i.e., not final). They by default call the
 * corresponding do-method and are typed accordingly, but they are
 * free for redefinition which can even result in synchronized being
 * omitted if mutual exclusion is ensured elsewhere (in enclosing
 * monitor, etc.). See also the implemented interfaces, which are
 * facilitating access to only one end of the buffer.
 * <p>
 * While events are stored in the buffer, the buffer is set to be the owner
 * of the events, see the <code>RTEvent</code> class. Only events owned by
 * the current thread are accepted for posting to the buffer. When events
 * are fetched, the fetching thread is assigned to be the new owner. This
 * is to be ensured also by subclasses, thereby maintaining the detection
 * of event producers keeping references to objects already sent. In some
 * cases, however, it is desirable to resend (typically immutable) objects.
 * Therefore, events with the owner equal to <code>null</code> is always
 * accepted. The concept of owners is from the OSE operating system.
 * <p>
 * In practice and by default, all methods can be considered being
 * synchronized, even if the methods as such are not declared to be
 * synchronized. Having the methods synchronized would agree with
 * common application programming practice, and together with the fact
 * that most methods are final, those synchronized methods would remain
 * synchronized is subclasses. However, synchronization is performed
 * on an internal lock object for the following reasons: <br>1) Other
 * object cannot interfere with the internal wait/notify by
 * synchronizing on an instance of this class (by using a synchronized
 * statement). <br>2) An external object (provided via the
 * constructor) can be used as the lock object, thereby permitting
 * synchronized use of several buffers/object.
 * <p>
 * If a blocked call is interrupted, an RTInterrupted (subclass of Error, not
 * an exception) is thrown, and that buffer object is most likely not useful
 * anymore.
 *
 * @see License
 * @see se.lth.cs.realtime.RTInterrupted
 */

public class BufferLegacy extends RTEventBuffer {
	
	/**
	 * Initializes the buffer with a size of <code>DEFAULT_SIZE</code>.
     */
	public BufferLegacy() {
		super();
	}

    /**
     * Initializes the queue (buffer) with a given maximal size. A size smaller
     * than one results in a buffer with size one.
     *
     * @param maxSize maximum queue size
     */
    public BufferLegacy(int maxSize) {
    	super(maxSize);
    	maxSizeR = maxSizeW = maxSize;
    }

    /**
     * Initializes the queue (buffer) with a given maximal size, and with a
     * lock object to be used for synchronization.
     *
     * @param maxSize maximum queue size
     * @param lock    external monitor/object to synchronized with
     */
    public BufferLegacy(int maxSize, Synchronized.Lock lock) {
    	super(maxSize, lock);
    	maxSizeR = maxSizeW = maxSize;
    }

    /**
     * Returns the next RTEvent in the queue, blocks if none available.
     *
     * @return a non-null event.
     */
    public RTEvent doFetch() {
    	synchronized (lock) {
    		
    		try {
    			while (currSize == 0)
    				lock.wait(); // Wait for non-empty
    		} catch (InterruptedException exc) {
    			error("doFetch interrupted.");
    		}
    		
    		RTEvent e = buffer[fetchAt];
    		buffer[fetchAt] = null; // For garbage collector
    		fetchAt = (fetchAt + 1) % maxSizeR;
    		
            if (currSize-- == maxSizeR)
            	lock.notifyAll(); // Non-full notification
            if (currSize == 0)
            	lock.notifyAll(); // Empty notification
      
            if (e.owner != null)
            	e.setOwner(/**/RTThread.currentThread());
            
            return e;
        }
    }

    /**
     * Returns the next available RTEvent in the queue, or null if the
     * queue is empty. In other words, this method always returns immediately,
     * even if the queue is empty (non-blocking).
     *
     * @return an RTEvent if any available, null otherwise
     */
    public RTEvent tryFetch() {
    	synchronized (lock) {
    		
    		if (isEmpty()) {
    			return null;
    		} else {
    			RTEvent e = buffer[fetchAt];
    			buffer[fetchAt] = null; // For garbage collector
    			fetchAt = (fetchAt + 1) % maxSizeR;
        
    			if (currSize-- <= maxSizeR)
    				lock.notifyAll(); // Non-full notification
    			if (currSize == 0)
    				lock.notifyAll(); // Empty notification
        
    			if (e.owner != null)
    				e.setOwner(/**/RTThread.currentThread());
        
    			return e;
    		}
    	}
    }

    /**
     * Get the next RTEvent in the queue, or block (up to the specified
     * timeout time) if no event is available.
     *
     * @param timeout the maximum blocking time time in milli-seconds.
     * @return an RTEvent if any available within the timeout time, null
     *         otherwise.
     */
    public RTEvent tryFetch(long timeout) {
    	long now;
    	long arrival = now = System.currentTimeMillis();
    	synchronized (lock) {
    		
    		RTEvent ans = null;
    		try {
    			while (currSize == 0 && now < arrival + timeout) {
    				lock.wait(timeout);
    				now = System.currentTimeMillis();
    			}
    			if (now < arrival + timeout)
    				ans = tryFetch();
    		} catch (InterruptedException exc) {
    			error("tryFetch(timeout) interrupted.");
    		}
    		
    		return ans;
    	}
    }

    /**
     * Adds an RTEvent to the queue, blocks caller if the queue is full.
     * @param e the event to enqueue
     */
    public final void doPost(RTEvent e) {
    	if (e.owner != null && e.owner != /**/RTThread.currentThread()) {
    		throw new OwnerError(e.owner);
    	}
    	synchronized (lock) {
    		
    		try {
    			while (currSize >= maxSizeW)
    				lock.wait();
    		} catch (InterruptedException exc) {
    			error("doPost interrupted.");
    		}
    		
    		if (e.owner != null)
    	        e.setOwner(this);
    	    buffer[postAt] = e;
    	    postAt = (postAt + 1) % maxSizeR;
      
    	    if (currSize++ == 0)
    	    	lock.notifyAll(); // Non-empty notification
    	    if (currSize == maxSizeW)
    	    	lock.notifyAll(); // Full notification
    	    
    	    return;
    	}
    }

    /**
     * Adds an EventObject to the queue, without blocking if the queue is full.
     * @param e the event to enqueue
     * @return null when event posted to non-full buffer, the event otherwise
     */
    public RTEvent tryPost(RTEvent e) {
    	if ( e.owner != null && e.owner != /**/
    		RTThread.currentThread()) {
    		throw new OwnerError(e.owner);
    	}
    	synchronized (lock) {
    		
    		if (currSize >= maxSizeW)
    			return e;
    		else {
    			
    			if (e.owner != null)
    				e.setOwner(this);
		        buffer[postAt] = e;
    			postAt = (postAt + 1) % maxSizeR;
		        
    			if (currSize++ == 0)
		          lock.notifyAll(); // Non-empty notification
		        if (currSize == maxSizeW)
		          lock.notifyAll(); // Full notification
		        
		        return null;
    		}
    	}
    }

    /**
     * Adds an RTEvent to the queue, blocks caller if the queue is full up to
     * the specified timeout time.
     *
     * @param e       the event to enqueue
     * @param timeout the maximum blocking time time in milli-seconds.
     * @return null when event posted within the timeout time to non-full
     *         buffer, the event otherwise.
     */
    public RTEvent tryPost(RTEvent e, long timeout) {
    	long now;
    	long arrival = now = System.currentTimeMillis();
    	synchronized (lock) {
    		
    		RTEvent ans = e;
    		try {
    			while (currSize >= maxSizeW && now < arrival + timeout) {
    				lock.wait(timeout);
    				now = System.currentTimeMillis();
    			}
    			if (now < arrival + timeout)
    				ans = tryPost(e);
    		} catch (InterruptedException exc) {
    			error("tryPost(timeout) interrupted.");
    		}
      
    		return ans;
    	}
    }

    /**
     * Checks if buffer is empty.
     * @return true if the queue is empty, false if it is non-empty
     */
    public boolean isEmpty() {
    	synchronized (lock) {
    		return currSize == 0;
    	}
    }

    /**
     * Waits for buffer to become empty.
     */
    public void awaitEmpty() {
    	synchronized (lock) {
    		
    		try {
    			while (currSize > 0)
    				lock.wait();
    		} catch (InterruptedException exc) {
    			error("BufferLegacy.awaitEmpty() interrupted");
    		}
    	}
    }

    /**
     * Waits for buffer to become non-empty.
     */
    public void awaitNotEmpty() {
    	synchronized (lock) {
      
    		try {
    			while (currSize == 0)
    				lock.wait();
    		} catch (InterruptedException exc) {
    			error("awaitNotEmpty interrupted.");
    		}
    	}
    }

    /**
     * Checks if buffer is full.
     * @return true if the queue is full, false if it is non-full
     */
    public boolean isFull() {
    	synchronized (lock) {
    		return currSize >= maxSizeW;
    	}
    }

    /**
     * Waits for buffer to become full. If the buffer has been resized to a size
     * less than the current number of events, the caller is blocked until the
     * buffer is exactly full.
     */
    public void awaitFull() {
    	synchronized (lock) {
      
    		try {
    			while (currSize != maxSizeW)
    				lock.wait();
    		}
    		catch (InterruptedException exc) {
    			error("BufferLegacy.awaitFull() interrupted");
    		}
    	}
    }

    /**
     * Waits for buffer to become non-full. If the buffer has been resized to a
     * size less than the current number of events, the caller is blocked until
     * the current size of the buffer is smaller than the new maximum size.
     */
    public void awaitNotFull() {
    	synchronized (lock) {
      
    		try {
    			while (currSize >= maxSizeW)
    				lock.wait();
    		} catch (InterruptedException exc) {
    			error("BufferLegacy.awaitNotFull() interrupted");
    		}
    	}
    }

    /**
     * The current size of the buffer.
     * @see RTEventBuffer#maxSizeR
     * @return the number of objects currently enqueued
     */
    public int currentSize() {
    	synchronized (lock) {
    		return currSize;
    	}
    }

    /**
     * Removes all buffered events.
     */
    public void flush() {
    	synchronized (lock) {
    		
    		for (int i = 0; i < currSize; i++) {
    			buffer[fetchAt] = null;
    			fetchAt = (fetchAt + 1) % maxSizeR;
    		}
    		currSize = 0;
    		lock.notifyAll();
    	}
    }

    /**
     * Get all buffered events leaving the buffer empty. The event that was the 
     * next to be fetched is returned as the first element (index 0), followed 
     * by the other events in FIFO order with the most recently posted event as
     * the last element.
     *
     * @return array containing buffered events. An empty array (not null) is
     *         returned if the buffer was empty.
     */
    public RTEvent[] fetchAll() {
    	synchronized (lock) {
      
    		int len = currSize;
    		RTEvent[] ans = new RTEvent[len];
    		for (int i = 0; i < len; i++) {
    			ans[i] = buffer[fetchAt];
    			if (ans[i].owner != null)
    				ans[i].setOwner(/**/RTThread.currentThread());
    			buffer[fetchAt] = null; // For garbage collector
    			fetchAt = (fetchAt + 1) % maxSizeR;
    		}
    		lock.notifyAll(); // Empty (and Non-full) notification
    		currSize = 0;
    		return ans;
    	}
    }

    /**
     * Gets the maximum size of the buffer.
     * @return the maximum number of objects that can be enqueued
     */
    public int getMaxSize() {
    	synchronized (lock) {
    		return maxSizeW;
    	}
    }

    /**
     * Changes size of buffer.
     * @param newSize the new maximum number of objects that can be enqueued
     */
    public void setMaxSize(int newSize) {
    	synchronized (lock) {
      
    		// Prepare and allocate.
    		try {
    			while (resizing)
    				lock.wait(); // Permit only one at a time.
    		} catch (InterruptedException exc) {
    			error("BufferLegacy.setMaxSize(): Pending call interrupted");
    		}
      
    		if (newSize == maxSizeW)
    			return; // Size correct already.
    		resizing = true;
    		newSize = (newSize > 0 ? newSize : 1); // The size must be at least 1.
    		RTEvent[] newbuff = new RTEvent[newSize];

    		// Notify if there is space left in larger buffer.
    		if (currSize >= maxSizeR && currSize < newSize)
    			lock.notifyAll();
	
    		// Wait for array to be long enough.
    		if (currSize > newSize) {
    			// New space not enough for all present events, prevent posting
    			maxSizeW = newSize; // No writing while overfilled.
    			try {
    				while (currSize > newSize)
    					lock.wait();
    			} catch (InterruptedException exc) {
    				resizing = false;
    				maxSizeW = maxSizeR;
    				error("BufferLegacy.setMaxSize(): Interrupt during buffer resizing.");
    			}
    		}
	      
    		// OK, store the not-to-many events in new buffer.
    		for (int i = 0; i < currSize; i++) {
    			newbuff[i] = buffer[fetchAt];
    			fetchAt = (fetchAt + 1) % maxSizeR;
    		}
      
    		fetchAt = 0;
    		postAt = currSize;
    		maxSizeR = maxSizeW = newSize;
    		buffer = newbuff;

    		// Release pending calls to this method, and notify for other calls too.
    		resizing = false;
    		lock.notifyAll();
    	}
    }

  /**
   * Maximal queue size for Reading events. Temporarily, during
   * decrease of full buffer with <code>setMaxSize</code>, this
   * attribute can differ from the ordinary buffer size.
   * @see RTEventBuffer#maxSizeW
   */
  protected int maxSizeR;

  /**
   * Maximal queue size, which is also the size for Writing/posting.
   */
  protected int maxSizeW;

  /**
   * Marks that <code>setMaxSize</code> is being evaluated.
   * More than one call would screw-up that operation because it need to
   * be blocked until all messages fit into the new (possibly smaller) size.
   */
  protected boolean resizing;
}