package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;

/**
 * A thread safe basic queue/buffer for <code>RTEvent</code> objects.
 * Essentially a protected bounded buffer with both blocking and non-blocking
 * methods for posting and fetching messages in terms of references to objects.
 * <p>
 * For the functionality provided by this basic class, the buffer could have
 * been typed to contain <code>EventObject</code>s (the base class of RTEvent),
 * but to permit subclasses of <code>RTEventBuffer</code> to utilize the
 * properties of the RTEvents that class is used here. For instance, one might
 * need a buffer that keeps track of the total delay of data by referring to
 * the time-stamps of the newest and oldest RTEvent.
 * <p>
 * The methods checking or awaiting buffer full/empty are useful for operation
 * supervision, capturing events until a buffer is filled, mode changes, etc.
 * Do not misuse them for polled (concurrently bad) or busy-wait solutions.
 * <p>
 * To meet the efficiency demands of some applications, most methods
 * are final.  The class as such is, however, not final since
 * subclassing is expected for adding more features, hence the
 * protected visibility of fields. Additionally, the methods post and
 * fetch are virtual (i.e., not final). They by default call the
 * corresponding do-method and are typed accordingly, but they are
 * free for redefinition which can even result in synchronized being
 * omitted if mutual exclusion is ensured elsewise (in enclosing
 * monitor, etc.). See also the implemented interfaces, which are
 * facilitating access to only one end of the buffer.
 * <p>
 * While events are stored in the buffer, the buffer is set to be the owner
 * of the events, see the <code>RTEvent</code> class. Only events owned by
 * the current thread are accepted for posting to the buffer. When events
 * are fetched, the fetching thread is assigned to be the new owner. This
 * is to be ensured also by subclasses, thereby maintaining the detection
 * of event producers keeping references to objects already sent. In some
 * cases, however, it is desirable to resend (typically immutable) objects.
 * Therefore, events with the owner equal to <code>null</code> is always
 * accepted. The concept of owners is from the OSE operating system.
 * <p>
 * In practice and by default, all methods can be considered to be
 * synchronized, even if the methods as such are not declared to be
 * synchronized. Having the methods synchronized would agree with
 * common application programming practice, and together with the fact
 * that most methods are final, those synchronized methods would remain
 * synchronized in subclasses. However, synchronization is performed
 * on an internal lock object for the following reasons: <br>1) Other
 * object cannot interfere with the internal wait/notify by
 * synchronizing on an instance of this class (by using a synchronized
 * statement). <br>2) An external object (provided via the
 * constructor) can be used as the lock object, thereby permitting
 * synchronized use of several buffers/object.
 * <p>
 * If a blocked call is interrupted, an RTInterrupted (subclass of Error, not
 * an exception) is thrown, and that buffer object is most likely not useful
 * anymore.
 *
 * @see se.lth.cs.realtime.RTInterrupted
 */

public class BufferPlain extends RTEventBuffer {
	  
	  /**
	   * Initalizes the buffer with a size of <code>DEFAULT_SIZE</code>.
	   */
	  public BufferPlain() {
		  super();
	  }

	  /**
	   * Initalizes the queue (buffer) with a given maximal size. A size smaller
	   * than one results in a buffer with size one.
	   *
	   * @param maxSize maximum queue size
	   */
	  public BufferPlain(int maxSize) {
		  super(maxSize);
	  }

	  /**
	   * Initalizes the queue (buffer) with a given maximal size, and with a
	   * lock object to be used for synchronization.
	   *
	   * @param maxSize maximum queue size
	   * @param lock    external monitor/object to synchronized with
	   */
	  public BufferPlain(int maxSize, Synchronized.Lock lock) {
		  super(maxSize, lock);
	  }

	  /**
	   * Returns the next RTEvent in the queue, blocks if none is available.
	   *
	   * @return a non-null event.
	   */
	  public synchronized RTEvent doFetch() {
		  
		  try {
			  while (currSize == 0) {
				  wait();
			  }
		  } catch (InterruptedException e) {
			  error("BufferPlain.doFetch() interrupted");
			  return null;
		  }
			  
		  RTEvent e = buffer[fetchAt];
		  buffer[fetchAt] = null; // For garbage collector
		  fetchAt = (fetchAt + 1) % buffer.length;
		  currSize--;
		  
		  if (e.owner != null)
			  e.setOwner(/**/RTThread.currentThread());
			  
		  notifyAll();
		  return e;
	 }

	  /**
	   * Returns the next available RTEvent in the queue, or null if the
	   * queue is empty. In other words, this method always returns immediately,
	   * even if the queue is empty (non-blocking).
	   *
	   * @return an RTEvent if any available, null otherwise
	   */
	  public synchronized RTEvent tryFetch() {
		  
		  if (isEmpty()) {
			  return null;
		  } else {
			  
			  RTEvent e = buffer[fetchAt];
			  buffer[fetchAt] = null; // For garbage collector
			  fetchAt = (fetchAt + 1) % buffer.length;
			  currSize--;
			  
			  if (e.owner != null)
				  e.setOwner(/**/RTThread.currentThread());
			  
			  notifyAll();
				  
			  return e;
		}
	}

	  /**
	   * Get the next RTEvent in the queue, or block (up to the specified
	   * timeout time) if no event is available.
	   *
	   * @param timeout the maximum blocking time time in milli-seconds.
	   *
	   * @return an RTEvent if any available within the timeout time, null
	   *         otherwise.
	   */
	  public RTEvent tryFetch(long timeout) {
		  
		  long now;
		  long arrival = now = System.currentTimeMillis();
		  
		  synchronized (this) {
			  
			  RTEvent ans = null;
	      
			  try {
				  while (currSize == 0 && now < arrival + timeout) {
					  wait(timeout);
					  now = System.currentTimeMillis();
				  }
			  } catch (InterruptedException e) {
				  error("BufferPlaint.tryFetch(timeout) interrupted");
				  return ans;
			  }
			  
			  if (now < arrival + timeout) {
				  try {
					  ans = tryFetch();
				  } catch (RTInterrupted e) {
					  error("BufferPlain.tryFetch(timeout) interrupted");
				  }
			  }
			  
			  return ans;
		  }
	  }

	  /**
	   * Adds an RTEvent to the queue, blocks caller if the queue is full.
	   *
	   * @param e the event to enqueue
	   */
	  public void doPost(RTEvent e) {
		  
		  if (e.owner != null && e.owner != /**/RTThread.currentThread()) {
			  throw new OwnerError(e.owner);
		  }
		  
		  synchronized (this) {
			
			  try {
				  while (currSize >= maxSize) {
					  wait();
				  }
			  } catch (InterruptedException ex) {
				  error("BufferPlain.doPost(RTEvent) interrupted");
				  return;
			  }
		
			  if (e.owner != null)
				  e.setOwner(this);
			  buffer[postAt] = e;
			  postAt = (postAt + 1) % buffer.length;
			  currSize++;
			  
			  notifyAll();
			  return;
		  }
	  }

	  /**
	   * Adds an EventObject to the queue, without blocking if the queue is full.
	   *
	   * @param e the event to enqueue
	   * @return null when event posted to non-full buffer, the event otherwise
	   */
	  public RTEvent tryPost(RTEvent e) {
		  
		  if ( e.owner != null && e.owner != /**/RTThread.currentThread()) {
			  throw new OwnerError(e.owner);
		  }
	    
		  synchronized (this) {
	    
			  if (currSize >= buffer.length) {
				  return e;
			  }
			  else {
	    		
				  if (e.owner != null)
					  e.setOwner(this);
				  buffer[postAt] = e;
				  postAt = (postAt + 1) % buffer.length;    
				  currSize++;
				  
				  notifyAll();
				  return null;
			  }
		  }
	  }

	  /**
	   * Adds an RTEvent to the queue, blocks caller if the queue is full up to
	   * the specified timeout time.
	   *
	   * @param e       the event to enqueue
	   * @param timeout the maximum blocking time time in milli-seconds.
	   *
	   * @return null when event posted within the timeout time to non-full
	   *         buffer, the event otherwise.
	   */
	  public RTEvent tryPost(RTEvent e, long timeout) {
		  
		  long now;
		  long arrival = now = System.currentTimeMillis();
	    
		  synchronized (this) {
	      
			  RTEvent ans = e;
	      
			  try {
				  while (currSize >= buffer.length && now < arrival + timeout) {
					  wait(timeout);
					  now = System.currentTimeMillis();
				  }
	    	
				  if (now < arrival + timeout) {
					  ans = tryPost(e);
				  }
			  } catch (InterruptedException ex) {
				  error("BufferPlain.tryPost(timeout) interrupted");
			  }
	      
			  return ans;
		  }
	  }

	  /**
	   * Checks if buffer is empty.
	   *
	   * @return true if the queue is empty, false if it is non-empty
	   */
	  public synchronized boolean isEmpty() {
		  return currSize == 0;
	  }

	  /**
	   * Checks if buffer is full.
	   *
	   * @return true if the queue is full, false if it is non-full
	   */
	  public synchronized boolean isFull() {
		  return currSize >= buffer.length;
	  }

	  /**
	   * The current size of the buffer.
	   * @see RTCondEventBuffer#maxSizeR
	   *
	   * @return the number of objects currently enqueued
	   */
	  public synchronized int currentSize() {
		  return currSize;
	  }
	  
	  /**
	   * Gets the maximum size of the buffer.
	   * @return the maximum number of objects that can be enqueued
	   */
	  public synchronized int getMaxSize() {
		  return maxSize;
	  }
	  
	  /**
	   * Removes all buffered events.
	   */
	  public synchronized void flush() {
		  
		  for (int i = 0; i < currSize; i++) {
			  buffer[fetchAt] = null;
			  fetchAt = (fetchAt + 1) % buffer.length;
		  }
		  currSize = 0;
		  
		  notifyAll();
	  }

	  /**
	   * Get all buffered events leaving the buffer empty. The event that was the 
	   * next to be fetched is returned as the first element (index 0), followed 
	   * by the other events in FIFO order with the most recently posted event as
	   * the last element.
	   *
	   * @return array containing buffered events. An empty array (not null) is
	   *         returned if the buffer was empty.
	   */
	  public synchronized RTEvent[] fetchAll() {
		  
		  RTEvent[] ans = new RTEvent[currSize];
		  
		  for (int i = 0; i < currSize; i++) {
			  ans[i] = buffer[fetchAt];
			  if (ans[i].owner != null)
				  ans[i].setOwner(/**/RTThread.currentThread());
			  buffer[fetchAt] = null; // For garbage collector
			  fetchAt = (fetchAt + 1) % buffer.length;
		  }  
		  currSize = 0;
			
		  notifyAll();
		  return ans;
	  }
}
