/* $Id: ExceptionEvent.java,v 1.4 2005/08/15 13:24:44 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;

/**
 * Event to be used by a thread that has encountered an error
 * situation (typically when an Exception has been caught) and other
 * threads need to be informed (typically to handle recovery,
 * shutdown, backup control, etc.). Do recall that throwing and
 * catching <code>Throwable</code>s (<code>Exception</code>s or
 * <code>Error</code>s) is something done within one thread of
 * execution, involving operations on the stack of that thread. In
 * concurrent systems, however, there is often the need to propagate
 * error situation from one (malfunctioning) thread to another, hence
 * this class.
 * <p>
 * As of JDK1.4, <code>java.lang.Throwable</code> also provides a 
 * method <code>getCause</code> as in this class, so successive calls 
 * (until null is returned) should lead to the original fault.
 * 
 * @see License
 */

public class ExceptionEvent extends RTEvent {

  /**
   * The cause of the fault situation.
   */
  private final Throwable cause;

  /**
   * Constructs an <code>ExceptionEvent</code> object with the specified
   * source thread.
   * 
   * @param cause the error or exception that is the reason for this event.
   * @param source the thread where the event originated.
   */
  public ExceptionEvent(Throwable cause, Activity source) {
    super(source);
    this.cause = cause;
  }

  /**
   * Constructs an <code>ExceptionEvent</code> object with the specified
   * cause, as in class <code>Exception</code> from the JDK1.4.
   *
   * @param cause the error or exception that is the reason for this event.
   */
  public ExceptionEvent(Throwable cause) {
    this.cause = cause;
  }

  /**
   * Return the reason for this event, which in turn can be thrown
   * by the event-recieving thread as a way of transferring exception
   * handling from one thread to another.
   *
   * @return the cause of this event or null if the cause is nonexistent
   *         or unknown.
   */
  public final Throwable getCause() {
    return cause;
  }
}
