/* $Id: PeriodicEvent.java,v 1.6 2005/08/19 09:37:16 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.Timebase;

/**
 * Generic event used to trigger periodic activities in an event-driven 
 * fashion, as an alternative to using <code>PeriodicThread</code> which 
 * is time driven. Both RTThread and other subclasses are candidates for
 * being triggered by <code>PeriodicEvent</code>s, whenever based on 
 * externally triggered cyclic behavior. Some timer or periodic thread is 
 * needed for the generation of periodic events.
 * <p>
 * A <code>PeriodicEvent</code> has no owner. That is, <code>owner</code>
 * is set to <code>null</code>, enabling the event to be sent over and over 
 * again from the same source. Subclasses are, however, free to assign 
 * ownership simply by assigning the (protected) <code>owner</code> field.
 *
 * @see License
 * @see se.lth.cs.realtime.PeriodicThread
 * @see java.util.Timer
 * @see RTEvent
 */

public class PeriodicEvent extends RTEvent {

  /**
   * Constructs an <code>PeriodicEvent</code> object with the current
   * thread as source and a time stamp from the current system time.
   */
  public PeriodicEvent() {
    super();
    owner = null;
  }

  /**
   * Constructs an <code>PeriodicEvent</code> object with the specified
   * source object and a time stamp from the current system time. 
   *
   * @param source The object where the event originated.
   */
  public PeriodicEvent(Object source) {
    super(source);
  }

  /**
   * Constructs an <code>PeriodicEvent</code> object with the current
   * thread as source and with the specified time stamp. 
   *
   * @param ts The time from which the event originates.
   */
  public PeriodicEvent(long ts) {
    super(Thread.currentThread(), ts);
  }

  /**
   * Constructs an <code>PeriodicEvent</code> object with the specified
   * source object and time stamp. 
   *
   * @param source The object where the event originated.
   * @param ts The time from which the event originates.
   */
  public PeriodicEvent(Object source, long ts) {
    super(source, ts);
  }

  /**
  * Constructs an <code>PeriodicEvent</code> object with the specified
  * source object and time-stamp with respect to the specified time-base. 
  *
  * @param source The object where the event originated.
  * @param ts The time from which the event originates.
  */
  public PeriodicEvent(Object source, long ts, Timebase timebase) {
    super(source, ts, timebase);
  }

}
