/* $Id: RTEvent.java,v 1.2 2005/08/22 08:18:22 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import java.util.EventObject;
import se.lth.cs.realtime.Timebase;
import se.lth.cs.realtime.RTSystem;
import se.lth.cs.realtime.ActivityImpl;
/*//*/
import se.lth.cs.realtime.RTThread;

/**
 * The base class of real-time/time-stamped events.
 * Every <code>RTEvent</code> comes from a source object referred to
 * in the <code>EventObject</code> base class. Additional IDs, such as
 * sequential numbers or the kind of event within a certain type, are
 * expected to be defined in subclasses together with other user
 * attributes. The time stamp is, however, fundamental to real-time
 * systems as a mean to denote the freshness of the data. It is
 * therefore an attribute here and the motivation for this class.
 * <p>
 * The source object, often referring to another thread, is not supposed
 * to keep any reference to the event after posting it. Therefore, methods
 * of this class should not need to be synchronized. To support detection of
 * kept references, for instance when the event is sent again, there is a
 * field <code>owner</code> which is set to the same as <code>source</code>
 * when the event is created, if the <code>source</code> is a thread. If 
 * the <code>source</code> is not a thread, the <code>owner</code> is 
 * set to be the currently executing (real-time of Java) thread. Since 
 * ownership is meaningful within a program only, it is <code>transient</code>
 * and an <code>Activity</code> that is not a thread should not be owner. 
 * Specifically, if the <code>source</code> is an instance of 
 * <code>ActivityImpl</code>, the represented thread object is set to be
 * the <code>owner</code>.
 * <p>
 * When an event is transferred, for instance via
 * a buffer, the ownership changes to the buffer and then to the receiving
 * thread. Using available classes, the user should not need to care about
 * the ownership, it is checked to detect certain programming errors.
 * Sometimes, however, it is desirable to keep a set of event objects that
 * are sent several times, for instance every time a certain state transition
 * occurs. In such cases the specific event subclasses are supposed to set
 * the (protected) owner field to <code>null</code>, which disables the
 * checking of ownership. As way of doing that, creating an <code>RTEvent</code>
 * with <code>source</code> specified to be <code>null</code> results in 
 * <code>source</code> being set to <code>RTThread.currentThread()</code>
 * and <code>owner</code> is set to <code>null</code>.
 * <p>
 * There is a <code>ticker</code> attribute referring to a time-base,
 * which can be given to the constructor. This is for special computers
 * or special applications; if you do not know that you need it, simply
 * ignore it and the default <code>System.currentTimeMillis()</code>
 * will be used as the actual ticker. However, obtaining the time via
 * <code>getMillis</code> or <code>getSeconds</code> will make your code
 * portable to such special applications where the ticker is used (e.g. 
 * to run an extra instance of a model-based control system ahead of time
 * to draw conclusions about optimal control actions).
 * 
 * @see License
 */

public class RTEvent extends EventObject {

  // ----------------------------------------------------------- CONSTRUCTOR

  /**
   * Constructs an <code>RTEvent</code> object with the current
   * thread as source and a time stamp from the current system time.
   */
  public RTEvent() {
    this(RTThread.currentThread(), RTSystem.currentTimeMillis());
  }

  /**
   * Constructs an <code>RTEvent</code> object with the specified
   * source object and a time stamp from the current system time.
   *
   * @param source The object where the event originated.
   */
  public RTEvent(Object source) {
    this(source, RTSystem.currentTimeMillis());
  }

  /**
   * Constructs an <code>RTEvent</code> object with the current
   * thread as source and with the specified time stamp.
   *
   * @param ts The time from which the event originates.
   */
  public RTEvent(long ts) {
    this(RTThread.currentThread(), ts);
  }

  /**
   * Constructs an <code>RTEvent</code> object with the specified
   * source object and time stamp.
   *
   * @param source The object where the event originated.
   * @param ts The time from which the event originates.
   */
  public RTEvent(Object source, long ts) {
    super(source == null ? RTThread.currentThread() : source);

    //
    owner = source;
    //
    
    /*
    if (source instanceof ActivityImpl) {
      owner = ((ActivityImpl)source).getThread();
    } else {
      owner = source; // Can still be null, but this.source is not null
    }
    */

    timestamp = ts;
  }

  /**
  * Constructs an <code>RTEvent</code> object with the specified
  * source object and time-stamp with respect to the specified time-base.
  *
  * @param source The object where the event originated.
  * @param ts The time from which the event originates.
  * @param timebase The base of time in which the time stamp is expressed.
  */
  public RTEvent(Object source, long ts, Timebase timebase) {
    this(source, ts);
    ticker = timebase;
  }

  // -------------------------------------------------------- PUBLIC METHODS

  /**
   * Obtain current owner. Mainly useful for debugging.
   *
   * @return the object which is currently the owner of this event.
   */
  public final Object getOwner() {
    return owner;
  }

  /**
   * Returns the event's time stamp. Use the other get-time methods to make
   * the management of time portable, e.g. <code>getSeconds()</code>.
   *
   * @return the time in number of (system dependent) ticks.
   */
  public long getTicks() {
    return timestamp;
  }

  /**
   * Returns the creation-time (or supplied time-stamp= of this event.
   *
   * @return the time-stamp expressed in seconds.
   */
  public double getSeconds() {
    if (ticker == null) {
      return (double) timestamp * 0.001;
    }
    else {
      return ticker.seconds(timestamp);
    }
  }

  /**
   * Returns the event's time stamp in number of milli-seconds.
   * The returned value is compliant with
   * <code>System.currentTimeMillis()</code>.
   *
   * @return the time-stamp expressed in milli-seconds.
   */
  public long getMillis() {
    if (ticker == null) {
      return timestamp;
    }
    else {
      return ticker.millis(timestamp);
    }
  }

  /**
   * Return the sub-milli part of the time-stamp, if any.
   * If permitted by numerical accuracy, using <code>getSeconds</code>
   * usually is a better alternative when working with times that contain
   * fractions of milli-seconds.
   *
   * @return any nano-second part of time-stamp.
   */
  public long getNanos() {
    if (ticker == null) {
      return 0;
    }
    else {
      return ticker.nanos(timestamp);
    }
  }

  /**
   * Obtains the time-base relating (possibly locally counted) ticks or
   * seconds to real or simulated time.
   *
   * @return the time-base, or <code>null</code> if
   *         <code>System.currentTimeMillis</code> is used.
   */
  public Timebase getTimebase() {
    return ticker;
  }

  /**
   * Gives a string representation of the event object.
   * <p>
   * The string is of the form:<p>
   * <i>ClassName</i>[<i>paramString</i>]<p>
   * where <i>ClassName</i> is the name of the event class,
   * <i>paramString</i> is the string returned by <code>paramString</code>.
   */
  public String toString() {
    return //getClass().getName() 
	    "RTEvent " 
	    + "[" + paramString() + "]";
  }

  // ----------------------------------------------------- PROTECTED METHODS

  /**
   * Return a string representation of the attributes/state of this object.
   * Useful for debugging purposes via <i>toString</i>. Subclasses may
   * simply use <i>super.paramString</i> and add its additional attributes.
   */
  protected String paramString() {
    return source + ", " + Long.toString(timestamp);
  }

  // ------------------------------------------------------- PACKAGE METHODS

  /**
   * Change owner, permitted within package.
   *
   * @param owner the object to take over ownership of this event.
   */
  final void setOwner(Object owner) {
    this.owner = owner;
  }

  // ------------------------------------------ PRIVATE/PROTECTED ATTRIBUTES

  /**
   * The time stamp of this event, in number of (system dependent)
   * ticks.
   */
  protected long timestamp;

  /**
   * The time-base relating number of ticks to real time.
   */
  protected Timebase ticker;

  /**
   * The current owner of the event, equal to the source when the object
   * is created. As with <code>source</code> (inherited from
   * <code>EventObject</code>, this attribute is declared
   * <code>transient</code> since it has no meaning outside this
   * JVM or program, and it is <code>volatile</code> to ensure
   * concurrently correct ownership despite unsynchronized
   * <code>setOwner</code> and <code>getOwner</code> methods.
   */
  protected volatile transient Object owner;
}
