/* $Id: RTEventBuffer.java,v 1.3 2005/09/12 10:03:09 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;

/**
 * A thread safe basic queue/buffer for <code>RTEvent</code> objects.
 * Essentially a protected bounded buffer with both blocking and non-blocking
 * methods for posting and fetching messages in terms of references to objects.
 * <p>
 * For the functionality provided by this basic class, the buffer could have
 * been typed to contain <code>EventObject</code>s (the base class of RTEvent),
 * but to permit subclasses of <code>RTEventBuffer</code> to utilize the
 * properties of the RTEvents that class is used here. For instance, one might
 * need a buffer that keeps track of the total delay of data by referring to
 * the time-stamps of the newest and oldest RTEvent.
 * <p>
 * The methods checking or awaiting buffer full/empty are useful for operation
 * supervision, capturing events until a buffer is filled, mode changes, etc.
 * Do not misuse them for polled (concurrently bad) or busy-wait solutions.
 * <p>
 * To meet the efficiency demands of some applications, most methods
 * are final.  The class as such is, however, not final since
 * subclassing is expected for adding more features, hence the
 * protected visibility of fields. Additionally, the methods post and
 * fetch are virtual (i.e., not final). They by default call the
 * corresponding do-method and are typed accordingly, but they are
 * free for redefinition which can even result in synchronized being
 * omitted if mutual exclusion is ensured elsewhere (in enclosing
 * monitor, etc.). See also the implemented interfaces, which are
 * facilitating access to only one end of the buffer.
 * <p>
 * While events are stored in the buffer, the buffer is set to be the owner
 * of the events, see the <code>RTEvent</code> class. Only events owned by
 * the current thread are accepted for posting to the buffer. When events
 * are fetched, the fetching thread is assigned to be the new owner. This
 * is to be ensured also by subclasses, thereby maintaining the detection
 * of event producers keeping references to objects already sent. In some
 * cases, however, it is desirable to resend (typically immutable) objects.
 * Therefore, events with the owner equal to <code>null</code> is always
 * accepted. The concept of owners is from the OSE operating system.
 * <p>
 * In practice and by default, all methods can be considered being
 * synchronized, even if the methods as such are not declared to be
 * synchronized. Having the methods synchronized would agree with
 * common application programming practice, and together with the fact
 * that most methods are final, those synchronized methods would remain
 * synchronized is subclasses. However, synchronization is performed
 * on an internal lock object for the following reasons: <br>1) Other
 * object cannot interfere with the internal wait/notify by
 * synchronizing on an instance of this class (by using a synchronized
 * statement). <br>2) An external object (provided via the
 * constructor) can be used as the lock object, thereby permitting
 * synchronized use of several buffers/object.
 * <p>
 * If a blocked call is interrupted, an RTInterrupted (subclass of Error, not
 * an exception) is thrown, and that buffer object is most likely not useful
 * anymore.
 *
 * @see License
 * @see se.lth.cs.realtime.RTInterrupted
 */

public abstract class RTEventBuffer 
implements RTEventBufferInput, RTEventBufferOutput, Synchronized {

  public final static int DEFAULT_SIZE = RTSystem.getDefaultEventBufferSize();

  
  // ------------------------------------------------------------ CONSTRUCTORS

  /**
   * Initializes the buffer with a size of <code>DEFAULT_SIZE</code>.
   */
  public RTEventBuffer() {
    this(DEFAULT_SIZE);
  }

  /**
   * Initializes the queue (buffer) with a given maximal size. A size smaller
   * than one results in a buffer with size one.
   *
   * @param maxSize maximum queue size
   */
  public RTEventBuffer(int maxSize) {
    this.maxSize = (maxSize > 0 ? maxSize : 1);
    buffer = new RTEvent[maxSize];
  }

  /**
   * Initializes the queue (buffer) with a given maximal size, and with a
   * lock object to be used for synchronization.
   *
   * @param maxSize maximum queue size
   * @param lock    external monitor/object to synchronized with
   */
  // Temp. removal of constructor containing Synchronized.Lock parameter !!
  public RTEventBuffer(int maxSize, Object lock) {
  //public RTEventBuffer(int maxSize, Synchronized.Lock lock) {
    this.maxSize = (maxSize > 0 ? maxSize : 1);
    buffer = new RTEvent[maxSize];
    this.lock = lock;
  }

  // ----------------------------- METHODS IMPLEMENTING RTEventBufferOutput:

  /**
   * Returns the next event object in the queue, by default by calling
   * <code>doFetch</code> which is synchronized (opposed to this
   * method) via an internal lock.  Subclasses overriding this method
   * should be very careful to maintain mutual exclusion, normally by
   * simply declaring <code>fetch</code> as <code>synchronized</code>
   * or by synchronizing on the <code>lock</code> attribute.
   *
   * @return the first event in queue.
   */
  public RTEvent fetch() {
    return doFetch();
  }

  /**
   * Returns the next RTEvent in the queue, blocks if none available.
   *
   * @return a non-null event.
   */
  public abstract RTEvent doFetch();
  
  /**
   * Returns the next available RTEvent in the queue, or null if the
   * queue is empty. In other words, this method always returns immediately,
   * even if the queue is empty (non-blocking).
   *
   * @return an RTEvent if any available, null otherwise
   */
  public abstract RTEvent tryFetch();
  
  /**
   * Get the next RTEvent in the queue, or block (up to the specified
   * timeout time) if no event is available.
   *
   * @param timeout the maximum blocking time time in milli-seconds.
   *
   * @return an RTEvent if any available within the timeout time, null
   *         otherwise.
   */
  public abstract RTEvent tryFetch(long timeout);

  
  // ------------------------------ METHODS IMPLEMENTING RTEventBufferInput:

  /**
   * Put the supplied event object last in the queue, by default by calling
   * <code>doPost</code> which is synchronized (opposed to this method).
   * Subclasses overriding this method should be very careful to maintain
   * mutual exclusion, normally by simply declaring <code>post</code>
   * synchronized or by locking some enclosing monitor/resource.
   *
   * The return value is not useful for the base-class implementation since
   * there is nothing to return from calling <code>doPost</code>. Therefore,
   * <code>null</code> is always returned. The intention of the return
   * value/type is to facilitate the use of <code>tryPost</code> or other
   * (for instance GC-saving) subclass implementations.
   *
   * @return null (or possibly some event according to subclass, if any).
   */
  public RTEvent post(RTEvent e) {
    doPost(e);
    return null;
  }

  /**
   * Adds an RTEvent to the queue, blocks caller if the queue is full.
   *
   * @param e the event to enqueue
   */
  public abstract void doPost(RTEvent e);
  
  /**
   * Adds an EventObject to the queue, without blocking if the queue is full.
   *
   * @param e the event to enqueue
   *
   * @return null when event posted to non-full buffer, the event otherwise
   */
  public abstract RTEvent tryPost(RTEvent e);
  
  /**
   * Adds an RTEvent to the queue, blocks caller if the queue is full up to
   * the specified timeout time.
   *
   * @param e       the event to enqueue
   * @param timeout the maximum blocking time time in milli-seconds.
   *
   * @return null when event posted within the timeout time to non-full
   *         buffer, the event otherwise.
   */
  public abstract RTEvent tryPost(RTEvent e, long timeout);

  
  // ------------------------------------------------ OTHER PUBLIC METHODS:

  /**
   * Checks if buffer is empty.
   *
   * @return true if the queue is empty, false if it is non-empty
   */
  public abstract boolean isEmpty();
  
  /**
   * Checks if buffer is full.
   *
   * @return true if the queue is full, false if it is non-full
   */
  public abstract boolean isFull();
  
  /**
   * The current size of the buffer.
   * @see RTEventBuffer#maxSizeR
   *
   * @return the number of objects currently enqueued
   */
  public abstract int currentSize();
  
  /**
   * Gets the maximum size of the buffer.
   *
   * @return the maximum number of objects that can be enqueued
   */
  public abstract int getMaxSize();
  
  /**
   * Removes all buffered events.
   */
  public abstract void flush();

  /**
   * Get all buffered events leaving the buffer empty. The event that was the 
   * next to be fetched is returned as the first element (index 0), followed 
   * by the other events in FIFO order with the most recently posted event as
   * the last element.
   *
   * @return array containing buffered events. An empty array (not null) is
   *         returned if the buffer was empty.
   */
  public abstract RTEvent[] fetchAll();

  
  // ---------------------------------------------------- PROTECTED METHODS:

  /**
   * Takes care of <code>InterruptedException</code>.
   */
  protected void error(String message) {
    // Use exception handling object here, which in turn should affect
    // the calling thread.

    throw new RTInterrupted(message);
  }

  // ------------------------------------------------- PROTECTED ATTRIBUTES:

  /**
   * The actual buffer.
   */
  protected RTEvent[] buffer;

  /**
   * Maximal buffer size.
   */
  protected int maxSize;

  /**
   * Current queue size. That is, the number of events currently buffered.
   */
  protected int currSize;

  /**
   * Free position to insert events at (the successor of last element)
   */
  protected int postAt;

  /**
   * Position of first element to fetch, if any.
   */
  protected int fetchAt;

  /**
   * The lock to this monitor.
   */
  // Temp. removal of Synchronized.Lock !! 
  protected Object lock = new Object();
  //protected Synchronized.Lock lock = new Synchronized.Lock(); // Can be replaced in constructor.
}
