/* $Id: RendezvousEvent.java,v 1.2 2005/08/22 08:18:22 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.Condition;
import se.lth.cs.realtime.RTSystem;
import se.lth.cs.realtime.RTThread;

/**
 * Support rendezvous between an event producer and an event consumer.
 * 
 * @see License
 */

public class RendezvousEvent extends RTEvent {

  /**
   * Create empty event, without shared data and without work to be done, 
   * which can be useful to only synchronize execution.
   */
  public RendezvousEvent() {
    super();
  }

  /**
   * The work to be carried out by the tread accepting the rendezvous. 
   * To be defined in subclasses; by default nothing. Data to be provided 
   * from the client side can be defined as attributes in the subclasses, 
   * e.g. by a reference to a shared-data object. (It it not good a good 
   * practise to use the event source for calling methods of the client 
   * thread object; store shared data in passive objects).
   * 
   * @param arg to be provided by the accepting thread, as an argument to
   *            the <code>accept</code> method, thereby giving access 
   *            data in the context of the accepting thread.
   */
  protected void call(Object arg) {}
  
  /**
   * Enter the rendezvous where blocking should occur while the accepting
   * thread carries out all transfers involving shared data.
   * 
   * @param worker the thread that is expected to accept the rendezvous.
   */
public final synchronized void entry(RTEventListener worker) {
    timestamp = RTSystem.currentTimeMillis();
    owner = source = RTThread.currentThread();
    worker.putEvent(this);
    Condition.await(this);
  }
  
  /**
   * Accept the rendezvous and call <code>call</code> to carry out any work
   * using the shared data as provided in subclasses of this class.
   *  
   * @param arg additional local data provided by the accepting thread.
   */
public final synchronized void accept(Object arg) {
    if (owner != RTThread.currentThread()) throw new OwnerError(owner);
    owner = source; // Permit reuse; can be prevented by assigning owner in call
    this.call(arg);
    Condition.signal(this);
  }
}
