/* $Id: BinarySem.java,v 1.2 2005/08/22 08:45:17 d02en Exp $          
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.semaphore;

import se.lth.cs.realtime.RTSystem;

/**
 * A Semaphore containing a boolean flag instead of a counter, to
 * support signaling in cases when waiting threads should <b>not</b>
 * catch up with successive missed calls of <code>give</code>. Another
 * reason to use this class, mainly applicable to small compiled
 * applications, is efficiency when this type of semaphore is the
 * simplest and most optimized one for the particular type of OS/kernel. 
 * Instead of a boolean flag, an implementation is free to use equivalent
 * solutions such as a counter with a maximum value of one, as
 * natively defined in some operating systems such as Windows.
 * 
 * <p>The order of the awakenings of the threads which called
 * <code>take</code> depends on the underlying OS and/or VM, but for
 * standard implementation in pure Java it will be the same as for the
 * wait/notify mechanism.
 *
 * @see License
 * @see se.lth.cs.realtime.semaphore.Semaphore
 * @see se.lth.cs.realtime.semaphore.MutexSem
 * @see se.lth.cs.realtime.semaphore.CountingSem
 * @see se.lth.cs.realtime.semaphore.MultistepSem 
 * @see se.lth.cs.realtime.semaphore.SemInterrupted
 */

public final class BinarySem implements Semaphore {

	/**
     * Object used by the pure Java implementation for blocking and to make methods atomic. 
     * Used instead of <code>this</code> object to prevent misuse of notify.
     */
    private Object queue;

    /**
     * The one-bit representation of this semaphore. Used by the pure Java implementation.
     */
    private boolean access;
    
    /**
     * Reference to the semaphore of the current RTOS, if semaphores are provided natively.
     * Implemented by a built-in type to hide it from the GC.
     */
    private long target;
    
    /**
     * Creates a new <code>BinarySem</code> with
     * the internal (boolean) flag set to false.
     */
    public BinarySem() {
    	if (RTSystem.NATIVE_SEM) {
    		target = createBinarySem(0);
    	} else {
    		queue = new Object();
    	}
    }   
    
    /**
     * Creates a new <code>BinarySem</code> with
     * the internal (boolean) flag set to
     * the specified value.
     *
     * @param access the initial value of the bit
     *		representing this semaphore, with <code>true</code> 
     *      meaning that the bit is set and the semaphore is 
     *      free to be taken..
     */ 
    public BinarySem(boolean access) {
    	if (RTSystem.NATIVE_SEM) {
    		target = createBinarySem(access ? 1 : 0);
    	} else {
    		queue = new Object();
    		this.access = access;
    	}
    }
    
    /**
     * Creates a new <code>BinarySem</code> with the boolean
     * initialized in a way compatible with the constructors of
     * <code>CountingSem</code>.
     *
     * @param count the initial value of the boolean it is set to
     * true if the argument is greater than zero, false otherwise.
     */
    public BinarySem(int count) {
    	if (RTSystem.NATIVE_SEM) {
    		target = createBinarySem(count);
    	} else {
    		queue = new Object();
    		this.access = count>0;
    	}
    }
    
    /**
     * Sets the internal flag to true and releases the first-in-queue
     * blocked thread, if any.
     */
    public void give() {
    	if (RTSystem.NATIVE_SEM) {
    		giveBinarySem(target);
    	} else {
    		synchronized (queue) {
    		    access = true;
    		    queue.notify();
    		}
    	}
    }
    
    /**
     * Causes the calling thread to block until the boolean flag
     * represented by this semaphore is set to true. On exiting, the
     * flag is set to false, telling that the semaphore has been
     * taken.
     * 
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. The caller is <b>not</b> expected to catch it since
     * it indicated severe violation in the handling of resources,
     * concurrency, or time. Thus, you should not catch such an error,
     * except for special cases like handling an emergency stop.
     */
    public void take() {
    	if (RTSystem.NATIVE_SEM) {
    		takeBinarySem(target);
    	} else {
    		try {
    		    synchronized (queue) {
    		    	while (!access) queue.wait();
    		    }
    		    access = false;
    		} catch (InterruptedException exc) {
		    // HERE BE DRAGONS! Include me again
		    System.out.println("BinarySem.take interrupted");
    		    //throw new SemInterrupted("BinarySem.take interrupted: " + exc);
    		}	
    	}
    }
    
    /* NOT NATIVELY IMPLEMENTED YET */
    /** 
     * Try to take the semaphore but give up after a certain time. Be
     * careful not to neglect the returned value, which could result in
     * concurrency faults.
     * 
     * @return true if the semaphore was successfully taken, and
     * false if the timeout occurred which means that the semaphore was
     * not taken.
     *
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. The caller is <b>not</b> expected to catch it since
     * it indicated severe violation in the handling of resources,
     * concurrency, or time. Thus, you should not catch such an error,
     * except for special cases like handling an emergency stop.
     */
    public boolean tryTake(long timeout) {
		if (RTSystem.NATIVE_SEM) {
			return takeBinarySem(target, timeout);
		} else {
			long now;
			long arrival = now = System.currentTimeMillis();
			try {
				synchronized (queue) {
					while (!access && now < arrival + timeout) {
						queue.wait(timeout - (now - arrival));
						now = System.currentTimeMillis();
					}
				}
				if (now >= arrival + timeout)
					return false;
				else {
					access = false;
					return true;
				}
			} catch (InterruptedException exc) {
				// HERE BE DRAGONS! Include me again
				System.out.println("BinarySem.tryTake interupted");
				//throw new SemInterrupted("BinarySem.tryTake interrupted: "+ exc);
			}
		}
	}
    
    private native long createBinarySem(int iValue);
    private native void giveBinarySem(long target);
    private native void takeBinarySem(long target);
    private native boolean takeBinarySem(long target, long timeout);
}
