/* $Id: CountingSem.java,v 1.10 2005/08/19 09:37:16 d02en Exp $          
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.semaphore;

import se.lth.cs.realtime.RTSystem;

/**
 * Basic type of semaphore containing a counter which is incremented
 * by one fore each call to <code>give</code>, and decremented by one
 * for each returning call of <code>take</code>. A counter value of
 * zero prevents the caller of <code>take</code> to return by blocking
 * it, until some other thread performs a <code>give</code> and the
 * blocked thread is the first one in the waiting queue associated
 * with the semaphore.
 *
 * <p>The order of the awakenings of the threads which called
 * <code>take</code> depends on the underlying OS and/or VM, but for
 * standard implementation in pure Java it will be the same as for the
 * wait/notify mechanism.
 *
 * <p>The intended use of this class is signaling. By means of the
 * semaphore methods, mutual exclusion can also be achieved. However,
 * that is not the intended usage since (native) implementations of
 * this class are <em>not required to provide any priority
 * inheritance</em>.
 *
 * @see License
 * @see se.lth.cs.realtime.semaphore.Semaphore
 * @see se.lth.cs.realtime.semaphore.BinarySem
 * @see se.lth.cs.realtime.semaphore.MutexSem
 * @see se.lth.cs.realtime.semaphore.MultistepSem
 * @see se.lth.cs.realtime.semaphore.SemInterrupted
 */

public final class CountingSem implements Semaphore {

	/**
     * Object used by the pure Java implementation for blocking and to make methods atomic.
     * Used instead of <code>this</code> object to prevent misuse of notify.
     */
//     private Object queue;

    /**
     * The integer representing the value of this semaphore. 
     * Used by the pure Java implementation.
     */
//     private int count;

    /**
     * Reference to the semaphore of the current RTOS, if semaphores are provided natively.
     * Implemented by a built-in type to hide it from the GC.
     */
    private int target;

    /**
     * Creates a new <code>CountingSem</code> with the internal counter
     * initialized to 0.
     */
    public CountingSem() {
//     	if (RTSystem.NATIVE_SEM) {
    		target = createCountingSem(0);
//     	} else {
//     		queue = new Object();
//     	}
    }

    /**
     * Creates a new <code>CountingSem</code>, initializing it to the
     * specified (possibly negative) number.
     *
     * @param init the initial value of the number represented by
     * this semaphore.
     */
    public CountingSem(int init) {
//     	if (RTSystem.NATIVE_SEM) {
    		target = createCountingSem(init);
//     	} else {
//     		queue = new Object();
//     		count = init;
//     	}
    }

    /**
     * Increments the internal counter by one and releases the
     * first-in-queue blocked thread, if any.
     */
    public void give() {
//     	if (RTSystem.NATIVE_SEM) {
    		giveCountingSem(target);
//     	} else {
//     		synchronized (queue) {
//     		    ++count;
//     		    queue.notify();
//     		}
//     	}
        
    }

    /**
     * Causes the calling thread to wait until the counter
     * represented by this semaphore assumes a positive value.
     * Before returning, this number is decremented by 1.
     * 
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. The caller is <b>not</b> expected to catch it since
     * it indicated severe violation in the handling of resources,
     * concurrency, or time. Thus, you should not catch such an error,
     * except for special cases like handling the termination of the
     * application, since continuing execution after an interrupted
     * resource allocation does normally result in the application
     * <b>not</b> being concurrently correct.
     */
    public void take() {
//     	if (RTSystem.NATIVE_SEM) {
    		takeCountingSem(target);
//     	} else {
//     		try {
//     		    synchronized (queue) {
//     		    	while (count < 1) queue.wait();
//     		    	--count;
//     		    }
//     		} catch (InterruptedException exc) {
// 		    // HERE BE DRAGONS! Include me again
// 		    System.out.println("CountingSem.take interrupted");
//     		    //throw new SemInterrupted("CountingSem.take interrupted: " + exc);
//     		}
//     	}
        
    }

    /* NOT NATIVELY IMPLEMENTED YET */
    /** 
     * Try to take the semaphore but give up waiting after the timeout
     * number of milliseconds.
     * 
     * @argument timeout specifies the time after which the calling
     * thread should be unblocked and return even if the semaphore was
     * not taken. Timeout values less than or equal to zero is legal
     * and results in a wait-free call.
     *
     * @return true if the semaphore was successfully taken, and
     * false if the timeout occurred which means that the semaphore was
     * not taken.
     *
     * @throws SemInterrupted as when a call of <code>take</code> is
     * interrupted.
     */
    public boolean tryTake(long timeout) {
		// FIXME: Not implemented!
		return false;
// 		if (RTSystem.NATIVE_SEM) {
			//return takeCountingSem(target, timeout);
// 		} else {
// 			long now;
// 			long arrival = now = System.currentTimeMillis();
// 			try {
// 				synchronized (queue) {
// 					while (count < 1 && now < arrival + timeout) {
// 						queue.wait(timeout - (now - arrival));
// 						now = System.currentTimeMillis();
// 					}
// 					if (now >= arrival + timeout)
// 						return false;
// 					else {
// 						--count;
// 						return true;
// 					}
// 				}
// 			} catch (InterruptedException exc) {
// 				// HERE BE DRAGONS! Include me again
// 				System.out.println("CountingSem.tryTake interrupted");
// 				// throw new SemInterrupted("CountingSem.tryTake interrupted: "+ exc);
// 			}
// 		}
	}
	
    private native int createCountingSem(int iValue);
    private native void giveCountingSem(int target);
    private native void takeCountingSem(int target);
    //private native boolean takeCountingSem(int target, long timeout);
}
