/* $Id: MultistepSem.java,v 1.8 2005/08/19 09:37:16 d02en Exp $          
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.semaphore;
 
import se.lth.cs.realtime.RTSystem;

/**
 * NOTE1: This class currently lacks native support, so you might not 
 * be able to run a compiled version on your target HW/RTOS.
 * <p>
 * NOTE2: In the next release, this class will be renamed to QuantitySem
 * and extended with a maximum value of the internal counter.
 * <p>
 * Counting semaphore permitting increments and decrements that are
 * greater than one. Apart from the cases when the ordinary
 * <code>CountingSem</code> can be used, a <code>MultistepSem</code>
 * is useful also in the following two situations: <ol>
 *
 * <li>Multiple resources need to be atomically allocated. For
 * instance, requesting exclusive use of two hardware ports can be
 * accomplished in one operation, but calling <code>take</code>twice
 * could result in a deadlock.
 *
 * <li> Several threads that have been blocked by calling
 * <code>take</code> all need to be released by another signaling
 * thread. Not to require that the signaling thread
 * need to know the number of waiting threads, the method <giveAll>
 * simply releases them all.
 *
 * </ol><p>The order of the awakenings of the threads which called
 * <code>take</code> depends on the underlying OS and/or VM, but for
 * standard implementation in pure Java it will be the same as for the
 * wait/notify mechanism.
 *
 * @see License
 * @see se.lth.cs.realtime.semaphore.Semaphore
 * @see se.lth.cs.realtime.semaphore.CountingSem
 * @see se.lth.cs.realtime.semaphore.BinarySem
 * @see se.lth.cs.realtime.semaphore.MutexSem 
 */

public final class MultistepSem implements Semaphore {

	/**
     * Object used by the pure Java implementation for blocking and to make methods atomic.
     * Used instead of <code>this</code> object to prevent misuse of notify.
     */
    private Object queue;

    /**
     * Counter for the number of blocked threads, i.e., the queue length.
     */
    private int inqueue;

    /**
     * The counter representing the value of this semaphore.
     * Used by the pure Java implementation.
     */
    private int count;

    /**
     * Reference to the semaphore of the current RTOS, if semaphores are provided natively.
     * Implemented by a built-in type to hide it from the GC.
     */
    private long target;

    /**
     * Creates a new <code>MultistepSem</code> with the internal counter
     * initialized to 0.
     */
    public MultistepSem() {
    	if (RTSystem.NATIVE_SEM) {
    		throw new UnsupportedOperationException("Native impl missing!");
    	} else {
    		queue = new Object();
    	}
    }

    /**
     * Creates a new <code>MultistepSem</code>, initializing it to the
     * specified (possibly negative) number.
     *
     * @param init the initial value of the number represented by
     * this semaphore.
     */
    public MultistepSem(int init) {
    	if (RTSystem.NATIVE_SEM) {
    		throw new UnsupportedOperationException("Native impl missing!");
    	} else {
    		queue = new Object();
    		count = init;
    	}
    }

    /**
     * Increments the internal counter by the supplied number of steps
     * and releases that many blocked thread, if any. This is
     * like an atomic multiple <code>give()</code>.
     *
     * @param howMany the number of steps which the internal counter of
     * this semaphore should be incremented.
     *
     * @throws IllegalArgumentException if the argument is not positive.
     *
     * @see se.lth.cs.realtime.semaphore.CountingSem#give
     */
	public void give(int howMany) {
		if (RTSystem.NATIVE_SEM) {
			throw new UnsupportedOperationException("Native impl missing!");
		} else {
			if (howMany <= 0) {
				throw new IllegalArgumentException(
						"MultistepSem.give: Argument must be positive");
			}
			synchronized (queue) {
				count += howMany;
				for (int i = howMany; i > 0; i--) {
					queue.notify();
				}
				inqueue = howMany>inqueue ? 0 : inqueue-howMany;
			}
		}
	}

    /**
     * The same as <code>give(1)</code>.
     */
    public void give() {
    	give(1);
    }

    /**
     * Increments the internal counter by the number of steps needed
     * to release all blocked threads, if any. This is like an atomic
     * multiple <code>give()</code> as many times as the (to the
     * caller unknown) number of threads waiting. 
     *
     * <p>Note that even if all threads are released, a (high
     * priority) thread arriving after <code>giveAll</code> but before
     * the last returning (previously blocked) thread can pass. That
     * is, the thread that has not returned will be blocked again, as
     * one would expect.  
     */
    public void giveAll() {
    	if (RTSystem.NATIVE_SEM) {
    		throw new UnsupportedOperationException("Native impl missing!");
    	} else {
    		synchronized (queue) {
    		    count += inqueue;
    		    queue.notifyAll();
    		    inqueue = 0;
    		}
    	}
    }

    /**
     * Causes the calling thread to block until the counter
     * represented by this semaphore assumes a value greater than
     * defined by the argument, that is, until that many resources are
     * available.  Before returning, this number is decremented as
     * many steps as requested.
     * 
     * @param howMany the number of which the internal counter of
     * this semaphore must exceed before returning.
     *
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. You should not catch such an error, except for
     * special cases like handling system shut-down.
     *
     * @throws IllegalArgumentException if the argument is not positive.
     */
	public void take(int howMany) {
		if (RTSystem.NATIVE_SEM) {
			throw new UnsupportedOperationException("Native impl missing!");
		} else {
			if (howMany <= 0) {
				throw new IllegalArgumentException(
						"MultistepSem.take: Argument must be positive.");
			}
			try {
				synchronized (queue) {
					inqueue++;
					while (count < howMany) {
						queue.wait();
					}
					inqueue--;
					count -= howMany;
				}
			} catch (InterruptedException exc) {
				// HERE BE DRAGONS! Include me again
				System.out.println("MultistepSem.take interrupted");
				// throw new SemInterrupted("MultistepSem.take interrupted: "+ exc);
			}
		}
	}

    /**
     * The same as <code>take(1)</code>.
     */
    public void take() {
    	take(1);
    }

    /**
     * Try to atomically take one or several resources within a
     * certain time. Like the method tryTake as defined in the
     * Semaphore interface, there is a timeout on the take operation,
     * as specified by the first argument. The second argument is for
     * the multi-step feature of this class. 
     *
     * @param timeout specifies the time after which the take
     * operation should be given up.
     * 
     * @param howMany the number of which the internal counter of
     * this semaphore must exceed before returning.
     *
     * @return true if the semaphore was successfully taken the
     * specified number of steps, and false if the timeout occured
     * which means that the semaphore was not taken at all.
     *
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. Usually, you should not catch such an error.
     *
     * @throws IllegalArgumentException if the second argument is not positive.
     */
	public boolean tryTake(long timeout, int howMany) {
		if (RTSystem.NATIVE_SEM) {
			throw new UnsupportedOperationException("Native impl missing!");
		} else {

			if (howMany <= 0) {
				throw new IllegalArgumentException(
						"MultistepSem.tryTake: Argument 2 must be positive.");
			}
			long now;
			long arrival = now = System.currentTimeMillis();
			try {
				synchronized (queue) {
					inqueue++;
					while (count < howMany && now < arrival + timeout) {
						queue.wait(timeout - (now - arrival));
						now = System.currentTimeMillis();
					}
					inqueue--;
					if (now >= arrival + timeout)
						return false;
					else {
						count -= howMany;
						return true;
					}
				}
			} catch (InterruptedException exc) {
				// HERE BE DRAGONS! Include me again
				System.out.println("MultistepSem.tryTake interrupted");
				// throw new SemInterrupted("MultistepSem.tryTake interrupted: " + exc);
			}
		}
	}

    /**
     * Calling <code>tryTake(maxtime)</code>is the same as
     * calling <code>tryTake(maxtime,1)</code>.  
     */
    public boolean tryTake(long timeout) {
    	return tryTake(timeout,1);
    }
}
