/* $Id: Semaphore.java,v 1.4 2005/08/19 09:37:16 d02en Exp $          
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.semaphore;

/**
 * Common methods of all types of semaphores, which implements the
 * well-known basic primitive for synchronization and mutual exclusion
 * within the field of concurrent and real-time programming.
 * Basically, a semaphore is an integer-valued variable which can be
 * atomically accessed by only one thread at a time. 
 *
 * <p>Accesses are either of the <b><code>take</code></b> type to be used
 * for taking (locking, requesting to continue execution using it,
 * etc.) a resource, or of the <b><code>give</code></b> type providing
 * access to the resource (or piece of code) for other threads. These
 * methods, <code>take</code> and <code>give</code>, are defined to have
 * no arguments here, implicitly assuming increments/decrements of
 * one. In addition, classes implementing this interface may define
 * additionally overloaded methods, for other increments or other
 * specific features. Furthermore, the semantics of the <code>take</code> 
 * and <code>give</code> with no arguments can differ in different classes.
 *
 * <p>Note, there is no method defined in this interface for getting the
 * value of the internal counter, and no such method is expected to be
 * provided in any implementing class. That would result in misuse of
 * semaphores, in particular among beginners in concurrent
 * programming. However, a method <code>tryTake</code> has been added
 * to permit a thread to check if a resource could be taken within a
 * certain time, and if so, the semaphore is actually taken. This
 * method has been given a different name (instead of simply
 * overloading take) to point out the different behavior, avoiding
 * programming errors resulting in threads that proceed after a timeout
 * as if the resource had been allocated.
 *
 * <hr><em>Note about implementations</em> (not for beginners):<br> The
 * methods <code>take</code> and <code>give</code> are normally not
 * declared synchronized in implementing classes for the following two
 * reasons: <ol>
 *
 * <li> If methods were synchronized and the implementation of
 * blocking is based on wait and notify (see class
 * <code>Object</code>) on the semaphore object itself, there is
 * nothing in Java preventing any thread with a reference to the
 * semaphore to call wait and notify directly in a synchronized block,
 * thereby interfering with the intended behavior of the semaphore
 * methods and probably causing the application to behave
 * erroneously. Instead, blocking/locking is accomplished using a
 * private lock object. It is up to native implementations to
 * accomplish the same thing in more efficient ways.
 *
 * <li>Disabling interrupts during <code>give</code> and
 * <code>take</code> can be the most appropriate/efficient way of
 * preventing a context switch over the small number of native
 * instructions. Even if a compiler (to native code) is free to
 * accomplish mutual exclusion for small <code>synchronized</code>
 * methods by disabling interrupts, this is from a programming and
 * portability point of view different from explicitly disabling
 * interrupts. A semaphore class for embedded systems, however, is
 * often optimized with respect to the particular platform (OS and HW)
 * and should be considered as a built-in mechanism independently of
 * wait and notify (which may not even be supported natively), and
 * hence the methods (although atomic) should not be viewed as
 * <code>synchronized</code>.
 *
 * <p> Another reason for disabling interrupts, mainly for natively
 * compiled applications or special VMs, is to permit
 * <code>give</code> to be called from an interrupt routine (which may
 * not be blocked even temporarily).
 *
 * </ol><hr> <em>Notes for implementations</em>:<br> Classes
 * implementing this interface should (for security and efficiency
 * reasons mentioned in the above note) be final.
 *
 * <p>Threads calling the blocking methods (<code>take</code> and
 * <code>tryTake</code>) are to be thrown an error if they are
 * interrupted, as described in the documentation of those methods. As
 * usual, errors (which implements throwable) can be caught, but in 
 * normal cases the user program is not expected to catch them.
 *
 * @see License
 * @see se.lth.cs.realtime.semaphore.CountingSem
 * @see se.lth.cs.realtime.semaphore.MutexSem
 * @see se.lth.cs.realtime.semaphore.MultistepSem
 * @see se.lth.cs.realtime.semaphore.BinarySem
 */

public interface Semaphore {

    /**
     * Increments by 1 the counter represented by this semaphore, and notify
     * at least one, if any, of the waiting threads.
     * Basically this means executing the following two lines, <br>
     * <code>&nbsp&nbsp  ++count;   </code><br>
     * <code>&nbsp&nbsp  notify();  </code><br>
     * <em>atomically</em> (synchronized, with interrupts disabled, or in 
     * hardware, as accomplished in a particular type of system).
     */
    void give();

    /**
     * Causes the calling thread to block until the counter that represents 
     * this semaphore obtains a positive value. On return the counter, named 
     * <code>count</code> below, is decremented by 1.
     * Basically this means executing<br>
     * <code>&nbsp&nbsp  while (count<1) wait(); </code><br>
     * <code>&nbsp&nbsp  --count;                </code><br>
     * <em>atomically</em> (synchronized, with interrupts disabled, or in 
     * hardware, as accomplished in each particular type of system).
     *
     * @throws SemInterrupted (which is an error and not an
     * exception) if the calling thread has been interrupted before or
     * during blocking. The caller is <b>not</b> expected to catch
     * this type of error, except for properly shut-down of the
     * application, since continued execution most likely will violate
     * resource allocation rules and result in concurrency faults.
     */
    void take();

    /**
     * Causes the calling thread to block until the counter that
     * represents this semaphore obtains a positive value (just like
     * the ordinary take method) or until the <code>timeout</code>
     * time has passed. 
     *
     * <p>If, due to other threads calling give, the semaphore is
     * possible to take, but the caller of <code>take</code> is not
     * rescheduled for execution until after the <code>timeout</code>
     * time after its call of take, it is considered as a timeout and
     * hence <code>false</code> is returned (and the internal state of
     * the semaphore is left unchanged).
     *
     * @argument timeout specifies the time after which the calling
     * thread should be unblocked and return even if the semaphore was
     * not taken. The time of the actual return depends on the
     * underlaying OS/VM/scheduler etc., just like for the calls
     * without a timeout argument.
     *
     * @return the success of take, that is, true if the semaphore was
     * taken and false if the return was due to a timeout. Hence, be
     * careful to make use of the returned value not to proceed using
     * a resource without really having reserved it.
     *
     * @throws SemInterrupted (which is an error and not an
     * exception) if the calling thread has been interrupted before or
     * during blocking. The caller is <b>not</b> expected to catch
     * this type of error, except for properly shut-down of the
     * application, since continued execution most likely will violate
     * resource allocation rules and result in concurrency faults.  
     */
    boolean tryTake(long timeout);

}
