package LabComm.External;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;



/**
 * The <code>UtilsExternal</code> class is a util class that contains methods
 * for converting externalized objects to arrays of bytes 
 * and vice versa.
 * 
 * It works using the interface Externalizable and let every externalizable object to convert its content
 * into array of bytes. While reading, the object is reconstructed from this content.
 * 
 * 
* @author Juan Periset, Ferran Carlas
*/

public final class UtilsExternal {
	/**
	 * Private buffer to store data.
	 */
	private byte[] b=new byte[BUFFERSIZE];
	/**
	 *Boolean used to syncronize methodes write and read of this class.
	 */
	private boolean availableRead;
	
	
	/**
	 * Piped output stream used for conversion.
	 */
	 private PipedOutputStream pout;
	
	
	/**
	 * Piped input stream used for conversion.
	 */
	 private PipedInputStream pin;
	
	
	/**
	 * Object output stream used for conversion of objects to byte arrays.
	 */
	private ObjectOutputStream out;
	
	
	/**
	 * Object input stream used for conversion of byte arrays to objects.
	 */
	private ObjectInputStream in;
	
	
	
	 static final int  PIPESIZE=1024;
	 public static final int  BUFFERSIZE=8192;
	
	/**
	 * Constructs an utility object and initializes all streams. 
	 *
	 * @exception IOException if an I/O error occurs.
	 */
	public UtilsExternal() throws IOException {
		initialize();
	}
	
	
	/**
	 * Initializes the streams.
	 *
	 * @exception IOException if an I/O error occurs.
	 */
	private void initialize() throws IOException {
		pout = new PipedOutputStream();
		pin = new PipedInputStream(pout);
		out = new ObjectOutputStream(pout);
		out.flush();
		in = new ObjectInputStream(pin);
			
	}
	/**
	 * Externalizes an object.It writes the object in a PipedOutputStream. 
	 * The object must implement externalizable.
	 * @exception IOException if an I/O error occurs.
	 */
	public void externalize(Externalizable ext) throws IOException {
 		out.reset();
 		ext.writeExternal(out);
 		out.flush();
 	}
	/**
	 * Externalizes the object.It reads the data from the PipedInputStream and converts it to an object 
	 * with the structure it had before. The object must implement externalizable.
	 * @exception IOException if an I/O error occurs and ClassNotFoundException when
	 * the class does't match with the class of the expected data.
	 */
	public void externalizeRead(Externalizable obj)
	throws IOException, ClassNotFoundException {
		pout.flush();
		obj.readExternal(in);
	}
	/**
	 * Writes the parameter buff ina private buffer.
	 * @param buff The buffer from whom we take the information.
	 * 
	 */
	public synchronized void write( byte[] buff){
		b=buff;
		availableRead=true;
		notifyAll();
	}
	
	/**
	 * Writes length number of bytes of the buffer starting from offset in PipedOutputStream.
	 * @param buff It is the buffer from whom we take the information.
	 * @param offset
	 * @param length
	 * @exception IOException if an I/O error occurs.
	 */
	public synchronized int  writePout( byte[] buff,int offset,int length){
		try{
			pout.write(buff,offset,length);
			
		}catch(IOException e){
// 		    System.out.println("Inside utils writePout"+e);
		    System.out.println("Inside utils writePout");
		}
		return length;
	
 	}
	/**
	 * Returns private buffer b.
	 * @return a byte[].
	 */
	public synchronized byte[] read(){
		try {
			while(!availableRead) wait();
		}catch(InterruptedException e){
// 		    System.out.println(e);
		    System.out.println("UtilsExternal.read: Caught Exception");
		}
		return b;
	}
	/**
	 * Reads from PipedInputStream and stores the bytes in the buffer b.
	 * @return the number of bytes read.
	 * @exception IOException if an I/O error occurs.
	 */
	public int readPin(byte[] b,int offset,int length){
		int bytes=-1;
		try {
			bytes=pin.read(b,offset,length);
				
		}catch(IOException e){
// 		    System.out.println("Inside utils readPin"+e);
		    System.out.println("UtilsExternal.readPin: Caught Exception");		    
		}
		return bytes;
	}
	
	/**
	 * Returns the number of bytes that can be read from PipedInputStream.
	 * @return the number of bytes that can be read from PipedInputStream.
	 * @exception IOException if an I/O error occurs.
	 */
	public int available(){
		
		int bytes=-1;
		try{
			bytes=pin.available();
		}catch(IOException e){
// 		    System.out.println(e);
		    System.out.println("UtilsExternal.available: Caught Exception");
		}
		return bytes;
	}
	
	/**
	 * It searches the starting bit (-1) to know the position where we have to read the range from
	 * and returns it.
	 * @param b  transmit buffer 
	 * @return the number of LabObjects in the transmitted LabStruct
	 */
	public int getSize(byte[] b){
		// k=3 is the first position that could contain the initial bit.
		int k=3;
		// searches the initial bit=-1;
		while(b[k]!=-1){
			k++;
		}
		int range=readInt(b,k+1);
		return range;
	}
	
	/**
	 *Reads an integer from the buffer b starting from offset. 
	 *@param b It is the buffer.
	 *@param off
	 */
	public int readInt(byte[] b,int off){
	return ((b[off + 3] & 0xFF) << 0) +
    ((b[off + 2] & 0xFF) << 8) +
    ((b[off + 1] & 0xFF) << 16) +
    ((b[off + 0] & 0xFF) << 24);
	}
	
}
