
/**
 * BoxPanel.java
 *
 *
 * Created: 1998-11-29
 * LastEditDate: 2000-09-04
 *
 * @author Leif Andersson
 * @version
 */
package se.lth.control;

import javax.swing.*;
import java.awt.Dimension;

/**
 * A <code>BoxPanel</code> is a subclass of <code>JPanel</code> with a
 * <code>BoxLayout</code> layout manager. It also has two convenience
 * methods to add fixed-size and glue-type empty space to the pane.
 * The layout will be specified as horizontal or vertical.
 *
 * @author <a href="mailto:leif@control.lth.se">Leif Andersson</a>
 * @version 1.0
 * @since 1.0
 * @see JPanel */
public class BoxPanel extends JPanel {
  
  /**
   * Specifies that the <code>BoxPanel</code> should be laid out vertically.
   */
  public static final Constant VERTICAL = new Constant() {
    { val = BoxLayout.Y_AXIS; }
    void addGlue(BoxPanel b) { b.add(Box.createVerticalGlue()); }
    void addFixed(BoxPanel b, int size) {
      b.add(Box.createRigidArea(new Dimension(0,size)));
    }
  };

  /**
   * Specifies that the <code>BoxPanel</code> should be laid out
   * horizontally.
   */
  public static final Constant HORIZONTAL = new Constant() {
    { val = BoxLayout.X_AXIS; }
    void addGlue(BoxPanel b) { b.add(Box.createHorizontalGlue()); }
    void addFixed(BoxPanel b, int size) {
      b.add(Box.createRigidArea(new Dimension(size,0)));
    }
  };

  private Constant direction;  

  /**
   * Creates a new <code>BoxPanel</code> with the specified layout direction.
   *
   * @param direction The direction of the layout
   * (<code>BoxPanel.HORIZONTAL</code> or
   * <code>BoxPanel.VERTICAL</code>)
   *
   */
  public BoxPanel(Constant direction) {
    this.direction = direction;
    setLayout(new BoxLayout(this,direction.val));
  }

  /**
   * Creates a new <code>BoxPanel</code> instance with unspecified
   * layout direction. Before components can be added to this
   * instance, the direction must be set with
   * <code>setDirection</code>.
   *
   */
  public BoxPanel() {
    this.direction = null;
  }

  /**
   * Sets the direction of this <code>BoxPanel</code> instance.
   *
   * @param direction The direction of the layout
   * (<code>BoxPanel.HORIZONTAL</code> or
   * <code>BoxPanel.VERTICAL</code>)
   *
   */
  public void setDirection(Constant direction) {
    this.direction = direction;
    setLayout(new BoxLayout(this,direction.val));
  }
    

  /**
   * Adds glue, an empty space with variable size, to the
   * <code>BoxPanel</code> instance.
   *
   */
  public void addGlue() {
    if (direction != null) { direction.addGlue(this); }
    else { throw(illState); }
  }

  /**
   * Adds a fixed size empty space to the <code>BoxPanel</code> instance.
   *
   * @param size The size of the empty space.
   */
  public void addFixed(int size) {
    if (direction != null) { direction.addFixed(this, size); }
    else { throw(illState); }
  }

  private static abstract class Constant {
    int val;
    abstract void addGlue(BoxPanel b);
    abstract void addFixed(BoxPanel b, int size);
  }
  
  private IllegalStateException illState =
    new IllegalStateException("No value for 'direction'");    

} // BoxPanel
