/**
 * DoubleField.java
 *
 *
 * Created: 1998-11-23
 * LastEditDate: 2001-10-06
 *
 * @author Leif Andersson
 * @version
 */

package se.lth.control;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;  

/**
 * A subclass of JTextField that accepts only text that can be parsed
 * to numbers of type double, and within a range that may be specified.
 * While the user is changing the text in the field, the text is in a
 * special color (red by default), and the user can then commit the
 * number with the ENTER key. If the number is legal, the internal value
 * is changed to this value, and the text color is changed to another
 * color (black by default). If the number is illegal, the internal value
 * is not changed, and the text in the field is changed back to a
 * representation of the internal value.
 */ 
public class DoubleField extends JTextField {
  private DoubleDocument doc;
  private Color normalForeground;
  private Color modifiedForeground = Color.red;
  private boolean isSettingValue = false;
  private boolean isModifying = false;

  
  public DoubleField(int positions, int numDigits) {
    super();
    doc = new DoubleDocument(positions, numDigits);
    normalForeground = getForeground();
    setDocument(doc);
    setColumns(positions + 1);
    setHorizontalAlignment(RIGHT);
    setFont(new Font("Monospaced",Font.PLAIN,14));
    setForeground(normalForeground);
  }

  public double getValue() {
    return doc.getValue();
  }

  public void setValue(double value) {
    isSettingValue = true;
    doc.setValue(value);
    isSettingValue = false;
  }
  
  public void setText(String str) {}

  public void setMinimum(double min) { doc.setMinimum(min); }

  public void setMaximum(double max) { doc.setMaximum(max); }  

  public Dimension getMaximumSize() { return getPreferredSize(); }

  public void setForeground(Color col) {
    normalForeground = col;
    super.setForeground(col);
  }

  private void handleForeground() {
    if (isSettingValue) return;
    if (!isModifying) {
      super.setForeground(modifiedForeground);
      isModifying = true;
    }
  }


  public void setModifiedForeground(Color col) {
    modifiedForeground = col;
  }

  public void postActionEvent() {
    double oldVal = doc.getValue();
    isModifying = false;
    super.setForeground(normalForeground);
    doc.update();
    if (doc.getValue() != oldVal) {
      super.postActionEvent();
    }
  }

 
  class DoubleDocument extends PlainDocument {
    private DoubleFormat form; 
    private double min = -Double.MAX_VALUE;
    private double max = Double.MAX_VALUE;
    private double value = 0.0D;
    // if true, then a setValue outside the limits means no change,
    // else setValue outside the limits means set value to the limit.
    private boolean illegalNoChange = true;

    DoubleDocument(int maxPositions, int numDigits) {
      super();
      form = new DoubleFormat(maxPositions, numDigits);
      setValue(value);
    }

    void setIllegalNoChange(boolean val) {
      illegalNoChange = val;
    }

    void setMinimum(double min) {
      if (min > max) { throw new IllegalArgumentException(); }
      this.min = min;
    }

    void setMaximum(double max) {
      if (max < min) { throw new IllegalArgumentException(); }
      this.max = max;
    }

//     void setPositions(int pos) {
//       form = new DoubleFormat(pos);
//     }

    private void setValue(double value) {
      if (illegalNoChange) {
	if ((value > min) && (value < max)) {
	  this.value = value;
	}
      } else {
	this.value = Math.max(min,value);
	this.value = Math.min(max,this.value);
      }
      try {
        super.remove(0,getLength());
        super.insertString(0, form.format(this.value), null);
      } catch (BadLocationException e) {
        getToolkit().beep();
      }
    }
    
    private void setValue(String str) {
      double temp = value;
      try {
	temp = (Double.valueOf(str)).doubleValue();
      } catch (NumberFormatException e) {}
      setValue(temp);
    }

    private double getValue() { return value; }

    private String getText() { return form.format(value); }

    public void insertString(int offs, String str, AttributeSet a) 
                                          throws BadLocationException {
      handleForeground();
      super.insertString(offs, str, a);
    }
    public void remove(int offs, int len) throws BadLocationException {
      handleForeground();
      super.remove(offs,len);
    }
    
    private void update() {
      try {
        String s = super.getText(0,super.getLength());
        setValue(s);
      } catch (BadLocationException e) {
        getToolkit().beep();
      }
    }

  } // DoubleDocument

} // DoubleField



