/**
 * DoubleFormat.java
 *
 *
 * Created: Wed Nov 25 10:27:51 1998
 * LastEditDate: 1999-09-16
 *
 */
package se.lth.control;

import java.text.*;

/**
 * DoubleFormat is a class for formatting numbers of type <tt>Double</tt>.
 * The format is specified using two numbers: <tt>maxPositions</tt>,
 * which is the maximum number of characters in the formatted output, and 
 * <tt>numDigits</tt>, which is the desired number of significant digits
 * in the formatted output. The output may contain both fewer and more
 * significant digits than <tt>numDigits</tt>, but <tt>maxPositions</tt>
 * will always be honored. Some examples:
 *  
 * <table>
 * <tbody>
 * <tr>
 * <td><tt>(maxPositions,numDigits)</tt></td><td><tt>(6,3)</tt></td><td><tt>(8,4)&nbsp;&nbsp;</tt></td><td><tt>(11,5)</tt></td>
 * </tr>
 * <tr></tr>
 * <tr>
 * <td>1.23456E-123&nbsp;</td>              <td> 0.###</td>            <td> 1.E-123</td>          <td> 1.234E-123</td>
 * </tr>
 * <tr>
 * <td>1.23456E-6</td>                      <td> 1.E-3</td>            <td> 1.23E-6</td>          <td> 1.2346E-3</td>
 * </tr>
 * <tr>
 * <td>0.123456</td>                        <td> 0.123</td>            <td> 0.1234</td>           <td> 0.12346</td>
 * </tr>
 * <tr>
 * <td>123.456</td>                         <td> 123.</td>             <td> 123.4</td>            <td> 123.46</td>
 * </tr>
 * <tr>
 * <td>12345.6</td>     <td> 1.E+4</td>            <td> 12346.</td>           <td> 12346.</td>
 * </tr>
 * <tr>
 * <td>1.23456E+8</td>  <td> #####</td>            <td> 1.23E+6</td>          <td>1.2346E+6</td>
 * </tr>
 * </tbody>
 * </table>
 *
 * As can be seen from the examples, if <tt>maxPositions</tt> is small
 * and the number to be formatted is very large or very small, then there
 * is no way to show the number and keep the significance. In that case
 * a sequence of hash marks (#) is shown instead. This can only occur if
 * <tt>maxPositions &lt; 8</tt>.
 *
 * @author Leif Andersson
 * @version 1.0
 * 
 */
public class DoubleFormat {
  private DecimalFormat[] nf;
  private DecimalFormat nfExp1 = null, nfExp2 = null, nfExp3 = null;
  private int maxPositions;

  /**
   * Create a DoubleFormat.
   * @param maxPositions The maximum number of positions for the
   *              output string. Minimum value: numDigits + 2.
   * @param numDigits The desired number of significant digits. 
   *                  Minimum value: 2.
   * @exception IllegalArgumentException if any of the parameter values
   * are too small
   */
  public DoubleFormat(int maxPositions, int numDigits) {
    int i, j, iexp, nfLength, numFrac; 
    this.maxPositions = maxPositions;
//
// Check the arguments
//
    if (numDigits < 2) throw new IllegalArgumentException();
    if (maxPositions < numDigits + 2) throw new IllegalArgumentException();
//
// First find out when we use an exponential form.
//
    iexp = numDigits;
    if (maxPositions > iexp + 2) { iexp++; }
    if (maxPositions > iexp + 2) { iexp++; }
    if (maxPositions > iexp + 2) { iexp++; }
//
// The vector of formats has a length of (iexp + 2)
// except for small values of maxPosition
//
    if (maxPositions > 5) { nfLength = iexp + 2; }
    else { nfLength = maxPositions; } 
//
// Compute the array of formats
//
    nf = new DecimalFormat[nfLength];
    for (i = 0, j = numDigits+1 ; i < nf.length; i++, j--) {
      numFrac = Math.max(0, Math.min(maxPositions-3, j));
      nf[i] = makeFormat(numFrac);
    }
    if (maxPositions == 4) { nf[0] = null; }
//
// Handle the formats with exponents
//
    numFrac = Math.min(maxPositions - 6, numDigits - 1);
    if (numFrac >= 0) {
      nfExp1 = makeFormat(numFrac);
    }
    numFrac = Math.min(maxPositions - 7, numDigits - 1);
    if (numFrac >= 0) {
      nfExp2 = makeFormat(numFrac);
    }
    numFrac = Math.min(maxPositions - 8, numDigits - 1);
    if (numFrac >= 0) {
      nfExp3 = makeFormat(numFrac);
    }
  }
  
  /**
   * Formats num.
   *
   */
  public String format(double num) {
    StringBuffer fail = null;
    char echar;
    DecimalFormat ftemp;
    int il, il2;
    double absnum = Math.abs(num);
    if (absnum < Double.MIN_VALUE) {
      return nf[2].format(num);
    }
    double lognum = log10(absnum);
    il = (int)Math.floor(lognum);
    if (il < 0) { echar = '-'; }
    else { echar = '+'; }
    il2 = il + 2;
    if ((il2 >= 0) && (il2 < nf.length)){
      ftemp = nf[il2];
      if (ftemp != null) return ftemp.format(num);
    } else {
      num /= Math.pow(10D,il);
      il2 = Math.abs(il);
      if (il2 < 10) { ftemp = nfExp1; } 
      else if (il2 < 100) { ftemp=nfExp2; }
      else { ftemp = nfExp3; }
      if (ftemp != null) {
	return (ftemp.format(num) + "E" + echar + il2);
      }
    } 
    return failFormat(num, il, maxPositions);      
  }

  private static final String failFormat(double num, int il,
					     int maxPositions) {
    StringBuffer fail = new StringBuffer(maxPositions);
    for (int i = 0; i < maxPositions; i++) { fail.append('#'); } 
    if (num < 0D) { fail.setCharAt(0,'-'); }
    else { fail.setCharAt(0,' '); }
    if (il < 0) {
      fail.setCharAt(1,'0'); 
      fail.setCharAt(2,'.');
    }
    return fail.toString();
  }

  private static final DecimalFormat makeFormat(int numFrac) {
    DecimalFormat df = new DecimalFormat(" 0.0;-0.0");
    df.setMinimumIntegerDigits(1);
    df.setMinimumFractionDigits(numFrac);    
    df.setMaximumFractionDigits(numFrac);
    df.setGroupingUsed(false);
    df.setDecimalSeparatorAlwaysShown(true);
    return df;
  }
  private static final double ln10 = Math.log(10D);
  private static final double log10(double arg) {
    return (Math.log(arg)/ln10);
  }

} // DoubleFormat
