
/**
 * IntegerField.java
 *
 *
 * Created: 1998-11-25
 * LastEditDate: 1999-09-20
 *
 * @author Leif Andersson
 * @version
 */

package se.lth.control;

import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * A subclass of JTextField that accepts only text that can be parsed
 * to numbers of type integer, and within a range that may be specified.
 * While the user is changing the text in the field, the text is in a
 * special color (red by default), and the user can then commit the
 * number with the ENTER key. If the number is legal, the internal value
 * is changed to this value, and the text color is changed to another
 * color (black by default). If the number is illegal, the internal value
 * is not changed, and the text in the field is changed back to a
 * representation of the internal value.
 */ 
public class IntegerField extends JTextField {
  private IntegerDocument doc = new IntegerDocument();
  private Color normalForeground;
  private Color modifiedForeground = Color.red;
  private boolean isSettingValue = false;
  private boolean isModifying = false;
  
   
  public IntegerField(int positions) {
    super();
    normalForeground = getForeground();
    setColumns(positions + 1); 
    setDocument(doc);
    setHorizontalAlignment(RIGHT);
    setFont(new Font("Monospaced",Font.PLAIN,14));
    setForeground(normalForeground);
  }
  
  public int getValue() {
    return doc.getValue();
  }

  public void setValue(int value) {
    isSettingValue = true;
    doc.setValue(value);
    isSettingValue = false;
  }
 
  public void setText(String str) {}

  public Dimension getMaximumSize() { return getPreferredSize(); }

  public void setMinimum(int min) { doc.setMinimum(min); }

  public void setMaximum(int max) { doc.setMaximum(max); }

  public void setForeground(Color col) {
    normalForeground = col;
    super.setForeground(col);
  }

  private void handleForeground() {
    if (isSettingValue) return;
    if (!isModifying) {
      super.setForeground(modifiedForeground);
      isModifying = true;
    }
  }

  public void setModifiedForeground(Color col) {
    modifiedForeground = col;
  }

  public void postActionEvent() {
    int oldVal = doc.getValue();
    isModifying = false;
    super.setForeground(normalForeground);
    doc.update();
    if (doc.getValue() != oldVal) {
      super.postActionEvent();
    }
  }
 
  class IntegerDocument extends PlainDocument {
    private int min = Integer.MIN_VALUE;
    private int max = Integer.MAX_VALUE;
    private int value = 0;
    // if true, then a setValue outside the limits means no change,
    // else setValue outside the limits means set value to the limit.
    private boolean illegalNoChange = true;

    IntegerDocument() {
      setValue(value);
    }
    
    void setIllegalNoChange(boolean val) {
      illegalNoChange = val;
    }

    void setMinimum(int min) {
      if (min > max) { throw new IllegalArgumentException(); }
      this.min = min;
      value = Math.max(min,value);
    }

    void setMaximum(int max) {
      if (max < min) { throw new IllegalArgumentException(); }
      this.max = max;
      value = Math.min(max,value);
    }
 
    private void setValue(int value) {
      if (illegalNoChange) {
        if ((value >= min) && (value <= max)) {
          this.value = value;
        }
      } else {
        this.value = Math.max(min,value);
        this.value = Math.min(max,this.value);
      }
      try {
	super.remove(0,getLength());
	super.insertString(0, Integer.toString(value), null);
      } catch (BadLocationException e) {
	getToolkit().beep();
      }
    }
 
    private void setValue(String str) {
      int temp = value;
      try {
	temp = Integer.parseInt(str);
      } catch (NumberFormatException e) {}
      setValue(temp);
    }

    private int getValue() { return value; }

    private String getText() { return Integer.toString(value); }

    public void insertString(int offs, String str, AttributeSet a)
                                          throws BadLocationException {
      handleForeground();
      super.insertString(offs, str, a);
    }

    public void remove(int offs, int len) throws BadLocationException {
      handleForeground();
      super.remove(offs,len);
    }

    void update() {
      try {
	String s = super.getText(0,super.getLength());
	setValue(s);
      } catch (BadLocationException e) {
	getToolkit().beep();
      }
    }

  } // IntegerDocument

} // IntegerField
