/*
     A Java Class for polynomial operations  
     By Johan Eker

     Version 1.3.1                             
     Copyright  2000 Department of Automatic Control   
     Department of Automatic Control                    
     Lund Institute of Technology                       
  
  Copyright (c) 2000 by Johan Eker, Department of
  Automatic Control, Lund Institute of Technology, Lund, Sweden
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

package se.lth.control.algorithms;
import java.io.*;
import java.lang.*;

class cmplx {
  double r;
  double i;
};

class BasePolynomial {
  double zero = 0.0;
  double NaN = zero/zero;
  static boolean verbose = true;
  static final int defaultDeg = 20; 
  static final double epsilon = 1.0E-7; 

  double[] coeff;
  int deg, allocatedDeg;

  public BasePolynomial() {
    deg = 0;
    allocatedDeg = defaultDeg;
    coeff = new double[defaultDeg];
  }

  /**
     Sets the degree of the Polynomial to <CODE>newDeg</CODE> and uses the 
     values in <CODE>newCoeff</CODE> as coefficients.
   */
  public void set(int newDeg, double[] newCoeff) {
    if ( newDeg > allocatedDeg ) {
      double[] tmp = new double[newDeg+1];
      coeff = tmp;
      allocatedDeg = newDeg;
    }
    deg = newDeg;
    
    for (int i=0; i<=deg; i++) {
      coeff[i] = newCoeff[i];
    }
  }

  /**
     Returns the degree of the polynomial.
  */
  public int getDegree() {
    return deg;
  }

  /**
     Sets the degree of the polynomial to newDeg.
   */
  void setDegree(int newDeg) {
    if ( newDeg > allocatedDeg ) {
      double[] tmp = new double[newDeg+1];
      for (int i=0; i<=deg; i++) {
	tmp[i] = coeff[i];
      }
      for (int i=deg+1; i<= newDeg; i++) {
	tmp[i] = 0;
      }
      coeff = tmp;
      allocatedDeg = newDeg;
    }
    
    deg = newDeg;
  }


  /**
     Returns the value of element i. A polynomial 
     <PRE>
     A=a_0 z^n + a_1 z^(n-1) + ... + a_n

     is organized as  
     A=[a_0, a_1, ... ,a_n]
     </PRE>
     This means that for example A.get(0) will return the value of the 
     highest power coefficient.
   */
  public double get(int elem) { 
    if ( (elem >= 0) && (elem <= deg) ) {
      return coeff[elem];
    } else { 
      System.out.println("(2) Index exceeding bounds!");
      return NaN;
    }
  }

  /**
     Sets the value of element i. A polynomial 
     <PRE>
     A=a_0 z^n + a_1 z^(n-1) + ... + a_n

     is organized as  
     A=[a_0, a_1, ... ,a_n]
     </PRE>
     This means that for example A.set(0, 1) will set the value of the 
     highest power coefficient to 1.
   */
  public void set(int elem, double value) { 
    if ( (elem >= 0) && (elem <= deg) ) {
      coeff[elem] = value;
    } else { 
      System.out.println("(2) Index exceeding bounds!");
    }
  }
}

/************************ Polynomial *************************/
/**
  A simple polynomial package designed for use in control applications.
  The internal representation of a polynomial is a vector where the
  coefficient of the highest power is stored at the first position,
  i.e. position zero. A polynomial 
  <PRE>
  A=a_0 z^n + a_1 z^(n-1) + ... + a_n

  will thus be represented as

  A=[a_0, a_1, ... ,a_n]
  </PRE>

  The data type of the polynomial coefficients is double.

 * @author <a HREF="mailto:johan.eker@control.lth.se">Johan Eker</a>
 * @version 1.0
 */


public class Polynomial extends BasePolynomial {

  static final int positive = 0; 
  static final int negative = 1; 

  String name;

  /**
     Constructs a polynomial of degree zero with the only
     coefficient equal to zero.
  */
  public Polynomial() {  
    name = null;
  }

 /**
     Constructs a polynomial identical to the argument <CODE>p</CODE>.
  */
  public Polynomial(Polynomial p) {
    setDegree(p.getDegree());
    for (int i = 0; i <= p.getDegree(); i++) {
      set(i, p.get(i));
    }
  }
  
  /**
     Constructs a polynomial of degree <CODE> Degree </CODE> with the 
     coefficients from <CODE> Coefficients</CODE>. The first element in
     both the input vector <CODE> Coefficients </CODE> and the resulting 
     polynomial represents the highest power coefficient.
  */

  public Polynomial(int Degree, double[] coefficients) {
    setDegree(Degree);
    for (int i=0; i<=getDegree(); i++) {
      set(i, coefficients[i]);
    }
    name = null;
  }
  
  /**
     This constructor works in the same way as the one above, but has in
     addition the ability to give a name to the polynomial. This feature is
     mainly used for debug purposes. The program example below uses this
     constructor to create the polynomial <CODE>A=q^2+1.5q-2</CODE>. 
     <PRE>
     <CODE>
     int   deg;
     double[] coeff = new double[3];
     Polynomial A;
     
     deg = 2;
     coeff[0] = 1;
     coeff[1] = 1.5;
     coeff[2] = -2;
     
     A = new Polynomial(deg, coeff, "A");
     A.display();
     </CODE>	

     This program generates the following output.
     A = 1.0 q^2 + 1.5 q - 2.0
     </PRE>
   */
  public Polynomial(int Degree, double[] coefficients, String str) {
    int i;
    
    setDegree(Degree);
    for (i=0; i<=getDegree(); i++) {
      set(i, coefficients[i]);
    }
    name = str;
  }
  

  /**
     Creates an new identical Polynomial object.
   */
  public Object clone() {
    return new Polynomial(this);
  }
 
  /**
     Set the degree of the Polyomial to Degree, the coefficients vals, and 
     the name to str.
   */
  public void set(int Degree, double[] vals, String str) {
    int i;    
    name = str;

    super.set(Degree, vals);
  }

  /**
     Sets the class attribute <CODE>name</CODE> to str.
   */
  public void setName(String str) {
    name = str;
  }

  String formatDouble(double d) {
    String tmp = "" + d;
    int end = tmp.indexOf('.');
    if ( end > 0 ) {
      if ((end + 3) < tmp.length()) {
	end = end + 3;
      } else {
	end = tmp.length();
      }
    } else {
      end = tmp.length();
    }
    return tmp.substring(0,end);
  }

  /**
     Displays the polynomial on the console. For example, let the variable
     <CODE>A</CODE> be defined and make the  following function call:
     <PRE>
     A.setName("A");
     A.display();
     
     The following output is generated:
     
     A = 1.0 q^2 + 1.5 q - 2.0
     </PRE>
  */
  public void display() {    
    display("");
  }
  
  void display(char c) {
    display(new Character(c).toString());
  }
  /**
     Displays the polynomial on the console. If the function is called with
     an argument it will write the string <CODE>str</CODE> followed by the
     polynomial to the console. For example, let the variable
     <CODE>A</CODE> be defined as in the program example above, and make the
     following function call:
     <PRE>
     A.display("This is polynomial A:");
     
     The following output is generated:
     
     This is polynomial A:1.0 q^2 + 1.5 q - 2.0
     </PRE>
  */
  public void display(String str) {
    if (!str.equals("")) {
      System.out.print(str);
    } else if (name != null) {
      System.out.print(name + " = ");
    }
    
    for (int i=0; i<=getDegree() ; i++) {
      System.out.print(formatDouble(get(i)));
      if ((getDegree() - i) > 0) {
	if ((getDegree() - i) > 1) {
	  System.out.print(" q^" + (getDegree() - i));
	} else {
	  System.out.print(" q");
	}
      }
      if (i != getDegree()) {
	if ( get(i+1) > 0 ) {
	  System.out.print(" +");
	} else {
	  System.out.print(" ");
	}
      }          
    }
    System.out.println("");
  }
  
  void display(char variable, int sign) {
    display(variable, sign, "");
  }

  /** By default the polynomial is presented in the variable
     <CODE>q</CODE>. This <CODE>Display()</CODE> method takes two new
     arguments. The first argument <CODE>variable</CODE> tells the function
     which variable to use, and the second argument <CODE>sign</CODE>
     contains the sign of this variable (1 or -1). The third argument is
     a text string and is used as in the other <CODE>Display()</CODE>methods.
  */
  void display(char variable, int sign, String str) {
    if (!str.equals("")) {
      System.out.print(str);
    } else if (name != null) {
      System.out.print(name + " = ");
    }
    
    for (int i=0; i<=getDegree() ; i++) {
      System.out.print(formatDouble(get(i)));
      if ((getDegree() - i) > 0) {
	if ((getDegree() - i) > 1) {
	  System.out.print(" " + variable + "^");
	  if (sign == negative) {
	    System.out.print("-");
	  }
	  System.out.print(getDegree() - i);
	} else {
	  System.out.print(" " + variable);
	}
      }
      
      if ( i != getDegree()) {
	if ( get(i+1) > 0 ) {
	  System.out.print(" +");
	} else {
	  System.out.print(" ");
	}
      }
    }  
  }

  /**
    Returns a string representation of the Polynomial
  */
   
  public String toString() {
    String result = new String();
    int n = getDegree();
    result = "";

    for (int i=0; i <= n; i++) {
      result += formatDouble(get(i)) + " ";
    }
    return result;
  }

  /**
     Sets all the coefficients to zero and the degree to <CODE>d</CODE>.
   */
  void Clear(int newDeg) {  
    setDegree(newDeg);
    for (int i=0; i<= newDeg ; i++) {
      set(i, 0);
    }
  }

  /**
     Returns the reciprocal polynomial. The reciprocal polynomial is
     defined as follows:
     <PRE>
     Let A=3.5q^2+1.1q+999, ten the reciprocal polynomial denoted A* is 
     A*=3.5+1.1q+999^2=q^2A(q^{-1})
     More formal:
     
     A^*(q)=1+a_1*q+\cdots+a_{nq}^{na}=q^{na}A(q^{-1})

     </PRE>
     See strm & Wittenmark: "Computer Controlled Systems", Prentice-Hall, 
     1997, for a full description.
  */
  public Polynomial reciprocal() {
    Polynomial r = new Polynomial();
    int n = getDegree();
    
    r.setDegree(n);
    for (int i = 0; i <= n; i++) {
      r.set(i, get(n - i));
    }
    
    return r;
  }
  /** 
      Calculates the value of the polynomial for q = r, i.e. the
      scalar A(r). 
  */
  public double eval(double r) {
    double sum = 0;
    
    for (int i=0; i<=getDegree(); i++) {
      sum = sum + get(i) * power(r ,getDegree() - i);
    }
    
    return sum;
  }

  double power(double r, int p) {
    double prod = 1;
    for (int i=1; i <= p; i++) {
      prod = prod * r; 
    }    
    return prod;
  }
    
  double polnorm() {
    double r,s;
    r = 0;
    for (int i = 0 ; i <= getDegree() ; i++) {
      s = abs(get(i));
      if (s > r) {
	r = s;
      }
    }
    return r;
  }
  
  void normalize(double reps) {
    double aeps;
    int align, i;
    
    aeps = reps * polnorm();
    align = 0;
    while ((align <= getDegree()) && (abs(get(align))<=aeps)) {
      align = align + 1;  
    }
  
    if (getDegree() >= align) {
      int newDegree = getDegree() - align; 
      
      for (i = 0; i <= newDegree ; i++) {
	set(i, get(i+align));
      }
      setDegree(newDegree);
    } else {
      setDegree(0);
      set(0, 0);
    }
  }

  /**
     Makes the polynomial monic by dividing all
     coefficients with the highest power coefficient. In a monic polynomial
     the highest power coefficient is always 1. 
     <PRE>
     For example let A=2q^2+q+10, then the following operation

     A.MakeMonic();

     sets A=q^2+0.5q+5.

   */
  public Polynomial makeMonic() {
    int n = getDegree();
    
    for (int i = 1; i <= n; i++) {
      set(i, get(i)/get(0));
    }
    set(0, 1);
    
    return this;
  }
 
  /**
     Normalizes the polynomial, i.e. all coefficients that are smaller than
     the largest coefficient times <CODE>epsilon</CODE> are removed. If 
     necessary the degree of the polynomial is changed.
   */
  public void cut(double reps) {
    normalize(reps);
    if (getDegree() < 0) {
      setDegree(0);
      set(0, 0);
    }
  }

  /**
    Gives the polynial the value of <CODE>p</CODE>.
   */
  public Polynomial assign(Polynomial p) {
    setDegree(p.getDegree());
    for (int i = 0; i <= p.getDegree(); i++) {
      set(i, p.get(i));
    }

    return this; 
  }

  /**
    Gives the polynial the value of <CODE>i</CODE>.
   */
  Polynomial assign(int i) {
    setDegree(0);
    set(0, i);

    return this; 
  }

  Polynomial addAndAssign(Polynomial p) {
    int align, i, j;
    Polynomial p2 = (Polynomial) p.clone();

    if  (getDegree() > p2.getDegree()) {
      align = getDegree() - p2.getDegree();
      
      for (j = 0; j <= p2.getDegree(); j++) {
	set(j+align, get(j+align) + p2.get(j));
      }
    } else {
      int oldDegree = getDegree();
      
      align = p2.getDegree() - getDegree();
      setDegree(p2.getDegree());
      
      for (j = oldDegree; j >= 0; j--) {
	set(j+align, p2.get(j+align) + get(j));
      }
      
      for (i = 0; i < align; i++) {
	set(i, p2.get(i));
      }
    }
    
    cut(epsilon);
    
    return this;
  }

  Polynomial subAndAssign(Polynomial p2) {
    int align, i, j;
    Polynomial p = (Polynomial) p2.clone();  

    if  (getDegree() > p.getDegree()) {
      align = getDegree() - p.getDegree();
    
      for (j = 0; j <= p.getDegree(); j++) {
	set(j+align, get(j+align) - p.get(j));
      }      
    
    } else {
      int oldDegree = getDegree();
    
      align = p.getDegree() - getDegree();
      setDegree(p.getDegree());
    
      for (j = oldDegree; j >= 0; j--) {
	if ( abs(get(j) - p.get(j+align)) > epsilon) {
	  set(j+align, get(j) - p.get(j+align));
	} else {
	  set(j+align, 0);
	}
      } 
    
      for (i = 0; i < align; i++) {
	set(i, -p.get(i));
      }      
    }
  
    cut(epsilon);  
    return this;
  }

  Polynomial mulAndAssign(Polynomial p2) {
    double[] oldCoeff;
    int  oldDegree,i,j;  
    Polynomial p = (Polynomial) p2.clone();      

    oldDegree = getDegree();
  
    oldCoeff = new double[oldDegree+1];
    setDegree(getDegree()+p.getDegree());

    for (i = 0; i <= oldDegree; i++) {
      oldCoeff[i] =  get(i);
    }
  
    for (i = 0; i <= getDegree(); i++) {
      set(i, 0);
    }  
    
    for (i = 0;i <= oldDegree; i++) {
      for (j = 0;j <= p.getDegree(); j++) {
	set(i+j, get(i+j) + oldCoeff[i]*p.get(j));     	
      }
    }  
    return this; 
  }

  Polynomial mulAndAssign(double r) {
    for (int i = 0; i <= getDegree(); i++) {
      set(i, get(i)*r);
    }  
  
    return this;
  }
  
  Polynomial divAndAssign(Polynomial p2) {
    int i, j;
    Polynomial denominator = (Polynomial) p2.clone();
    Polynomial numerator = new Polynomial();
    Polynomial quotient  = new Polynomial();
    Polynomial remainder = new Polynomial();
  
    // remainder = numerator
    numerator.setDegree(getDegree());
    for (i = 0 ; i <= getDegree(); i++) {
      numerator.set(i, get(i));
    }
    
    polDiv(numerator, denominator, quotient, remainder);
    assign(quotient);
    return this;
  }
  
  Polynomial divAndAssign(double r) {
    for (int i = 0; i <= getDegree(); i++) {
      set(i, get(i)/r);
    }  
    
    return this;
  }

  Polynomial modAndAssign(Polynomial p2) {
    int i, j;
    Polynomial quotient = new Polynomial();
    Polynomial remainder = new Polynomial();
    Polynomial p = (Polynomial) p2.clone();  

    // remainder = numerator
    remainder.setDegree(getDegree());
    for (i = 0 ; i<=getDegree(); i++) {
      remainder.set(i, get(i));
    }

    if (getDegree() < p.getDegree()) {
      assign(remainder);      
      return this;
    } else {
      quotient.setDegree(getDegree() - p.getDegree());
      for (j = 0; j <= quotient.getDegree(); j++) {
	quotient.set(j, remainder.get(0)/p.get(0));
	for (i = 0; i <= p.getDegree(); i++) {
	  remainder.set(i, remainder.get(i) -quotient.get(j)*p.get(i));
	}
	remainder.set(0, 0);
	remainder.cut(epsilon);
      }
      remainder.cut(epsilon);
      assign(remainder); 
      return this;
    }
  }

  /**
     This method shifts the coefficients of the polynomial to the left
     and sets the last coefficient to <CODE>r</CODE>. Let A=2.3q^2+3.5q+1.1, 
     then the following operation 
     <PRE>
     A.ShiftForward(999);
     </PRE>
     changes A to A=3.5q^2+1.1q+999
  */
  public void shiftForward(double newLastcoeff) {
    for (int i = 0; i < getDegree(); i++) {
      set(i, get(i+1));
    }
    set(getDegree(), newLastcoeff);
  }

  /**
     This method shifts the coefficients of the polynomial to the right
     and sets the highest power coefficient to <CODE>r</CODE>.\\
   */
  public void shiftBackward(double newFirstcoeff) {
    for (int i = getDegree(); i>0 ; i--) {
      set(i, get(i-1));
    }
    set(0, newFirstcoeff);
  }

  /**
     Returns a Polynomial whose value is (this * r).
   */
  public Polynomial mul(double r)  {
    return mul(this, r);
  } 

  /**
     Returns a Polynomial whose value is (p * r).
  */
  public static Polynomial mul(Polynomial p, double r)  {
    Polynomial result = new Polynomial();
  
    result.setDegree(p.getDegree());
    for (int i=0; i <= p.getDegree(); i++) {
      result.set(i, r*p.get(i));
    } 
  
    return result;
  }

  /**
     Returns a Polynomial whose value is (r * p).
  */
  public static Polynomial mul(double r, Polynomial p) {
    return mul(p, r);
  }

  /**
     Returns a Polynomial whose value is (this * p2).
  */
  public Polynomial mul(Polynomial p2) {
    return mul(this, p2);
  }


  /**
     Returns a Polynomial whose value is (p1 * p2).
  */
  public static Polynomial mul(Polynomial p1, Polynomial p2) {
    Polynomial r = (Polynomial) p1.clone();
    r.mulAndAssign(p2);
    return r;
  }


  /**
     Returns a Polynomial whose value is the quotient of (this / r).
  */
  public Polynomial div(double r) {
    return div(this, r); 
  }

  /**
     Returns a Polynomial whose value is (p / r).
  */
  public static Polynomial div(Polynomial p, double r) {
    Polynomial result = new Polynomial();
  
    result.setDegree(p.getDegree());
    for (int i=0;i<=p.getDegree();i++) {
      result.set(i, p.get(i)/r);
    }
  
    return result;
  }

  /**
     Returns a Polynomial whose value is the quotient of (this / p2).
  */
  public Polynomial div(Polynomial p2) {
    return div(this, p2); 
  }

  /**
     Returns a Polynomial whose value is the quotient of (p1 / p2).
  */
  public static Polynomial div(Polynomial p1, Polynomial p2) {
    Polynomial q = new Polynomial();
    Polynomial r = new Polynomial();
    polDiv(p1, p2, q, r);
    return q;
  }

  /**
     Returns a Polynomial whose value is the remainder of (this / p2).
  */
  public Polynomial mod(Polynomial p2) {
    return mod(this, p2); 
  }

  /**
     Returns a Polynomial whose value is the remainder of (p1 / p2).
  */
  public static Polynomial mod(Polynomial p1, Polynomial p2) {
    Polynomial q = new Polynomial();
    Polynomial r = new Polynomial();
    polDiv(p1, p2, q, r);
   return r; 
  }
 
  static void polDiv(Polynomial numerator, Polynomial denominator,
		     Polynomial quotient, Polynomial remainder) {
    int	i, j;
    Polynomial tmp = new Polynomial();
   
    /* remainder = numerator */
    remainder.setDegree(numerator.getDegree());
    for (i = 0 ; i <= numerator.getDegree() ; i++) {
      remainder.set(i, numerator.get(i));
    }
    if (numerator.getDegree() < denominator.getDegree()) {
      quotient.setDegree(0);
    } else {
      quotient.setDegree(numerator.getDegree() - denominator.getDegree());
      for (i = 0 ;  i <= quotient.getDegree() ; i++) {
	quotient.set(i, 0);
      }

      while (remainder.getDegree() >= denominator.getDegree() && 
	     !remainder.isNull() ) {
	i = quotient.getDegree() - 
	  (remainder.getDegree() - denominator.getDegree()); 
	tmp.setDegree(remainder.getDegree() - denominator.getDegree());
	for (j=1; j <= tmp.getDegree(); j++) {
	  tmp.set(j, 0);
	}
	tmp.set(0, 1);
	quotient.set(i, remainder.get(0)/denominator.get(0));
	remainder.assign(sub(remainder,			     
			     mul(mul(tmp, quotient.get(i)), denominator)));
	remainder.cut(epsilon);
      }
      quotient.cut(epsilon);
      remainder.cut(epsilon);
    }
  }
 
  /**
     Returns a Polynomial whose value is the quotient of (this + p2).
  */
  public Polynomial add(Polynomial p2) {
    return add(this, p2);
  }

  /**
     Returns a Polynomial whose value is the quotient of (p1 + p2).
  */
  public static Polynomial add(Polynomial p1, Polynomial p2) {
    Polynomial r = (Polynomial) p1.clone();
    r.addAndAssign(p2);
    return r;
  }

  /**
     Returns a Polynomial whose value is the quotient of (this - p2).
  */
  public Polynomial sub(Polynomial p2) {
    return sub(this, p2);
  }

  /**
     Returns a Polynomial whose value is the quotient of (p1 + p2).
  */
  public static Polynomial sub(Polynomial p1, Polynomial p2) {
    Polynomial r = (Polynomial) p1.clone();
    r.subAndAssign(p2);
    return r;
  }
 
  boolean isNull() {
    if ((getDegree() == 0) && 
	(get(0) < epsilon) && (get(0) > -epsilon)) {
      return true;
    } else {
      return false;
    }
  }

  /**
     This method uses Jury's Criterion to determine if the polynomial is
     stable. The algorithm is described in strm & Wittenmark: "Computer 
     Controlled Systems", Prentice-Hall. If the
     polynomial is stable true is returned otherwise false.
   */
  public boolean isStable() {
    Polynomial a, b;
    double alfa;
    int i, n;
   
    a = (Polynomial) clone();
    n = getDegree();
   
    while ((a.get(0) > 0) && ((n-1) >= 0)) {
      n = a.getDegree();
      b = a.reciprocal();
     
      alfa = a.get(n)/a.get(0);

      for (i=0; i<n; i++) {
	a.set(i, a.get(i) - alfa*b.get(i));
      }
    
      if ((n-1) >= 0) {
	a.setDegree(n-1);
      }
    }

    if (a.get(0) > 0) {
      return true;
    } else {
      return false;
    }
  }

  /**
      Compares this Polynomial with the specified Polynomial for equality.
      Returns true if and only if the Polynomials are numerically equal to
      each other. 
  */
  public boolean equals(Polynomial p2) {
    Polynomial p1 = this;
    if (p1.getDegree() != p2.getDegree()) {
      System.out.println("Error! Operands not of same size!");
	return false;
    } else {
      double tmp1 = p1.polnorm();
      double tmp2 = p2.polnorm();
      double rel_epsilon;   
      if ( tmp1 > tmp2 ) {
	rel_epsilon = epsilon * tmp2;
      } else {
	rel_epsilon = epsilon * tmp1;
      }
      if ((sub(p1, p2)).polnorm() < rel_epsilon) {
	return true;
      } else {
	return false;
      }
    }
  }
  /** 
      This function calculates the greatest common divisor of polynomials
      <CODE>A</CODE> and <CODE>B</CODE> using Euclid's algorithm. 
      <PRE>
      a*x + b*y = g

      a*u + b*v = 0
      </PRE>
      g is the greatest common divisor of a and b. 

   */ 
  public static void GCD(Polynomial a, Polynomial b, double reps, 
			 Polynomial g, Polynomial x, Polynomial y, 
			 Polynomial u, Polynomial v)
  {
    /*
      Calculates greatest common divisor, g, and also x, y, u, v such that
      a*x + b*y = g
      a*u + b*v = 0
    */
    
    boolean   swapped;
    double    nothing;
    Polynomial h = new Polynomial();
    Polynomial hold = new Polynomial();
    Polynomial uold = new Polynomial();
    Polynomial vold = new Polynomial();
    Polynomial temp = new Polynomial();
    Polynomial a2 = new Polynomial(a);
    Polynomial b2 = new Polynomial(b);
    
    if (b.getDegree() > a.getDegree()) {
      swapped = true;
    } else {
      swapped = false;
    }
    
    polydef(1, x);
    polydef(0, y);
    polydef(0, u);
    polydef(1, v);
    
    swapConditionally(a2, b2, swapped);
    swapConditionally(x, y, swapped);
    swapConditionally(u, v, swapped);
    
    g.assign(a2);
    h.assign(b2);
    
    nothing = reps*g.polnorm();
    while (h.polnorm() > nothing) {
      hold.assign(h);
      uold.assign(u);
      vold.assign(v);
      
      temp.assign(g);
      g.assign(div(g, h));
      
      h.assign(mod(temp,h));       
            
      u.assign(mul(g, u));
      u.assign(sub(x, u));
      v.assign(mul(g, v));
      v.assign(sub(y, v));
      g.assign(hold);
      x.assign(uold);
      y.assign(vold);
    }
    swapConditionally(x,y,swapped);
    swapConditionally(u,v,swapped);    
  }

  /**
     DiophantineMDS finds the minimal degree solution to the equation
     AR+BS=Ac. The equation is solved in two steps. First the two
     equations 
     <PRE>
     AX+BY=G 
     AU+BV=0 
     </PRE>
     are solved using the GCD function. Then in the second step the general 
     solution is found through 
     <PRE>
     R = R0 + Q*U
     S = S0 + Q*V

     with
     
     R0 = X*Ac div G 
     S0 = Y*Ac div G
     </PRE>

     where G is the greatest common divisor of A and B.
     The minimal degree solution is now given by simply choosing 
     Q=S0 div V. The greatest common G must divide
     Ac otherwise the equation has no solution.

     <PRE>Example

     Polynomial A, B, C, XX, YY;

     A = new Polynomial();
     B = new Polynomial();
     C = new Polynomial();
     XX = new Polynomial();
     YY = new Polynomial();

     real A_data[]= {1, 2.3, 3.5};
     A.Set(2, A_data, "A");
     
     real B_data[] = {1, 1.45};
     B.Set(1, B_data, "B");
     
     real C_data[] = {1, 3.4, 0.8, 2};
     C.Set(3, C_data, "C");
     
     DiophantineMDS(A, B, C, XX, YY);
     
     XX.Display("XX = ");
     YY.Display("YY = ");

     </PRE>
     When running the program above the following output is produced:
     <PRE>
     
     XX = 1.000 q + 3.629
     YY = -2.529 q -7.379
     
     </PRE>
   */

  public static int diophantineMDS(Polynomial  a, Polynomial  b, Polynomial c,
				   Polynomial xx, Polynomial yy) {
    Polynomial g = new Polynomial();
    Polynomial x = new Polynomial();
    Polynomial y = new Polynomial();
    Polynomial u = new Polynomial();
    Polynomial v = new Polynomial(); 
    Polynomial q = new Polynomial();

    Polynomial g2 = new Polynomial();
    Polynomial c2 = new Polynomial();
    double nothing;
 
    GCD(a,b,epsilon,g,x,y,u,v);  

    if (b.getDegree() > a.getDegree()) {
      if ( verbose ) {   
	System.out.println("Error in DiophantineMDS - deg(B) > deg(A)");
      }
      return 1;
    }
    
    nothing = epsilon * c.polnorm();
    if (g.getDegree() > c.getDegree() || (g.polnorm() < nothing)) {
      if ( verbose ) {
	System.out.println("Error in DiophantineMDS" +  
			   " - Attempted division by zero!"); 
      }
      return 2;
    }  
    polDiv(c,g,g2,c2);
    
    if (c2.polnorm() > nothing) {
      if ( verbose ) {
	System.out.print("Error in DiophantineMDS" +
			 "- Non-zero-remainder after GCD");
	return 3;
      }
    }

    xx.assign(mul(x, g2)); // where g = g div c
    yy.assign(mul(y, g2));
    
    // Seek minimal degree solution.
    
    q.assign(div(yy, v));
    
    xx.assign(sub(xx, mul(q, u))); 
    yy.assign(sub(yy, mul(q, v)));
    
    return 0;
  }

  public static void integralDesign(Polynomial A, Polynomial B, Polynomial Ac,
				    Polynomial R, Polynomial S, double x0) {
    /*  
	A*R0 + B*S0 = Ac
	  
	X = q - x0
	Y = y0
	R = X*R0 + Y*B
	S = X*S0 - Y*A
	  
	A*R + B*S = X*Ac
	R(1) = 0  {i.e. R contains an integrator.}
    */
    
    double y0;
    Polynomial X = new Polynomial(); 
    Polynomial Y = new Polynomial(); 
    Polynomial R0 = new Polynomial(); 
    Polynomial S0 = new Polynomial();;
  
    X.setDegree(1);
    X.set(0, 1);
    X.set(1, -x0);
  
    diophantineMDS(A, B, Ac, R0, S0);
  
   y0 = -X.eval(1)*R0.eval(1)/B.eval(1);
  
    R.assign(add(mul(X, R0), mul(y0, B)));
    S.assign(sub(mul(X, S0), mul(y0, A)));
  }

  public static void robustDesign(Polynomial  A, Polynomial  B, Polynomial  Ac,
				  Polynomial  R, Polynomial  S, double x0) {
  
    //   A*R0 + B*S0 = Ac0
    //
    //   X = q - x0
    //   Y = q - y0    
    //   R = X*R0 + Y*B
    //   S = X*S0 - Y*A
    //   
    //   A*R + B*S = X*Ac0
    //   S(-1) = 0   {S(-1)=S(e^i*w*h)+S(i*pi)=0  
    // i.e. S/R zero at nyquist freq.)
    double y0;
    Polynomial X = new Polynomial();
    Polynomial Y = new Polynomial();
    Polynomial R0 = new Polynomial();
    Polynomial S0 = new Polynomial();

    X.setDegree(1);
    X.set(0, 1);
    X.set(1, -x0);

    diophantineMDS(A, B, Ac, R0, S0);
  
    y0 =- X.eval(-1)*S0.eval(-1)/A.eval(-1);
    Y.setDegree(0);
    Y.set(0, -y0);
  
    R.assign(add(mul(X, R0), mul(Y, B)));
    S.assign(sub(mul(X, S0), mul(Y, A)));
  }


  public static void robustIntegralDesign(Polynomial A,  Polynomial B, 
				   Polynomial Ac, Polynomial R, 
				   Polynomial S, double x1, double x0) {
    /*
      A*R0 + B*S0 = Ac
  
      X = (q - x0)*(q - x1)
    
      Let Y = y0*(q + y1) and use
      R = X*R0 + Y*B      R(1)  = 0  {i.e. R contains an integrator.} 
      S = X*S0 - Y*A      S(-1) = 0  {i.e. S robust.} 
      => y0 & y1.
  
      A*R + B*S = X*Ac
    */

    double y0, y1, I, J;
    Polynomial X = new Polynomial();
    Polynomial Y = new Polynomial(); 
    Polynomial R0 = new Polynomial();
    Polynomial S0 = new Polynomial();;
    
    diophantineMDS(A, B, Ac, R0, S0);
  
    X.setDegree(2);
    X.set(0, 1);   //  X[0] = 1;
    X.set(1, x1);  //  X[1] = -(x0+x1);
    X.set(2, x0);  //  X[2] = x0*x1;

    I = - X.eval(1)*R0.eval(1)/B.eval(1);    // I = Y(1)  <=> R(1)  = 0    
    J =  X.eval(-1)*S0.eval(-1)/A.eval(-1);  // J = Y(-1) <=> S(-1) = 0 

    y0 = (I-J)/2;
    y1 = (I+J)/(2*y0);

    Y.setDegree(1);
    Y.set(0, y0);
    Y.set(1, y0*y1);
    
    R.assign(add(mul(X, R0), mul(Y, B)));
    S.assign(sub(mul(X, S0), mul(Y, A)));
  }
  
  /**
   Calculates a LQG-controller for the system $A$, $B$, and $C$ with the
   loss function coefficient $\rho$. The computational procedure is
   described in detail in section 12.5 in \cite{ast+wit90}. This
   implementation only handles the 
   case where $A(0)\neq0$. Below is an
   example of how a LQG-problem can be solved using this function. It is
   taken from \cite{ast+wit90}, Example 12.7, with $A(z)=z+0.5$,
   $B(z)=0.5$, and $C=z+0.8$.   
   <PRE><B>Example</B>  
      
   Polynomial A, B, C, R, S;
   A = new Polynomial();
   B = new Polynomial();
   C = new Polynomial();
   R = new Polynomial();
   S = new Polynomial();

   System.out.println("Example 12.7 in CCS");
   System.out.println("LQG Design with rho = 0.5.");
 
   A.SetDegree(1);
   A[0] = 1; 
   A[1] = 0.5;  // a = 0.5
   B.SetDegree(0);
   B[0] = 0.5;  // b = 0.5
   C.SetDegree(1);   
   C[0] = 1; 
   C[1] =0.8;   // c = 0.8
  
   LQGDesign(A, B, C, R, S, 0.5);  
   
   R.Display("R = ");
   S.Display("S = "); 

   The following output is generated:

   R = 1.000 q + 0.614
   S = -0.112
   </PRE>
   */
  public static void LQGDesign(Polynomial A, Polynomial B, Polynomial C, 
			       Polynomial R, Polynomial S, double rho) {
    Polynomial A1 = new Polynomial();
    Polynomial A2 = new Polynomial();
    Polynomial B1 = new Polynomial(); 
    Polynomial D = new Polynomial();
    Polynomial P = new Polynomial();
    Polynomial P1 = new Polynomial();
    Polynomial U = new Polynomial();
    Polynomial V = new Polynomial();
    Polynomial X = new Polynomial();
    Polynomial Y = new Polynomial();;
    double indata[], r;
    int d;
  
    GCD(A, B, epsilon, A2, X, Y, U, V);  
  
    if (A2.getDegree() <= 1) {
      // if no common factor in A and B 
      d = A.getDegree() - B.getDegree();
      indata = new double[d+1];
    
      indata[0] = 1;
      for (int i=1; i <= d; i++) {
	indata[i] = 0;
      }
      D.set(d, indata, "D");
    
      sfactorize(add(mul(rho, mul(A, A.reciprocal())), 
		     mul(D, mul(B, B.reciprocal()))), P);
      r = P.get(0)*P.get(0);
      P.makeMonic();
    
      diophantineMDS(A, B, mul(P, C), R, S);

    } else { 
      // A and B have a common factor.
      if (A2.isStable()) {
	A1.assign(div(A, A2));
	B1.assign(div(B, A2)); 
	P1.assign(div(P, A2));
      
	d = A1.getDegree() - B1.getDegree();
	indata = new double[d+1];

	indata[0] = 1;
	for (int i=1; i <= d; i++) {
	  indata[i] = 0;
	}

	D.set(d, indata);

	sfactorize(add(mul(rho,mul(A1,A1.reciprocal())),
		       mul(D, mul(B1, B1.reciprocal()))), P1);
	r = (P1.get(0)*P1.get(0));
	P1.makeMonic();

	diophantineMDS(A1, B1, mul(P1, C), R, S);
      } else {
	System.out.println("Error in LQGDesign --- Common factor not stable!");
      }
    }
  }

  /**
     MDPPNZCDesign stands for Minimal Degree Pole Placement with No Zero
     Cancelation. The function chooses B+ = 1 and B- = B. Furthermore
     Bm is chosen so that the stationary gain will be unity.  
     <PRE>
     Bm(q) = A_m(1)/B(q) B(1)
     </PRE>
     The closed-loop characteristic equation to be solved now becomes
     <PRE>
     A R + B S = A_o A_m
     </PRE>
     The T polynomial is given by
     <PRE>
     T(q) = A_m(1) / A_o(q) B(1)
     </PRE>
     The usage of MDPPNZCDesign-method  is used is demonstrated below.
     <PRE> <B>Example</B>
     
     Polynomial A, B, Am, Ao, R, S, T;
     A = new Polynomial();
     B = new Polynomial();
     Am = new Polynomial();
     Ao = new Polynomial();
     R = new Polynomial();
     S = new Polynomial();
     T = new Polynomial();

     B.SetDegree(1);
     B[0] = 1;
     B[1] = 0.7;
     A.SetDegree(2);
     A[0] = 1;
     A[1] = -1.8;
     A[2] = 0.81;
     Am.SetDegree(2);
     Am[0] = 1;
     Am[1] = -1.5;
     Am[2] = 0.7;
     Ao.SetDegree(1);
     Ao[0] = 1;
     Ao[1] = 0;
     
     MDPPNZCDesign(A, B, Am, Ao, R, S, T);
     
     R.Display("R = ");
     S.Display("S = ");
     T.Display("T = ");
     }

     From this program the following output is generated:
     
     R = 1.000 q + 0.088
     S = 0.213 q + -0.101
     T = 0.118 q + 0.000
     </PRE>
  */
  public static int MDPPNZCDesign(Polynomial A, Polynomial B, 
				  Polynomial Am, Polynomial Ao,
				  Polynomial R, Polynomial S, Polynomial T) {
    /* 
       __MDPP__No zero cancellation:
       B+ = 1, B- = B
       Condition: degAo = degA - 1 @@
       Bm = beta*B
    */
    Polynomial Ac = new Polynomial();
    double beta;
  
    Ac.assign(mul(Ao, Am));
    if ( diophantineMDS(A,B,Ac,R,S) == 0) {
      beta = Am.eval(1)/B.eval(1);
      T.assign(mul(beta, Ao));
      return 1;
    } else {
      return 0;
    }
  }

  static double abs(double x) {
    if (x < 0) {
      return (-x);
    } else {
      return x;
    }
  }

  static void polydef(double r, Polynomial p) {
    p.setDegree(0);
    p.set(0, r);
  }


  static void swapConditionally(Polynomial p1, Polynomial p2, 
				boolean swapped) {
    Polynomial p;
    if (swapped) {
      p  = p1;
      p1 = p2;
      p2 = p;
    }
  }

  /**
     The reciprocal polynomial of A, denoted A* is obtained by
     reversing the order of the coefficients of A. 
     <PRE> <B>Example</B>
     Let  A(z) = z^(na)+a_1*z^(na-1) + ... + a_(na).
     
     The it reciprocal polynomial of A is the given as
     
     A*(z) = 1 + a_1*z + ...  + a_(na) * z^(na) = z^(na) * A(z^(-1)).
     </PRE>
     Sfactorize takes a polynomial B = e e* and returns a stable
     polynomial A so that A A* = e e*
   */
  public static void sfactorize(Polynomial b, Polynomial a) {
    int maxiter = 20;
  
    int i, iter, j, n;
    double[] alfa, beta, cr, il, s, sq, d, x;
    double[][] p, sr;
    
    n = b.getDegree()/2;                      /* Initialize a */

    a.setDegree(n);

    alfa = new double[n+1];
    beta = new double[n+1];
    cr   = new double[n+1];
    il   = new double[n+1];
    s    = new double[n+1];
    sq   = new double[n+1];
    d    = new double[n+1];
    x    = new double[n+1];
  
    p = new double[n+1][];
    sr = new double[n+1][];
    for (i=0; i<=n; i++) {
      p[i] = new double[n+1];
      sr[i] = new double[n+1];
    }

    if (b.get(n) < 0) {
      b.assign(mul(-1, b));
    }

    a.set(n, b.get(n)/Math.sqrt(b.get(n)));

    for (i = n-1; i >= 0; i--) {
      a.set(i, b.get(i)/Math.sqrt(b.get(n)));
    }
  
    for (iter = 1; iter <= maxiter; iter++) {
      /*Iteration loop*/ 
      sq[0] = b.get(0)/a.get(0); 
      for (i = 1; i <= n-1 ; i++) {
	sq[i] = b.get(i);
	for (j = 1; j <= i; j++) {
	  sq[i] = sq[i]-sq[i-j]*a.get(j); 
	}
    
	sq[i] = sq[i]/a.get(0);
      }
  
      for (i = 0; i <= n-1; i++) {
	s[i] = b.get(n-i); 
	for (j = 0; j <=  n-i-1; j++) {
	  s[i] = s[i]-sq[i+j]*a.get(n-j); 
	}
      }
      s[n] = b.get(0);

      for (i = 0; i <= n; i++) {
	/* Table */
	p[0][i] = a.get(i);
	sr[0][i] = s[i];
      }
   
      for (i = 0; i <= n-1; i++) {
	for (j = 0; j <= n-i; j++) {
	  cr[j] = p[i][n-i-j];
	}

	alfa[i] = p[i][n-i]/cr[n-i];
	beta[i] = sr[i][n-i]/cr[n-i];
	for (j = 0; j <= n-i-1; j++) {
	  p[i+1][j] = p[i][j]-alfa[i]*cr[j];
	  sr[i+1][j] = sr[i][j]-beta[i]*cr[j];
	}
      }

      beta[n] = sr[n][0]/p[n][0];
      
      /* Integral */
      for (i = 0; i <= n; i++) {               
	beta[i] = beta[i]/p[0][0];
      }

      il[0] = beta[n];
      for (i = 1; i <= n; i++) {
	il[i] = beta[n-i];
	for (j = 0;j <= i-1; j++) {
	  il[i] = il[i]-p[n-i][i-j]*il[j]/p[n-i][0];
	}
      }

      x[0] = a.get(0)*il[0];              /* Solution */    
      for (i = 1; i <= n; i++) {
	x[i] =a.get(i)*il[0];
	for (j = 1; j <= i; j++) {
	  x[i] = x[i]+2*a.get(i-j)*il[j];
	}
      }
      for (i = 0; i <= n; i++) {
	a.set(i, 0.5*(a.get(i)+x[i]));
      }
    }
    a.assign(a.reciprocal());
  }

  static void fixDegree(Polynomial p, int SpecDeg)
  {
    // Change a Polynomialial so that it has a specified degree 
  
    int i; 
    if ((p.getDegree() < SpecDeg) && (p.polnorm()>0)) { 
      for (i = p.getDegree(); i >= 0; i--) {
	p.set(i+SpecDeg-p.getDegree(), p.get(i));
      } 
    
      for (i = SpecDeg - p.getDegree()-1 ; i>= 0 ; i--) {
	p.set(i, 0); 
      }
    } 
    p.setDegree(SpecDeg); 
  }


  /**
     Given vectors
     <PRE>
     a = [1 a_2 ... a_n]^T
     b = [1 b_2 ... b_n]^T
     </PRE>
     and scalars alpha and beta, find vectors
     <PRE>
     a' = [1 a'_2 ... a'_n]^T
     b' = [1 b'_2 ... b'_n]^T
     </PRE>
     such that 
     <PRE>
     alpha a a^T + \beta b b^T = alpha' a' a'^T + beta' b' b'^T
     </PRE>
     The vectors a' and b' can be found using dyadic decomposition. 
     DyadicReduction is very useful when doing square root   
     recursive least square estimations. For a closer look at the algorithm
     and its applications see strm & Wittenmark: "Adaptive Control", 
     Addison-Wesley, 1995.
   */
  public static void dyadicReduction(double[] a, double[] b, double[] alpha,
				     double[] beta, int i0, int i1, int i2) {
    // Performs a dyadic reduction of a, b, alpha and beta, so that
    //
    // alpha*(a a') + beta*(b b') = alpha2*(a2 a2') + beta2* (b2 b2') 
    //
    // constraint:
    // a is monic.
    //
    // Result:
    //   a2 will be monic and the first element of b2 will be zero
    //
    double gamma;
    if (beta[0] < epsilon) {
      beta[0] = 0.0;
    }
    double b1 = b[i0];
    double w1 = alpha[0];
    double w2 = beta[0] * b1;
    alpha[0] += w2 * b1;
    if (alpha[0] > epsilon) {
      beta[0] *= w1/alpha[0];
      gamma = w2/alpha[0];
      for (int i = i1; i <= i2; i++) {
	b[i] -= b1 * a[i];
	a[i] += gamma * b[i]; 
      }
    }  
  }

  /** 
  LFilter s an implementation of an recursive estimator and would
  together with any of the design functions above form a complete adaptive
  controller. The algorithms behind LDFilter is taken from strm & 
  Wittenmark: "Adaptive Control", Addison-Wesley, 1995. 

  Let the system to be estimated be on the form 
  <PRE>
  y(t) = phi^T(t) theta 
  </PRE>
  where theta is a parameter vector and phi is a vector of
  signals. The following recursive least-square estimation algorithm is used
  <PRE>
  theta'(t) = theta'(t-1) + K(t)(y(t)-phi^T(t)theta'(t-1)
  K(t) = P(t) phi(t)
  P(t) =(I-K(t) phi^T(t))P(t-1)
  </PRE>
  The covariance matrix P has a decomposition P=LDL^T, where L is
  a lower triangular matrix and D is a diagonal matrix. Initially set 
  L=I, which gives that P(0)=D. 
  @param l A matrix of doubles with the size (degree+1)^2.   
  @param d An array of doubles with the size (degree+1). 
  @param Phi The Phi vector contains old process values and old 
  control signals. The polynomial is arranged on the following format:
  <PRE>
  Phi = [y(t), -y(t-1), \ldots , -y(t-n), u(t-d), \dots , u(t-d-m)]
  </PRE>
  where n=deg(A), m=deg(B) and d=n-m and is of size (degree+1).
  @param theta This is a vector with the size set to  
  (deg(A) + deg(B) + 2) and with the coefficients to be estimated on the 
  following format: 
  <PRE>
  theta = [a_0, a_1, \ldots, a_n,b_0, ..., b_m]
  </PRE>
  and is of size (degree+1).
  @param l This parameter is the forgetting factor $\lambda$.
  @param degree This parameter is the dimension of the problem 
  degree = deg(A) + deg(B) + 1. 
  */
  public static void LDFilter(double[] l, double[] d, double[] phi,
		double[] theta, double lambda, int degree) {
    // New estimation of Theta based on the new information
    // in the regressor Phi.
    // The algorithm is described in the third chapter of
    //     Astrom & Wittenmark: Adaptive Control
    //     A-W Publishing 
    // degree = degA + degB + 1

    double[] alpha = new double[1]; 
    double[] beta = new double[1]; 
    int i, j;
    double w;
    int nrofcols = degree + 1;

    d[0] = lambda;
 
    // Form (phi)'l and calculate the residual 
    // e(t) = yf(t) - (theta * phi)[1..n]
    double e = phi[0];
    for (i = 1; i <= degree; i++) {
      e -= theta[i] * phi[i];
      w = phi[i];
      for (j = i+1; j <= degree ; j++) {
	w += phi[j] * l[i*nrofcols+j];
      }
      l[i] = 0.0;
      l[i*nrofcols] = w;
    }
 
    // Reduction to lower triangular form
    alpha[0] = d[0];
    for ( i = degree; i >= 1; i--) { 
      double[] tmp = new double[nrofcols];
      for (j=0; j<nrofcols; j++) {
	tmp[j] = l[i*nrofcols+j];
      }
      beta[0] = d[i];
      dyadicReduction(l,tmp,alpha,beta,0,i,degree);
      d[i] = beta[0];
      for (j=i; j<nrofcols; j++) {
	l[i*nrofcols+j] = tmp[j];
      }
     }
    d[0] = alpha[0];

    // Update d and calculate the new theta
    for ( i = 1; i <= degree; i++) {
      theta[i] += l[i] * e ;
      d[i] /= lambda;
    }
  }
}
