package se.lth.control.io;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ProtocolException;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketImpl;

public class UnixSocketImpl extends SocketImpl {

  private int unixFd = -1;
  private String path;

  public UnixSocketImpl() {
  }

  private UnixSocketImpl(int unixFd) {
    this.unixFd = unixFd;
  }

  private native int connect(String path);

  private native int bind(String path, int backlog);

  private native int accept(int fd);

  private native void close(int fd, String path);
	
  private native int writeBytes(int fd, byte[] data, int offset, int length);

  private native int readBytes(int fd, byte[] data, int offset, int length);

  @Override
  protected void accept(SocketImpl arg0) throws IOException {
    System.out.println("accept");
  }

  public UnixSocketImpl accept() {
    return new UnixSocketImpl(accept(unixFd));
  }


  @Override
  protected int available() throws IOException {
    System.out.println("available");
    return 0;
  }

  @Override
  protected void bind(InetAddress arg0, int arg1) throws IOException {
    throw new ProtocolException("bind() is not supported");
  }

  public void bindAndListen(UnixSocketAddress address, int backlog) throws IOException {
    String absolutePath = address.getAbsolutePath();
    unixFd = bind(absolutePath, backlog);
    if (unixFd < 0) {
      throw new IOException("Failed to bind " + address.getAbsolutePath() + " (" + unixFd + ")");
    }
    path = absolutePath;
  }

  @Override
  protected void close() throws IOException {
    close(unixFd, path);
  }

  @Override
  protected void connect(String host, int port) throws IOException {
    throw new ProtocolException("Can only connect with UnixSocketAddress");
  }

  @Override
  protected void connect(InetAddress host, int port) throws IOException {
    throw new ProtocolException("Can only connect with UnixSocketAddress");
  }

  @Override
  protected void connect(SocketAddress address, int timeout) throws IOException {
    UnixSocketAddress a = (UnixSocketAddress)address;
    if (a == null) {
      throw new ProtocolException("Can only connect with UnixSocketAddress");
    } else {
      if (a.getAbsolutePath() == null) {
	if (unixFd < 0) {
	  throw new IOException("Can not connect to an empty path");
	}
      } else {
	unixFd = connect(a.getAbsolutePath());
	if (unixFd < 0) {
	  throw new IOException("Failed to connect " + a.getAbsolutePath() + " (" + unixFd + ")");
	}
      }
    }
  }

  @Override
  protected void create(boolean stream) throws IOException {
    if (! stream) {
      throw new ProtocolException("Only stream sockets allowed");
    }
  }

  @Override
  protected InputStream getInputStream() throws IOException {
    return new UnixSocketInputStream(this);
  }

  @Override
  protected OutputStream getOutputStream() throws IOException {
    return new UnixSocketOutputStream(this);
  }

  @Override
  protected void listen(int arg0) throws IOException {
    System.out.println("listen");

  }

  @Override
  protected void sendUrgentData(int arg0) throws IOException {
    System.out.println("sendUrgentData");
  }

  public Object getOption(int arg0) throws SocketException {
    System.out.println("getOption");
    return null;
  }

  public void setOption(int arg0, Object arg1) throws SocketException {
    System.out.println("setOption");
  }

  public void write(int data) throws IOException {
    byte b[] = new byte[]{(byte) data};
    int written = writeBytes(unixFd, b, 0, 1);
    if (written != 1) {
      throw new IOException("Failed to write: " + written);
    }
  }

  public void write(byte[] b, int off, int len) throws IOException {
    int written = writeBytes(unixFd, b, off, len);
    if (written != len) {
      throw new IOException("Failed to write: " + written);
    }
  }

  public int read(byte[] data, int offset, int length) throws IOException {
    int read = readBytes(unixFd, data, offset, length);
    if (read <= 0) {
      throw new EOFException();
    }
    return read;
  }

  public int read() throws IOException {
    byte data[] = new byte[1];
    int read = read(data, 0, 1);
    if (read <= 0) {
      throw new EOFException();
    }
    int result = data[0];
    if (result < 0) {
      result += 256;
    }
    return result;
  }

  static {
    System.loadLibrary("se_lth_control_io_UnixSocketImpl");
  }



}
