/**
 * PlotComponent.java
 *
 *
 * Created: 1998-12-20
 * LastEditDate: 2002-09-06
 *
 * @author Leif Andersson
 * @version
 */

package se.lth.control.plot;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;
import se.lth.control.plot.*;
import se.lth.control.*;

public class PlotComponent extends JComponent {
  private static final int inset = 6;
  private static final int tickLength = 6;
  private static final int tickToText = 3;
  private static Color defaultColor = Color.black;

  protected boolean updateFlag;
  protected int channels;

  protected Dimension preferredSize = new Dimension(600,200);
  protected Dimension minimumSize = new Dimension(100,100);
 
  protected String title, titleTemp;
  private boolean titleChange = false;

  protected Axis xAxis = null, xAxisTemp = null;
  private boolean xAxisChange = false;
  protected Axis yAxis = null, yAxisTemp = null;
  private boolean yAxisChange = false;
 
  protected StringPosition ct =
               new StringPosition(StringPosition.CENTER_TOP);
  protected StringPosition rm =
               new StringPosition(StringPosition.RIGHT_MIDDLE);
  protected StringPosition rb =
               new StringPosition(StringPosition.RIGHT_BOTTOM);

  protected Color gridColor = Color.lightGray; 

//  protected DoublePoint[] plotChannel;
  protected DoublePoint[][] plotData;
  protected List[]  plotDataTemp;  // An array of List of DoublePoint.
  private boolean plotDataChange = false;

  protected Color[] plotColor;

// labelPattern is used to determine the space needed for the labels
// on the axes.
  protected String labelPattern = "-00.0";

// The following members are recomputed by paintComponent each time
// the actual size changes.
  protected Dimension lastSize = new Dimension(0,0);
  protected Dimension size;
  protected Point origin = new Point(0,0);
  protected Rectangle plotRect = new Rectangle();
  protected int xlength = 0, ylength = 0;

// Translation and scaling of the data points
  protected double xmin, xscale, ymin, yscale;

  protected int[] xv = new int[1], yv = new int[1];
  Color saveColor;

  /**
   * Creates a new <code>PlotComponent</code> instance.
   *
   * @param channels The number of channels for this
   * <code>PlotComponent</code>
   */
  public PlotComponent(int channels) {
    int i;
    if (channels <= 0) {
      throw(new IllegalArgumentException("'channels' must be > 0")); }
    this.channels = channels;
    plotColor = new Color[channels];
      
    setDoubleBuffered(true);
    plotData = new DoublePoint[channels][];
    for (i = 0; i < channels; i++) {
      plotData[i] = new DoublePoint[0];
      plotColor[i] = Color.black;
    }
    if (channels > 1) plotColor[1] = Color.red;
    if (channels > 2) plotColor[2] = Color.green;
    if (channels > 3) plotColor[3] = Color.blue;
    plotDataTemp = new List[channels];
  }

  /**
   * Creates a new <code>PlotComponent</code> instance with one channel
   *
   */
  public PlotComponent() { this(1); }

    public void setColor(int nr,Color col) {
        plotColor[nr] = col;
    }

    public Color getColor(int nr) {
        return plotColor[nr];
    }

  public boolean isOpaque() { return true; }

  /**
   * Sets the x-axis of the <code>PlotComponent</code>
   *
   * @param x an <code>Axis</code> value
   */
  public synchronized void setXAxis(Axis x) {
    xAxisTemp = x;
    xAxisChange = true;
  }

  /**
   * Sets the y-axis of the <code>PlotComponent</code>
   *
   * @param y an <code>Axis</code> value
   */
  public synchronized void setYAxis(Axis y) {
    yAxisTemp = y;
    yAxisChange = true;
  }

  /**
   * Sets the plot data for one channel of the <code>PlotComponent</code>
   *
   * @param channel The channel to set.
   * @param p The plot data, a <code>List</code> of <code>DoublePoint</code>
   */
  public synchronized void setPlot(int channel, List p) {
    if (channel >= channels)  { throw(new IllegalArgumentException()); }
    plotDataTemp[channel] = p;
    plotDataChange = true;
  }

  /**
   * Sets the plot data for channel 0 of the <code>PlotComponent</code>
   *
   * @param p a <code>List</code> value
   */
  public void setPlot(List p) { this.setPlot(0,p); }

  public void setTitle(String title) { this.title = title; }

  public void update() {
    updateFlag = true;
    repaint();
  }

  protected synchronized void updateInternal() {
    List p;
    DoublePoint[] data; 
    if (xAxisChange) {
      xAxis = xAxisTemp; xAxisChange = false;
      xmin = xAxis.getMin();
      xscale = 1.0D/(xAxis.getMax() - xmin);
    }
    if (yAxisChange) {
      yAxis = yAxisTemp; yAxisChange = false;
      ymin = yAxis.getMin();
      yscale = 1.0D/(yAxis.getMax() - ymin);
    }
    if (titleChange) { title = titleTemp; titleChange = false; }
    if (plotDataChange) {
      for (int i = 0; i < channels; i++) {
        p = plotDataTemp[i];
        data = plotData[i];
        if (p != null) {
          int psize = p.size();
          if (data.length != psize) {
            data = new DoublePoint[psize];
          }
          data = (DoublePoint[])p.toArray(data);
        } else {
          data = new DoublePoint[0];
        }
        plotData[i] = data;
        if (xv.length < data.length) {
          xv = new int[data.length];
          yv = new int[data.length];
        }
      }
      plotDataChange = false;
    }
    updateFlag = false;
  }
    
  public Dimension getPreferredSize() { return preferredSize; }

  public Dimension getMinimumSize() { return minimumSize; }
 


  public void paintComponent(Graphics g) {
    int x,y;
    if (updateFlag) { updateInternal(); }
    size = getSize();
    if ((size.width != lastSize.width) || (size.height != lastSize.height)) {
      lastSize = size;
      int sw = g.getFontMetrics().stringWidth(labelPattern);
      int sh = g.getFontMetrics().getHeight();
      x = inset + sw + tickToText + tickLength;
      y = inset + sh;
      plotRect.setLocation(x, y);
      x = size.width - x - inset - sw/2;
      y = size.height - y - sh - inset - tickToText - tickLength;
      plotRect.setSize(x, y);
//      origin.x = plotRect.x;
//      origin.y = plotRect.y + plotRect.height;
    }
    saveColor=g.getColor();
    g.setColor(Color.white);
    g.fillRect(0, 0, size.width, size.height);
    g.translate(plotRect.x,plotRect.y);
    paintGrid(g);
    g.setColor(saveColor);
    if (title != null) { paintTitle(g); }
    if (xAxis != null) { paintXAxis(g); }
    if (yAxis != null) { paintYAxis(g); }
    if (plotData != null) { paintData(g); }
    g.translate(-plotRect.x, -plotRect.y);
  }

  protected void paintGrid(Graphics g) {
    int i, x, y;
    int gridSize;
    double dd;
    int height = plotRect.height;
    int width = plotRect.width;
    g.setColor(gridColor);
    if (xAxis != null) {
      gridSize = xAxis.getGridSize();
      for (i = 0; i < gridSize; i++) {
        dd = xAxis.getGridPoint(i);
        x = (int)Math.round(dd * width);
        g.drawLine(x, 0, x, height);
      }
    }
    if (yAxis != null) {
      gridSize = yAxis.getGridSize();
      for (i = 0; i < gridSize; i++) {
        dd = yAxis.getGridPoint(i);
        y = height - (int)Math.round(dd * height);
        g.drawLine(0, y, width, y);
      }
    }
  }

  protected void paintData(Graphics g) {
    DoublePoint[] data;
    DoublePoint dp;
    double x, y;
    int i, j, np = 0;
    for (j = 0; j < channels; j++) {
      data = plotData[j];
      if (data != null) {
        np = data.length;
        for (i = 0; i < np; i++) {
          dp = data[i];
          x = xscale * (dp.x - xmin);
          y = yscale * (dp.y - ymin);
          xv[i] = (int)Math.round(x * plotRect.width); 
          yv[i] = plotRect.height - (int)Math.round(y * plotRect.height);
        }
        g.setColor(plotColor[j]);
        g.drawPolyline(xv,yv,np);
      }
    }
  }

  protected void paintTitle(Graphics g) {
    rb.put(title, g, plotRect.width, 0);
  }

  protected void paintXAxis(Graphics g) {
    int i, x, y, yt, ytext;
    AxisPoint ap;
    int axisSize = xAxis.getAxisSize();
    y = plotRect.height;
    yt = y + tickLength;
    ytext = yt + tickToText;
    int width = plotRect.width;
    g.drawLine(0, y, width, y);
    for (i = 0; i < axisSize; i++) {
      ap = xAxis.getAxisPoint(i);
      x = (int)Math.round(ap.position * width);
      g.drawLine(x, y, x, yt);
      ct.put(ap.label, g, x, ytext);
    }
  }
 
  protected void paintYAxis(Graphics g){
    AxisPoint ap;
    int i, x, y;
    int xtext = -tickLength - tickToText;
    int height = plotRect.height;
    int axisSize = yAxis.getAxisSize();
    g.drawLine(0, 0, 0, height);
    for (i = 0; i < axisSize; i++) {
      ap = yAxis.getAxisPoint(i);
      y = height - (int)Math.round(ap.position * height);
      g.drawLine(0, y,  -tickLength, y);
      rm.put(ap.label, g, xtext, y);
    }
  }
  
} // PlotComponent
