package se.lth.control.plot;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import se.lth.control.*;
import se.lth.control.plot.*;
import se.lth.control.realtime.*;

/**
 * Created: 2002-10-10
 * Edited:  2003-11-20
 *
 *  A plotter panel with a built-in thread to handle the plotting.
 * @author Dan Henriksson <a href="mailto:dan@control.lth.se">dan@control.lth.se</a>
 */
public class PlotterPanel extends JPanel {

    private PlotComponent plotter;
    private int channels;
    private PlotThread plotThread;
    private ArrayList[] plotData;
    private LinearAxis xAxis, yAxis;
    private double xBottom = 0;
    private double xRange = 10;
    private int xDivTicks = 5;
    private int xDivGrid = 5;
    private double yBottom = -10;
    private double yRange = 20;
    private int yDivTicks = 4;
    private int yDivGrid = 4;
    private boolean doIt = true;
    private int priority;
    private double time;
    private int freqCount = 0;
    private int freq = 1;

    private class PlotThread extends Thread {
	
	public void run() {
	    boolean remove = false;
	    int i;

	    setPriority(priority);
	    
	    while (doIt) {
		synchronized(this) {
		    try {
			wait();
		    } catch (Exception e) {}
		    	
		    xBottom =  time - xRange;
		    synchronized(plotter){
		      while (((DoublePoint)plotData[0].get(0)).x < xBottom) {
			for (i = 0 ; i < plotData.length ; i++) {
			    plotData[i].remove(0);
			}
		      }
		    }

		    plotter.setXAxis(new LinearAxis(xRange,xBottom,xDivTicks,xDivGrid));
		    
		    for (i = 0 ; i < plotData.length ; i++) {
			plotter.setPlot(i, plotData[i]);
		    }

		}
		plotter.update();
	    }
      	}
    }
    
    /**
     * Creates a plotter panel with specified number of channels 
     * and thread priority
     */ 
    public PlotterPanel(int channels, int threadPriority) {

	super();
	this.channels = channels;
	priority = threadPriority;
	plotData = new ArrayList[channels];
	for (int i = 0 ; i < channels ; i++) {
	    plotData[i] = new ArrayList();
	}

	plotter = new PlotComponent(channels);
	plotter.setYAxis(new LinearAxis(yRange,yBottom,yDivTicks,yDivGrid));
	plotter.setXAxis(new LinearAxis(xRange,xBottom,xDivTicks,xDivGrid));
	plotter.update();
	add(plotter);

	addComponentListener(new ComponentListener() {
		public void componentHidden(ComponentEvent e) {}
		public void componentMoved(ComponentEvent e) {}
		public void componentShown(ComponentEvent e) {}

		public void componentResized(ComponentEvent e) {
		    repaint();
		}
	    });
	
    }

    public void paintComponent(Graphics g) {
	super.paintComponent(g); //paint background
	Dimension d = getSize();
	
	plotter.setBounds(3, 3, d.width-6, d.height-6);
    }

    /** Sets the x-axis of the plotter */
    public synchronized void setXAxis(double range, int divTicks, int divGrid) {
	plotter.setXAxis(new LinearAxis(range,xBottom,divTicks,divGrid));
	plotter.update();
        xRange = range;
        xDivTicks = divTicks;
        xDivGrid = divGrid;
    }

    /** Sets the y-axis of the plotter */
    public synchronized void setYAxis(double range, double bottom, int divTicks, int divGrid) {
	plotter.setYAxis(new LinearAxis(range,bottom,divTicks,divGrid));
	plotter.update();
	yBottom = bottom;
        yRange = range;
        yDivTicks = divTicks;
        yDivGrid = divGrid;
    }

    /** Sets the update freqency of the plotter. */
    public synchronized void setUpdateFreq(int newFreq) {
	    freq = newFreq;
    }

    /** Sets the title of the plotter panel. */
    public void setTitle(String s) {
	plotter.setTitle(s);
    }

    private synchronized int getFreq() {
	return freq;
    }

    /** Starts the built-in plotter thread */
    public void start() {
	plotThread = new PlotThread();
	doIt = true;
	plotThread.start();
    }

    /** Stops the built-in plotter thread */
    public void stopThread() {
      doIt = false;
      synchronized(plotThread) {
	plotThread.notify();
      }
    }

  /** May only be called from synchronized(plotThread) environment */
    private void updatePlot(double x) {
        time = x;
	freqCount++;
	if (freqCount >= getFreq()) {
	  plotThread.notify();
	  freqCount = 0;
	}
    }
    


    /** Puts a one-channel data point */
    public void putData(double x, double y1) {
	synchronized(plotThread) {
	  synchronized(plotter) {
	    plotData[0].add(new DoublePoint(x, y1));
	  }
	  updatePlot(x);
	}
    }

    /** Puts a two-channel data point */
    public void putData(double x, double y1, double y2) {
	synchronized(plotThread) {
	  synchronized(plotter) {
	    plotData[0].add(new DoublePoint(x, y1));
	    plotData[1].add(new DoublePoint(x, y2));
	  }
	  updatePlot(x);
	}
    }

    /** Puts a three-channel data point */
    public void putData(double x, double y1, double y2, double y3) {
	synchronized(plotThread) {
	  synchronized(plotter) {
	    plotData[0].add(new DoublePoint(x, y1));
	    plotData[1].add(new DoublePoint(x, y2));
	    plotData[2].add(new DoublePoint(x, y3));
	  }
	  updatePlot(x);
	}
    }

    /** Puts a four-channel data point */
    public void putData(double x, double y1, double y2, double y3, double y4) {
	synchronized(plotThread) {
	  synchronized(plotter) {
	    plotData[0].add(new DoublePoint(x, y1));
	    plotData[1].add(new DoublePoint(x, y2));
	    plotData[2].add(new DoublePoint(x, y3));
	    plotData[3].add(new DoublePoint(x, y4));
	  }
	  updatePlot(x);
	}
    }

    /** Sets the plot color of a specified plot channel. */
    public void setColor(int nr, Color col) {
    	plotter.setColor(nr - 1, col);
    }

    /** Returns the plot color of the specified plot channel. */
    public Color getColor(int nr) {
    	return plotter.getColor(nr - 1);
    }


}
