/*
 * BeanDescriptor.java
 * 
 * Created on 18/03/2004
 *
 * Copyright (C) Anders Blomdell <anders.blomdell@control.lth.se>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package se.lth.control.xml;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;


/**
 * @author andersb
 *
 * Helper class that holds attribute/list getter/setter methods for 
 * a Bean class.
 */
public class BeanDescriptor {
	
	private Class beanClass;
	private Constructor constructor;
	private Attribute text = new Attribute("[Text]");
	private Map attribute = new TreeMap();
	private Map child = new TreeMap();
	private Set transientField = null;
	
	/**
	 * @author andersb
	 *
	 * To change the template for this generated type comment go to
	 * Window - Preferences - Java - Code Generation - Code and Comments
	 */
	
	abstract class Element {

		protected String name;
		protected Method getter;
		protected Method setter;

		protected void checkBean(Bean b, Method m, String caller) throws Exception {
			if (m == null) {
				throw new Exception("'" + beanClass.getName() + 
						"' has no " + caller + " method for '" + name + "'");
			}
			if (b.getClass() != beanClass) {
				throw new Exception("Expected '" + beanClass.getName() + 
						"', but got a '" + b.getClass().getName() + "' in " +
						caller + " method for '" + name + "'");
			}
		}

		public void setSetter(Method setter) {
				this.setter = setter;
		}

		public void setGetter(Method getter) {
			this.getter = getter;
		}

		public String getName() {
			return name;
		}
		
	}
	
	public class Child extends Element  {
		
		public Child(String name) {
			this.name = name;
		}
		
		public Class getChildClass() {
			return setter.getParameterTypes()[0];
		}
		
		public void addChild(Bean b, Bean child) throws Exception {
			checkBean(b, setter, "child set");
			setter.invoke(b, new Object[] { child });
		}
	
		public Bean[] getChildren(Bean b) throws Exception {
			checkBean(b, getter, "child get");
			return (Bean[])getter.invoke(b, null);
		}

	
	}	
	
	/**
	 * @author andersb
	 *
	 * To change the template for this generated type comment go to
	 * Window - Preferences - Java - Code Generation - Code and Comments
	 */
	public class Attribute extends Element {
		
		public Attribute(String name) {
			this.name = name;
		}
		
		public void setSetter(Method setter) {
			if (this.setter == null || 
					setter.getParameterTypes()[0].equals(String.class)) {
				this.setter = setter;
			}
		}
		
		private Object[] args(String value) throws Exception {
			Class type = setter.getParameterTypes()[0];
			Object result[] = new Object[1];
			if (type.equals(int.class)) {
				type = Integer.class;
			} else if (type.equals(long.class)) {
				type = Long.class;
			} else if (type.equals(short.class)) {
				type = Short.class;
			} else if (type.equals(byte.class)) {
				type = Byte.class;
			} else if (type.equals(boolean.class)) {
				type = Boolean.class;
			} else if (type.equals(float.class)) {
				type = Float.class;
			} else if (type.equals(double.class)) {
				type = Double.class;
			} else if (type.equals(char.class)) {
				type = Character.class;
			}
			Constructor c = type.getConstructor(new Class[] {String.class});
			result[0] = c.newInstance(new Object[] { value } );	
	
			return result;	
		}
		
		public String getValue(Bean b) throws Exception {
			String result = null;
			checkBean(b, getter, "attribute get");
			Object o = getter.invoke(b, null);
			if (o != null) { 
				result = o.toString();
			}
			return result;
		}
		
		public void setValue(Bean b, String value) throws Exception {
			checkBean(b, setter, "attribute set");
			setter.invoke(b, args(value));
		}
		
	}
	
	private String normalizeName(String name) {
		return name.toLowerCase();
	}

	public Attribute getAttribute(String name) throws Exception {
		Attribute result = (Attribute)attribute.get(normalizeName(name));
		if (result == null) {
			throw new Exception(beanClass + " has no attribute named " + name);
		}
		return result;
	}
	
	private Attribute createAttribute(String name) {
		Attribute result = (Attribute)attribute.get(normalizeName(name));
		if (result == null) {
			result = new Attribute(normalizeName(name));
			if (! transientField.contains(normalizeName(name))) {
				// Transient fields are silently thrown away
				attribute.put(normalizeName(name), result);
			}
		}
		return result;
	}
	
	public Child getChild(String name) throws Exception {
		Child result = (Child)child.get(normalizeName(name));
		if (result == null) {
			throw new Exception(beanClass + " has no children named " + name);
		}
		return result;
	}
	
	private Child createChild(String name) {
		Child result = (Child)child.get(normalizeName(name));
		if (result == null) {
			result = new Child(normalizeName(name));
			child.put(normalizeName(name), result);
		}
		return result;
	}
	
	/**
	 * @param c
	 * @throws Exception
	 * 
	 * Construct a Bean description for the given class.
	 */
	public BeanDescriptor(Class bc) throws Exception {
		beanClass = bc;
	
		// Collect transient fields for later filtering of attributes
		transientField = new TreeSet();
		Field fl[] = beanClass.getDeclaredFields();
		for (int i = 0; i < fl.length; i++) {
			if ((fl[i].getModifiers() & Modifier.TRANSIENT) != 0) {
				transientField.add(normalizeName(fl[i].getName()));
			}
		}
		
		// Save default constructor
		constructor = beanClass.getConstructor(null);
		
		// Collect bean attributes and children
		Method ml[] = beanClass.getMethods();
		for (int i = 0; i < ml.length; i++) {
			Method m = ml[i];
			if (Bean.class.isAssignableFrom(m.getDeclaringClass())) {
				// Collect all bean methods
			  String name = m.getName();
			  Class p[] = m.getParameterTypes();
				if (name.startsWith("set") && p.length == 1) {
						createAttribute(name.substring(3)).setSetter(m);
				} else if (name.startsWith("add") &&
						       p.length == 1 && Bean.class.isAssignableFrom(p[0])) {				
					createChild(name.substring(3)).setSetter(m);
				} else if (name.startsWith("append") &&
			       p.length == 1 && Text.class.isAssignableFrom(p[0])) {				
					text.setSetter(m);
				} else if (name.startsWith("get") && p.length == 0) {
					if (m.getReturnType().isArray()) {				
						createChild(name.substring(3)).setGetter(m);
					} else if (m.getReturnType().isAssignableFrom(Text.class)) {
						text.setGetter(m);
					} else {
						createAttribute(name.substring(3)).setGetter(m);
					}
				}
			}
		}
		for (Iterator i = attribute.keySet().iterator(); i.hasNext();) {
			Attribute a = (Attribute) (attribute.get(i.next()));
			if (a.setter == null || a.getter == null) {
				throw new Exception(beanClass  + " asymmetric attribute: " + a.getName());
			}
		}
		for (Iterator i = child.keySet().iterator(); i.hasNext();) {
			Child c = (Child) (child.get(i.next()));
			if (c.setter == null || c.getter == null) {
				throw new Exception(beanClass  + " asymmetric child: " + c.getName());
			}
		}
		if (text.setter != null && text.getter == null ||
				text.setter == null && text.getter != null) {
			throw new Exception(beanClass  + " asymmetric text: " + text.getName());
		}
		
		// No more need to check transient fields
		transientField = null;
	}
	
	/**
	 * @return Create a new instance of the described Bean.
	 */
	public Bean construct() throws Exception {
		return (Bean)constructor.newInstance(null);
	}
	
	/**
	 * @return All attributes of the described Bean.
	 */
	public Attribute[] getAttributes() {
		return (Attribute[])attribute.values().toArray(new Attribute[0]);
	}

	/**
	 * @return All children of the described Bean.
	 */
	public Child[] getChildren() {
		return (Child[])child.values().toArray(new Child[0]);
	}

	/**
	 * @param string
	 */
	public void appendText(Bean b, String string) throws Exception {
		String stripped = string.replaceAll("^\\s*", "").replaceAll("\\s*$", "");
		if (stripped.length() != 0) {
			if (text.setter != null) {
				text.setValue(b, stripped);
			} else {
				throw new Exception(beanClass + " has no text setter");
			}
		}
	}	
	
	public String getText(Bean b) throws Exception {
		String result = null;
		if (text.getter != null) {
			result = (String)text.getValue(b);
		}
		return result;
	}
	
}