/*
 * BeanFactory.java
 * 
 * Created on 17/03/2004
 *
 * Copyright (C) Anders Blomdell <anders.blomdell@control.lth.se>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package se.lth.control.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.Reader;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * @author andersb
 *
 * To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
public class BeanFactory {
	
	private static Map descriptor = new HashMap();
	
	static synchronized BeanDescriptor getDescriptor(Class c) throws Exception {
		BeanDescriptor result = (BeanDescriptor)descriptor.get(c);
		if (result == null) {
			result = new BeanDescriptor(c);
			descriptor.put(c, result);	
		}	
		return result;
	}

	static synchronized BeanDescriptor getDescriptor(Bean b) throws Exception {
		return getDescriptor(b.getClass());
	}

	/**
	 * @param name
	 * @return A new bean of tope <code>name</code>
	 */
	public static Bean create(Class c) throws Exception {
		BeanDescriptor d = getDescriptor(c);
		return d.construct();
	}
	
	/**
	 * @param result
	 * @param map
	 */
	private static void setAttributes(Bean b, NamedNodeMap a) throws Exception {
		if (a != null) {
			for (int i = 0; i < a.getLength() ; i++) {
				Node attr = a.item(i);
				getDescriptor(b).getAttribute(attr.getNodeName()).setValue(b, attr.getNodeValue());
			}
		}
	}
	
	private static Class getChildClass(Bean b, String name) throws Exception {
		return getDescriptor(b).getChild(name).getChildClass();
	}
	
	private static void addChildren(Bean b, NodeList nl) throws Exception {
		if (nl != null) {
			BeanDescriptor d = getDescriptor(b);
			for (int i = 0 ; i < nl.getLength() ; i++) {
				Node n = nl.item(i);
				if (n.getNodeType() == Node.ELEMENT_NODE) { 
					BeanDescriptor.Child c = d.getChild(n.getNodeName());
					c.addChild(b, parse(c.getChildClass(), n));	
				} else {
					d.appendText(b, n.getNodeValue());
				}
			}	
		}
	}

	public static Bean parse(Class c, Node n) throws Exception {
		Bean result = null;
		switch (n.getNodeType()) {
//				case Node.DOCUMENT_TYPE_NODE: { parse(n.getFirstChild()); } break;
			case Node.DOCUMENT_NODE: { 
				result = parse(c, n.getFirstChild());
			} break;
			case Node.ELEMENT_NODE: { 
				result = create(c);
				setAttributes(result, n.getAttributes());
				addChildren(result, n.getChildNodes());
			} break;
			default: { 
				throw new Exception("Invalid XML node: " + n.getNodeType() + " " + n.toString()); 
			}
		}
		return result;
	}	
	
	public static Bean parse(Class c, InputSource s) throws Exception {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document document = builder.parse(s);
		return parse(c, document);
	}
	
	public static Bean parse(Class c, File f) throws Exception {
		return parse(c, new InputSource(new FileInputStream(f)));
	}

	public static Bean parse(Class c, Reader r) throws Exception {
		return parse(c, new InputSource(r));
	}

	private static Node unparse(Bean b, String tag, Document doc) throws Exception {
		BeanDescriptor d = getDescriptor(b);
		Element result = doc.createElement(tag);
		// Set all attributes
		BeanDescriptor.Attribute attr[] = d.getAttributes();
		for (int i = 0; i < attr.length; i++) {
			BeanDescriptor.Attribute a = attr[i];
			String value = a.getValue(b);
			if (value != null) {
				result.setAttribute(a.getName(), value);
			}
		}
		// Append all children
		BeanDescriptor.Child child[] = d.getChildren();
		if (child.length > 0) {
			for (int i = 0; i < child.length; i++) {
				BeanDescriptor.Child c = child[i];
				String name = c.getName();
				Bean values[] = c.getChildren(b);
				if (values != null && values.length > 0) {
					for (int j = 0; j < values.length; j++) {
						result.appendChild(doc.createTextNode("\n"));
						result.appendChild(unparse(values[j], name, doc));
					}
				}
			}	
			result.appendChild(doc.createTextNode("\n"));
		}
		// Append text
		String text = d.getText(b);
		if (text != null) {
			result.appendChild(doc.createTextNode(text + "\n"));
		}
		return result;
	}

	public static void unparse(Bean b, String tag, StreamResult target) throws Exception {
	  DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document document = builder.newDocument();
    document.appendChild(unparse(b, tag, document));
    TransformerFactory tFactory = TransformerFactory.newInstance();
    Transformer transformer = tFactory.newTransformer();
    DOMSource source = new DOMSource(document);
    transformer.transform(source, target);
 
	}
	
	public static void unparse(Bean b, String tag, PrintStream s) throws Exception {
		unparse(b, tag, new StreamResult(s));
		s.println();
	}

}