package se.lth.cs;


/**
 * The <code>Vector</code> class implements a growable array of objects. Like an array, it
 * contains components that can be accessed using an integer index. However,
 * the size of a Vector can grow or shrink as needed to accommodate adding and
 * removing items after the Vector has been created.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.8, 12/11/03
 * @since   ThrottleNet0.6
 */
public class Vector {
	
	
	/**
	 * The default size of this vector.
	 */
	public static final int DEFAULT_INIT_SIZE = 100;
	
	
	/**
	 * The amount by which the capacity of the vector is automatically
	 * incremented when its size becomes greater than its capacity.
	 */
	private static final int CAPACITY_INCREMENT = 100;
	
	
	/**
	 * The collection of elements in this vector.
	 */
	private Object[] elements;
	
	
	/**
	 * The number of elements in this vector.
	 */
	private int nbr;
	
	
	/**
	 * Default contructor. Creates a vector initialized to default size.
	 */
	public Vector() {
		this(DEFAULT_INIT_SIZE);
	}
	
	
	/**
	 * Constructs a vector initialized to specific size.
	 * 
	 * @param	initSize the size of this vector.
	 */
	public Vector(int initSize) {
		elements = new Object[initSize];
		nbr = 0;
	}

	
	/**
	 * Returns the component at the specified index.
	 *
	 * @param	pos	an index into this vector.
	 * @return	the component at the specified index.
	 * 
	 * @exception	ArrayIndexOutOfBoundsException	if the pos is negative or
	 *					not less than the current size of this Vector object.
	 */
	public Object elementAt(int pos) {
		return elements[pos];
	}
	
	
	/**
	 * Appends the specified element to the end of this Vector. 
	 * 
	 * @param	o	element to be appended to this Vector. 
	 */
	public void add(Object o) {
		if(nbr == elements.length)
			extendVector();
		elements[nbr++] = o;
	}
	
	
	/**
	 * Increases the capacity of this vector.
	 */
	private void extendVector() {
		Object[] tmp = new Object[elements.length + CAPACITY_INCREMENT];
		for(int i = 0; i < elements.length; i++)
			tmp[i] = elements[i];
		elements = tmp;
	}
	
	
	/**
	 * Returns the number of components in this vector. 
	 *
	 * @return	the number of components in this vector.
	 */
	public int size() {
		return nbr;
	}
	
	
	/**
	 * Removes the element at the specified position in this Vector. The last
	 * object in the vector will be moved to the empty position. This means
	 * that the order will be changed if the removed object wasn't the last.
	 * Returns the element that was removed from the Vector. 
	 * 
	 * @param	pos	 the index of the element to removed. 
	 * @return	element that was removed.
	 * 
	 * @exception	ArrayIndexOutOfBoundsException index out of range 
	 *                                      (index < 0 || index >= size()).
	 */
	public Object remove(int pos) throws ArrayIndexOutOfBoundsException {
		Object tmp = elements[pos];
		elements[pos] = elements[--nbr];
		return tmp;
	}
	
	
	/**
	 * Removes the first occurrence of the specified element in this Vector.
	 * If the Vector does not contain the element, it is unchanged. More
	 * formally, removes the element with the lowest index i such that
	 * (o==null ? get(i)==null : o.equals(get(i))) (if such an element exists).
	 * The last object in the vector will be moved to the empty position.
	 * 
	 * @param	o	element to be removed from this Vector, if present.
	 * @return	true if the Vector contained the specified element.
	 */
	public boolean remove(Object o) {
		for(int i = 0; i < nbr; i++) {
			if((o == null ? elementAt(i) == null : o.equals(elementAt(i)))) {
				remove(i);
				return true;
			}
		}
		return false;
	}
	
	
	/**
	 * Returns a string representation of this Vector, containing the String
	 * representation of each element.
	 *
	 * @return	a string representation of this collection.
	 */
	public String toString() {
		String s = "trottlenet.Vector[";
		for(int i = 0; i < nbr-1; i++)
			s += elementAt(i) + ",";
		if(nbr > 0)
			s += elementAt(nbr-1);
		return (s +  "]");
	}
	
}
