/* $Id: CyclicThread.java,v 1.8 2005/08/19 09:37:16 d02en Exp $ 
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * This class models a real-time thread with cyclic behavior.
 * <p>
 * Such a thread loops performing an action and waiting for some time to
 * pass. The action performed and the waiting policy are not specified
 * in this class, which is so declared abstract.
 *
 * @see License
 * @see PeriodicThread
 */

public abstract class CyclicThread extends RTThread implements Runnable {

  /**
   * The flag checked to exit the run method.
   */
  private volatile boolean stopReq = false;

  /**
   * The flag set by <code>suspend</code>.
   * It is checked at every cycle, before calling <code>perform</code>.
   */
  private volatile boolean suspendReq = false;

  /**
   * The flag set by this thread when it enters the "suspended" state.
   * Checked by <code>resume</code> to know if a call to <code>notifyAll</code>
   * should be issued.
   */
  private volatile boolean suspended = false;

  /**
   * The object used for mutual exclusion for the <code>suspend</code>/<code>
   * resume</code> mechanism.
   */
  Object suspendLock;

  /**
   * Allocates a new <code>CyclicThread</code>.
   * <p>
   * Equivalent to <code>CyclicThread(null, null, true)</code>.
   *
   * @see CyclicThread#CyclicThread(RTThreadGroup, java.lang.String, boolean)
   */
  public CyclicThread() {
    this(null, null);
  }

  /**
   * Allocates a new <code>CyclicThread</code>.
   * <p>
   * Equivalent to <code>CyclicThread(null, name, true)</code>.
   *
   * @param name this thread's name.
   *
   * @see CyclicThread#CyclicThread(RTThreadGroup, java.lang.String, boolean)
   */
  public CyclicThread(String name) {
    this(null, name);
  }

  /**
   * Allocates a new <code>CyclicThread</code>.
   * <p>
   * Equivalent to <code>CyclicThread(group, null, true)</code>.
   *
   * @param group this thread's <code>RTThreadGroup</code>.
   *
   * @see CyclicThread#CyclicThread(RTThreadGroup, java.lang.String)
   */
  public CyclicThread(RTThreadGroup group) {
    this(group, null);
  }

  /**
   * Allocates a new <code>CyclicThread</code>.
   * <p>
   * Equivalent to <code>CyclicThread(group, name, true)</code>.
   *
   * @param group this thread's <code>RTThreadGroup</code>.
   * @param name this thread's name.
   *
   * @see CyclicThread#CyclicThread(RTThreadGroup, java.lang.String)
   */
  public CyclicThread(RTThreadGroup group, String name) {
	super(group, name);
  	suspendLock = new Object();
  }

  /**
   * Defines the action performed by this thread.
   * A <b>CyclicThread</b> will execute cyclically
   * this function.
   */
  protected abstract void perform();

  /**
   * Defines the waiting policy of this thread.
   * A <b>CyclicThread</b> will call this method (possibly more than once)
   * before every call to <code>perform</code>.
   * <p>
   * The implementor is given the possibility of throwing or passing an
   * <code>InterruptedException</code>, which may cause the acknowledgment
   * of a <code>suspend</code> or a <code>stop</code>.
   *
   * @exception InterruptedException if another thread interrupts this thread.
   */
  protected abstract void waitToPerform() throws InterruptedException;

  /**
   * Asks the thread to suspend its execution. May not have an
   * immediate effect, since a CyclicThread is allowed to suspend its
   * execution only before each call to <code>perform</code>, to prevent
   * deadlocks. It is possible to suspend a thread which has not yet been started.
   * Calling <code>suspend</code> on a suspended thread has no effect.
   */
  public final void suspend() {
    /*
     * Note that this method is not synchronized, since it basically sets a one-way boolean,
     * which is guaranteed to be done atomically.
     */
    if (!suspended && !suspendReq) {
      suspendReq = true;
    }
  }

  /**
   * Checks whether this thread is suspended.
   * A thread is considered to be suspended since it acknowledges
   * a suspension request until the method
   * <code>onResuming</code> is called.
   *
   * @return <b>true</b> if the thread is suspended;
   *		<b>false</b> otherwise.
   *
   * @see CyclicThread#suspend()
   */
  public boolean isSuspended() {
    return suspended;
  }

  /**
   * Asks a suspended thread to resume execution.
   * If acknowledged, <code>onResuming</code> will be called.
   * It is possible to resume a thread which has not yet been started,
   * to undo a previous call to <code>suspend</code>..
   * If the thread is not suspended, <code>resume</code> has no effect.
   *
   * @see CyclicThread#suspend()
   */
  public final void resume() {
    suspendReq = false;
    if (suspended) {
      synchronized (suspendLock) {
        if (suspended) { // check again after having acquired the lock...
          suspendLock.notifyAll();
        }
      }
    }
  }

  /**
   * Asks this thread to stop its execution.
   * May not have an immediate effect, but is irrevocable.
   * Stopping a not yet started thread will cause it
   * to stop as soon as started. Stopping a stopped thread
   * has no effect.
   */
  public void stop() {
    stopReq = true;
    interrupt();
  }

  /**
   * Perform any adjustments required when resuming a thread.
   * This method is executed by the resumed thread after a
   * resumption acknowledgement. This default implementation of
   * <code>onResuming</code> does nothing.
   *
   * @see CyclicThread#resume()
   */
  protected void onResuming() {}

  /**
   * Defines the behaviour of a cyclic thread.
   * A <b>CyclicThread</b> acts like this:
   * <p>
   * 1. if has been asked to stop, it stops. <p>
   * 2. if has been asked to suspend itself, waits to be resumed
   *   and then calls <code>onResuming</code>. <p>
   * 3. calls <code>waitToPerform</code> (if an InterruptedException
   *   is caught, it goes to 1). <p>
   * 4. calls <code>perform</code>. <p>
   * 5. goes to 1.
   *
   * @see RTThread#run()
   */
  public final void run() {
	  System.out.println("Cyclic thread run()");
      if (currentThread() != this) {
	  throw new RTError("Cyclic thread trying to use run from another thread object.");
      }
mainloop: 
      {
	  while (!stopReq) {
	      if (suspendReq) {
		  synchronized(suspendLock) {
		      suspended = true;
		      while (suspendReq) {// check again after having acquired the lock...

			  
			  try {
			      suspendLock.wait();
			  } catch (InterruptedException e) {
			      if (stopReq) {
				  break mainloop;
			      }
			  }
			  
		      };
		      suspended = false;
		  }
		  onResuming();
		  // Here used to be: interrupted(); continue main loop;
		  // to clear the interrupted-flag and to continue with next 
		  // sample. But, interrupt is not useful to immediately let
		  // suspend have effect (since it interferes with catching
		  // InterrurptedException in libraries and user code). In fact,
		  // explicit checking of suspend in the running thread is the
		  // only(?) reliable way to suspend threads.
	      }
	  }

	  try {
	      try {
			  System.out.println("Wait to perform");
		  waitToPerform();
	      }
	      catch (InterruptedException exc) {
		  if (stopReq) {
		      break mainloop;
		  }
	      }
	  } catch (RTDelayed exc) {
	      //        System.out.println(exc.toString() + " in " + getName());
	      System.out.println(" in " + getName());
	  }

	  if (!suspendReq) {
	      perform();
	  }
      }
  }

  /**
   * Returns a string representation of this thread, including the
   * thread's name and priority.
   *
   * @return  a string representation of this thread.
   */
  public String toString() {
    return "CyclicThread[" + getName() + ", " + getPriority() + "]";
  }

}
