/* $Id: JThread.java,v 1.2 2005/08/22 08:18:22 d02en Exp $       
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

import se.lth.cs.realtime.event.*;

/**
 * Thread of concurrent execution according to the Java platform, but
 * extended with the facilities provided by the <code>RTThread</code>
 * class. Using the <code>JThread</code> class, the event processing
 * facilities provided by the <code>RTThread</code> is made available
 * in a subclass of <code>java.lang.Thread</code> (which the
 * <code>RTThread</code> is not). Another supported feature is the
 * method <code>sleepUntil</code>. Preferably, threads are based on
 * the <code>RTThread</code> which will run well on any Java platform
 * and enable real-time execution on a real-time (Java)
 * platform. However, for compatibility with existing software
 * (requiring <code>java.lang.Thread</code>s) in cases when real-time
 * execution is known not to be needed, sub-classing
 * <code>JThread</code> is useful. Thus, use <code>RTThread</code> for
 * real-time threads and use <code>JThread</code> for concurrent (Java)
 * threads.
 *
 * @see License
 * @see java.lang.Thread
 * @see RTThread
 */

public class JThread
  extends Thread
  implements Activity, RTEventListener, RTEventSource {

  private char name[];
  protected long maxTimeLag;

  /** The event input buffer. */
  protected volatile RTEventBuffer mailbox;

  /** Default size of the event buffer, used by constructor possibly after
   *  being changed by some subclass constructor. */
  protected int defaultEventBufferSize = RTSystem.defaultEventBufferSize;

  /**
   * The list of the listeners of this thread.
   *
   * @see se.lth.cs.realtime.event.RTEventDispatcher
   * @see se.lth.cs.realtime.event.RTEventSource
   */
  protected RTEventDispatcher dispatcher = new RTEventDispatcher(this);

  /**
   * Put the time-stamped event in the input buffer of this thread.
   * If the buffer is full, the caller is blocked.
   *
   * @return null, but subclasses are free to return an event if desired
   *         to inform about blocking or results.
   */
  public RTEvent putEvent(RTEvent ev) {
    mailbox.doPost(ev);
    return null;
  }

  /**
   * Obtain the object that internally stores the subscriptions of events,
   * often referred to as the listener list. Opposed to listener lists in
   * GUI packages (such as AWT and Swing), where the listener typically gets
   * its processing method called and fully performed in the context of the
   * GUI thread, the subscribers/listeners of <code>RTEvent</code>s gets a
   * call of <code>putEvent</code> which typically only stores/buffers the
   * event for later processing by the receiving thread (object). Thus,
   * the <code>EventSource</code> simply dispatches the event distribution,
   * hence the name dispatcher of the lister list storage object.
   */
  public RTEventDispatcher getDispatcher() {
    return dispatcher;
  }

  /**
   * Register <code>subscriber</code> as listener of application events
   * emitted from this thread object. An <code>Activity</code> calling
   * <code>OtherThread.addListener(this);</code> is the same as
   * <pre>
     OtherThread.getDispatcher().addListener(
         this, RTThread.currentThread().getPriority(), null);
   * </pre>
   */
  public void addListener(RTEventListener subscriber) {
    RTThread thread = RTThread.currentRTThread();
    if (thread == null) {
      dispatcher.addListener(
        subscriber,
        Thread.currentThread().getPriority(),
        null);
    }
    else {
      dispatcher.addListener(subscriber, thread.getPriority(), null);
    }
  }

  /**
   * Causes the currently executing thread to sleep (temporarily cease
   * execution) until the specified moment in time.
   * Time is measured in milliseconds, starting from 00am, January 1st
   * 1970, UTC. The thread does not lose ownership of any monitors.
   *
   * @param  wakeUpTime   the time to wake up.
   * @throws InterruptedException if another thread has interrupted
   *         this thread.
   * @throws RTDelayed if execution continues more than
   *         <code>maxTimeLag</code> too late.
   */
  public static void sleepUntil(long wakeUpTime)
    throws RTInterrupted, RTDelayed {
    RTSystem.sleepUntil(wakeUpTime);
  }

  /**
   * Causes the currently executing thread to sleep (temporarily cease
   * execution) for the specified number of milliseconds. The thread
   * does not lose ownership of any monitors.
   *
   * @param      millis   the length of time to sleep in milliseconds.
   * @exception  IllegalArgumentException  if the value of millis is
   *             negative.
   * @exception  InterruptedException  if another thread has interrupted
   *               this thread.
   * @see        java.lang.Object#notify()
   */
  public static void sleep(long millis) throws InterruptedException {
    if (millis < 0) {
      throw new IllegalArgumentException("timeout value is negative");
    }
    long wakeUp = System.currentTimeMillis() + millis;
    sleepUntil(wakeUp);
  }

  /**
   * Causes the currently executing thread to sleep (cease execution)
   * for the specified number of milliseconds plus the specified number
   * of nanoseconds. The thread does not lose ownership of any monitors.
   *
   * @param      millis   the length of time to sleep in milliseconds.
   * @param      nanos    0-999999 additional nanoseconds to sleep.
   * @exception  IllegalArgumentException  if the value of millis is
   *             negative or the value of nanos is not in the range
   *             0-999999.
   * @exception  InterruptedException  if another thread has interrupted
   *               this thread.
   */
  public static void sleep(long millis, int nanos)
    throws InterruptedException {
    if (millis < 0) {
      throw new IllegalArgumentException("timeout value is negative");
    }

    if (nanos < 0 || nanos > 999999) {
      throw new IllegalArgumentException("nanosecond timeout value out of range");
    }

    if (RTSystem.tickMillis > 0) {
      if (nanos != 0)
        millis++;
    }
    else {
      nanos = nanos * (nanos / RTSystem.tickNanos + 1);
      if (nanos > 999999) {
        nanos -= RTSystem.tickNanos;
        millis++;
      }
    }
    if (nanos == 0) {
      long wakeUp = RTSystem.currentTimeMillis() + millis;
      sleepUntil(wakeUp);
    }
    else {
      throw new UnsupportedOperationException("nano sleep not available");
    }
  }

  /**
   * Allocates a new <code>JThread</code> object with
   * a newly generated name in its father's <code>ThreadGroup</code>.
   * Automatically generated names are of the
   * form <code>"Thread-"+</code><i>n</i>, where <i>n</i> is an integer.
   */
  public JThread() {
    super();
    if (defaultEventBufferSize > 0) {
      mailbox = new Buffer(defaultEventBufferSize);
    }
  }

  /**
   * Allocates a new <code>JThread</code> object with
   * the specified name in its parent's <code>ThreadGroup</code>.
   *
   * @param   name   the name of the new thread.
   */
  public JThread(String name) {
    super((ThreadGroup) null, name);
    if (defaultEventBufferSize > 0) {
      mailbox = new Buffer(defaultEventBufferSize);
    }
  }

  /**
   * Allocates a new <code>JThread</code> object with a newly generated name
   * in the specified <code>ThreadGroup</code>.
   * <p>
   * Automatically generated names are of the
   * form <code>"Thread-"+</code><i>n</i>, where <i>n</i> is an integer.
   *
   * @param   group   the <code>ThreadGroup</code> of the new thread.
   */
  public JThread(ThreadGroup group) {
    super(group, (String) null);
    if (defaultEventBufferSize > 0) {
      mailbox = new Buffer(defaultEventBufferSize);
    }
  }

  /**
   * Allocates a new <code>JThread</code> object with
   * the specified name in the specified <code>ThreadGroup</code>.
   *
   * @param   name   the name of the new thread.
   * @param   group   the <code>ThreadGroup</code> of the new thread.
   */
  public JThread(ThreadGroup group, String name) {
    super(group, name);
    if (defaultEventBufferSize > 0) {
      mailbox = new Buffer(defaultEventBufferSize);
    }
  }

  /**
   * Default non-empty <code>run</code> method, first calling <code>init</code>, 
   * then calling <code>perform</code> while not interrupted, and finally calling 
   * exit. Hence, to define a repetitive thread, you can keep this run method and
   * instead override the <code>perform</code> method. As well known, creation of 
   * a new thread is accomplished by calling <code>start</code>, which in turn 
   * results in (in the context of the new thread) <code>run</code> being called.
   */
  public void run() {
    init();
    while (!isInterrupted()) {
      perform();
    }
    exit();
  }
  
  /**
   * Simple trigger method which puts a TriggerEvent object in this.mailbox.
   */
  protected void trigger() {
	  mailbox.doPost(new TriggerEvent());
  }

  /**
   * Empty perform method, to be replaced in subclasses.
   */
  protected void perform() {};

  /**
   * Called by typical run methods, such as the default one, to give a this thread
   * a chance to clean up before it actually exits.
   */
  protected void init() {};

  /**
   * Called by typical run methods, such as the default one, to give a this thread
   * a chance to clean up before it actually exits.
   */
  protected void exit() {};

  /**
   * Time bases not implemented yet.
   * @return null.
   */
  public Timebase getTimebase() {
    return null;
  }

  /**
   * Environments not implemented yet.
   * @return null.
   */
  public Environment getEnv() {
    return null;
  }

  /**
   * Interrupt and join.
   */
  public void terminate() {
    if (isAlive() && !isInterrupted()) {
      interrupt();
      try {
        join();
      }
      catch (InterruptedException exc) {
        throw new RTInterrupted(
          "Termination (join) of " + String.copyValueOf(name) + " interrupted");
      }
    }
  }
}
