/* $Id: Main.java,v 1.6 2005/08/19 09:37:16 d02en Exp $       
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

import java.io.*;
import java.lang.reflect.*;
import org.objectweb.asm.*;

/**
 * Main program providing the method/entry main, always to be used to start the
 * application (class containing main provided as the first argument to this
 * Main.main method), thereby permitting an enhanced class-loader to be used.
 * 
 * @see License
 */
public class Main {

	private static String currentClassName = "";
	private static final String mutex = "Monitor$mutex";

	/**
	 * Modified ClassLoader transforming a class before loading.
	 */
	private static final class TransformClassLoader extends ClassLoader {
		
		private ClassLoader cl;
    
		public TransformClassLoader(ClassLoader cl) {
			super(cl);
			this.cl = cl;
		}
    
		public Class loadClass(String name) throws ClassNotFoundException {
			
			try {
				currentClassName = name.replace('.', '/');
				String fileName = currentClassName + ".class";
				
				InputStream is = cl.getResourceAsStream(fileName); 
				
				if (name == null || is == null) {
					System.err.println("Couldn't find class");
				}
			
				ClassReader r = new ClassReader(is);
				ClassWriter w = new ClassWriter(true, true);
				Transformer t = new Transformer(w);
			
				r.accept(t, false);
				
				byte[] bytecode = w.toByteArray();
				
				Class c = null;
				try {
					c = super.defineClass(name, bytecode, 0, bytecode.length);
				} catch (SecurityException e) {
					c = cl.loadClass(name);
				}
				
				return c;
			}
			catch (IOException e0) {
				System.err.println("EXCEPTION [" + name + "] : " + e0);
				return cl.loadClass(name);
			}
		    catch (ClassFormatError e1) {
		    	//- If the data did not contain a valid class
		    	System.err.println("EXCEPTION [" + name + "] : " + e1);
				return cl.loadClass(name);
		    }
		    catch (IndexOutOfBoundsException e2) {
		    	//- If either off or len is negative, or if off+len is greater than b.length.
		    	System.err.println("EXCEPTION [" + name + "] : " + e2);
				return cl.loadClass(name);
		    }
		    catch (NullPointerException e4) {
				System.err.println("EXCEPTION [" + name + "] : " + e4);
				return cl.loadClass(name);
			}
		}
	}
  
	
	/**
	 * ClassVisitor which looks for classes implementing Monitor.
	 */
	private static class Transformer extends ClassAdapter {
	  
		public static boolean isMonitorClass;
		public static boolean isSynchMethod;
		
		/**
		 * Creates a Transformer which will modify classes implementing
		 * Monitor by removing the synchronized keyword and creating
		 * mutual exclusion using a MutexSem instead.
		 */
		public Transformer(ClassVisitor cv) {
			super(cv);
			isSynchMethod = false;
			isMonitorClass = false;
		}
    
		/**
		 * Searches for classes implementing Monitor. If such a 
		 * class is found a MutexSem field is added to the class.
		 */
		public void visit(int version, int access, String name, 
				String signature, String superName, String[] interfaces) {
			
			String monitor = "se/lth/cs/realtime/Monitor";
			isMonitorClass = false;
			
			for (int i = 0; i < interfaces.length; i++) {
	
				if(interfaces[i].equals(monitor)) {
				
					isMonitorClass = true;
					
					// ADD: MutexSem mutex;
					visitField(Opcodes.ACC_PRIVATE, mutex, "Lse/lth/cs/realtime/semaphore/MutexSem;", null, null);
					
					break;
				}
			}
			
			super.visit(version, access, name, signature, superName, interfaces);
			
			return;
		}
		
		/**
		 * Modifies synchronized methods within a class implementing Monitor
		 * by removing the synchronized keyword and adding a mutex.take() call
		 * in the beginning of the method.
		 */
		public MethodVisitor visitMethod(int access, String name, 
					String desc, String signature, String[] exceptions) {
		
			isSynchMethod = (access & Opcodes.ACC_SYNCHRONIZED) != 0; 
			
			if (isMonitorClass) {
 			
				if (isSynchMethod) {
			
					// REMOVE: synchronized
					MethodVisitor cd = cv.visitMethod(access & ~Opcodes.ACC_SYNCHRONIZED,
							name, desc, signature, exceptions);
				
					if (cd == null)
						return null;
				
					// ADD: mutex.take();
					cd.visitVarInsn(Opcodes.ALOAD, 0);
					cd.visitFieldInsn(Opcodes.GETFIELD, currentClassName, mutex, "Lse/lth/cs/realtime/semaphore/MutexSem;");
					cd.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "se/lth/cs/realtime/semaphore/MutexSem", "take", "()V");
				
					// ADD: System.out.println("mutex.take()");
					//cd.visitFieldInsn(Opcodes.GETSTATIC, "java/lang/System", "out", "Ljava/io/PrintStream;");
					//cd.visitLdcInsn(mutex + ".take()");
					//cd.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/String;)V");
				
					return new MethodCodeVisitor(cd);
					
				}
				
				MethodVisitor cd = cv.visitMethod(access, name, desc, signature, exceptions);
				
				if (cd == null) 
					return null;
				
				if (name.equals("<init>")) {
					
					// ADD: mutex  = new MutexSem();
					cd.visitVarInsn(Opcodes.ALOAD, 0);
					cd.visitTypeInsn(Opcodes.NEW, "se/lth/cs/realtime/semaphore/MutexSem");
					cd.visitInsn(Opcodes.DUP);
					cd.visitMethodInsn(Opcodes.INVOKESPECIAL, "se/lth/cs/realtime/semaphore/MutexSem", "<init>", "()V");
					cd.visitFieldInsn(Opcodes.PUTFIELD, currentClassName, mutex, "Lse/lth/cs/realtime/semaphore/MutexSem;");
					
					// ADD: System.out.println("mutex = new MutexSem()");
					//cd.visitFieldInsn(Opcodes.GETSTATIC, "java/lang/System", "out", "Ljava/io/PrintStream;");
					//cd.visitLdcInsn(mutex + " = new MutexSem()");
					//cd.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/String;)V");
				}
				
				return new MethodCodeVisitor(cd);
			}
			
			MethodVisitor cd = cv.visitMethod(access, name, desc, signature, exceptions);
			return cd == null ? null : new MethodAdapter(cd);
		}
	}

  
	/**
	 * Method visitor changing the following:
	 * - If a Constructor.Factory was created with a Monitor as argument
	 *   the argument is changed to the mutex created in the Class visitor.
	 * - A call to mutex.give() is added at each return call in synchronized 
	 *   methods.
	 */
	private static class MethodCodeVisitor extends MethodAdapter {
	  
		public MethodCodeVisitor(MethodVisitor cv) {
			super(cv);
		}

		/**
		 * Modifies call to Condition.Factory constructors if needed.
		 */
		public void visitMethodInsn(int opcode, String owner, String name, String desc) {
			
			if (opcode == Opcodes.INVOKESPECIAL && name.equals("<init>") && 
					owner.equals("se/lth/cs/realtime/Condition$Factory") &&
					desc.equals("(Lse/lth/cs/realtime/Monitor;)V")) {
				
				// ADD: System.out.println("new Condition.Factory(mutex)");
				//visitFieldInsn(Opcodes.GETSTATIC, "java/lang/System", "out", "Ljava/io/PrintStream;");
				//visitLdcInsn("new Condition.Factory(" + mutex + ")");
				//visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/String;)V");
				
				// MODIFY: new Condition.Factory(Monitor) -> new Condition.Factory(MutexSem)
				visitFieldInsn(Opcodes.GETFIELD, currentClassName, mutex, "Lse/lth/cs/realtime/semaphore/MutexSem;");
				super.visitMethodInsn(opcode, owner, name, "(Lse/lth/cs/realtime/semaphore/MutexSem;)V");
			} 
			else {
				super.visitMethodInsn(opcode, owner, name, desc);
			}
		}

		/**
		 * Modifies return instructions in synchronized methods with in 
		 * a class implementing Monitor.
		 */
		public void visitInsn(int opcode) {

			if (Transformer.isMonitorClass && Transformer.isSynchMethod &&
					(opcode == Opcodes.RETURN || opcode == Opcodes.IRETURN ||
							opcode == Opcodes.ARETURN || opcode == Opcodes.DRETURN ||
							opcode == Opcodes.FRETURN || opcode == Opcodes.LRETURN)) {
				
				// ADD: System.out.println("mutex.give()");
				//visitFieldInsn(Opcodes.GETSTATIC, "java/lang/System", "out", "Ljava/io/PrintStream;");
				//visitLdcInsn("mutex.give()");
				//visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/String;)V");
				
				// ADD: mutex.give();
				visitVarInsn(Opcodes.ALOAD, 0);
				visitFieldInsn(Opcodes.GETFIELD, currentClassName, mutex, "Lse/lth/cs/realtime/semaphore/MutexSem;");
				visitMethodInsn(Opcodes.INVOKEVIRTUAL, "se/lth/cs/realtime/semaphore/MutexSem", "give", "()V");
			}
			
			super.visitInsn(opcode);
		}
	}

	
	public static void main(String[] args) {
	
		try {
			
			if(args.length == 0) {
				
	          System.err.println("Usage: java Main main-class");
	          System.exit(1);
	        }
			
			System.out.println("args[0]=" + args[0]);
			
			// Load class
			String className = args[0];
	        ClassLoader cl = 
	        	new TransformClassLoader(Thread.currentThread().getContextClassLoader());
	        Thread.currentThread().setContextClassLoader(cl);
	        Class c =  cl.loadClass(className);
	        
	        // Invoke class
	        Method m = c.getMethod("main", new Class[] { String[].class });
	        String[] newArgs = new String[args.length - 1];
	        System.arraycopy(args, 1, newArgs, 0, args.length - 1);
	        m.invoke(null, new Object[] { newArgs });
	        
	      }
	      catch (Exception e) {
	    	  e.printStackTrace();
	      }	
	}
}