/* $Id: PeriodicThread.java,v 1.6 2005/08/19 09:37:16 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * This class models a real-time thread with periodic behavior.
 * <p>
 * Such a thread loops performing an action once every <code>period
 * </code> milliseconds (if <code>FixedPeriod</code> is implemented,
 * the period can be modified only before the thread is started).
 * <p>
 * Being a <code>CyclicThread</code>,
 * a <code>PeriodicThread</code> may be suspended and resumed
 * by other threads. By default, the calls to <code>perform</code>
 * which would have been executed during the suspension time are 
 * issued after the resumption.
 * 
 * @see License
 */

public class PeriodicThread extends CyclicThread {

  /**
   * The period of this thread, in milliseconds.
   */
  private long period;

  /**
   * Incremented when the period has been requested to be changed.
   */
  private volatile short periodChange;

  /**
   * Incremented when the period has been changed.
   */
  private short periodChanged;

  /**
   * The temporary storage for the new period.
   * This thread will copy this value into <code>period</code> if
   * it finds that <code>periodChanged</code> is set.
   * In this way we avoid synchronizing the access to <code>period</code>,
   * which is read at every cycle by this thread, even though we must
   * check <code>periodChanged</code> at every cycle.
   */
  private long newPeriod;

  /**
   * The absolute time (calculated in ms since midnight, January 1, 1970 UTC)
   * used to compute the time of the next call to <code>perform</code>.
   * <p>
   * <b>PLEASE NOTE:</b> in this implementation, this field is given
   * protected visibility, thus allowing easy modifications from subclasses,
   * but also, possibly, from within other threads of execution, which could
   * result in inconsistent values of <code>baseTime</code>, being a <code>long
   * </code>.
   * Hence, care must be taken that <code>baseTime</code> be accessed only 
   * by this thread itself (i.e. in <code>onResuming</code>).
   */
  protected long baseTime;
  /*
   * The above problem could be solved with the policy used in setPeriod,
   * but a real need to allow changes from other threads
   * seems quite unlikely.
   * "baseTime" has not been declared volatile to further discourage any
   * concurrent access to it.
   */

  /**
   *  Time when next sample should start, basically only needed within 
   * <code>waitToPerform</code> but hereby also exposed for use in the 
   * <code>setPeriod</code> method.
   */
  private volatile long nextPerformanceTime;

  /**
   * Allocates a new <code>PeriodicThread</code> object.
   * <p>
   * Equivalent to <code>PeriodicThread(null, null, true, period)</code>.
   *
   * @param	period	    the period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the period is not positive.
   
   * @see PeriodicThread#PeriodicThread(RTThreadGroup, 
   *                                    java.lang.String, boolean, long)
   */
  public PeriodicThread(long period) {
    this(null, null, period);
  }

  /**
   * Allocates a new <code>PeriodicThread</code> object.
   * <p>
   * Equivalent to <code>PeriodicThread(null, name, true, period)</code>.
   *
   * @param   name	    the name of the new thread.
   * @param	period	    the period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the period is not positive.
   *
   * @see PeriodicThread#PeriodicThread(RTThreadGroup, 
   *                                    java.lang.String, boolean, long)
   */
  public PeriodicThread(String name, long period) {
    this(null, name, period);
  }

  /**
   * Allocates a new <code>PeriodicThread</code> object.
   * <p>
   * Equivalent to <code>PeriodicThread(group, null, true, period)</code>.
   *
   * @param	group	    this thread's <code>RTThreadGroup</code>.
   * @param	period	    the period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the period is not positive.
   *
   * @see PeriodicThread#PeriodicThread(RTThreadGroup, 
   *                                    java.lang.String, long)
   */
  public PeriodicThread(RTThreadGroup group, long period) {
    this(group, null, period);
  }

  /**
   * Allocates a new <code>PeriodicThread</code> object.
   * <p>
   * Equivalent to <code>PeriodicThread(group, name, true, period)</code>.
   *
   * @param   name	    the name of the new thread.
   * @param	group	    this thread's <code>RTThreadGroup</code>.
   * @param	period	    the period of the new thread, in milliseconds.
   *
   * @exception IllegalArgumentException if the period is not positive.
   *
   * @see PeriodicThread#PeriodicThread(RTThreadGroup, 
   *                                    java.lang.String, long)
   */
  public PeriodicThread(RTThreadGroup group, String name, long period) {
  	super(group, name);
    try {
      setPeriod(period);
    }
    catch (FixedPeriodError e) {/* impossible */}
    this.setMaxTimeLag((int)period);
  }

  /**
   * Overrides <code>RTThread.start</code>. Perform the settings
   * needed to ensure a correct periodical behavior and calls
   * <code>RTThread.start</code>, which in turns causes the call
   * of <code>CyclicThread.run</code>.
   *
   * @see RTThread#start()
   * @see CyclicThread#run()
   */
  public void start() {
    baseTime = System.currentTimeMillis();
    super.start();
  }

  /**
   * Does nothing.
   * This empty implementation allows the use of a <b>PeriodicThread</b>
   * as a <code>PeriodicEvent</code> generator only.
   *
   * @see se.lth.cs.realtime.event.PeriodicEvent
   * 
   */
  protected void perform() {}

  /**
   * Wait until the beginning of the next period.
   *
   * @exception InterruptedException if another thread has interrupted this one.
   */



	/**
	 * UGLY HACK!  Simplest possible implementation which it makes it
	 * possible to use PeriodicThread with the LJRT system.
	 */
	public void run() {
		long nextPerf = baseTime;
		while (true) {
			perform();
			nextPerf += period;
			sleepUntil(nextPerf);
		}
	}



  protected void waitToPerform() throws InterruptedException {
  	if (this instanceof FixedPeriod) {
  		nextPerformanceTime = baseTime + period;
  		sleepUntil(nextPerformanceTime);
  	} else {
  		boolean isNewPeriod = false;
  		if (periodChange != periodChanged) {
  			synchronized (this) {
  				period = newPeriod;
  				periodChanged = periodChange;
  				notifyAll();
  			}
  		}
  		nextPerformanceTime = baseTime + period;
  		try {
  			sleepUntil(nextPerformanceTime);
  		} catch (RTInterrupted exc) {
  			if (periodChange != periodChanged) {
  				isNewPeriod = true;
  				interrupted(); // Clear interrupted flag.
  			}
  		}
  		if (isNewPeriod || periodChange != periodChanged) {
  			synchronized (this) {
  				period = newPeriod;
  				periodChanged = periodChange;
  				notifyAll();
  			}
  		}
  	}
    baseTime = nextPerformanceTime;
  }

  /**
   * Returns the period of this thread, in milliseconds.
   *
   * @return the period in ms.
   */
  public synchronized long getPeriod() {
    return period;
  }

  /**
   * Sets the period, in milliseconds, of this thread.
   * If this thread implements <code>FixedPeriod</code> and is alive, an
   * exception is thrown.
   *
   * @param period the new period of this thread, in ms.
   *
   * @exception IllegalArgumentException if the period is not positive.
   * @exception FixedPeriodError if this thread implements <code>
   *		    FixedPriority</code> and is alive.
   */
  public synchronized void setPeriod(long period) throws FixedPeriodError {
  	if (period <= 0) {
  		throw new IllegalArgumentException("period value must be positive.");
  	}
//   	if (isAlive()) {
//   		if (this instanceof FixedPeriod) {
//   			throw new FixedPeriodError();
//   		}
//   		newPeriod = period;
//   		periodChanged++;
//   		while (periodChanged != periodChange 
//   				&& nextPerformanceTime > RTSystem.currentTimeMillis()) {
//   			// Thread is sleeping, waiting for next sample...
//   			interrupt(); // To immediately activate the change
//   			Condition.await(this);
//   		}
//   	}
//   	else {
  		this.period = period;
//   	}
  }

  /**
   * Returns a string representation of this thread, including the 
   * thread's name, priority and period in milliseconds.
   *
   * @return  a string representation of this thread.
   */
  public String toString() {
    return "PeriodicThread["
      + getName()
      + ", "
      + getPriority()
      + ", "
      + getPeriod()
      + "]";
  }

}
