/* $Id: RTSystemProperties.java,v 1.4 2005/08/19 09:37:16 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * Compile-time general system settings and properties, reflecting 
 * execution on an ordinary JVM on an ordinary operating system. 
 * <p>
 * For JVMs with different properties, create your own version of this 
 * class. According to plans for future implementations, user version
 * should be situated in the default package and should replace this 
 * default implementation by means of the class loading according to
 * class <code>Main</code>. Until that is implemented, you have to 
 * create your own se.lth.cs.RTSystemProperties object and make sure
 * the system class loader loads your version before the default.
 * <p>
 * For compilation to non-JVM real-time platforms, the compiler
 * should support changing these constants before the binary is created
 * (for the specific target that constants should reflect). For 
 * translation via C this means that there should be a header file
 * where compile-time constants can be redefined before compilation.
 * 
 * Note that there is no setting stating if the system has real-time
 * properties. The application source should be real-time compliant if
 * RTThreads are used, if the application is deployed on a real-time
 * platform. If such an application is run on an OS without real-time
 * capabilities, there will be no real-time properties anyhow, but the
 * application should behave correctly from a concurrency point of view.
 * 
 * @see License
 */

public interface RTSystemProperties {

	/**
	 * During development, debugging support (for instance by means of a 
	 * real-time clock that can be stopped at breakpoints) is desired and activated
	 * by <code>DEPLOYED==false</code>. The final software to be deployed/shipped
	 * is optimized (at compile/load/run-time) by setting <code>DEPLOYED==true</code>.
	 */
	boolean DEPLOYED = false;

	/**
	 * The real-time clock tick of the currently used kernel or operating system. 
	 * This should be the minimum quantity that the value returned by 
	 * <code>System.currentTimeMillis</code> could change, that is, the 
	 * quantization of system time in a tick-based system. This means that 
	 * two readings of the system time returning the same value could refer 
	 * to actual times that differ at the most up to <code>OS_TICK_MICRO</code>
	 * micro second. A value of 10000 (10ms) is then typical to be compliant 
	 * with most (all?) existing computer platforms, even if many platforms 
	 * after year 2000 have a tick of 1ms (which is OK since time differences 
	 * are allowed to be less than <code>OS_TICK_MICRO</code>). 
	 * <p>
	 * To better model
	 * a specific target platform, change the value accordingly. To improve 
	 * runtime efficiency during debugging (when <code>DEPLOYED==false</code>),
	 * increase the value to increase the granularity of the virtual time.
	 * <p>
	 * A value of zero will give a virtual-time increment (when 
	 * <code>DEPLOYED==false</code>) that is system dependent (based on
	 * <code>Runtime.getProperties</code>) but should be suitable for debugging.
	 * When <code>OS_TICK_MICRO==0</code>) for a <code>DEPLOYED</code> system,
	 * the timing of the OS/kernel is not tick based and the system time 
	 * (for wait, sleep, and current time) has a granularity of 1 micro second
	 * or less.
	 */
	int OS_TICK_MICRO = 10000;

	/**
	 * When <code>PRIORITY_BASED==true</code>, the system scheduling is based on 
	 * priorities, as is the case for a standard JVM or RTOS. When 
	 * <code>PRIORITY_BASED==false</code>, priorities still (for Java compliance)
	 * exist and express some kind of importance, but the scheduling in this case 
	 * is based on other attributes such as deadlines.
	 */
	boolean PRIORITY_BASED = true;

	/**
	 * The scheduler of the runtime system never permits lower priority threads to run if
	 * any thread with higher priority is ready to run, if <code>PRIORITY_STRICT==true</code>. 
	 * For typical desktop and server platforms (Linux, Solaris, Windows, Mac-OSX) as 
	 * supported by normal JVMs, this setting will be false.
	 * Set to true if (you for sure knows and needs to know that) deployment applies to
	 * a target platform that provides strict priorities. Recall that concurrency-correct
	 * code needs to work correctly for any (in the source) permitted interleaving of 
	 * the execution, and that strict priorities are not required in the Java standard.
	 * Making use of <code>PRIORITY_STRICT</code> can be useful to switch in quite special
	 * and optimized alternative implementations (in addition to the default thread-safe 
	 * code that works for any platform).
	 */
	boolean PRIORITY_STRICT = false;

	/**
	 * Since <code>Object.wait(timeout)</code> with a timeout in milliseconds is standard for 
	 * any Java platform this setting should always by true. However, with
	 * <code>TIMEOUT_MS==false</code> you state that there is no timer thread 
	 * handling timeouts, which could be the case for some dedicated controller
	 * binaries (for micro-controllers, for single-threaded plug-ins, for Linux 
	 * kernel modules without RTAI, etc.), and the compilation or deployment 
	 * procedure could optimize the system further.
	 */
	boolean TIMEOUT_MS = true; // Always true for JVM execution.

	/**
	 * Some methods, such as <code>java.lang.Object.wait(timeout_ms,timeout_ns)</code>,
	 * accept a nano-second timeout value. In most (all up till 2005) JVMs the nano 
	 * value is, however, not supported in the actual timeout handling. Even worse, the 
	 * JVMs as provided by Sun rounds the nano value to the nearest millisecond value, 
	 * effecting the supported timeout value in millisecond. This means that with the Sun 
	 * classes with <code>TIMEOUT_NS==false</code> and when using nano timeouts of less than 
	 * 500000, a java.lang.Thread.sleep could actually return half a millisecond before
	 * the provided sleeping time. Both with targets supporting <code>TIMEOUT_NS==true</code>
	 * and when using only the classes of this package, such flaws are avoided.
	 */
	boolean TIMEOUT_NS = false;

	/**
	 * When <code>NATIVE_SEM==false</code> semaphores are implemented by other
	 * pure Java mechanisms (that is, using synchronized). There is then no
	 * native libraries needed when using the
	 * <code>se.lth.cs.realtime.semaphore</code> package, but semaphore
	 * operations are less efficient and semaphores cannot be given (signalled)
	 * from interrupt routines (which exists natively only anyhow).
	 * <p>
	 * With <code>NATIVE_SEM==true</code> the Java compiler (or translator to
	 * C-code) is expected to replace semaphore operations by the corresponding
	 * native call, or the semaphore implementations will explicitly load/use
	 * native functions in the case of JVM execution.
	 */
	boolean NATIVE_SEM = false;
}
