/* $Id: Synchronized.java,v 1.6 2005/08/19 09:37:16 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * Empty interface stating that the implementing class is thread safe, typically by 
 * means of synchronized methods. An object instance could then (by use of special 
 * tools) be checked that it conceptually comprises a monitor [Hoare 1974], with some
 * differences that can be further avoided by using the <code>Monitor</code> interface.
 * <p>
 * There are five ways of implementing thread-safe objects:
 * <ol>
 * <li> Mutual exclusion by means of the Java built-in construct synchronized.
 * <li> Mutual exclusion by other mechanisms such as semaphores.
 * <li> Atomic operation by disabling interrupts or preventing preemption.
 * <li> The method only uses local data (allocated on the stack of each thread) or 
 *      external data that comprises constants (such as Pi in Math) or monitors (such
 *      as the <code>mailbox</code> monitor inside <code>JThread.putEvent</code>.
 * <li> Attributes are public (or package visible) but declared <volatile> and the logic
 *      of the code is written such that object interaction is thread safe.
 * </ol>
 * Items 1 and 4 represent the normally acceptable alternatives. Classes that 
 * provide methods that are all thread safe according to these alternatives could/should 
 * implement <code>Synchronized</code>, thereby permitting additional checking of the 
 * code (by special tools).
 * <p>
 * <b>Standard simple programming:</b> 
 * Implement this (empty) interface, declare all methods synchronized, have attributes 
 * protected or private, and beware (for method implementations) that threads using 
 * other objects may run code that calls (and interferes with) wait and notify on this 
 * object.
 * <p>
 * <b>Advanced programming:</b> 
 * In Java where the mutual exclusion of monitors are accomplished by the use of the keyword
 * <code>synchronized</code>, whereas monitors as defined by Hoare are not explicitly
 * supported in the language or in the standard libraries (but the name of the   
 * byte-codes refer to monitors). Due to the lack of language support, 
 * external tools or test suits have to check that a class that implements Synchronized
 * really comprises a monitor, with some refinements due to object orientation. 
 * Specifically, this means that the class implementing this interface should:
 * <ul>
 * <li> Have all non-private methods <code>synchronized<code>.
 * <li> Have no public or package-visible fields.
 * <li> Any protected (which implies package visibility) field may not be referenced 
 *      from within the same package.
 * <li> Methods that are synchronized, without being declared synchronized according to
 *      the first item, should use an internal non-public object as an attribute for locking. 
 *      That is, with a private <code>lock</code> an unsynchronized method can contain 
 *      <code>synchronized(lock){...}</code>, which prevents 
 *      wait/notify-interference from outside the monitor. A method that is synchronized 
 *      this way may not access monitor fields outside the scope of synchronized, but 
 *      local data and calls of other thread-safe methods are permitted. For example, 
 *      the following is also acceptable since <code>arrival</code> will be allocated 
 *      on the stack of each calling thread and <code>currentTimeMillis()</code> is 
 *      thread safe.
 *      <pre>
 *      public void monitorMethod() {
 *          long arrival = RTSystem.currentTimeMillis(); // This line is OK!!
 *          synchronized (lock) {
 *              // Using monitor attributes here....
 *          }
 *      }</pre>
 * <li> Any <code>static</code> non-final attributes may only be referenced by 
 *      <code>static synchronized</code> methods, or by synchronizing (inside the 
 *      non-static monitor object method) on the class object for the entire scope 
 *      of the non-static method. 
 * </ul>
 * Since <code>synchronized</code> is not inherited, is is very important that subclasses 
 * of classes implementing <code>Synchronized</code> also obey the items above, remembering
 * to implement each (virtual synchronized) method according to the rules above.
 * 
 * @see License 
 */

public interface Synchronized {
	
	final static class Lock implements Synchronized {}
}
