/* $Id: Timebase.java,v 1.2 2005/08/22 08:45:17 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * The time-base class, mapping ticks to time.
 *
 * This class is needed for the following reasons: <ul>
 * <li> To obtain necessary resolution of the time variable, both for ordinary
 *      PCs/workstations and fast embedded controllers, even if the time is
 *      stored in a variable of a built in data type (long in this case). </li>
 * <li> To permit applications to run in a local timescale using a
 *      real-time clock that is not calibrated w.r.t. real-world time. This
 *      is the case in many small control applications. </li>
 * <li> To permit subsystems (e.g. for planning or simulation) to run in
 *      a local time-scale. This will permit hierarchical time-bases (via the
 *      <code>envTimebase</code>) but that is not yet supported.
 * </ul>
 *
 * Ordinary applications/users: Use default constructor .... this class
 * currently only supports the default System time-base. A related interface
 * is the <code>javax.media.TimeBase</code>
 * 
 * @see License
 * @see se.lth.cs.realtime.event.RTEvent
 */

public class Timebase {

  static final long REL_TIME = Integer.MIN_VALUE;

  /**
   * The start of time, in milliseconds relative to beginning of 1970.
   */
  protected final long start;

  /**
   * Seconds per tick.
   */
  protected final double scale;

  /**
   * Ticks per second.
   */
  protected final int rate;

  /**
   * Ticks per milli-second.
   */
  protected final int milliRate;

  /**
   * Ticks per nano-seconds.
   */
  protected final int nanoRate;

  /**
   * The time base of the surrounding system, such as the real world.
   * If null, the system time is used. That is, System.currentTimeMillis()
   * is valid in the scale of time since this is an ordinary RT application.
   */
  protected Timebase envTimebase;

  /**
   * Constructor for default system time, using ticks for milli-seconds
   * from beginning of 1970 as for System.currentTimeMillis().
   */
  public Timebase() {
    this(0, 1000);
  }

  /**
   * Constructor for a time base starting with time zero at <code>origin</code>
   * and with <code>pace</code> system dependent ticks per second.
   * To represent the time in an embedded controller without battery supported
   * clock (time realative to start of system) use <code>REL_TIME</code> as
   * the <code>origin</code> argument.
   */
  public Timebase(long origin, long pace) {
    start = origin;
    rate = (int) pace;
    milliRate = (int)(pace / 1000);
    nanoRate = (int) (pace / 1000 / 1000 / 1000);
    scale = 1.0 / (double) pace;
  }

  /**
   * Obtain physical real-world time. In ordinary applications the returned
   * value will be equal to the value returned by <code>seconds</code>;
   * that method should be used except in special cases of simulation, any-time
   * algorithms, self-aware robots, etc.
   *
   * @see Timebase#seconds
   */
  public double physicalSeconds(long timestamp) throws IllegalStateException {
    if (envTimebase == null) {
      if (start == REL_TIME) {
        throw new IllegalStateException("physical time not available");
      }
      else {
        return (double) (timestamp - start) * scale;
      }
    }
    else {
      throw new IllegalStateException("hierarchical time not implemented.");
    }

  }
  /**
   * Express the time stamp (that is, number of clock ticks) in seconds.
   */
  public double seconds(long timestamp) {
    return (double) timestamp * scale;
  }

  /**
   * Express the time stamp (number of ticks) in milli-seconds.
   */
  public long millis(long timestamp) {
    return timestamp * milliRate; // Note the computation order.
  }

  /**
   * Express the time stamp in nano-seconds. Under construction...
   */
  public long nanos(long timestamp) {
    return timestamp * nanoRate;
  }

  /**
   * The start of this time base, possibly relative to an enclosing time base,
   * and expressed in seconds from beginning 1970.
   */
  public long time0() {
    return start;
  }

  /**
   * Rate of clock ticks per second.
   */
  public long pace() {
    return rate;
  }

}
