package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;

public class Buffer extends RTEventBuffer {

  /**
   * Initializes the buffer with a size of <code>DEFAULT_SIZE</code>.
   */
  public Buffer() {
	  super();
  }

  /**
   * Initializes the queue (buffer) with a given maximal size. A size smaller
   * than one results in a buffer with size one.
   * 
   * @param maxSize maximum queue size
   */
  public Buffer(int maxSize) {
	  super(maxSize);
	  maxSizeR = maxSizeW = (maxSize > 0 ? maxSize : 1);
    
	  Condition.Factory factory = new Condition.Factory();
	  isFull = factory.newCondition();
	  notFull = factory.newCondition();
	  isEmpty = factory.newCondition();
	  notEmpty = factory.newCondition();
      lock = factory.obtainLock();
  }

  /**
   * Initializes the queue (buffer) with a given maximal size, and with a
   * lock object to be used for synchronization.
   *
   * @param maxSize maximum queue size
   * @param lock    external monitor/object to synchronized with
   */
  public Buffer(int maxSize, Synchronized.Lock lock) {
    
	  super(maxSize, lock);
	  maxSizeR = maxSizeW = (maxSize > 0 ? maxSize : 1);
	  this.lock = lock;
	  
	  Condition.Factory factory = new Condition.Factory(lock);
	  isFull = factory.newCondition();
	  notFull = factory.newCondition();
	  isEmpty = factory.newCondition();
	  notEmpty = factory.newCondition();
  }


  /**
   * Returns the next RTEvent in the queue, blocks if none available.
   * @return a non-null event.
   */
  public RTEvent doFetch() {
	  
// 	  synchronized (lock) {
		  
		  try {
			  while (currSize == 0) {
				  notEmpty.await();
			  }
		  }
		  catch (RTInterrupted e) {
			  error("Buffer.doFetch(): Interrupted");
			  return null;
		  }
		  
		  RTEvent e = buffer[fetchAt];
		  buffer[fetchAt] = null; // For garbage collector
		  fetchAt = (fetchAt + 1) % maxSizeR;
          currSize--;
		  
		  if (currSize < maxSizeW) {
			  notFull.signal();
		  }
		  if (currSize == 0) {
			  isEmpty.signal();
		  }
		  
		  if (e.owner != null)
			  e.setOwner(/**/RTThread.currentThread());
		  
		  return e;
// 	  }
  }

  /**
   * Returns the next available RTEvent in the queue, or null if the
   * queue is empty. In other words, this method always returns immediately,
   * even if the queue is empty (non-blocking).
   *
   * @return an RTEvent if any available, null otherwise
   */
  public RTEvent tryFetch() {
	  
	  synchronized (lock) {
		  
		  if (isEmpty()) {
			  return null;
		  }
		  else {
			  
			  RTEvent e = buffer[fetchAt];
			  buffer[fetchAt] = null; // For garbage collector
			  fetchAt = (fetchAt + 1) % maxSizeR;
			  currSize--;
			  
			  if (currSize < maxSizeW) {
				  notFull.signal();
			  }
			  if (currSize == 0) {
				  isEmpty.signal();
			  }
			
			  if (e.owner != null)
				  e.setOwner(/**/RTThread.currentThread());
			  
			  return e;
		  }
	  }
  }

  /**
   * Get the next RTEvent in the queue, or block (up to the specified
   * timeout time) if no event is available.
   *
   * @param timeout the maximum blocking time time in milli-seconds.
   *
   * @return an RTEvent if any available within the timeout time, null
   *         otherwise.
   */
  public RTEvent tryFetch(long timeout) {
	  
	  long now;
	  long arrival = now = System.currentTimeMillis();
	  
	  synchronized (lock) {
		  
		  RTEvent ans = null;
      
		  if (currSize == 0 && now < arrival + timeout) {
			  notEmpty.await(timeout);
			  now = System.currentTimeMillis();
		  }
		  
		  if (now < arrival + timeout) {
			  try {
				  ans = tryFetch();
			  } catch (RTInterrupted e) {
				  error("Buffer.tryFetch(timeout); Interrupted");
			  }
		  }
		  
		  return ans;
	  }
  }

  
  /**
   * Adds an RTEvent to the queue, blocks caller if the queue is full.
   * @param e the event to enqueue
   */
  public void doPost(RTEvent e) {
	  
	  if (e.owner != null && e.owner != /**/RTThread.currentThread()) {
		  throw new OwnerError(e.owner);
	  }
	  
	  synchronized (lock) {
		
		  try {
			  while (currSize >= maxSizeW) {
				  notFull.await();
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.doPost(): Interrupted");
			  return;
		  }
		  
		  buffer[postAt] = e;
		  if (e.owner != null)
			  e.setOwner(this);
		  postAt = (postAt + 1) % maxSizeR;
		  currSize++;
		  
		  notEmpty.signal();
		  if (currSize == maxSizeW) {
			  isFull.signal();
		  }
		  
		  return;
	  }
  }

  /**
   * Adds an EventObject to the queue, without blocking if the queue is full.
   * @param e the event to enqueue
   * @return null when event posted to non-full buffer, the event otherwise
   */
  public RTEvent tryPost(RTEvent e) {
	  
	  if ( e.owner != null && e.owner != /**/
		  RTThread.currentThread()) {
		  throw new OwnerError(e.owner);
	  }
    
	  synchronized (lock) {
    
		  if (currSize >= maxSizeW) {
			  return e;
		  }
		  else {
    		
			  buffer[postAt] = e;
			  if (e.owner != null)
				  e.setOwner(this);
			  postAt = (postAt + 1) % maxSizeR;
			  currSize++;
			  
			  notEmpty.signal();
			  if (currSize == maxSizeW) {
				  isFull.signal();
			  }
  		  	
			  return null;
		  }
	  }
  }

  /**
   * Adds an RTEvent to the queue, blocks caller if the queue is full up to
   * the specified timeout time.
   *
   * @param e       the event to enqueue
   * @param timeout the maximum blocking time time in milli-seconds.
   * @return null when event posted within the timeout time to non-full
   *         buffer, the event otherwise.
   */
  public RTEvent tryPost(RTEvent e, long timeout) {
	  long now;
	  long arrival = now = System.currentTimeMillis();
	  synchronized (lock) {
      
		  RTEvent ans = e;
      
		  try {
			  while (currSize >= maxSizeW && now < arrival + timeout) {
				  notFull.await(timeout);
				  now = System.currentTimeMillis();
			  }
    	
			  if (now < arrival + timeout) {
				  ans = tryPost(e);
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.tryPost(timeout): Interrupted");
		  }
      
		  return ans;
	  }
  }

  /**
   * Checks if buffer is empty.
   * @return true if the queue is empty, false if it is non-empty
   */
  public boolean isEmpty() {
    synchronized (lock) {
    	return currSize == 0;
    }
  }

  /**
   * Waits for buffer to become empty.
   */
  public void awaitEmpty() {
	  synchronized (lock) {
    	
		  try {
			  while (currSize > 0) {
				  isEmpty.await();
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.awaitEmpty(): Interrupted");
		  }
	  }
  }

  /**
   * Waits for buffer to become non-empty.
   */
  public void awaitNotEmpty() {
      synchronized (lock) {
   
		  try {
			  while (currSize == 0) {
				  notEmpty.await();
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.awaitNotEmpty(): Interrupted");
		  }
	  }
  }

  /**
   * Checks if buffer is full.
   * @return true if the queue is full, false if it is non-full
   */
  public boolean isFull() {
      synchronized (lock) {
		  return currSize >= maxSizeW;
	  }
  }

  /**
   * Waits for buffer to become full. If the buffer has been resized to a size
   * less than the current number of events, the caller is blocked until the
   * buffer is exactly full.
   */
  public void awaitFull() {
	  synchronized (lock) {
    	
		  try {
			  while (currSize != maxSizeW) {
				  isFull.await();
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.awaitFull(): Interrupted");
		  }
	  }
  }

  /**
   * Waits for buffer to become non-full. If the buffer has been resized to a
   * size less than the current number of events, the caller is blocked until
   * the current size of the buffer is smaller than the new maximum size.
   */
  public void awaitNotFull() {
	  synchronized (lock) {
		  
		  try {
			  while (currSize == maxSizeW) {
				  notFull.await();
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.awaitNotFull(): Interrupted");
		  }
	  }
  }

  /**
   * The current size of the buffer.
   * @see RTCondEventBuffer#maxSizeR
   *
   * @return the number of objects currently enqueued
   */
  public int currentSize() {
	  synchronized (lock) {
		  return currSize;
	  }
  }

  /**
   * Removes all buffered events.
   */
  public void flush() {
	  synchronized (lock) {
    	
		  for (int i = 0; i < currSize; i++) {
			  buffer[fetchAt] = null;
			  fetchAt = (fetchAt + 1) % maxSizeR;
		  }
		  currSize = 0;
      
		  notFull.signal();
		  isEmpty.signal();
	  }
  }

  /**
   * Get all buffered events leaving the buffer empty. The event that was the 
   * next to be fetched is returned as the first element (index 0), followed 
   * by the other events in FIFO order with the most recently posted event as
   * the last element.
   *
   * @return array containing buffered events. An empty array (not null) is
   *         returned if the buffer was empty.
   */
  public RTEvent[] fetchAll() {
	  synchronized (lock) {
    	
		  int len = currSize;
		  RTEvent[] ans = new RTEvent[len];
      
		  for (int i = 0; i < len; i++) {
			  ans[i] = buffer[fetchAt];
			  if (ans[i].owner != null)
				  ans[i].setOwner(/**/
						  RTThread.currentThread());
			  buffer[fetchAt] = null; // For garbage collector
			  fetchAt = (fetchAt + 1) % maxSizeR;
		  }
		  
		  notFull.signal();
		  isEmpty.signal();
      
		  currSize = 0;
		  return ans;
	  }
  }

  /**
   * Gets the maximum size of the buffer.
   * @return the maximum number of objects that can be enqueued
   */
  public int getMaxSize() {
	  synchronized (lock) {
		  return maxSizeW;
	  }
  }

  /**
   * Changes size of buffer.
   *
   * @param newSize the new maximum number of objects that can be enqueued
   */
  public void setMaxSize(int newSize) {
	  synchronized (lock) {
      
		  // Prepare and allocate.
    	  try {
			  while (resizing) { 
				  Condition.await(lock);
			  }
		  } catch (RTInterrupted ex) {
			  error("Buffer.setMaxSize(): Pending call interrupted");
			  return;
		  }
    	
		  if (newSize == maxSizeW)
			  return; // Size correct already.
      
		  resizing = true;
      
		  newSize = (newSize > 0 ? newSize : 1); // The size must be at least 1.
		  RTEvent[] newbuff = new RTEvent[newSize];
      
		  // Wait for array to be long enough.
		  if (currSize >= newSize) {
			  // New space not enough for all present events, prevent posting
			  maxSizeW = newSize; // No writing while overfilled.
			  
			  try {
				  while (currSize >= newSize) {
					  notFull.await();
				  }
			  } catch (RTInterrupted e) {
				  resizing = false;
				  maxSizeW = maxSizeR;
				  error("Buffer.setMaxSize(): Interrupt during buffer resizing");
				  return;
			  }
		  }
      
		  // OK, store the not-to-many events in new buffer.
		  for (int i = 0; i < currSize; i++) {
			  newbuff[i] = buffer[fetchAt];
			  fetchAt = (fetchAt + 1) % maxSizeR;
		  }
		  fetchAt = 0;
		  postAt = currSize;
		  maxSizeR = maxSizeW = newSize;
		  buffer = newbuff;

		  // Release pending calls to this method, and notify for other calls too.
		  resizing = false;
		  Condition.signal(lock);
      }
  }

  /**
   * Maximal queue size for Reading events. Temporarily, during
   * decrease of full buffer with <code>setMaxSize</code>, this
   * attribute can differ from the ordinary buffer size.
   * @see RTCondEventBuffer#maxSizeW
   */
  protected int maxSizeR;

  /**
   * Maximal queue size, which is also the size for Writing/posting.
   */
  protected int maxSizeW;

  /**
   * Marks that <code>setMaxSize</code> is being evaluated.
   * More than one call would screw-up that operation because it need to
   * be blocked until all messages fit into the new (possibly smaller) size.
   */
  protected boolean resizing;

  /**
   * Condition variables.
   */
  protected Condition isFull;
  protected Condition notFull;
  protected Condition isEmpty;
  protected Condition notEmpty;
}