/* $Id: ConfigureEvent.java,v 1.3 2005/08/15 13:24:44 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.Activity;
import se.lth.cs.realtime.RTThread;

/**
 * Basic support for in-process and buffered reconfiguration of threads, as defined 
 * by providing a subclass with suitable <code>configure</code> method(s). The usage 
 * scenario is explained by the following example: <br>
 * Assume a master thread M sends orders to a slave thread S, which normally runs 
 * with priority NORM_PRIORITY during the processing of it. After having processed 
 * a certain event e0, S should be instructed to change priority to MAX_PRIORITY. 
 * M cannot call S.setPriority since it may not be done before e0 is processed, and 
 * M must continue producing new orders. Hence, the set-priority order has to be 
 * buffered along with the application events, and carried out within the context of
 * S. In code this means for M:
 * <pre>
 *     S.putEvent(e0);
 *     S.putEvent(new ConfigureEvent(){
 *         public void configure(Thread self){
 *             self.setPriority(Thread.MAX_PRIORITY);
 *         }
 *         public void configure(RTThread self){
 *             self.setPriority(RTThread.MAX_PRIORITY);
 *         }
 *     )
 * </pre>
 * If we know what type of thread M is, one of these methods could (of course) be omitted.
 * Then in S, no matter if S is an RT thread or not, we trust M and call configure:
 * <pre>
 *     ev = mailbox.doFetch();
 *     while (ev instanceof ConfigureEvent) {
 *         ((ConfigureEvent)ev).configure(this);
 *         ev = mailbox.doFetch();
 *     }
 * </pre>
 * These lines could of course comprise a getEvent method, but to reduce code 
 * and complexity in the simple/normal case, threads accepting event-driven 
 * reconfigurations are to include this or similar code where appropriate.
 * <p>
 * Another examples could be change of period, and the configure method could send 
 * back an AckEvent when done. The ActivityEvent is designed to be independent of 
 * specific thread methods and application needs.
 * 
 * @see License
 */

public abstract class ConfigureEvent extends ActivityEvent {

  /**
   * Constructs a <code>ConfigureEvent</code> object with the specified
   * source thread and identifier.
   *
   * @param source the thread where the event originated.
   */
  public ConfigureEvent(Activity source) {
    super(source);
  }
  
  /**
   * Configuration or reconfiguration of Java/concurrent threads, as defined by 
   * subclassing in the event-posting thread.
   * 
   * @param self the thread object to be (re)configured, typically this in the 
   * event-fetching thread.
   */
  public void configure(Thread self){}
  
  /**
   * Configuration or reconfiguration of real-time threads.
   * 
   * @param self the real-time thread to be (re)configured, typically this in the 
   * event-fetching thread.
   */
  public void configure(RTThread self){}

}
