/* $Id: OwnerError.java,v 1.2 2005/08/22 08:18:22 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;

/**
 * Error to be thrown when a thread tries to post an event that is owned by 
 * another thread. This type of error represents an event access conflict, 
 * typically detected when the event object is sent via a buffer from one 
 * thread to another. With event types statically defined, and message-based 
 * inter-thread communication and access control built into the OS, such access
 * conflicts can be detected earlier (for each access). That is the case for the 
 * OSE RTOS (@link http://www.ose.com), from where the concept of event/message 
 * ownership was taken. As stated in the OSE documentation, <i>"This important 
 * feature eliminates the common error of memory access conflicts that can happen
 * so easily in designs based on conventional operating systems."</i> Here, for 
 * portability, flexibility, and efficiency, we by default only check ownership
 * during buffering. Since shared events normally are communicated this way, this 
 * should capture almost all such access faults. In addition, a null owner can be
 * used to permit shared (read-only or monitor protected) events.
 *
 * @see License
 * @see RTEvent
 * @see RTEventBuffer
 */

public final class OwnerError extends RTError {

  /** 
   * The actual owner of the object/event that caused this error to be thrown.
   * The claimed 
   */
  transient final Object owner;

  /**
   * The (RT)Thread trying to use/post an object/event belonging to someone else.
   */
  transient final Object claimer;
  
  public OwnerError(Object owner) {
  	super();
  	this.owner = owner;
  	claimer = RTThread.currentThread();
  }
  
  /**
   * Obtain the claiming (RT)Thread.
   * 
   * @return the activity that tried to use the event it did not own.
   */
  public Object getClaimer() {
    return claimer;
  }

  /**
   * Obtain the Thread or RTThread actually owning the event possessed.
   * 
   * @return the object that was the actual owner.
   */
  public Object getOwner() {
    return owner;
  }

  public String toString() {
  	return "OwnerError: "+claimer+" attempted to use an object owned by "+owner;
  }
}
