/* $Id: RTCondEventBuffer.java,v 1.1 2005/08/22 08:18:22 d02en Exp $
 *  
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */
package se.lth.cs.realtime.event;

import se.lth.cs.realtime.Condition;

/**
 * A thread safe basic queue/buffer for <code>RTEvent</code> objects.
 * Essentially a protected bounded buffer with both blocking and non-blocking
 * methods for posting and fetching messages in terms of references to objects.
 * <p>
 * For the functionality provided by this basic class, the buffer could have
 * been typed to contain <code>EventObject</code>s (the base class of RTEvent),
 * but to permit subclasses of <code>RTEventBuffer</code> to utilize the
 * properties of the RTEvents that class is used here. For instance, one might
 * need a buffer that keeps track of the total delay of data by referring to
 * the time-stamps of the newest and oldest RTEvent.
 * <p>
 * The methods checking or awaiting buffer full/empty are useful for operation
 * supervision, capturing events until a buffer is filled, mode changes, etc.
 * Do not misuse them for polled (concurrently bad) or busy-wait solutions.
 * <p>
 * To meet the efficiency demands of some applications, most methods
 * are final.  The class as such is, however, not final since
 * subclassing is expected for adding more features, hence the
 * protected visibility of fields. Additionally, the methods post and
 * fetch are virtual (i.e., not final). They by default call the
 * corresponding do-method and are typed accordingly, but they are
 * free for redefinition which can even result in synchronized being
 * omitted if mutual exclusion is ensured elsewise (in enclosing
 * monitor, etc.). See also the implemented interfaces, which are
 * facilitating access to only one end of the buffer.
 * <p>
 * While events are stored in the buffer, the buffer is set to be the owner
 * of the events, see the <code>RTEvent</code> class. Only events owned by
 * the current thread are accepted for posting to the buffer. When events
 * are fetched, the fetching thread is assigned to be the new owner. This
 * is to be ensured also by subclasses, thereby maintaining the detection
 * of event producers keeping references to objects already sent. In some
 * cases, however, it is desirable to resend (typically immutable) objects.
 * Therefore, events with the owner equal to <code>null</code> is always
 * accepted. The concept of owners is from the OSE operating system.
 * <p>
 * In practice and by default, all methods can be considered to be
 * synchronized, even if the methods as such are not declared to be
 * synchronized. Having the methods synchronized would agree with
 * common application programming practice, and together with the fact
 * that most methods are final, those synchronized methods would remain
 * synchronized in subclasses. However, synchronization is performed
 * on an internal lock object for the following reasons: <br>1) Other
 * object cannot interfere with the internal wait/notify by
 * synchronizing on an instance of this class (by using a synchronized
 * statement). <br>2) An external object (provided via the
 * constructor) can be used as the lock object, thereby permitting
 * synchronized use of several buffers/object.
 * <p>
 * If a blocked call is interrupted, an RTInterrupted (subclass of Error, not
 * an exception) is thrown, and that buffer object is most likely not useful
 * anymore.
 *
 * @see se.lth.cs.realtime.RTInterrupted
 *
 * @author Klas.Nilsson@cs.lth.se
 * @version 1.0, LastEditDate: 2002-12-11
 */

import se.lth.cs.realtime.RTInterrupted;
/*//*/
import se.lth.cs.realtime.RTThread;
import se.lth.cs.realtime.RTSystem;
import se.lth.cs.realtime.Synchronized;

public class RTCondEventBuffer implements RTEventBufferInput, RTEventBufferOutput, Synchronized {

  public final static int DEFAULT_SIZE = RTSystem.getDefaultEventBufferSize();

  // ------------------------------------------------------------ CONSTRUCTORS

  /**
   * Initalizes the buffer with a size of <code>DEFAULT_SIZE</code>.
   */
  public RTCondEventBuffer() {
    
	  this(DEFAULT_SIZE);
  }

  /**
   * Initalizes the queue (buffer) with a given maximal size. A size smaller
   * than one results in a buffer with size one.
   *
   * @param maxSize maximum queue size
   */
  public RTCondEventBuffer(int maxSize) {
    
	maxSizeR = maxSizeW = (maxSize > 0 ? maxSize : 1);
    buffer = new RTEvent[maxSizeR];
    
    Condition.Factory factory = new Condition.Factory();
    
    isFull = factory.newCondition();
    notFull = factory.newCondition();
    isEmpty = factory.newCondition();
    notEmpty = factory.newCondition();
    
    lock = factory.obtainLock();
  }

  /**
   * Initalizes the queue (buffer) with a given maximal size, and with a
   * lock object to be used for synchronization.
   *
   * @param maxSize maximum queue size
   * @param lock    external monitor/object to synchronized with
   */
  public RTCondEventBuffer(int maxSize, Synchronized.Lock lock) {
    
	maxSizeR = maxSizeW = (maxSize > 0 ? maxSize : 1);
    buffer = new RTEvent[maxSizeR];
    
    Condition.Factory factory = new Condition.Factory(lock);
    
    isFull = factory.newCondition();
    notFull = factory.newCondition();
    isEmpty = factory.newCondition();
    notEmpty = factory.newCondition();
    
    this.lock = lock;
  }

  // ----------------------------- METHODS IMPLEMENTING RTEventBufferOutput:

  /**
   * Returns the next event object in the queue, by default by calling
   * <code>doFetch</code> which is synchronized (opposed to this
   * method) via an internal lock.  Subclasses overriding this method
   * should be very careful to maintain mutual exclusion, normally by
   * simply declaring <code>fetch</code> as <code>synchronized</code>
   * or by synchronizing on the <code>lock</code> attribute.
   *
   * @return the first event in queue.
   */
  public RTEvent fetch() {
	  
	  return doFetch();
  }

  /**
   * Returns the next RTEvent in the queue, blocks if none available.
   *
   * @return a non-null event.
   */
  public final RTEvent doFetch() {
	  
	  synchronized (lock) {
		  
		  try {
			  while (currSize == 0) {
				  notEmpty.await();
			  }
		  }
		  catch (RTInterrupted e) {
			  error("doFetch interrupted");
			  return null;
		  }
		  
		  System.out.println("fetchAt=" + fetchAt + ", currSize=" + currSize + 
		  		  ", maxSizeW=" + maxSizeW + ", maxSizeR=" + maxSizeR);
    	
		  RTEvent e = buffer[fetchAt];
		  buffer[fetchAt] = null; // For garbage collector
		  fetchAt = (fetchAt + 1) % maxSizeR;
        
		  currSize--;
		  
		  if (currSize < maxSizeW) {
			  notFull.signal();
		  }
		  
		  if (currSize == 0) {
			  isEmpty.signal();
		  }
		  
		  if (e.owner != null)
			  e.setOwner(/**/
					  RTThread.currentThread());
		  
		  return e;
	  }
  }

  /**
   * Returns the next available RTEvent in the queue, or null if the
   * queue is empty. In other words, this method always returns immediately,
   * even if the queue is empty (non-blocking).
   *
   * @return an RTEvent if any available, null otherwise
   */
  public final RTEvent tryFetch() {
	  
	  synchronized (lock) {
		  
		  if (isEmpty()) {
			  return null;
		  }
		  else {
			  
			  RTEvent e = buffer[fetchAt];
			  buffer[fetchAt] = null; // For garbage collector
			  fetchAt = (fetchAt + 1) % maxSizeR;
        	
			  currSize--;
			  
			  if (currSize < maxSizeW) {
				  notFull.signal();
			  }
			  
			  if (currSize == 0) {
				  isEmpty.signal();
			  }
			
			  if (e.owner != null)
				  e.setOwner(/**/
						  RTThread.currentThread());
			  return e;
		  }
	  }
  }

  /**
   * Get the next RTEvent in the queue, or block (up to the specified
   * timeout time) if no event is available.
   *
   * @param timeout the maximum blocking time time in milli-seconds.
   *
   * @return an RTEvent if any available within the timeout time, null
   *         otherwise.
   */
  public final RTEvent tryFetch(long timeout) {
	  
	  long now;
	  long arrival = now = System.currentTimeMillis();
	  
	  synchronized (lock) {
		  
		  RTEvent ans = null;
      
		  if (currSize == 0 && now < arrival + timeout) {
			  notEmpty.await(timeout);
			  now = System.currentTimeMillis();
		  }
		  
		  if (now < arrival + timeout) {
			  try {
				  ans = tryFetch();
			  } 
			  catch (RTInterrupted e) {
				  error("tryFetch(timeout) interrupted.");
			  }
		  }
		  
		  return ans;
	  }
  }

  // ------------------------------ METHODS IMPLEMENTING RTEventBufferInput:

  /**
   * Put the supplied event object last in the queue, by default by calling
   * <code>doPost</code> which is synchronized (opposed to this method).
   * Subclasses overriding this method should be very careful to maintain
   * mutual exclusion, normally by simply declaring <code>post</code>
   * synchronized or by locking some enclosing monitor/resource.
   *
   * The return value is not useful for the base-class implementation since
   * there is nothing to return from calling <code>doPost</code>. Therefore,
   * <code>null</code> is always returned. The intention of the return
   * value/type is to facilitate the use of <code>tryPost</code> or other
   * (for instance GC-saving) subclass implementations.
   *
   * @return null (or possibly some event according to subclass, if any).
   */
  public RTEvent post(RTEvent e) {
    
	  doPost(e);
	  return null;
  }

  /**
   * Adds an RTEvent to the queue, blocks caller if the queue is full.
   *
   * @param e the event to enqueue
   */
  public final void doPost(RTEvent e) {
	  
	  if (e.owner != null && e.owner != /**/RTThread.currentThread()) {
		  throw new OwnerError(e.owner);
	  }
	  
	  synchronized (lock) {
		
		  try {
			  while (currSize >= maxSizeW) {
				  notFull.await();
			  }
		  }
		  catch (RTInterrupted ex) {
			  error("doPost() interrupted.");
			  return;
		  }
		  
		  System.out.println("postAt=" + postAt + ", currSize=" + currSize +
				  ", maxSizeW=" + maxSizeW + ", maxSizeR=" + maxSizeR);
	
		  buffer[postAt] = e;
		  if (e.owner != null)
			  e.setOwner(this);
		  postAt = (postAt + 1) % maxSizeR;
		  
		  /*
		   * if only the following is used: 
		   * if (currSize-- == 0) {
		   * 	notEmpty.signal();
		   * }
		   * the signal can be caugth by a thread
		   * waiting on awaitNotEmpty which means the
		   * doFetch call will be blocked forever.
		   * 
		   * Each post will result in a notEmpty signal.
		   */ 
		  currSize++;
		  notEmpty.signal();
		  
		  if (currSize == maxSizeW) {
			  isFull.signal();
		  }
		  
		  return;
	  }
  }

  /**
   * Adds an EventObject to the queue, without blocking if the queue is full.
   *
   * @param e the event to enqueue
   *
   * @return null when event posted to non-full buffer, the event otherwise
   */
  public RTEvent tryPost(RTEvent e) {
	  
	  if ( e.owner != null && e.owner != /**/
		  RTThread.currentThread()) {
		  throw new OwnerError(e.owner);
	  }
    
	  synchronized (lock) {
    
		  if (currSize >= maxSizeW) {
			  return e;
		  }
		  else {
    		
			  buffer[postAt] = e;
			  if (e.owner != null)
				  e.setOwner(this);
			  postAt = (postAt + 1) % maxSizeR;
        
			  currSize++;
			  notEmpty.signal();
			  
			  if (currSize == maxSizeW) {
				  isFull.signal();
			  }
  		  	
			  return null;
		  }
	  }
  }

  /**
   * Adds an RTEvent to the queue, blocks caller if the queue is full up to
   * the specified timeout time.
   *
   * @param e       the event to enqueue
   * @param timeout the maximum blocking time time in milli-seconds.
   *
   * @return null when event posted within the timeout time to non-full
   *         buffer, the event otherwise.
   */
  public RTEvent tryPost(RTEvent e, long timeout) {
	  
	  long now;
	  long arrival = now = System.currentTimeMillis();
    
	  synchronized (lock) {
      
		  RTEvent ans = e;
      
		  try {
			  while (currSize >= maxSizeW && now < arrival + timeout) {
				  notFull.await(timeout);
				  now = System.currentTimeMillis();
			  }
    	
			  if (now < arrival + timeout) {
				  ans = tryPost(e);
			  }
		  }
		  catch (RTInterrupted ex) {
			  error("tryPost(timeout) interrupted.");
		  }
      
		  return ans;
	  }
  }

  // ------------------------------------------------ OTHER PUBLIC METHODS:

  /**
   * Checks if buffer is empty.
   *
   * @return true if the queue is empty, false if it is non-empty
   */
  public boolean isEmpty() {
	  
    synchronized (lock) {
      
    	return currSize == 0;
    }
  }

  /**
   * Waits for buffer to become empty.
   */
  public void awaitEmpty() {
	  
	  synchronized (lock) {
    	
		  try {
			  while (currSize > 0) {
				  isEmpty.await();
			  }
		  } 
		  catch (RTInterrupted ex) {
			  error("awaitEmpty() interrupted.");
		  }
	  }
  }

  /**
   * Waits for buffer to become non-empty.
   */
  public void awaitNotEmpty() {
    
	  synchronized (lock) {
   
		  try {
			  while (currSize == 0) {
				  notEmpty.await();
			  }
		  }
		  catch (RTInterrupted ex) {
			  error("awaitNotEmpty() interrupted.");
		  }
	  }
  }

  /**
   * Checks if buffer is full.
   *
   * @return true if the queue is full, false if it is non-full
   */
  public boolean isFull() {
    
	  synchronized (lock) {
		  
		  return currSize >= maxSizeW;
	  }
  }

  /**
   * Waits for buffer to become full. If the buffer has been resized to a size
   * less than the current number of events, the caller is blocked until the
   * buffer is exactly full.
   */
  public void awaitFull() {
	  
	  synchronized (lock) {
    	
		  try {
			  while (currSize != maxSizeW) {
				  isFull.await();
			  }
		  } 
		  catch (RTInterrupted ex) {
			  error("awaitFull() interrupted.");
		  }
	  }
  }

  /**
   * Waits for buffer to become non-full. If the buffer has been resized to a
   * size less than the current number of events, the caller is blocked until
   * the current size of the buffer is smaller than the new maximum size.
   */
  public void awaitNotFull() {
    
	  synchronized (lock) {
		  
		  try {
			  while (currSize == maxSizeW) {
				  notFull.await();
			  }
		  }
		  catch (RTInterrupted ex) {
			  error("awaitNotFull() interrupted.");
		  }
	  }
  }

  /**
   * The current size of the buffer.
   * @see RTCondEventBuffer#maxSizeR
   *
   * @return the number of objects currently enqueued
   */
  public int currentSize() {
	  
	  synchronized (lock) {
		
		  return currSize;
	  }
  }

  /**
   * Removes all buffered events.
   */
  public void flush() {
	  
	  synchronized (lock) {
    	
		  for (int i = 0; i < currSize; i++) {
			  buffer[fetchAt] = null;
			  fetchAt = (fetchAt + 1) % maxSizeR;
		  }
		  currSize = 0;
      
		  notFull.signal();
		  isEmpty.signal();
	  }
  }

  /**
   * Get all buffered events leaving the buffer empty. The event that was the 
   * next to be fetched is returned as the first element (index 0), followed 
   * by the other events in FIFO order with the most recently posted event as
   * the last element.
   *
   * @return array containing buffered events. An empty array (not null) is
   *         returned if the buffer was empty.
   */
  public RTEvent[] fetchAll() {
	  
	  synchronized (lock) {
    	
		  int len = currSize;
		  RTEvent[] ans = new RTEvent[len];
      
		  for (int i = 0; i < len; i++) {
			  ans[i] = buffer[fetchAt];
			  if (ans[i].owner != null)
				  ans[i].setOwner(/**/
						  RTThread.currentThread());
			  buffer[fetchAt] = null; // For garbage collector
			  fetchAt = (fetchAt + 1) % maxSizeR;
		  }
		  
		  notFull.signal();
		  isEmpty.signal();
      
		  currSize = 0;
		  return ans;
	  }
  }

  /**
   * Gets the maximum size of the buffer.
   *
   * @return the maximum number of objects that can be enqueued
   */
  public int getMaxSize() {
    
	  synchronized (lock) {
		  
		  return maxSizeW;
	  }
  }

  /**
   * Changes size of buffer.
   *
   * @param newSize the new maximum number of objects that can be enqueued
   */
  public void setMaxSize(int newSize) {
    
	  synchronized (lock) {
      // Prepare and allocate.
    
		  try {
			  while (resizing) { 
				  Condition.await(lock);
			  }
		  }
		  catch (RTInterrupted ex) {
			  error("setMaxSize(): Pending call interrupted.");
			  return;
		  }
    	
		  if (newSize == maxSizeW)
			  return; // Size correct already.
      
		  resizing = true;
      
		  newSize = (newSize > 0 ? newSize : 1); // The size must be at least 1.
		  RTEvent[] newbuff = new RTEvent[newSize];
      
		  // Wait for array to be long enought.
		  if (currSize >= newSize) {
			  // New space not enought for all present events, prevent posting
			  maxSizeW = newSize; // No writing while overfilled.
			  
			  try {
				  while (currSize >= newSize) {
					  notFull.await();
				  }
			  }
			  catch (RTInterrupted e) {
				  resizing = false;
				  maxSizeW = maxSizeR;
				  error("setMaxSize(): Interrupt during buffer resizing.");
				  return;
			  }
		  }
      
		  System.out.println("Resizing - currSize=" + currSize + ", maxSizeR=" +
				  maxSizeR + ", maxSizeW=" + maxSizeW);
		   
		  // OK, store the not-to-many events in new buffer.
		  for (int i = 0; i < currSize; i++) {
			  newbuff[i] = buffer[fetchAt];
			  fetchAt = (fetchAt + 1) % maxSizeR;
		  }
		  fetchAt = 0;
		  postAt = currSize;
		  maxSizeR = maxSizeW = newSize;
		  buffer = newbuff;

		  // Release pending calls to this method, and notify for other calls too.
		  resizing = false;
		  Condition.signal(lock);
      }
  }

  // ---------------------------------------------------- PROTECTED METHODS:

  /**
   * Takes care of <code>InterruptedException</code>.
   */
  protected void error(String message) {
    // Use exception handling object here, which in turn should affect
    // the calling thread.

	  throw new RTInterrupted(message);
  }

  // ------------------------------------------------- PROTECTED ATTRIBUTES:

  /**
   * The actual buffer.
   */
  protected RTEvent[] buffer;

  /**
   * Maximal queue size for Reading events. Temporarily, during
   * decrease of full buffer with <code>setMaxSize</code>, this
   * attribute can differ from the ordinary buffer size.
   * @see RTCondEventBuffer#maxSizeW
   */
  protected int maxSizeR;

  /**
   * Maximal queue size, which is also the size for Writing/posting.
   */
  protected int maxSizeW;

  /**
   * Marks that <code>setMaxSize</code> is being evaluated.
   * More than one call whould screw-up that operation because it need to
   * be blocked until all messages fit into the new (possibly smaller) size.
   */
  protected boolean resizing;

  /**
   * Current queue size. That is, the number of events currently buffered.
   */
  protected int currSize;

  /*
   * Free position to insert events at (the successor of last element)
   */
  protected int postAt;

  /**
   * Position of first element to fetch, if any.
   */
  protected int fetchAt;

  /**
   * The synchronization lock for this object.
   */
  protected Synchronized.Lock lock;
  
  /**
   * Condition variables.
   */
  protected Condition isFull;
  protected Condition notFull;
  protected Condition isEmpty;
  protected Condition notEmpty;
}
