/* $Id: RTEventDispatcher.java,v 1.9 2005/09/30 08:15:45 klas Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

/**
 * A collection of <code>RTEventListener</code>s, organized in three sets and
 * ordered according to priority within each set. Event dispatching (by calling
 * the listener's <code>putEvent</code>) starts with the high priority
 * listers. The priority is specified when adding a listener.
 * <p>
 * Having three sets (internal lists) facilitated the separation between
 * <ul>
 * <li> ordinary application level events,
 * <li> events that have to do with concurrency and timing control,
 * <li> the events that have to do with exceptions and errors.
 * </ul>
 * By default, using the <code>addListener</code> and
 * <code>removeListener</code>, the listener only subscribes on
 * <code>RTEvent</code>s that are not due to errors or thread control.
 * <p>
 * For the ordinary events, there is the possibility to specify a subclass of
 * RTEvent as a pattern for the subscription, resulting in that only such events
 * (and subclasses thereof) result in notification of the listener. This is
 * similar to the <code>javax.swing.event.EventListenerList</code>.
 * <p>
 * Within each set of subscriptions, given as an argument when listeners are added 
 * or removed, it can be requested to only listen for events of a certain type. <br>
 * NOTE: Currently this feature is disabled (to comply with a limitation of our 
 * Java2C translated), so each listener in it's <code>putEvent</code> method needs to
 * check the type with <code>instanceof</code> (which can be translated to C).
 * 
 * @see se.lth.cs.realtime.License
 */

public class RTEventDispatcher
{

  private Object owner;

  protected static class Subscription
  {
    RTEventListener listener;

    int priority;

    Object type;

    Subscription(RTEventListener address, int priority, Object type)
    {
      this.listener = address;
      this.priority = priority;
      this.type = type;
    }
  }

  protected final static Subscription[] NONE = new Subscription[0];

  protected Subscription[] appListener = NONE; // Application

  protected Subscription[] actListener = NONE; // Activity

  protected Subscription[] excListener = NONE; // Exception&Error

  /**
   * Create a dispatcher with its creator (the event source) registered as the
   * owner. Only the owner is permitted to dispatch events, which is checked for
   * at run time.
   */
  public RTEventDispatcher(Object source)
  {
    owner = source;
  }

  /**
   * Return the number of application listeners
   */
  public int getAppListenerCount()
  {
    return appListener.length;
  }

  /**
   * Return the number of activity listeners
   */
  public int getActListenerCount()
  {
    return actListener.length;
  }

  /**
   * Return the number of exception listeners
   */
  public int getExcListenerCount()
  {
    return excListener.length;
  }

  /**
   * Synchronized part of <code>addListener</code>.
   */
  protected synchronized final void addApp(RTEventListener dest, int priority,
      Object mask)
  {
    Subscription record = new Subscription(dest, priority, mask);

    int i;
    for (i = 0; i < appListener.length; i++) {
      if (priority >= appListener[i].priority) {
        break;
      }
    }
    Subscription[] tmp = new Subscription[appListener.length + 1];
    System.arraycopy(appListener, 0, tmp, 0, i);
    tmp[i] = record;
    System.arraycopy(appListener, i, tmp, i + 1, appListener.length - i);
    appListener = tmp;
  }

  /**
   * Synchronized part of <code>addActListener</code>.
   */
  protected synchronized final void addAct(RTEventListener dest, int priority,
      Object mask)
  {

    Subscription record = new Subscription(dest, priority, mask);

    int i;
    for (i = 0; i < actListener.length; i++) {
      if (priority >= actListener[i].priority) {
        break;
      }
    }

    Subscription[] tmp = new Subscription[actListener.length + 1];
    if (i > 0) System.arraycopy(actListener, 0, tmp, 0, i);
    tmp[i] = record;
    if (i < actListener.length)
      System.arraycopy(actListener, i, tmp, i + 1, actListener.length - i);
    actListener = tmp;
  }

  /**
   * Synchronized part of <code>addExcListener</code>.
   */
  protected synchronized final void addExc(RTEventListener dest, int priority,
      Object mask)
  {

    Subscription record = new Subscription(dest, priority, mask);

    int i;
    for (i = 0; i < excListener.length; i++) {
      if (priority >= excListener[i].priority) {
        break;
      }
    }

    Subscription[] tmp = new Subscription[excListener.length + 1];
    if (i > 0) System.arraycopy(excListener, 0, tmp, 0, i);
    tmp[i] = record;
    if (i < excListener.length)
      System.arraycopy(excListener, i, tmp, i + 1, excListener.length - i);
    excListener = tmp;
  }

  /**
   * Add the listener as a listener of application events, with the specified
   * priority, and possibly with an event mask to limit the amount of event
   * notifications. Typically, this method is called directly from a thread
   * object, that is, <code>dest</code> typically is a thread object. Then the
   * supplied priority may not exceed the maximum priority of that thread. If
   * the destination is not a thread object, such as a global mailbox, the
   * registered priority may not exceed the current priority of the calling
   * thread. These restrictions are to prevent low-priority activities causing
   * delays for high priority ones. If the dispatched call of
   * <code>putEvent</code> attempts to add or remove a listener, an exception
   * is thrown since such a self-referencing call is not supported (for
   * efficiency).
   * 
   * @param dest the type of the listener to be added
   * @param priority the priority of the listener.
   * @param type the subtype of desired events, null means all RTEvents.
   * @exception IllegalArgumentException if the arguments are null or
   *              incompatible.
   * @exception UnsupportedOperationException if called from the dispatching
   *              thread.
   */
  public final void addListener(RTEventListener dest, int priority, Object type)
      throws UnsupportedOperationException
  {
    // if (Thread.holdsLock(this)) {
    // throw new UnsupportedOperationException("Self-ref dispatching!");
    // }
    if (dest == null) {
      throw new IllegalArgumentException("Listener is null!");
    }
    /**
     * <!--
     * 
     * <pre>
     * Class klass = type;
     * try {
     *   if (type == null) klass = Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;);
     *   else if (!(type instanceof Class)) klass = type.getClass();
     *   else if (!Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;).isAssignableFrom(
     *       klass)) {
     *     throw new IllegalArgumentException(&quot;Listener mask to a subclass of RTEvent&quot;);
     *   }
     * }
     * catch (ClassNotFoundException exc) {
     *   throw new UnsupportedOperationException(&quot;Class RTEvent not found&quot;);
     * }
     * </pre>
     * 
     * -->
     */

    addApp(dest, priority, type);
  }

  /**
   * Default method to add activity listeners.
   */
  public final void addActListener(RTEventListener dest, int priority,
      Object type) throws UnsupportedOperationException
  {

    // if (Thread.holdsLock(this)) {
    // throw new UnsupportedOperationException("Self-ref dispatching!");
    // }
    if (dest == null) {
      throw new IllegalArgumentException("Listener is null!");
    }

    /**
     * <!--
     * 
     * <pre>
     * Class klass = type;
     * try {
     *   if (type == null) klass = Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;);
     *   else if (!(type instanceof Class)) klass = type.getClass();
     *   if (!Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;)
     *       .isAssignableFrom(klass)) {
     *     throw new IllegalArgumentException(&quot;Listener mask to a subclass of RTEvent&quot;);
     *   }
     * }
     * catch (ClassNotFoundException exc) {
     *   throw new UnsupportedOperationException(&quot;Class RTEvent not found&quot;);
     * }
     * </pre>
     * 
     * -->
     */

    addAct(dest, priority, type);
  }

  /**
   * Method to add exception listeners.
   */
  public final void addExcListener(RTEventListener dest, int priority,
      Object type) throws UnsupportedOperationException
  {

    // if (Thread.holdsLock(this)) {
    // throw new UnsupportedOperationException("Self-ref dispatching!");
    // }
    if (dest == null) {
      throw new IllegalArgumentException("Listener is null!");
    }

    /**
     * <!--
     * 
     * <pre>
     * Class klass = type;
     * try {
     *   if (type == null) klass = Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;);
     *   else if (!(type instanceof Class)) klass = type.getClass();
     *   if (!Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;)
     *       .isAssignableFrom(klass)) {
     *     throw new IllegalArgumentException(&quot;Listener mask to a subclass of RTEvent&quot;);
     *   }
     * }
     * catch (ClassNotFoundException exc) {
     *   throw new UnsupportedOperationException(&quot;Class RTEvent not found&quot;);
     * }
     * </pre>
     * 
     * -->
     */

    addExc(dest, priority, type);
  }

  /**
   * Remove the listener as a listener of the specified type. If multiple copies
   * of the same listener are present, the one with the highest priority is
   * removed.
   * 
   * @param dest the listener to be removed
   * @param type the type of the listener to be removed, null for any type.
   * @exception IllegalArgumentException if the arguments are null or
   *              incompatible.
   */
  public synchronized void removeListener(RTEventListener dest, Object type)
  {
    /**
     * <!--
     * 
     * <pre>
     *      
     *      Class klass = type;(type instanceof Class ? (Class) type : type.getClass());
     *      try {
     *        if (klass == null) klass = Class.forName(&quot;RTEvent&quot;);
     *      }
     *      catch (ClassNotFoundException exc) {
     *        throw new UnsupportedOperationException(&quot;Class RTEvent not found&quot;);
     *      }
     *      if (!(type instanceof Class)) klass = type.getClass();
     *      if (dest == null) {
     *        throw new IllegalArgumentException(&quot;Listener &quot; + dest + &quot; is null&quot;);
     *      }
     *      
     *      // Search...
     *      int i = 0;
     *      for (; i &lt; appListener.length; i++) {
     *        if ((appListener[i].type == klass) &amp;&amp; (appListener[i].listener == dest)) {
     *          break;
     *        }
     *      }
     *      if (i &lt; appListener.length) { // found.
     *        Subscription[] tmp;
     *        if (appListener.length == 1) {
     *          tmp = NONE;
     *        } else {
     *          tmp = new Subscription[appListener.length - 1];
     *          System.arraycopy(appListener, 0, tmp, 0, i);
     *          System.arraycopy(appListener, i + 1, tmp, i, appListener.length - (i + 1));
     *        }
     *        appListener = tmp;
     *      }
     * </pre>
     * 
     * -->
     */

    if (dest == null) {
      throw new IllegalArgumentException("Listener " + dest + " is null");
    }
    // Search...
    int i = 0;
    for (; i < appListener.length; i++) {
      if (appListener[i].listener == dest) {
        break;
      }
    }
    if (i < appListener.length) { // found.
      Subscription[] tmp;
      if (appListener.length == 1) {
        tmp = NONE;
      } else {
        tmp = new Subscription[appListener.length - 1];
        System.arraycopy(appListener, 0, tmp, 0, i);
        System.arraycopy(appListener, i + 1, tmp, i, appListener.length
            - (i + 1));
      }
      appListener = tmp;
    }
  }

  /**
   * Method to remove an activity listener.
   */
  public synchronized void removeActListener(RTEventListener dest, Object type)
  {
    /**
     * <!--
     * 
     * <pre>
     * Class klass = type;
     * try {
     *   if (klass == null) klass = Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;);
     * }
     * catch (ClassNotFoundException exc) {
     *   throw new UnsupportedOperationException(&quot;Class RTEvent not found&quot;);
     * }
     * if (!(type instanceof Class)) klass = type.getClass();
     * if (dest == null) {
     *   throw new IllegalArgumentException(&quot;Listener &quot; + dest + &quot; is null&quot;);
     * }
     * 
     * // Search...
     * int i = 0;
     * for (; i &lt; actListener.length; i++) {
     *   if ((actListener[i].type == klass) &amp;&amp; (actListener[i].listener == dest)) {
     *     break;
     *   }
     * }
     * if (i &lt; actListener.length) { // found.
     *   Subscription[] tmp;
     *   if (actListener.length == 1) {
     *     tmp = NONE;
     *   } else {
     *     tmp = new Subscription[actListener.length - 1];
     *     System.arraycopy(actListener, 0, tmp, 0, i);
     *     System.arraycopy(actListener, i + 1, tmp, i, actListener.length - (i + 1));
     *   }
     *   actListener = tmp;
     * }
     * </pre>
     * 
     * -->
     */

    if (dest == null) {
      throw new IllegalArgumentException("Listener " + dest + " is null");
    }
    // Search...
    int i = 0;
    for (; i < actListener.length; i++) {
      if (actListener[i].listener == dest) {
        break;
      }
    }
    if (i < actListener.length) { // found.
      Subscription[] tmp;
      if (actListener.length == 1) {
        tmp = NONE;
      } else {
        tmp = new Subscription[actListener.length - 1];
        System.arraycopy(actListener, 0, tmp, 0, i);
        System.arraycopy(actListener, i + 1, tmp, i, actListener.length
            - (i + 1));
      }
      actListener = tmp;
    }

  }

  /**
   * Method to remove an exception listener.
   */
  public synchronized void removeExcListener(RTEventListener dest, Object type)
  {
    /**
     * <!--
     * 
     * <pre>
     * Class klass = type;
     * try {
     *   if (klass == null) klass = Class.forName(&quot;se.lth.cs.realtime.event.RTEvent&quot;);
     * }
     * catch (ClassNotFoundException exc) {
     *   throw new UnsupportedOperationException(&quot;Class RTEvent not found&quot;);
     * }
     * if (!(type instanceof Class)) klass = type.getClass();
     * if (dest == null) {
     *   throw new IllegalArgumentException(&quot;Listener &quot; + dest + &quot; is null&quot;);
     * }
     * 
     * // Search...
     * int i = 0;
     * for (; i &lt; excListener.length; i++) {
     *   if ((excListener[i].type == klass) &amp;&amp; (excListener[i].listener == dest)) {
     *     break;
     *   }
     * }
     * if (i &lt; excListener.length) { // found.
     *   Subscription[] tmp;
     *   if (excListener.length == 1) {
     *     tmp = NONE;
     *   } else {
     *     tmp = new Subscription[excListener.length - 1];
     *     System.arraycopy(excListener, 0, tmp, 0, i);
     *     System.arraycopy(excListener, i + 1, tmp, i, excListener.length - (i + 1));
     *   }
     *   excListener = tmp;
     * }
     * </pre>
     * 
     * -->
     */

    if (dest == null) {
      throw new IllegalArgumentException("Listener " + dest + " is null");
    }
    // Search...
    int i = 0;
    for (; i < excListener.length; i++) {
      if (excListener[i].listener == dest) {
        break;
      }
    }
    if (i < excListener.length) { // found.
      Subscription[] tmp;
      if (excListener.length == 1) {
        tmp = NONE;
      } else {
        tmp = new Subscription[excListener.length - 1];
        System.arraycopy(excListener, 0, tmp, 0, i);
        System.arraycopy(excListener, i + 1, tmp, i, excListener.length
            - (i + 1));
      }
      excListener = tmp;
    }
  }

  /**
   * Dispatches an application message.
   */
  public synchronized void dispatch(RTEvent evt)
  {
    /**
     * <!--
     * 
     * <pre>
     * 
     * for (int i = 0; i &lt; appListener.length; i++) {
     *   if (appListener[i].type.isInstance(evt)) {
     *     appListener[i].listener.putEvent(evt);
     *   }
     * }
     * </pre>
     * 
     * -->
     */
    for (int i = 0; i < appListener.length; i++) {
      appListener[i].listener.putEvent(evt);
    }
  }

  /**
   * Dispatches an activity message.
   */
  public synchronized void dispatchActivity(ActivityEvent evt)
  {
    /**
     * <!--
     * 
     * <pre>
     * 
     * for (int i = 0; i &lt; actListener.length; i++) {
     *   if (actListener[i].type.isInstance(evt))
     *     actListener[i].listener.putEvent(evt);
     * }
     * </pre>
     * 
     * -->
     */
    for (int i = 0; i < actListener.length; i++) {
      actListener[i].listener.putEvent(evt);
    }
  }

  /**
   * Dispatches an exception message.
   */
  public synchronized void dispatchException(ExceptionEvent evt)
  {
    /**
     * <!--
     * 
     * <pre>
     * 
     * for (int i = 0; i &lt; excListener.length; i++) {
     *   if (excListener[i].type.isInstance(evt))
     *     excListener[i].listener.putEvent(evt);
     * }
     * </pre>
     * 
     * -->
     */
    for (int i = 0; i < excListener.length; i++) {
      excListener[i].listener.putEvent(evt);
    }
  }

  /**
   * Return a string representation of the RTEventListenerList.
   */
  public String toString()
  {
    String s = "RTEventDispatcher\n\n";
    s += "Application Listeners [" + appListener.length + "]: ";
    for (int i = 0; i < appListener.length; i++) {
      s += "\n\tlistener " + appListener[i].listener.toString();
      s += " (priority " + appListener[i].priority + ")";
      // s += ", type " + appListener[i].type.getName() + ")";
    }
    s += "\nActivity Listeners [" + actListener.length + "]: ";
    for (int i = 0; i < actListener.length; i++) {
      s += "\n\tlistener " + actListener[i].listener.toString();
      s += " (priority " + actListener[i].priority + ")";
      // s += ", type " + actListener[i].type.getName() + ")";
    }
    s += "\nException Listeners [" + excListener.length + "]: ";
    for (int i = 0; i < excListener.length; i++) {
      s += "\n\tlistener " + excListener[i].listener.toString();
      s += " (priority " + excListener[i].priority + ")";
      // s += ", type " + excListener[i].type.getName() + ")";
    }
    s += "\n";
    return s;
  }
}
