/* $Id: RTEventSource.java,v 1.5 2005/08/19 09:37:16 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

/**
 * Source of real-time events that are to be passed to
 * <code>RTEventListener</code> objects that have subscribed on such
 * events. In the simplest case, the listener register itself to
 * receive all ordinary (not <code>ExceptionEvent</code>s or
 * <code>ActivityEvent</code>s), by calling <code>addListener</code>
 * with the lister itself as a the subscriber argument. With a
 * <code>RTEventSource</code> called producer the call will be <pre>
 *
 *     producer.addListener(this);
 *
 * </pre> This should be equivalent to  <pre>
 *
 *     producer.getDispatcher().addListener(this);
 *
 * </pre>if the other method defined by this interface,
 * <code>getDispatcher</code>, returns a non-null object.
 * <p>
 * The dispatcher object provides additionally functionality that
 * could have been part of this interface.  The reason for this simple
 * two-method interface, instead of declaring all event subscription
 * methods, is to simplify the implementation of event sources. That
 * is, an event source can simply reuse an available dispatcher and
 * add a two-line getter method returning it. 
 * <p>
 * For simple statically defined one-to-one communication (one
 * producer to one consumer), the listener (consumer) is typically
 * created first (avoiding timing dependent problems during startup)
 * and then given to the constructor of the event source
 * (producer). Using an event dispatcher is good for more flexible
 * situations with run-time changes or one-to-many communication.
 * 
 * @see License
 * @see RTEventListener
 * @see RTEventDispatcher 
 */

public interface RTEventSource {

  /**
   * Subscribe on RTEvents from this source.
   *
   * @param subscriber the object interested in getting the events
   * via its <code>putEvent</code> method.
   */
  void addListener(RTEventListener subscriber);

  /**
   * Obtain object used for event subscription. The same object, using
   * other methods, is used by the issuing thread for dispatching the
   * events by calling <code>putEvent</code> of the subscribing object.
   *
   * @return the object providing registration of <code>RTEvent</code>
   * listeners. The returned dispatcher corresponds to the
   * <code>javax.swing.event.EventListenerList</code>, but enhanced
   * for real-time usage.
   */
  RTEventDispatcher getDispatcher();
}
