/* $Id: TriggerEvent.java,v 1.6 2005/08/19 09:37:16 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.Timebase;

/**
 * Generic event used to trigger one-shot, cyclic, or sporadic activities in an
 * event-driven fashion. Some timer thread is useful to generate a single
 * triggering event or periodic events, but periodic events sent to a
 * <code>PeriodicThread</code> should instead be of type
 * <code>PeriodicEvent</code> to enable dedicated support for time-based actions.
 * <p>
 * A <code>TriggerEvent</code> has no owner. That is, <code>owner</code> is
 * set to <code>null</code>, enabling the event to be sent over and over
 * again from the same source. Subclasses are, however, free to assign ownership
 * simply by assigning the (protected) <code>owner</code> field.
 * 
 * @see License
 * @see PeriodicEvent
 * @see java.util.Timer
 * @see RTEvent
 */

public class TriggerEvent extends RTEvent {

  /**
   * Constructs an <code>TriggerEvent</code> object with the current
   * thread as source and a time stamp from the current system time.
   */
  public TriggerEvent() {
    super();
    owner = null;
  }

  /**
   * Constructs an <code>TriggerEvent</code> object with the specified
   * source object and a time stamp from the current system time. 
   *
   * @param source The object where the event originated.
   */
  public TriggerEvent(Object source) {
    super(source);
  }

  /**
   * Constructs an <code>TriggerEvent</code> object with the current
   * thread as source and with the specified time stamp. 
   *
   * @param ts The time from which the event originates.
   */
  public TriggerEvent(long ts) {
    super(Thread.currentThread(), ts);
  }

  /**
   * Constructs an <code>TriggerEvent</code> object with the specified
   * source object and time stamp. 
   *
   * @param source The object where the event originated.
   * @param ts The time from which the event originates.
   */
  public TriggerEvent(Object source, long ts) {
    super(source, ts);
  }

  /**
  * Constructs an <code>TriggerEvent</code> object with the specified
  * source object and time-stamp with respect to the specified time-base. 
  *
  * @param source The object where the event originated.
  * @param ts The time from which the event originates.
  */
  public TriggerEvent(Object source, long ts, Timebase timebase) {
    super(source, ts, timebase);
  }

}
