/* $Id: MutexSem.java,v 1.7 2005/08/19 09:37:16 d02en Exp $          
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.semaphore;

import se.lth.cs.realtime.RTSystem;

/**
 * A Semaphore with special support for MUTual EXclusion. Basically,
 * this is equivalent with other semaphores, such as
 * <code>CountingSem</code> or <code>BinarySem</code>, except for:<ol>
 *
 * <li>The implementation is required to provide priority inheritance
 * in order to avoid priority inversion. Other semaphores are all
 * allowed to provide priority inheritance but they are not required
 * to do it.
 *
 * <li>It is required that a thread calling give already owns the
 * semaphore. That is, the same thread has to call take (or tryTake
 * returning true) first and then give. Violation of this rule results
 * in a <code>SemViolation</code> error being thrown.

 * </ol>This class should map well onto the Mutex primitive In the
 * POSIX standard.
 * 
 * <p>The order of the awakenings of the threads which called
 * <code>take</code> depends on the underlying OS and/or VM, but for
 * standard implementation in pure Java it will be the same as for the
 * wait/notify mechanism, which should provide priority inheritance
 * (even if priorities are not strictly obeyed).
 *
 * @see License
 * @see se.lth.cs.realtime.semaphore.Semaphore
 * @see se.lth.cs.realtime.semaphore.MutexSem
 * @see se.lth.cs.realtime.semaphore.CountingSem
 * @see se.lth.cs.realtime.semaphore.MultistepSem 
 * @see se.lth.cs.realtime.semaphore.SemInterrupted 
 * @see se.lth.cs.realtime.semaphore.SemViolation 
 */

public final class MutexSem implements Semaphore {

	/**
     * Object used by the pure Java implementation for blocking and to make methods atomic. 
     * Used instead of <code>this</code> object to prevent misuse of notify.
     */
//     private Object lock;

    /**
     * Integer keeping track of the number of recursive take() calls.
     */
//     private int takeCount;
    
    /**
     * The currently registered owner of the locked resource. Used by the pure Java implementation. 
     * The ownership is possibly not checked in the case of a native mutex, depending on the RTOS.
     */
//     private Thread owner;
    
    /**
     * Reference to the semaphore of the current RTOS, if semaphores are provided natively.
     * Implemented by a built-in type to hide it from the GC.
     */
    private long target;

    /**
     * Creates a new <code>MutexSem</code> with the internal (boolean)
     * access flag set to true. To enforce the rule of calling take
     * before give, this is the only way a <code>MutexSem</code> can be
     * created.
     */
    public MutexSem() {
//     	if (RTSystem.NATIVE_SEM) {
    		target = createMutexSem(1);
//     	} else {
//     		lock = new Object();	
//     		takeCount = 0;
//     	}
    }   
    
    /**
     * Sets the mutex to its unoccupied state and releases the
     * first-in-queue blocked thread, if any. In some systems, this
     * operation is referred to as <em>unlock</em>.
     *
     * @throws SemViolation if the calling thread tries
     * to give away a lock that it does not own, that is, if the
     * semaphore has not been taken first.
     */
    public void give() {
//     	if (RTSystem.NATIVE_SEM) {
    		giveMutexSem(target);
//     	} else {
//     		synchronized (lock) {
//     		    if (Thread.currentThread() != owner) {
//     		    	throw new SemViolation("MutexSem.give: Not owner!");
//     		    }
    		    
//     		    if (--takeCount <= 0) { 
//     		    	owner = null;
//     		    }
    		    
//     		    lock.notify();
//     		}
//     	}
    }
    
    /**
     * Causes the calling thread to block until the mutex semaphore
     * can be taken. When returning, the calling thread is registered
     * as the owner of the semaphore. In some systems, this operation
     * is referred to as <em>lock</em> since it is used to lock some
     * resource. This mutex is recursive which means that a thread
     * can take this mutex several times in a row without deadlocking.
     * However a certain amount of take calls must be acompanied by an 
     * equal amount of give calls to make sure the mutex is freed 
     * properly.
     * 
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. The caller is <b>not</b> expected to catch it since
     * it indicated severe violation in the handling of resources,
     * concurrency, or time. Thus, you should not catch such an error,
     * except for special cases like handling an emergency stop.
     */
    public void take() {
//     	if (RTSystem.NATIVE_SEM) {
    		takeMutexSem(target);
//     	} else {
//     		try {
//     		    synchronized (lock) {
    		
//     		    	while (takeCount > 0 && owner != Thread.currentThread()) {
//     		    		lock.wait();
//     		    	}
    		    
//     		    	if (takeCount <= 0) { 
//     		    		takeCount = 1;
//     		    		owner = Thread.currentThread();
//     		    	} else takeCount++;
//     		    }
//     		} catch (InterruptedException exc) {
// 		    // HERE BE DRAGONS! Include me again.
// 		    System.out.println("MutexSem.take interrupted");
//     		    // throw new SemInterrupted("MutexSem.take interrupted: " + exc);
//     		}	

//     	}
    }
    
    /* NOT NATIVELY IMPLEMENTED YET */
    /** 
     * Try to take the semaphore but give up after a certain time. Be
     * careful not to neglect the returned value; the resource has
     * been locked only if true is returned.
     * 
     * @return true if the semaphore was successfully taken, and
     * false if the timeout occurred which means that the semaphore was
     * not taken.
     *
     * @throws SemInterrupted (which is an error and not an exception)
     * if the calling thread has been interrupted before or during
     * blocking. The caller is <b>not</b> expected to catch it since
     * it indicated severe violation in the handling of resources,
     * concurrency, or time. Thus, you should not catch such an error,
     * except for special cases like handling an emergency stop.
     */
	public boolean tryTake(long timeout) {
	    takeMutexSem(target);
	    return true;
// 		if (RTSystem.NATIVE_SEM) {
// 			return takeMutexSem(target, timeout);
// 		} else {
// 			long now;
// 			long arrival = now = System.currentTimeMillis();
// 			try {
// 				synchronized (lock) {
// 					while (takeCount > 0 && now < arrival + timeout) {
// 						lock.wait(timeout - (now - arrival));
// 						now = System.currentTimeMillis();
// 					}
// 				}
// 				if (now >= arrival + timeout)
// 					return false;
// 				else {
					
// 					if (takeCount <= 0) { 
//     		    		takeCount = 1;
//     		    		owner = Thread.currentThread();
//     		    	} else takeCount++;
					
// 					return true;
// 				}
// 			} catch (InterruptedException exc) {
// 				//HERE BE DRAGONS! Include me again
// 				System.out.println("MutexSem.tryTake interrupted");
// 				//throw new SemInterrupted("MutexSem.tryTake interrupted: " + exc);
// 			}
// 		}
	}
     
    private native long createMutexSem(int iValue);
    private native void giveMutexSem(long target);
    private native void takeMutexSem(long target);
    private native boolean takeMutexSem(long target, long timeout);

}
