// vim:ts=4:
package se.lth.cs.robot.servo;

import se.lth.cs.realtime.Monitor;
import se.lth.cs.realtime.event.RTEvent;

/**
 * PID regulator, with a parameter for selecting which derivative of the 
 * feedback signal to use, and internal saturation
 */
public class CascadePID extends ClassicPID implements CascadePID_Enum {

    protected String name = "CascadePID";

	protected float limit(float v) {
			if( v > p.val[maxv_]) {
					return p.val[maxv_];
			} else if( v < p.val[minv_]) {
					return p.val[minv_];
			} else {
					return v;
			}
	}
	
	public Sample reference(Sample setpoint) {
			//System.err.println("PID:("+name+") reference: "+setpoint.siso);
			theDesiredReference = setpoint;
			float v = setpoint.siso;
			float u = limit(v);
			if(u != v) {
					theAdmittedReference.siso = u;
					theAdmittedReference.saturated = true;
			} else {
					theAdmittedReference = setpoint;
					theAdmittedReference.saturated = false;
			}
//			if(setpoint.siso > p.val[maxv_]) {
//					//System.err.println("PID:("+name+") saturating: "+theAdmittedReference.siso+" > "+p.val[maxv_]);
//					theAdmittedReference.siso = p.val[maxv_];
//					theAdmittedReference.saturated = true;
//			} else if (setpoint.siso < p.val[minv_]) {
//					//System.err.println("PID:("+name+") saturating: "+theAdmittedReference.siso+" < "+p.val[minv_]);
//					theAdmittedReference.siso = p.val[minv_];
//					theAdmittedReference.saturated = true;
//			} else {
//					theAdmittedReference = setpoint;
//					theAdmittedReference.saturated = false;
//			}
			theAdmittedReference.backward = true;
			return theAdmittedReference;
	}
	
    public Sample compute(Sample feedback) {
        Sample out = theDesiredOutput;
		out.setTimestamp(feedback.getMillis());
		// may be needed to get backwards prop. to work in the general case
		//out.setSource(this);	
        theDynamicsInput = feedback;
		// select derivative to use as input
		float input;
		if(p.val[ndot_] == 0.0f ){
			input = theDynamicsInput.siso;
		} else if(theDynamicsInput.derived >= 1 && p.val[ndot_] == 1.0f ){
			input = theDynamicsInput.siso1Dot;
		} else if(theDynamicsInput.derived >= 2 && p.val[ndot_] == 2.0f ){
			input = theDynamicsInput.siso2Dot;
		} else if(theDynamicsInput.derived >= 3 && p.val[ndot_] == 3.0f ){
			input = theDynamicsInput.siso3Dot;
		} else {
			System.out.println("CascadePID.compute(): strange ndot");
			//System.err.println("CascadePID("+name+"):compute: strange ndot:"+p.val[ndot_]+">"+feedback.derived);
			input=0;
		}
		//System.err.println("PID("+name+"):compute: ref="+((Sample)theAdmittedReference).siso+" di="+feedback.siso);
        out.siso = dynamics.compute(input, theAdmittedReference.siso);
		//System.err.println("PID("+name+"): out="+out.siso);
		out.forward = true;
		out.backward = out.saturated = false;

		// Alternative 1:
		dispatcher.dispatch(out);
		//dispatcher.dispatch(feedback);
		// Alternative 2:
		//dispatcher.dispatch(new ReferenceEvent(this, feedback.getMillis(),out));
		//dispatcher.dispatchActivity(new FeedbackEvent(this, feedback.getMillis(),feedback));
		feedback.forward = true;
		feedback.backward = feedback.saturated = false;
        return feedback;
    }

    public SDSblock update(Sample tracked) {
        theAdmittedOutput = tracked;
        dynamics.update(tracked.siso);
		if(prevBlock != null) {
			if(tracked.forward && tracked.backward && tracked.saturated) {
				prevBlock.update(tracked);
			} else {
				prevBlock.update(theAdmittedReference);
			}
			prevBlock = null;
		} else {
		}
        return null;
    }

    protected SDSblock prevBlock;
    public RTEvent putEvent(RTEvent e) {
	// Alternative 1:
	  if(e instanceof SensorSample) {
//		  obsolete
//		  compute((Sample)e);
	  } else if(e instanceof Sample) {
		  // keep source of the reference to allow
		  // backwards propagation
		  if(e.getSource() instanceof SDSblock) {
			  prevBlock = (SDSblock)e.getSource();
		  } else {
				System.out.println("CascadePID.putEvent(): strange source");
				//System.err.println("PID("+name+"): strange source="+e.getSource());
		  }
		  reference((Sample) e);
		  //System.out.println("CascaePID.putEvent(): sample.siso=" + ((Sample)e).siso);
	  }
	  // Alternative 2:
//	  if(e instanceof ReferenceEvent) {
//		  if(e.getSource() instanceof SDSblock) {
//			  prevBlock = (SDSblock)e.getSource();
//		  } 
//		  reference( ((ReferenceEvent) e).ref);
//	  } else if(e instanceof FeedbackEvent) {
//		  compute( ((FeedbackEvent) e).di);
//	  }
//
//	  old stuff:
//        if(e instanceof Sample) {
//	    prevBlock = (SDSblock)e.getSource();
//	    reference((Sample) e);
//           compute(theDynamicsInput); // temporary hack
////	    if(e.getSource() instanceof SDSblock) {
////                ((SDSblock)e.getSource()).update(theAdmittedOutput);
////            }
//        }
        return null;
    }

    public CascadePID() {
		dynamics = new PID();
		p = new Parameters(ClassicPID_Enum.PARAM_DIM+CascadePID_Enum.PARAM_DIM, 
						   paramName);
		p.val[K_] = -0.30f;
		p.val[Ti_] = 2.20f;
		p.val[Tr_] = 1.30f;
		p.val[Td_] = 0.75f;
		p.val[N_] = 10f;
		p.val[beta_] = 0.40f;
		p.val[h_] = 0.05f;
		p.val[I_on_] = 0f;
        p.val[ndot_] = 0.0f;
        p.val[maxv_] = 10.0f;
        p.val[minv_] = -10.0f;
		dynamics.setParameters(p);
    }

    public CascadePID(String name) {
		dynamics = new PID();
		p = new Parameters(ClassicPID_Enum.PARAM_DIM+CascadePID_Enum.PARAM_DIM, 
						   paramName);
		p.val[K_] = -0.30f;
		p.val[Ti_] = 2.20f;
		p.val[Tr_] = 1.30f;
		p.val[Td_] = 0.75f;
		p.val[N_] = 10f;
		p.val[beta_] = 0.40f;
		p.val[h_] = 0.05f;
		p.val[I_on_] = 0f;
        p.val[ndot_] = 0.0f;
        p.val[maxv_] = 10.0f;
        p.val[minv_] = -10.0f;
		dynamics.setParameters(p);
		this.name = name;
    }
    
    public CascadePID(Parameters p) {
		super(p);
    }

}
