/*
 * $Id: ClassicPID.java,v 1.6 2005/09/15 09:06:50 sven Exp $
 */
package se.lth.cs.robot.servo;

import se.lth.cs.realtime.Monitor;

/**
 * PID regulator, classig text-book style together with generic block interface.
 */
public class ClassicPID extends SDSblock implements ClassicPID_Enum {

    private static final double eps = 0.000001;

    protected String name = "ClassicPID";

    public ClassicPID() {
        dynamics = new PID();
        p = new Parameters(PARAM_DIM, paramName);
        p.val[K_] = 0.30f;
        p.val[Ti_] = 2.20f;
        p.val[Tr_] = 1.30f;
        p.val[Td_] = 0.75f;
        p.val[N_] = 10f;
        p.val[beta_] = 0.40f;
        p.val[h_] = 0.05f;
        p.val[I_on_] = 1f;
        dynamics.setParameters(p);
    }

    public ClassicPID(String name) {
        this();
        this.name = name;
    }
    
    public ClassicPID(Parameters p) {
        dynamics = new PID();
        dynamics.setParameters(p);
    }

    // TEXT-BOOK CODE BEGIN
    /**
     * Classical text-book version of a PID regulator.
     */
    class PID implements Monitor, Algorithm {

        // States (persist between samples):
        float I, D, yold;

        // Attributes (by compute for update):
        float y, e, v;

        // Exposed parameters that are adjustable by the operator:
        float K, Ti, Tr, Td, N, beta, h;

        // Internal parameters, computed from the exposed ones:
        private float ad, bd;

        private boolean integratorOn;

        /**
         * Calculate output.
         */
        public synchronized float compute(float feedback, float r) {
            y = feedback;
            e = r - y;
            D = ad * D - bd * (y - yold);
            v = K * (beta * r - y) + D;
            if (integratorOn) {
                v = v + I;
            }
            return v;
        }

        /**
         * Update state.
         */
        public synchronized void update(float u) {
            if (integratorOn) {
                I = I + h * K / Ti * e + h * (1 / Tr) * (u - v);
            } else {
                I = 0f;
            }
            yold = y;
        }

        /**
         * Reset states, including the I-part, to zero.
         */
        public synchronized void reset() {
            I = 0f;
            D = 0f;
            yold = 0f;
        }
        
        // TEXT-BOOK CODE END
        // Remaining part of this inner class is block interfacing.

        /**
         * Assign new control parameters.
         */
        public synchronized void setParameters(Parameters p) {
            K = p.val[K_];
            Ti = p.val[Ti_];
            Tr = p.val[Tr_];
            Td = p.val[Td_];
            N = p.val[N_];
            beta = p.val[beta_];
            h = p.val[h_];
            if (Ti > eps && p.bool(I_on_)) {
                integratorOn = true;
            } else {
                integratorOn = false;
                I = 0f;
            }
            ad = Td / (N * h + Td);
            bd = N * K * Td / (N * h + Td);
        }

        /**
         * Get state vector, as updated content of argument.
         */
        public synchronized void getState(StateVector x) {
            x.setState(I, I_);
            x.setState(D, D_);
            x.setState(yold, yold_);
        }

        /**
         * Set the algorithm-internal states according to provide state vector.
         */
        public synchronized void setState(StateVector x) {
            I = x.getState(I_);
            D = x.getState(D_);
            yold = x.getState(yold_);
        }        
    }

    public Sample compute(Sample feedback) {
        Sample out = theDesiredOutput;
	out.setTimestamp(feedback.getMillis());
        theDynamicsInput = feedback;
        out.siso = dynamics.compute( // ...
				theDynamicsInput.siso, theAdmittedReference.siso);
        out.forward = true;
        out.backward = out.saturated = false;
	theDesiredOutput = out;
	dispatcher.dispatch(theDesiredOutput);
	feedback.forward = true;
	feedback.backward = feedback.saturated = false;
        return feedback;
    }

    public SDSblock update(Sample tracked) {
        theAdmittedOutput = tracked;
        dynamics.update(tracked.siso);
        return null;
    }

    public void setParameters(Parameters newPars) {
        dynamics.setParameters(p = newPars);
    }

    public Parameters getParameters() {
        return (Parameters) p.clone();
    }
}
