/*
 * $Id: SDSblock.java,v 1.6 2005/09/15 09:06:50 sven Exp $
 */
package se.lth.cs.robot.servo;

import se.lth.cs.realtime.*;
import se.lth.cs.realtime.event.*;

/**
 * Sampled Dynamic System block with properties permitting composition such that
 * regulator objects or other sampled dynamic systems comprise entities with the
 * same compositional properties as their aggregated parts. SDS blocks are
 * mathematically causal (outputs are well defined and depend on former and
 * current inputs, and time as defined by the timestamps of the dynamics input
 * is monotonously increasing).
 * 
 */

public abstract class SDSblock implements SDSinput, SDSoutput, RTEventListener, RTEventSource, Activity {

    static {
        // Hide ASCII graphics from (stupid) eclipse comment formatting:
        if (false) {
            System.out.println("                 |     |                  ");
            System.out.println("             +---v-----v---+              ");
            System.out.println("             |  LD    par  |              ");
            System.out.println("             |             |              ");
            System.out.println("          --->- DR     DO ->---           ");
            System.out.println("             |             |              ");
            System.out.println("          ---<- AR     AO -<---           ");
            System.out.println("             |             |              ");
            System.out.println("             |  SO     DI  |              ");
            System.out.println("             +---v-----^---+              ");
            System.out.println("                 |     |                  ");
        }
    };

    /**
     * Set to <code>true</code> if this regulator is driven by its own thread.
     * Otherwise, the regulator object is driven by the callers of the methods,
     * e.g. when setpoints and samples values propagate through this block.
     */
    protected volatile boolean activity;

    /**
     * The <code>performer</code> is the thread calling the
     * <code>compute</code> (and <code>update</code>) methods, to be used
     * for propagation of mode changes, thread control, and errors/exception. If
     * <code>activity</code> is true the <code>performer</code> should be
     * the driving thread, but otherwise it may be <code>null</code>.
     */
    protected volatile Activity performer;

    /**
     * The control/dynamics/mathematical content, to be implemented in a subclass 
     * as an inner class implementing an Algorithm, and then assigned to this field.
     */
    protected Algorithm dynamics;
    
    protected String name;

    /**
     * The <code>theDynamicsInput</code> (DI) comprises the sampled data
     * driving the advancement in time is the most recently obtained measured
     * value, as provided via the <code>compute</code> method.
     */
    protected volatile Sample theDynamicsInput; // DI

    /**
     * The <code>theDesiredReference</code> (DR) is the most recently
     * requested setpoint (sometimes called the reference) value. If no
     * back-propagation of saturation is done to the setpoint, i.e. restricting
     * the admitted setpoint, this object is identical to the
     * <code>admittedControl</code>.
     */
    protected volatile Sample theDesiredReference; // DR

    /**
     * The <code>theAdmittedReference</code> (AR) is the setpoint that is
     * accepted to be used by this regulator. When being different from the
     * <code>theDesiredOutput</code>, this signal should be propagated
     * backward as a argument to the <code>update</code> of the
     * <code>SDSblock</code> (see argument to <code>admitted</code> method)
     * that calls <code>command</code> method of this regulator.
     */
    protected volatile Sample theAdmittedReference; // AR

    /**
     * The <code>theDesiredOutput</code> (DO) signal is the nominal control
     * output from this regulator, before effected by any saturation or tracking
     * of manual control.
     */
    protected volatile Sample theDesiredOutput; // DO

    /**
     * The <code>theAdmittedOutput</code> (AO) is the control output after
     * being effected by saturation or tracking of manual control. Provided via
     * <code>update</code>.
     */
    protected volatile Sample theAdmittedOutput; // AO

    /**
     * Uncontrollable input that has been possible to measure or compute
     * (externally to this block), such as a Load Disturbance (LD). Regulation
     * with good disturbance rejection should as much as possible make the
     * effect of the <code>theLoadDisturbance</code> small. This input is
     * assumed to be piece-wise constant and changing asynchronously with the
     * computing of the internal states, and has therefore no direct (time
     * critical) computations based on it, and there are (as seen from the
     * outside) no internal saturations effecting the value.
     */
    protected volatile Sample theLoadDisturbance; // LD

    /**
     * As some disturbances can actually be helful for the control with respect
     * to control signal saturation (such as friction making it easier to stop a
     * motion), the most helpful (to obtain best-case performance) disturbance
     * is saved in the <code>theDisturbanceMinimum</code> (when the
     * <code>minimum</code> flag of the disturbance signal is
     * <code>true</code>).
     */
    protected volatile Sample theDisturbanceMinimum;

    /**
     * The worst-case disturbance (possibly locally if this is being updated
     * during run time, e.g. when the payload of the robot changes) is stored in
     * <code>theDisturbanceMaximum</code>. Used for determining the admitted
     * value without having to compute and output the full control value, which
     * is better performed based on the sampled value for feedback.
     */
    protected volatile Sample theDisturbanceMaximum;

    /**
     * The currently on-purpose added sequence of disturbances, to get persitent excitation
     * of the system dynamics for process identification, is stored in
     * <code>theStateExcitation</code>.
     */
    protected volatile Sample[] theStateExcitation;

    /**
     * The <code>state</code> is the control state of this regulator, and
     * <code>state.length</code> is the so called order of the regulator.
     */
    protected volatile StateVector x;

    /**
     * The <code>param</code> contains tunable and internal parameters.
     */
    protected volatile Parameters p;

    /**
     * Command a new target value for the control by providing a nominal
     * setpoint.
     * 
     * @param setpoint
     *            the desired setpoint without considering control saturations.
     * @return the possibly saturated (admitted) setpoint (as also obtainable by
     *         calling <code>admitted</code>), or the setpoint object if no
     *         saturation occured.
     */
    public Sample reference(Sample setpoint) {
        theDesiredReference = theAdmittedReference = (Sample) setpoint.clone();
        setpoint.backward = true;
        return setpoint;
    }

    /**
     * Return the admitted setpoint according to saturations, operating ranges,
     * disturbances, etc. Normally the admitted setpoint is returned by the
     * <code>reference</code> method as a <code>Sample</code> with
     * <code>backward==true</code>, or saturation only effects the control
     * error integration locally (anti-reset-windup), which means this method is
     * not used. But to to carry out the state update of the commanding
     * regulator this method is useful internally in subclasses, and it also
     * permits having the setpoint command separated from the backward
     * propagation of the admitted value, which permits other computations (of
     * disturbances, joint coupling effects, etc.) to be carried out in between
     * (without blocking the commanding thread).
     * 
     * @param commander
     *            the regulator computing the setpoint of this one, which in
     *            practise will be the outer loop in a cascade control.
     * @return the accually accepted setpoint.
     */
    public Sample admitted(SDSblock commander) {
        Sample ans = theAdmittedReference;
        if (commander != null) {
            commander.update(ans);
        }
        return ans;
    }

    /**
     * Inform about all known load disturbances, by providing the measured and
     * possibly derived disturbance values. In case of many measurable
     * disturbances, they are to be merged into the array of the
     * <code>disturbance</code> object.
     * 
     * @param load
     *            the measured/known disturbances to suppress.
     */
    public void disturbance(Sample load) {
        if (load.minimum || load.maximum) {
            if (load.minimum) {
                theDisturbanceMinimum = load;
            }
            if (load.maximum) {
                theDisturbanceMaximum = load;
            }
        } else {
            theLoadDisturbance = load;
        }
    }

    /**
     * Compute the states of the dynamics of this block based on the input data sample, including its timestamp. 
     * For a regulator this means computing the control signal based on the sampled process-output value.
     * This method should execute the core of the control algorithm, and this is the
     * most time-critical part of the code since this is the parts of the control that is evaluated as part of the control loop. Thus, 
     * delays here will increase phase lag of the feedback control and jepordize damping and stability.
     * 
     * @param input The driving input signal, typically denoted u in a block diagram.
     * @return the or a sample object that can be used for next sample (normally reducing GC workload).
     */
    public Sample compute(Sample input) {
        input.forward = true;
        theDynamicsInput = input;
        
        return input;
    }

    /**
     * Based on possible saturations according to the tracked control signal,
     * update the state of this regulator, and possibly adjust the admitted
     * setpoint according to the effective control signal. The return value is
     * expected to be used as an argument to <code>admitted</code>, if direct
     * (not waiting for next <code>command</code>) backward propagation is
     * desired (<code>null</code> not returned).
     * 
     * @param tracked
     *            the control output that was actuated, being equal to the
     *            desired control if there were no saturations.
     * @return the setpoint provider to be informed about the saturation of the
     *         control signal. Usually this will be <code>null</code> since
     *         saturation propagation to the commanding regulator is normally
     *         done directly in the call of <code>command</code> based on a
     *         nominal model and the maximum disturbance.
     */
    public SDSblock update(Sample tracked) {
        return null;
    }

    public Sample sampleOutput() {
        return theAdmittedOutput;
    }
    
    public Sample sampleOutput(Sample s) {
        theAdmittedOutput.copy(s);
	return s ;
    }

    public void excite(Sample[] pertubation) {
        theStateExcitation = pertubation;
    }
    
    /**
     * The list of regulators that get input from this one.
     * 
     * @see se.lth.cs.realtime.event.RTEventDispatcher
     */
    protected RTEventDispatcher dispatcher = new RTEventDispatcher(this);

    /**
     * Get the dispatcher object to use for subscription, by the regulators that
     * obtain their setpoint from the output of this regulator.
     * 
     * @see se.lth.cs.realtime.event.RTEventSource#getDispatcher()
     */
    public RTEventDispatcher getDispatcher() {
        return dispatcher;
    }

    /**
     * Subscribe on ordinary application events, with listeners ordered
     * according to the priority of the caller. It is recommended to use the
     * featured methods of the dispatcher instead.
     * 
     * @see se.lth.cs.realtime.event.RTEventSource#addListener(se.lth.cs.realtime.event.RTEventListener)
     */
    public void addListener(RTEventListener subscriber) {
	//try {
            //addListener(subscriber, Class.forName("se.lth.robot.servo.Sample"));
            //addListener(subscriber, Class.forName("se.lth.robot.servo.ReferenceEvent"));
	    addListener(subscriber, null);
	//} catch (ClassNotFoundException e) {
        //    throw new RTError("ClassNotFoundException");
	//}
    }
    
    public void addListener(RTEventListener subscriber, Class type) {

        RTThread rtt = RTThread.currentRTThread();
        if (rtt != null) {
            dispatcher.addListener(subscriber, RTThread.currentRTThread()
                    .getPriority(), type);
        } else {
            dispatcher.addListener(subscriber, Thread.currentThread()
                    .getPriority(), type);
        }
	//System.err.println("addListener: "+dispatcher);
    }
    

    /**
     * Generic event-based version of command input. By default, simply calls
     * <pre>
     * return reference((Sample) event);
     * </pre>
     * 
     * @see se.lth.cs.realtime.event.RTEventListener#putEvent(se.lth.cs.realtime.event.RTEvent)
     */
    public RTEvent putEvent(RTEvent event) {
        return reference((Sample) event);
    }

    public void setParameters(Parameters p) {
        this.p = p;
    }

    public Parameters getParameters() {
        return p;	    
    }

    public SDSblock() {
	    // Instantiate Sample objects to put the block object in a consistent
	    // state. Perhaps some of these could reference the same Sample; e.g
	    // theAdmittedOutput = theDesiredOutput?
	    theDynamicsInput = new Sample(this,0);
	    theDesiredReference = new Sample(this,0);
	    theAdmittedReference = new Sample(this,0);
	    theDesiredOutput = new Sample(this,0);
	    theAdmittedOutput = new Sample(this,0);
	    theLoadDisturbance = new Sample(this,0);
	    theDisturbanceMinimum = new Sample(this,0);
	    theDisturbanceMaximum = new Sample(this,0);
    }

    public void addActListener(RTEventListener dest) {
	    try {
	    getDispatcher().addActListener(dest, 5, Class.forName("se.lth.cs.robot.servo.FeedbackEvent"));
	    } catch( ClassNotFoundException e ) {
		    throw new RTError("Class not found...");
	    }
    }

    public Timebase getTimebase() {
	    return null;
    }

    public Environment getEnv() {
	    return null;
    }

    public boolean isAlive() {
	    return true;
    }

    public void start() {}
    public void join() throws InterruptedException{}
    public void terminate() {}
}
