package se.lth.cs.robot.servo;

import se.lth.cs.realtime.event.RTEventSource;

public interface SDSinput extends RTEventSource{

   /**
     * Command a new target value for the control by providing a nominal
     * setpoint.
     * 
     * @param setpoint
     *            the desired setpoint without considering control saturations.
     * @return the possibly saturated (admitted) setpoint (as also obtainable by
     *         calling <code>admitted</code>), or the setpoint object if no
     *         saturation occured.
     */
    public Sample reference(Sample setpoint);

    /**
     * Return the admitted setpoint according to saturations, operating ranges,
     * disturbances, etc. Normally the admitted setpoint is returned by the
     * <code>reference</code> method as a <code>Sample</code> with
     * <code>backward==true</code>, or saturation only effects the control
     * error integration locally (anti-reset-windup), which means this method is
     * not used. But to to carry out the state update of the commanding
     * regulator this method is useful internally in subclasses, and it also
     * permits having the setpoint command separated from the backward
     * propagation of the admitted value, which permits other computations (of
     * disturbances, joint coupling effects, etc.) to be carried out in between
     * (without blocking the commanding thread).
     * 
     * @param commander
     *            the regulator computing the setpoint of this one, which in
     *            practise will be the outer loop in a cascade control.
     * @return the accually accepted setpoint.
     */
    public Sample admitted(SDSblock commander);


    /**
     * Compute the states of the dynamics of this block based on the input data sample, including its timestamp. 
     * For a regulator this means computing the control signal based on the sampled process-output value.
     * This method should execute the core of the control algorithm, and this is the
     * most time-critical part of the code since this is the parts of the control that is evaluated as part of the control loop. Thus, 
     * delays here will increase phase lag of the feedback control and jepordize damping and stability.
     * 
     * @param input The driving input signal, typically denoted u in a block diagram.
     * @return the or a sample object that can be used for next sample (normally reducing GC workload).
     */
    public Sample compute(Sample input); 
}
