package se.lth.cs.throttlenet;

import se.lth.cs.throttlenet.util.*;
import java.net.*;


/**
 * The <code>ThrottleNetPacket</code> class describes a ThrottleNet packet.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.4, 24/10/03
 * @since   ThrottleNet0.1
 */
class ThrottleNetPacket {
	
	
	/**
	 * The data array of this ThrottleNet packet.
	 */
	private byte[] theData;
	
	
	/**
	 * The header size of the first fragment.
	 */
	public static final int HEADER_FIRST = 8;
	
	
	/**
	 * The header size of the remaining fragments.
	 */
	public static final int HEADER_REST = 4;
	
	
	/**
	 * The size of the data.
	 */
	private int dataSize;
	
	
	/**
	 * The size of the fragment.
	 */
	private int fragSize;
	
	
	/**
	 * Number of fragments in this ThrottleNet packet.
	 */
	private int nbrFragments;
	
	
	/**
	 * The sequence number of the first fragment belonging to the current packet.
	 */
	private int fragmentStartNbr;
	
	
	/**
	 * Constructs a ThrottleNet packet with a given size of data.
	 * 
	 * @param   size     the amount of data in the ThrottleNet packet.
	 * @param   fragSize the amount of data in each fragment.
	 *
	 */
	public ThrottleNetPacket(int size, int fragSize) {
		this(new byte[size], size, fragSize, 0);
	}
	
	
	/**
	 * Constructs a ThrottleNet packet with a given data and its size.
	 *
	 * @param   realdata      the array of data.
	 * @param   size          the amount of data in the ThrottleNet packet.
	 * @param   fragSize      the amount of data in the ThrottleNet packet.
	 * @param   fragStartNbr  the sequence number in the first fragment of
	 *								  this packet.
	 *
	 */
	public ThrottleNetPacket(byte[] realdata, int size, int fragSize, 
																				int fragStartNbr) {
		this.fragSize = fragSize;
		dataSize = size;
		theData = realdata;
		
		int firstFragDataSize = fragSize - HEADER_FIRST;
		int rest = size - firstFragDataSize;
		int fragDataSize = fragSize - HEADER_REST;
		// Always round up
		nbrFragments = 1 + (rest + (fragDataSize - 1)) / fragDataSize; 
		fragmentStartNbr = fragStartNbr;
	}
	
	
	/**
	 * Returns the byte array representation of the requested fragment.
    *
	 * @param  i the number of the requested fragment
	 *
    * @return the byte array representation of the requested fragment.
    */
	public byte[] getFragment(int i) {
		if(i >= nbrFragments)
			return null;
		
		int dataOffset = 0;
		if(i > 0) {
			dataOffset = fragSize - HEADER_FIRST;
			dataOffset += (i-1) * (fragSize - HEADER_REST);
		}
		// Adjust fragment number offset
		int fragOffset = (i == 0) ? 4 : 0;
		
		// Data size
		int copyLen = Math.min(theData.length - dataOffset, fragSize - fragOffset 
																					- HEADER_REST);
		
		// This fragment size
		int size = Math.max(46, copyLen + fragOffset + HEADER_REST);
		
		byte[] frag = new byte[size];
		if(i == 0) {
			// adding data size (1) 4byte
			// |----------------------------------------------------------------|
			// |  1  | 2 | 3 |                      4                           |
			// |----------------------------------------------------------------|
			System.arraycopy(Utils.intToBytes(dataSize), 0, frag, 0, 4);
			
			// Set msb in frag[0]
			frag[0] |= 0x80;

		}
		// adding fragment nbr (2) 2byte
		// |----------------------------------------------------------------|
		// |  1  | 2 | 3 |                      4                           |
		// |----------------------------------------------------------------|
		int fragNbr = (fragmentStartNbr + i) % 0x8000;
		int tmp = (fragNbr & 0xFF00) >> 8;
		frag[fragOffset]   = (byte)(tmp);
		tmp = fragNbr & 0x00FF;
		frag[fragOffset+1] = (byte)(tmp);
		// adding last fragment nbr (3) 2byte
		// |----------------------------------------------------------------|
		// |  1  | 2 | 3 |                      4                           |
		// |----------------------------------------------------------------|
		int lastFragNbr = (fragmentStartNbr + nbrFragments - 1) % 0x8000;
		frag[fragOffset+2] = (byte)((lastFragNbr & 0xFF00) >> 8);
		frag[fragOffset+3] = (byte)(lastFragNbr & 0x00FF);
		// copying data (4)
		// |----------------------------------------------------------------|
		// |  1  | 2 | 3 |                      4                           |
		// |----------------------------------------------------------------|
		
		System.arraycopy(theData, dataOffset, frag, fragOffset + HEADER_REST, 
																							  copyLen);
		return frag;
	}
	
	
	/**
	 * Initializes a fragment with the specified number.
    *
	 * @param frag   the fragment array.
	 * @param iFrag  the number of the fragment.
    */
	public void setFragment(byte[] frag, int iFrag) {
		
		int fragOffset = iFrag == 0 ? HEADER_FIRST : HEADER_REST;
		int len;
		int dataOffset = 0;
		if(iFrag > 0) {
			dataOffset = fragSize - HEADER_FIRST;
			dataOffset += (iFrag-1) * (fragSize - HEADER_REST);
		}
		len = Math.min(dataSize - dataOffset, fragSize - fragOffset);
		
		System.arraycopy(frag, fragOffset, theData, dataOffset, len);
	}
	
	
	/**
	 * Returns the data size of this ThrottleNet packet.
    *
	 * @return the data size of this ThrottleNet packet.
    */
	public int getDataSize() {
		return dataSize;
	}
	
	
	/**
	 * Returns the data array of this ThrottleNet packet.
    *
	 * @return the data array of this ThrottleNet packet.
    */
	public byte[] getData() {
		return theData;
	}
	
	
	/**
	 * Returns the number of fragments for this packet.
	 * 
	 * @return the number of fragments for this packet.
	 */
	public int getFragmentCount() {
		return nbrFragments;
	}
}
