package se.lth.cs.throttlenet;

import java.net.*;
import java.io.*;


/**
 * The <code>ThrottleNetReceiveSocket</code> class describes a socket for
 * receiving data.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.8, 12/11/03
 * @since   ThrottleNet0.1
 */
 
 public final class ThrottleNetReceiveSocket extends ThrottleNetSocket {
	
	 
	/**
	 * Time to deliver each packet in microseconds.
	 */
	private int deadline;
	
	
	/**
	 * Constructs and initializes a throttlenet receive socket.
	 * 
	 * @param   name			the name of the connection.
	 * @param   size			the amount of data to be received in the throttlenet
	 * 							packet.
	 * @param   period_us	the time between two packets in microseconds.
	 * 
	 * @exception SocketException if the socket could not be opened.
 	 * @exception IOException if an I/O error occurs.
	 */
	public ThrottleNetReceiveSocket(String name, int size, int period_us)
												throws SocketException, IOException {
		this(name, size, period_us, 2000000000); // ungefr 2^31-1
	}
	
	
	/**
	 * Constructs and initializes a throttlenet receive socket.
	 * 
	 * @param   name			the name of the connection.
	 * @param   size			the amount of data to be received in the throttlenet
	 * 							packet.
	 * @param   period_us	the time between two packets in microseconds.
	 * @param   deadline		time to deliver each packet in microseconds.
	 * 
	 * @exception SocketException if the socket could not be opened.
 	 * @exception IOException if an I/O error occurs.
	 */
	public ThrottleNetReceiveSocket(String name, int size, int period_us,
								int deadline) throws SocketException, IOException {
		super(name, size, period_us);
		this.deadline = deadline;
		register(name, size, period_us, deadline);
	}
	
	
	/**
	 * Constructs a default throttlenet receive socket.
	 * 
	 * @exception SocketException if the socket could not be opened.
 	 * @exception IOException if an I/O error occurs.
	 */
	public ThrottleNetReceiveSocket() throws SocketException, IOException {
		super(null, 0, 0);
		socket = 0;
		deadline = 0;
	}
	
	
	/**
	 * Initializes the throttlenet receive socket.
	 *
	 * @param   name			the name of the connection.
	 * @param   size			the amount of data to be received in the throttlenet
	 * 							packet.
	 * @param   period_us	the time between two packets in microseconds.
	 * @param   deadline		time to deliver each packet in microseconds.
	 * 
	 * @exception IOException if an I/O error occurs.
	 */
	public void register(String name, int size, int period_us, int deadline)
																		throws IOException {
		this.name = name;
		this.size = size;
		this.period_us = period_us;
		this.deadline = deadline;
		socket = initReceiveSocket(name, size, period_us, deadline);
	}
	
	
	/**
	 * Receives an externalized object reconstructed from a byte array.
    *
    * @param obj	the externalized object to be reconstructed.
	 * 
	 * @exception IOException if an I/O error occurs.
	 * @exception ClassNotFoundException if no definition for the class could be
	 *            found.
	 * @exception PacketLossException if a packet is lost. 
    */
	public final void receive(Externalizable obj)
			throws IOException, PacketLossException /*, ClassNotFoundException*/ {
		if(socket == 0) throw new SocketException("Socket not initialied");
		// Lacks implementation
		// utils.externalize(obj, receive());
	}
	
	
	/**
	 * Receives a throttlenet packet and returns its data.
    *
    * @return the received data.
	 * 
	 * @exception IOException if an I/O error occurs.
	 * @exception PacketLossException if a packet is lost.
    */
	public final byte[] receive() throws IOException, PacketLossException {
		int size = this.size;
		byte[] tmp = new byte[size];
		int n;
		n = receiveData(socket, tmp, tmp.length);
		if(n == size)
			return tmp;
		byte[] data = new byte[n];
		byte tmp_b = 0;
		//System.arraycopy(tmp, 0, data, 0, n); // Not implemented in Java2C
		for(int i = 0; i < n; i++) { // Own implementation of copying
			data[i] = tmp[i];
		}
		return data; 
	}
	
	
	/**
	 * Closes this socket.
    */
	public final void close() {
		closeSocket(socket);
		
	}

	
	private native void closeSocket(int socket);
	
	
	private native int initReceiveSocket(String dstName, int size, int period_us, int deadline);
	
	
	private native int receiveData(int socket, byte[] data, int length) throws IOException;
	
}
