package se.lth.cs.throttlenet;

import java.net.*;
import java.io.*;

/**
 * The <code>ThrottleNetTransmitSocket</code> class describes a socket for
 * transmitting data.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.8, 12/11/03
 * @since   ThrottleNet0.1
 */
 public final class ThrottleNetTransmitSocket extends ThrottleNetSocket {

	
	/**
	 * Constructs and initializes a throttlenet transmit socket.
	 * 
	 * @param   name			the name of the connection.
	 * @param   size 			the amount of data to be sent in the throttlenet
	 * 							packet.
	 * @param   period_us 	the time between two packets in microseconds.
	 * 
	 * @exception SocketException if the socket could not be opened.
	 * @exception IOException if an I/O error occurs.
	 */
	public ThrottleNetTransmitSocket(String name, int size, int period_us)
											throws SocketException, IOException {
		super(name, size, period_us);
		socket = initTransmitSocket(name, size, period_us);
	}
	
	
	/**
	 * Constructs a throttlenet transmit socket.
	 * 
	 * @exception SocketException if the socket could not be opened.
	 * @exception IOException if an I/O error occurs.
	 */
	public ThrottleNetTransmitSocket() throws SocketException, IOException {
		super(null, 0, 0);
		socket = 0;
	}	
	
	
	/**
	 * Opens a connection with specified properites.
	 * 
	 * @param   name			the name of the connection.
	 * @param   size 			the amount of data to be sent in the throttlenet
	 * 							packet.
	 * @param   period_us 	the time between two packets in microseconds.
	 * 
	 * @exception IOException if an I/O error occurs.
	 */
	public void connect(String name, int size, int period_us) throws IOException {
		this.size = size;
		socket = initTransmitSocket(name, size, period_us);
	}
	
	
	/**
	 * Transmits an externalized object over throttlenet.
    *
    * @param obj the externalized object to transmit.
	 * 
	 * @exception InterruptedException if the thread has been interrupted.
	 * @exception Exception if an error occurs.
    */
	 public final void transmit(Externalizable obj)
												throws InterruptedException, Exception {
		if(socket == 0)
			throw new Exception("Socket not initialized");
		// Lacks implementation
		//		byte[] b = utils.externalize(obj);
		//		transmit(b);
	}
	
		
	/**
	 * Transmits an array of bytes over throttlenet.
    *
    * @param data the byte array to transmit.
	 * 
	 * @exception InterruptedException if the thread has been interrupted.
	 * @exception Exception if an error occurs.
    */
	 public final void transmit(byte[] data)
							throws InterruptedException, Exception {
		if(socket == 0)
			throw new Exception("Socket not initialized");
		if(data.length > size) {
			transmitData(socket, data, size);
		} else {
			transmitData(socket, data, data.length);
		}
	}
	
	
	/**
	 * Closes this socket.
	 */
	public void close() {
		closeSocket(socket);
	}
	
	
	private native int initTransmitSocket(String dstName, int size, int period_us);
	
	
	private native int transmitData(int socket, byte[] data, int length) throws IOException;
	
	
	private native void closeSocket(int socket);
	
}



