package se.lth.cs.throttlenet.globethrottle;


/**
 * The <code>ConnectionProperties</code> class describes connection properties
 * for two nodes.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.4, 24/10/03
 * @since   ThrottleNet0.3
 */
class ConnectionProperties {
	
	
	/**
	 * The size of the header of each datagram packet in bits, included the
	 * inter-frame gap.
	 */
	public static final int HEADER_SIZE_UDP = 528;
	
	
	/**
	 * The size of the header of each ethernet packet in bits, included the
	 * inter-frame gap.
	 */
	public static final int HEADER_SIZE_ETH = 304;
	
	
	/**
	 * The size of the header of each packet in bits, included the inter-frame
	 * gap.
	 */
	public static int HEADER_SIZE = HEADER_SIZE_UDP;
	
	
	/**
	 * The address of sending node.
	 * 
	 * @serial
	 */
	private String srcAddress;
	
	
	/**
	 * The address of receiving node.
	 * 
	 * @serial
	 */
	private String dstAddress;
	
	
	/**
	 * The name of the connection.
	 * 
	 * @serial
	 */
	private String name;
	
	
	/**
	 * The time between two packets in microseconds.
	 */
	private int period_us;
	
	
	/**
	 * The amount of data in a ThrottleNet packet in bytes.
	 */
	private int size;
	
	
	/**
	 * The current time between two fragments in microseconds.
	 */
	private int fragmentTime;
	
	
	/**
	 * The current size of each fragment.
	 */
	private int fragmentSize;
	
	
	/**
	 * Constructs properties for a connection.
	 * 
	 * @param   srcAddress      the address of the sending node.
	 * @param   dstAddress      the address of the receiving node.
	 * @param   name      		 the name of the connection.
	 * @param   period_us 		 the time between two packets in microseconds.
	 * @param   size      		 the amount of data in the ThrottleNet packet.
	 * 
	 */
	public ConnectionProperties(String srcAddress, String dstAddress, String name,
																int period_us, int size) {
		this.srcAddress = srcAddress;
		this.dstAddress = dstAddress;
		this.name = name;
		this.period_us = period_us;
		this.size = size;
		fragmentSize = 0;
	}
	
	
	/**
	 * Returns the address of the sending node.
    *
    * @return the address of the sending node.
    */
	public String getSrcAddress() {
		return srcAddress;
	}


	/**
	 * Returns the address of the receiving node.
    *
    * @return the address of the receiving node.
    */
	public String getDstAddress() {
		return dstAddress;
	}
	
	
	/**
	 * Returns the name of the connection.
    *
    * @return the name of the connection.
    */
	public String getName() {
		return name;
	}
	
	
	/**
	 * Returns the period time in microseconds.
    *
    * @return the period time in microseconds.
    */
	public int getTime() {
		return period_us;
	}
	
	
	/**
	 * Returns the size of the data in a ThrottleNet packet in bytes.
    *
    * @return the size of the data in a ThrottleNet packet in bytes.
    */
	public int getSize() {
		return size;
	}
	
	
	/**
	 * Returns the number of packets after fragmentation.
    *	
	 * @param size the data size in the packet
	 * 
    * @return the number of packets after fragmentation.
    */
	public static int nbrPacket(int size) {
		int nPackets = size / GlobeThrottle.MIN_FRAGMENT_DATA;
		nPackets += (size % GlobeThrottle.MIN_FRAGMENT_DATA) == 0 ? 0 : 1;
		return nPackets;
	}
	
	
	/**
	 * Returns the bandwidth for this node.
    *
    * @return the bandwidth for this node
    */
	public long getBandWidth() {
		return getBandWidth(size, period_us);
	}
	
	
	/**
	 * Returns the bandwidth for this node.
    * 
	 * @param size      the size of the data in the ThrottleNet packet in bytes.
	 * @param period_us the period time in microseconds.
	 *
    * @return the bandwidth for this node
    */
	public static long getBandWidth(int size, int period_us) {
		long nPackets = nbrPacket(size);
		//              period_us i # us => mul med 1000000
		return (long)(1.0 / (double)period_us * 1000000.0 * (double)nPackets 
				* (double)(GlobeThrottle.MIN_FRAGMENT_DATA * 8 + HEADER_SIZE)) ;
	}
	
	
	/**
	 * Returns the current fragment size.
	 * 
	 * @return the current fragment size.
	 */
	public int getFragmentSize() {
		return fragmentSize;
	} 
	
		
	/**
	 * Returns the current fragment time.
	 *
	 * @return the current fragment time.
	 */
	public int getFragmentTime() {
		return fragmentTime;
	}
	
		
	/**
	 * Sets the fragment size.
	 *
	 * @param size the fragment size.
	 */
	public void setFragmentSize(int size) {
		fragmentSize = size;
	}
	
	
	/**
	 * Sets the fragment time.
	 * 
	 * @param time the fragment time.
	 */
	public void setFragmentTime(int time) {
		fragmentTime = time;
	}
	

	/**
	 * Returns the string representation of this object.
    *
    * @return the string representation of this object.
    */
	public String toString() {
		String s = "";
		s += "{srcAddress = " + srcAddress + ",";
		s += "dstAddress = " + dstAddress + ",";
		s += "name = " + name + ",";
		s += "period_us = " + period_us + ",";
		s += "size = " + size + ",";
		s += "fragmentSize = " + fragmentSize + ",";
		s += "fragmentTime = " + fragmentTime + "}";
		return s;
	}
}
