package se.lth.cs.throttlenet.globethrottle;

import se.lth.cs.throttlenet.util.*;
import se.lth.cs.throttlenet.event.*;
import java.util.*;

/**
 * The <code>DeadlineController</code> class contains the implementation of 
 * computing ThrottleNet parameters depending on the deadlines. 
 */
class DeadlineController {

	
	/**
	 * Contains all registered connections.
	 */
	private Hashtable endPoints;
	
	
	/**
	 * Contains all transmitting nodes in ThrottleNet.
	 */
	private Hashtable srcNodes;
	
	
	/**
	 * Contains all receiving nodes in ThrottleNet.
	 */
	private Hashtable dstNodes;
	
	
	/**
	 * The flag signaling the state changes in ThrottleNet.
	 */
	private boolean stateChanged;
	
	
	/**
	 * The deadline event occured when the state has been changed.
	 */
	private DeadlineEvent event;
	
	
	/**
	 * The minimum data size of each fragment.
	 */
	private static final int MIN_FRAGMENT_SIZE = 46; // bytes
	
	
	/**
	 * The maximum data size of each fragment.
	 */
	private static final int MAX_FRAGMENT_SIZE = 1500; // bytes
	
	
	/**
	 * The desired weight between the smallest and the largest fragment.
	 */
	private static double WEIGHT = 0.5;
	
	
	/**
	 * Constructs the deadline controller.
	 * 
	 * @param   endPoints		all registered connections.
	 * @param   srcNodes			all transmitting nodes in ThrottleNet.
	 * @param   dstNodes			all receiving nodes in ThrottleNet.
	 * 
	 */
	public DeadlineController(Hashtable endPoints, Hashtable srcNodes, 
																			Hashtable dstNodes) {
		this.endPoints = endPoints;
		this.srcNodes = srcNodes;
		this.dstNodes = dstNodes;
		stateChanged = false;
	}
	
	
	/**
	 * Calculates the optimal fragment size.
	 * 
	 * @param name	the connection name.
	 *
	 * @return the calculated fragment size.
	 */
	public synchronized int calcFragmentSize(String name) {
		EndPoint endP = (EndPoint)endPoints.get(name);
		String dstAddress = endP.getAddress();
		
		// All connections to the endpoint.
		Vector dstV = (Vector)dstNodes.get(dstAddress);
	
		int minFrag = calcMinimumSize(dstV);
		int maxFrag = calcMaximumSize(dstV);
		// Calculate the weighted size p, min < p < max
		// WEIGHT = 0.0 => p = min,     WEIGHT = 1.0 => p = max
		int fragmentSize = (int)((double)(maxFrag - minFrag) * WEIGHT + 
																					(double)minFrag);
		// In microseconds
		int fragmentTime = calcFragmentTime(fragmentSize, dstV.size());
		Vector v = new Vector();
//		System.out.println("Vector size = " + dstV.size());
		for(int i = 0; i < dstV.size(); i++) {
			ConnectionProperties cp = (ConnectionProperties)dstV.elementAt(i);
			if(cp.getFragmentSize() != fragmentSize || cp.getFragmentTime() != 
																					fragmentTime) {
				String n = cp.getName();
				cp.setFragmentSize(fragmentSize);
				cp.setFragmentTime(fragmentTime);
				NewConnectionSettings newC = new NewConnectionSettings(n,
														fragmentSize, fragmentTime);
				v.add(newC);
			}
		}
		event = new DeadlineEvent(v);
		stateChanged = true;
		notifyAll();
		return fragmentSize;
	}
	
	
	/**
	 * Calculates the lowest possible fragment size for each node. 
	 * 
	 * @param dstV	the vector with all connections to the specific end-point.
	 * 
	 * @return the lowest possible fragment size.
	 */
	private int calcMinimumSize(Vector dstV) {
		int min = MIN_FRAGMENT_SIZE;
		int nNodes = dstV.size();
		int G = MIN_FRAGMENT_SIZE + ConnectionProperties.HEADER_SIZE / 8;
		
		for(int i = 0; i < nNodes; i++) {
			ConnectionProperties cp = (ConnectionProperties)dstV.elementAt(i);
			String n = cp.getName();
			EndPoint ep = (EndPoint)endPoints.get(n);
			int deadline = ep.getDeadline();
			// 100 Mbit / #us/s = 100.0 (= 10^8 / 10^6)
			int deadlineBytes = (int)((double)deadline * 100.0 / 8.0); 
			int packSize = ep.getPacketSize();

			double a = (double)(G + nNodes * ConnectionProperties.HEADER_SIZE / 8);
			double b = (double)deadlineBytes / (double)packSize - (double)nNodes;
			double minSizeD = (double)(a / b);
			int minSize = (int)minSizeD;
			// Rounding
			minSize = (int)(-Math.floor(-minSize));
			if(minSize > min)
				min = minSize;
		}
		return min;
	}
	
	
	/**
	 * Calculates the largest possible fragment size for each node. 
	 * 
	 * @param dstV	the vector with all connections to the specific endpoint.
	 * 
	 * @return	the largest possible fragment size.
	 */
	private int calcMaximumSize(Vector dstV) {
		int G = MIN_FRAGMENT_SIZE + ConnectionProperties.HEADER_SIZE / 8;
		int max = MAX_FRAGMENT_SIZE;
		int nNodes = dstV.size();
		
		for(int i = 0; i < nNodes; i++) {
			ConnectionProperties cp = (ConnectionProperties)dstV.elementAt(i);
			String n = cp.getName();
			EndPoint ep = (EndPoint)endPoints.get(n);
			int deadline = ep.getDeadline();
			// 100 Mbit / #us/s = 100.0 (= 10^8 / 10^6)
			int deadlineBytes = (int)((double)deadline * 100.0 / 8.0); 
			int size = (int)((double)(deadlineBytes - G) / (double)nNodes);
			if(size < max)
				max = size;
		}
		return max;
	}
	
	
	/**
	 * Calculates the fragment time to meet the deadline.
	 *
	 * @param fragmentSize	the fragment size.
	 * @param nNodes			the number of nodes involved in a specific
	 *								connection.
	 *
	 * @return	the fragment time.
	 */
	private int calcFragmentTime(int fragmentSize, int nNodes) {
		int G = MIN_FRAGMENT_SIZE + ConnectionProperties.HEADER_SIZE / 8;
		int fragTimeB = (fragmentSize + ConnectionProperties.HEADER_SIZE / 8) * 
																							nNodes + G;  
		return (int)(-Math.floor(-(8.0 * (double)fragTimeB / 100.0)));
	}
	
	
	/**
	 * Finds the lowest deadline in a collection of connections.
	 *
	 * @param dstV	the vector with all connections to the specific endpoint.
	 *
	 * @return the	deadline.
	 */
	private int getLowestDeadline(Vector dstV){
		int lowestDeadline = Integer.MAX_VALUE;
		for(int i = 0; i < dstV.size(); i++) {
			ConnectionProperties cp = (ConnectionProperties)dstV.elementAt(i);
			String n = cp.getName();
			EndPoint ep = (EndPoint)endPoints.get(n);
			int deadline = ep.getDeadline();
			if(deadline < lowestDeadline)
				lowestDeadline = deadline;
		}
		return lowestDeadline;
	}
	
	
	/**
	 * Awaits changes.
	 *
	 * @return the deadline event containing the changes.
	 *
	 * @see DeadlineEvent
	 */
	public synchronized DeadlineEvent waitForChanges() {
		while(!stateChanged)
			try {
				wait();
			} catch(InterruptedException e) {
				e.printStackTrace();
			}
		stateChanged = false;
		return event;
	}
}
