package se.lth.cs.throttlenet.globethrottle;


/**
 * The <code>EndPoint</code> class describes the connection properties for an
 * end-point node.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.4, 24/10/03
 * @since   ThrottleNet0.2
 */
class EndPoint {
	
	/**
	 * The name of the connection.
	 */
	private String name;
	
	
	/**
	 * The address to the receiving node.
	 */
	private String address;
	
	
	/**
	 * The port the receiving node is listening on.
	 */
	private int port;
	
	
	/**
	 * The time between two packets in microseconds.
	 */
	private int period_us;
	
	
	/**
	 * The amount of data in a ThrottleNet packet in bytes.
	 */
	private int packet_size;
	
	
	/**
	 * The time in which the packet has to be delivered (in microseconds).
	 */
	private int deadline;
	
	
	/**
	 * Constructs properties for a node.
	 * 
	 * @param   name      	the name of the connection.
	 * @param   address     the address of the receiving node.
	 * @param   port        the port of the receiving node.
	 * @param   period_us 	the time between two packets in microseconds.
	 * @param   packet_size the amount of data in the ThrottleNet packet.
	 * @param   deadline		the time in which the packet has to be delivered in
	 *								microseconds.
	 */
	public EndPoint(String name, String address, int port,
							int period_us, int packet_size, int deadline) {
		this.name = name;
		this.address = address;
		this.port = port;
		this.period_us = period_us;
		this.packet_size = packet_size;
		this.deadline = deadline;
	}
	
	
	/**
	 * Returns the name of the connection.
    *
    * @return the name of the connection.
    */
	public String getName() {
		return name;
	}
	
	
	/**
	 * Returns the address of the receiving node.
    *
    * @return the address of the receiving node.
    */
	public String getAddress() {
		return address;
	}
	
	
	/**
	 * Returns the port of the receiving node.
    *
    * @return the port of the receiving node.
    */
	public int getPort() {
		return port;
	}
	
	
	/**
	 * Returns the period time in microseconds.
    *
    * @return the period time in microseconds.
    */
	public int getPeriodTime() {
		return period_us;
	}
	
	
	/**
	 * Returns the size of data in a ThrottleNet packet in bytes.
    *
    * @return the size of data in a ThrottleNet packet in bytes.
    */
	public int getPacketSize() {
		return packet_size;
	}
	
	
	/**
	 * Returns the time to deliver each packet.
	 *
	 * @return the time to deliver each packet.
	 */
	public int getDeadline() {
		return deadline;
	}
	
	
	/**
	 * Sets the period time for this connection in microseconds.
    *
    * @param period_us the period time for this connection in microseconds.
    */
	public void setPeriodTime(int period_us) {
		this.period_us = period_us;
	}
	
	
	/**
	 * Sets the data size for this connection in bytes.
    *
    * @param packet_size the data size for this connection in bytes.
    */
    public void setPacketSize(int packet_size) {
		this.packet_size = packet_size;
	}
	
	
	/**
	 * Sets the time to deliver each packet in microsconds.
	 *
	 * @param deadline the time to deliver each packet in microsconds.
	 */
    public void setDeadline(int deadline) {
		this.deadline = deadline;
	}
	
	
	/**
	 * Returns the string representation of this object.
    *
    * @return the string representation of this object.
    */
    public String toString() {
		return "{name=" + name + ",address=" + address + ",port=" + port + 
				 ",period_us=" + period_us + ",packet_size=" + packet_size +
				 ",deadline=" + deadline + "}";
	}
}
