package se.lth.cs.throttlenet.globethrottle;

import se.lth.cs.throttlenet.util.*;
import se.lth.cs.throttlenet.event.*;
import se.lth.cs.realtime.*;
import java.net.*;
import java.io.*;
import java.util.Hashtable;
import java.util.Vector;



/**
 * The <code>GlobeThrottle</code> class describes the center node in ThrottleNet.
 * This node receives commands from other nodes in ThrottleNet through TCP/IP.
 * For every node there is one realtime thread handling the requests.
 * 
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.7, 4/11/03
 * @since   ThrottleNet0.2
 */
public class GlobeThrottle {
	
	
	/**
	 * Constant representing the <tt>add</tt> command sent from a node.
	 */
	public static final int ADD = 1;
	
	
	/**
	 * Constant representing the <tt>lookup</tt> command sent from a node.
	 */
	public static final int LOOKUP = 2;
	
	
	/**
	 * Constant representing the <tt>remove</tt> command sent from a node.
	 */
	public static final int REMOVE = 3;
	
	
	/**
	 * Constant representing the <tt>error</tt> code sent to a node.
	 */
	public static final int ERROR = 11;
	
	
	/**
	 * Constant representing the <tt>host</tt> command sent to a node.
	 */
	public static final int HOST = 12;
	
	
	/**
	 * Constant representing the <tt>change period</tt> command sent to a node.
	 */
	public static final int PERIOD = 13;
	
	
	/**
	 * Constant representing the <tt>change packet size</tt> command sent to a
	 * node.
	 */
	public static final int PACKET_SIZE = 14;
	
	
	/**
	 * Constant representing the <tt>fragment packet size</tt> command sent to a
	 * node
	 */
	public static final int FRAGMENT_SIZE = 15;
	
	
	/**
	 * Contains all registered connections.
	 */
	private Hashtable endpoints;


	/**
	 * Contains all transmitting nodes in ThrottleNet.
	 */
	private Hashtable srcNodes;
	
	
	/**
	 * Contains all receiving nodes in ThrottleNet.
	 */
	private Hashtable dstNodes;
	
	
	/**
	 * Contains all TCP sockets listening for ThrottleNet state changes.
	 */
	private Hashtable sockets;


	/**
	 * Defines number of bytes data in each packet fragment.
	 */
	public static final int MIN_FRAGMENT_DATA = 46;
	
	
	/**
	 * Defines the maximum bandwidth in ThrottleNet (100Mbps).
	 */
	public static final long MAX_BANDWIDTH = 100000000; // 100 Mbit/s
	
	
	/**
	 * Defines the default address to the center node in ThrottleNet.
	 */
	public static String globAddress = "192.168.25.56";
	//public static String globAddress = "torvalds.cs.lth.se";
	
	/**
	 * Defines the port that the center node is listening on.
	 */
	public static final int port = 4444;
	
		
	/**
	 * The server socket that listens for new nodes.
	 */
	private ServerSocket serverSocket;
	
	
	/**
	 * The client node's connection socket to center node.
	 */
	static Socket clientSocket = null;
	
	
	/**
	 * The output stream for the client socket.
	 */
	static DataOutputStream clientOutputStream = null;
	
	
	/**
	 * The input stream for the client socket.
	 */
	static DataInputStream clientInputStream = null;

	
	/**
	 * The center node's collection of realtime threads, servicing requests from
	 * ThrottleNet nodes.
	 */
	private Vector throttleThreads;
	
	
	/**
	 * The class contains the information about how to compute ThrottleNet 
	 * settings.
	 */
	private DeadlineController dlController;
	
	
	/**
	 * The thread which detectes changes in ThrottleNet.
	 */
	private FetchEventThread fetchThread;
	
	
	/**
	 * The thread which notifies ThrottleNet nodes about changes in the
	 * ThrottleNet state.
	 */
	private PostEventThread postThread;
	
	
	/**
	 * Constructs a GlobeThrottle node, the center node in ThrottleNet.
	 */
	public GlobeThrottle() {
		endpoints = new Hashtable();
		srcNodes = new Hashtable();
		dstNodes = new Hashtable();
		sockets = new Hashtable();
		dlController = new DeadlineController(endpoints, srcNodes, dstNodes);
		fetchThread = new FetchEventThread();
		postThread = new PostEventThread();
		
		fetchThread.start();
		postThread.start();
		
		try {
			serverSocket = new ServerSocket(port);
		} catch(Exception e) {
			e.printStackTrace();
		}
		throttleThreads = new Vector();
		while(true) {
			
			try {
				Socket clientSocket = serverSocket.accept();
				ThrottleThread throttleThread = new ThrottleThread(clientSocket);
				throttleThread.start();
				addThread(throttleThread);
			} catch(IOException e) {
				e.printStackTrace();
			} 
		}
	}
	
	
	/**
	 * Transmits an array of bytes to the center node using TCP.
	 * 
	 * @param data the array of bytes to transmit.
	 *
	 * @exception  IOException if an I/O error occurs.
	 */
	public static synchronized void throttleNetTunnelTransmit(byte[] data)
																				throws IOException {
		if(clientSocket == null)
			initSocket();
		clientOutputStream.write(data, 0, data.length);
		clientOutputStream.flush();
	}
	
	
	/**
	 * Receives an array of bytes from the center node using TCP.
	 * 
	 * @return  the received array of bytes.
	 * 
	 * @exception  IOException if an I/O error occurs.
	 */
	public static synchronized byte[] throttleNetTunnelReceive() 
																				throws IOException {
		if(clientSocket == null)
			initSocket();
		String obj = null;
		byte[] data = new byte[1000];
		int n = clientInputStream.read(data, 0, data.length);
		return Utils.subarray(data, 0, n);
	}
	
	
	/**
	 * Initializes the client connection to the center node.
	 */
	private static void initSocket() {
		try {
			InetAddress inetGlobAddr = InetAddress.getByName(globAddress);
			clientSocket = new Socket(inetGlobAddr, port);
			
		
			
			OutputStream outputS = clientSocket.getOutputStream();
			BufferedOutputStream buffOpStream = new BufferedOutputStream(outputS);
			clientOutputStream = new DataOutputStream(buffOpStream);
			
			InputStream inputS = clientSocket.getInputStream();
			BufferedInputStream buffInputStream = new BufferedInputStream(inputS);
			clientInputStream = new DataInputStream(buffInputStream);
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

		
	/**
	 * The <code>ThrottleThread</code> class serves requests from other nodes in
	 * ThrottleNet.
	 * 
	 * @author  Patricia Grudziecka
	 * @author  Daniel Nyberg
	 * @version 0.7, 4/11/03
	 * @since   ThrottleNet0.2
	 */
	private class ThrottleThread extends RTThread {
		
		
		/**
		 * The socket connected to client node.
		 */
		private Socket client;
		
		
		/**
		 * Constructs a real-time thread serving the new client node's requests.
		 */
		public ThrottleThread(Socket clientSock) {
			this.client = clientSock;
		}
		
		
		/**
		 * This thread's main method. Receives requests from the node and replies.
		 *
		 * @see RTThread
		 */
		public void run() {
			DataInputStream in = null;
			DataOutputStream out = null;
			try {
				OutputStream outputS = client.getOutputStream();
				BufferedOutputStream buffOut = new BufferedOutputStream(outputS);
				out = new DataOutputStream(buffOut);
				
				InputStream inputS = client.getInputStream();
				BufferedInputStream buffIn = new BufferedInputStream(inputS);
				in = new DataInputStream(buffIn);
				
			} catch(IOException e) {
				e.printStackTrace();
			}
			
			while(true) {
				try {
					byte[] data = new byte[40];
					int n = in.read(data, 0, 40);
					if(n == -1) // End of stream 
						break;
					data = Utils.subarray(data, 0, n);
					int choice = Utils.bytesToInt(data);
					if(choice == ADD) {
						
						int port = Utils.bytesToInt(Utils.subarray(data, 4, 4));
						int period = Utils.bytesToInt(Utils.subarray(data, 8, 4));
						int size = Utils.bytesToInt(Utils.subarray(data, 12, 4));
						int deadline = Utils.bytesToInt(Utils.subarray(data, 16, 4));
						int deadlinePort = Utils.bytesToInt(
																	Utils.subarray(data,20,4));
						String name = new String(Utils.subarray(data, 24,
																		data.length-24)).trim();
						String address = client.getInetAddress().getHostName();
						SocketInfo socketInfo = new SocketInfo(address,deadlinePort);
						sockets.put(name,socketInfo);
						add(name, address, port, period, size, deadline);
					} else if(choice == REMOVE) {
						remove(new String(Utils.subarray(data, 4,
																		data.length-4)).trim());
					} else if(choice == LOOKUP) {
						int period = Utils.bytesToInt(Utils.subarray(data, 4, 4));
						int size = Utils.bytesToInt(Utils.subarray(data, 8, 4));
						int deadlinePort = Utils.bytesToInt(
																	Utils.subarray(data,12,4));
						String name = new String(Utils.subarray(data, 16,
																		data.length-16)).trim();
						String srcAddress = client.getInetAddress().getHostName();
						byte[] response = lookup(name, period, size, srcAddress,
																					deadlinePort);
						/*for(int i = 0; i < response.length; i++)
							System.out.print(" " + response[i]);
						System.out.println();*/
						
						out.write(response, 0, response.length);
						out.flush();
					} else {
						System.out.println("Damn you hackers!!!!");
					}
				} catch(IOException e) {
					System.out.println(Thread.currentThread());
					e.printStackTrace();
					return;
				}
			}
			try {
				in.close();
				out.close();
				client.close();
			} catch(IOException e) {
				e.printStackTrace();
			}
			removeThread(this);
		}
	}
	
	
	/**
	 * Adds a real-time thread to the center node's collection of threads.
	 *
	 * @param thread the real-time thread.
	 */
	private synchronized void addThread(RTThread thread) {
		throttleThreads.add(thread);
	}

	
	/**
	 * Removes a real-time thread from the center node's collection of threads.
	 *
	 * @param thread the real-time thread.
	 */
	private synchronized void removeThread(RTThread thread) {
		throttleThreads.remove(thread);
	}
	
	
	/**
	 * Adds a new connection in ThrottleNet.
	 * 
	 * @param name 		 the connection name.
	 * @param address 	 the address of the receiving node.
	 * @param port			 the port of the receiving node.
	 * @param period_us	 the time between two packets in microseconds.
	 * @param packet_size the data size in bytes in the ThrottleNet packet.
	 * @param deadline	 the time in which the packet has to be delivered in
	 *						 	 microseconds.
	 */
	private synchronized void add(String name, String address, int port,
										int period_us, int packet_size, int deadline)  {
		EndPoint ePoint = new EndPoint(name, address,
												 port, period_us, packet_size, deadline);
		endpoints.put(name, ePoint);
		System.out.println(name + " (" + ePoint + ") added");
	}
	
	
	/**
	 * Removes a specific connection from ThrottleNet.
	 * 
	 * @param name the connection name.
	 */
	private synchronized void remove(String name) {
		EndPoint ep = (EndPoint)endpoints.remove(name);
		Vector dstVector = (Vector)dstNodes.get(ep.getAddress());
		String srcHost = "";
		
		// Search the destination table
		for(int i = 0; i < dstVector.size(); i++) {
			ConnectionProperties cprop =
											(ConnectionProperties)dstVector.elementAt(i);
			if(cprop.getName().equals(name)) {
				dstVector.remove(i);
				srcHost = cprop.getSrcAddress();
				break;
			}
		}
		Vector srcVector = (Vector)srcNodes.get(srcHost);

		// Search the source table
		for(int i = 0; i < srcVector.size(); i++) {
			ConnectionProperties cprop =
											(ConnectionProperties)srcVector.elementAt(i);
			if(cprop.getName().equals(name)) {
				srcVector.remove(i);
				break;
			}
		}
		System.out.println(name + " removed");
	}
	
	
	/**
	 * Finds a connection in ThrottleNet and creates the response message.
	 * 
	 * @param name 		  the connection name.
	 * @param srcPeriod	  the requesting node's minimum time between two packets.
	 * @param srcSize		  the requesting node's data size of ThrottleNet packets.
	 * @param srcAddress   the address of the transmitting node.
	 * @param deadlinePort the TCP port number used by the requesting node.
	 *
	 * @return the response message for the requesting node.
	 * 
	 * @exception IOException if an I/O error occurs.
	 */
	private synchronized byte[] lookup(String name, int srcPeriod, int srcSize,
							String srcAddress, int deadlinePort) throws IOException {
		System.out.println("Lookup '" + name + "'");
		EndPoint ePoint = (EndPoint)endpoints.get(name);
		
		
		byte[] responseMsg = null;
		if(ePoint == null) { // If no name
			String errorMsg = "'" + name +"' does not exist";
			responseMsg = Utils.concat(Utils.intToBytes(ERROR),
																			errorMsg.getBytes());
			System.out.println("Error: " + errorMsg);
		} else {
			String dstAddress = ePoint.getAddress();
			byte[] addr = new byte[0];
			try {
				addr = InetAddress.getByName(dstAddress).getAddress();
			} catch(Exception e) {}
		
			int dstPort = ePoint.getPort();
			int dstPeriod = ePoint.getPeriodTime();
			int dstSize = ePoint.getPacketSize();
			
			responseMsg = new byte[0];
			if(srcSize != dstSize) { // If different packet sizes
				responseMsg = Utils.intToBytes(PACKET_SIZE);
				responseMsg = Utils.concat(responseMsg, Utils.intToBytes(dstSize));
			}

			// Avoid network overloading
			int newPeriod = checkPeriod(name, dstPeriod, dstSize, srcAddress,
																					dstAddress);
			if(srcPeriod < newPeriod) {
				responseMsg = Utils.concat(responseMsg, Utils.intToBytes(PERIOD));
				responseMsg = Utils.concat(responseMsg,
																Utils.intToBytes(newPeriod));
			}
			
			SocketInfo socketInfo = (SocketInfo)sockets.get(name);
			socketInfo.setAddress(srcAddress);
			socketInfo.setPort(deadlinePort);
			
			int fragmentSize = dlController.calcFragmentSize(name);
			responseMsg = Utils.concat(responseMsg,
															Utils.intToBytes(FRAGMENT_SIZE));
			responseMsg = Utils.concat(responseMsg,
															Utils.intToBytes(fragmentSize));
			
			responseMsg = Utils.concat(responseMsg, Utils.intToBytes(HOST));
			responseMsg = Utils.concat(responseMsg, Utils.intToBytes(dstPort));
			responseMsg = Utils.concat(responseMsg, addr);
		}
		return responseMsg;
	}
	
	
	/**
	 * Checks the period time and adjusts it so ThrottleNet does not
	 * overload.
	 * 
	 * @param name			the connection name.
	 * @param period_us	the receiving node's minimum time between two packets.
	 * @param size			the recieving node's data size of ThrottleNet packets.
	 * @param srcAddress	the address of the transmitting node.
	 * @param dstAddress	the address of the receiving node.
	 * 
	 * @return the lowest time between two packets to avoid overloading in
	 *			  ThrottleNet.
	 */
	private synchronized int checkPeriod(String name, int period_us, int size,
													String srcAddress, String dstAddress) {
		
//		System.out.println("dstAddress = " + dstAddress + "   srcAddress = " + srcAddress);
		// Check if the receiving node's network isn't overloded
		long dstBandWidth = 0;
		Vector dstVector = (Vector)dstNodes.get(dstAddress);
		if(dstVector != null) {
			for(int i = 0; i < dstVector.size(); i++) {
				ConnectionProperties cprop =
											(ConnectionProperties)dstVector.elementAt(i);
				dstBandWidth += cprop.getBandWidth();
			}
		} else
			dstVector = new Vector();
		
		period_us = calcLowestPeriod(dstBandWidth, size, period_us);

		// Check if the transmitting node's network isn't overloded
		long srcBandWidth = 0;
		Vector srcVector = (Vector)srcNodes.get(srcAddress);
		if(srcVector != null) {
			for(int i = 0; i < dstVector.size(); i++) {
				ConnectionProperties cprop =
											(ConnectionProperties)srcVector.elementAt(i);
				srcBandWidth += cprop.getBandWidth();
			}
		} else
			srcVector = new Vector();
			
		period_us = calcLowestPeriod(srcBandWidth, size, period_us);
		
		// Update the tables
		ConnectionProperties newCP = new ConnectionProperties(
										srcAddress, dstAddress, name, period_us, size);
		dstVector.add(newCP);
		srcVector.add(newCP);
		dstNodes.put(dstAddress, dstVector);
		srcNodes.put(srcAddress, srcVector);
		return period_us;
	}
	
	
	/**
	 * Calculates the lowest alowed period time to avoid overloading.
	 * 
	 * @param bandWidth	the occupied bandwidth.
	 * @param size			the packet data size.
	 * @param period_us	the time between two packets.
	 *
	 * @return the lowest alowed period time to avoid overloading.
	 */
	private synchronized int calcLowestPeriod(long bandWidth, int size,int period_us) {
		long freeBandWidth = MAX_BANDWIDTH - bandWidth;
		int nPackets = ConnectionProperties.nbrPacket(size);
		long sum = ConnectionProperties.getBandWidth(size, period_us);
		// The relation between the requested and available bandwidth
		double k = (double)sum / (double)freeBandWidth;
		int newPeriodTime = period_us;
		if(k > 1.0) { // k > 1.0 => overloading
			newPeriodTime = (int)((double)period_us * k);
			// Round up
			if((double)period_us * k - (double)newPeriodTime > 0.0)
				newPeriodTime++;
		}
		return newPeriodTime;
	}
	

   /**
	 * Gets info from deadline controller.
	 */
	private class FetchEventThread extends RTThread{

		public void run(){
			while(true){
				DeadlineEvent event = dlController.waitForChanges();
				postThread.putEvent(event);
	 		}
   	}
	}
	
	
	/**
	 * Sends changes to the nodes.
	 */
	private class PostEventThread extends RTThread{

		public void run(){
			DeadlineEvent event = null;
			DataOutputStream out = null;
			Socket socket = null;
	    	try{
				while(true){
					event = (DeadlineEvent)mailbox.doFetch();
					Vector changes = event.getChanges();
					for(int i=0;i<changes.size();i++){
						NewConnectionSettings ncs =
											(NewConnectionSettings)changes.elementAt(i);
						
						String connName = ncs.getName();
						int fragmentSize = ncs.getFragmentSize();
						int fragmentTime = ncs.getFragmentTime();

						byte[] fs = Utils.intToBytes(fragmentSize);
						byte[] ft = Utils.intToBytes(fragmentTime);
						byte[] data = Utils.concat(fs,ft);

						// Fetch SocketInfo from the table sockets
						SocketInfo socketInfo = (SocketInfo)sockets.get(connName);
						
//						System.out.println(connName);
						
						String dstAddress = socketInfo.getDstAddress();
						String srcAddress = socketInfo.getSrcAddress();
						int dstPort = socketInfo.getDstPort();
						int srcPort = socketInfo.getSrcPort();
						
						OutputStream outStream = null;
						BufferedOutputStream buffOut = null;
						
						// Send message to the receiving node
						socket = new Socket(InetAddress.getByName(dstAddress),dstPort);
						outStream = socket.getOutputStream();
						buffOut = new BufferedOutputStream(outStream);
						out = new DataOutputStream(buffOut);
						out.write(data,0,data.length);
						out.flush();
						out.close();

						// Send message to the sending node
						socket = new Socket(InetAddress.getByName(srcAddress),srcPort);
						outStream = socket.getOutputStream();
						buffOut = new BufferedOutputStream(outStream);
						out = new DataOutputStream(buffOut);
						out.write(data,0,data.length);
						out.flush();
						out.close();
						socket.close();
					}
				}
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Sends changes to the nodes.
	 */
	public static void main(String[] atgs) {
		GlobeThrottle world = new GlobeThrottle();
   }	
}
