#include <ljrt_threading.h>
#include <class_indices.h>
#include <java_lang_Thread.h>
#include <java_lang_Runnable.h>

#ifdef JAVA_LANG_THROWABLE_INDEX
#include <java_lang_Throwable.h>
#endif

#include <java_lang_String.h>
#include <iface.h>
#include <exception.h>
#include <gci_common.h>
#include <pthread.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

// Only(?) needed for cygwin. Handle in a better way? AndersN
#include <sys/time.h>


/*
 * Design considerations:
 * 
 *   Since handling of sleep/wait/interrupt requires a lot of state 
 *   to be consistent between the the object lock state, and the 
 *   thread state, hence:
 *     1. all lock states are protected with the global_mutex
 *     2. all cond_wait is done within the global_mutex, hence
 *        all neccessary object locks releases are protected by 
 *        the global_mutex.
 */


//FIXME: These should be declared in the respective .h files
extern GC_REF(LayoutHead,java_lang_String_Layout);
extern java_lang_String_Class;
extern GC_REF(LayoutHead,java_lang_Thread_Layout);
extern java_lang_Thread_Class;
#ifdef JAVA_LANG_THROWABLE_INDEX
extern java_lang_Throwable_Class;
#endif

#ifdef JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX
extern java_lang_InterruptedException_Class;
extern GC_REF(LayoutHead,java_lang_InterruptedException_Layout);
#endif

extern JAVAMAIN();




typedef struct ljrt_thread_t {
  struct ljrt_monitor_t *monitor;
  enum { thread_state_start, thread_state_run, thread_state_wait, 
	 thread_state_sleep, thread_state_terminate } state;
  pthread_t pthread;
  pthread_cond_t cond;
  GC_REF(java_lang_Thread,javathread);
} ljrt_thread_t;

typedef struct ljrt_monitor_t {
  int count;
  int seq;
  int waiters;
  struct ljrt_thread_t *owner;
  pthread_mutex_t mutex;
  pthread_cond_t cond;
} ljrt_monitor_t;

/* Function translates Java priority in an equivalent scheduler's priority
   specified by the parameter sched */

int javaPr2schedPr(int java_prio,int sched) {
	const int pr_step_java = 10;
	int pr_step_posix;
	pr_step_posix = sched_get_priority_max(sched) - 
	  sched_get_priority_min(sched);
	return (((10-(java_prio-1)) * (abs(pr_step_posix)))/(pr_step_java))-
	  (abs(sched_get_priority_min(sched)));
}

int schedPr2javaPr(int sched_prio,int sched) {
	const int pr_step_java = 10;
	int pr_step_posix;
	pr_step_posix = sched_get_priority_max(sched) - 
	  sched_get_priority_min(sched);
	return (int)((float)11- 
		     ((float)(sched_prio+sched_get_priority_min(sched))*
		      ((float)(abs(pr_step_java))/(pr_step_posix))));
}


// FIXME: these should be forced to always be generated
#ifndef JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX
#define JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX -1
#endif

#ifndef JAVA_LANG_ILLEGALARGUMENTEXCEPTION_INDEX
#define JAVA_LANG_ILLEGALARGUMENTEXCEPTION_INDEX -1
#endif


/* static pthread_mutex_t global_mutex = PTHREAD_MUTEX_INITIALIZER; */
/* static pthread_cond_t global_cond = PTHREAD_COND_INITIALIZER; */
/* static pthread_key_t ljrt_tsd_key; */
/* static int thread_count; */
 pthread_mutex_t global_mutex = PTHREAD_MUTEX_INITIALIZER;
 pthread_cond_t global_cond = PTHREAD_COND_INITIALIZER;
 pthread_key_t ljrt_tsd_key;
 int thread_count;

void inc_threads() {
  pthread_mutex_lock(&global_mutex);
  thread_count++;
  pthread_mutex_unlock(&global_mutex);
}

/* Function decreasing the number of C threads alive */
void dec_threads() {	
  pthread_mutex_lock(&global_mutex);
  thread_count--;
  if (thread_count == 0) {
    pthread_cond_broadcast(&global_cond);  
  }
  pthread_mutex_unlock(&global_mutex);
}

/* 
 * Function in order to simulate Java main thread behavior, it 
 * waits until every non-daemon thread has finished its work 
 */
 
//FIXME: implement daemon threads
void wait_for_all_threads_to_die() {
  pthread_mutex_lock(&global_mutex);
  while (thread_count) {
    pthread_cond_wait(&global_cond, &global_mutex);
  }
  pthread_mutex_unlock(&global_mutex);  
}

/*
 * Monitor support routines
 */

GC_PROC_BEGIN(ljrt_monitor_init,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    GC_SET(this, lock, 0); // Lazy allocation
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(ljrt_monitor_init)

/*FIXME: modify GC so this works (symmetry)
GC_PROC_BEGIN(ljrt_monitor_free,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_monitor_t *lock;

//    printf("%s\n", __FUNCTION__);
    GC_GET(lock, this, lock);
    if (lock) {
      GC_SET(this, lock, 0);
      pthread_cond_destroy(&lock->cond);
      pthread_mutex_destroy(&lock->mutex);
      free(lock);
//      printf("Freed! %s\n", __FUNCTION__);
    }
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(ljrt_monitor_free)
*/

void ljrt_monitor_free(java_lang_Object *this)
{
  ljrt_monitor_t *lock;
  lock = this->lock;
  if (lock) {
    //FIXME: seems to be some race condition with the GC
    //       at termination time (i.e printout not always present)
    //       Possible cause: gc_thread stopped after object is
    //       removed from heap, but before finalizers are run
//    printf("Freed! %s %p\n", __FUNCTION__, lock);
    this->lock = 0;
    pthread_cond_destroy(&lock->cond);
    pthread_mutex_destroy(&lock->mutex);
    free(lock);
  }
}

GC_PROC_BEGIN(ljrt_monitor_enter,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_monitor_t *lock;
    ljrt_thread_t *currentThread;

//    printf("%s\n", __FUNCTION__);
    currentThread = (ljrt_thread_t*)pthread_getspecific(ljrt_tsd_key);
    GC_GET(lock, this, lock);
    if (!lock) {
      lock = malloc(sizeof(ljrt_monitor_t));
      lock->count = 0;
      lock->seq = 0;
      lock->waiters = 0;
      pthread_mutex_init(&lock->mutex, 0);
      pthread_cond_init(&lock->cond, 0);
      GC_SET(this, lock, lock);
    }
    if (!lock) { 
      printf("No monitor allocated\n"); 
    } else {

      pthread_mutex_lock(&global_mutex);
/* 	  printf("Enter: %x %x %d\n",currentThread,lock->owner,lock->count); */
      do {
	if (lock->count == 0) {
	  // Acquire free lock 
	  lock->count = 1;
	  lock->owner = currentThread;
	  pthread_mutex_lock(&lock->mutex);
	} else if (lock->owner == currentThread) {
	  // Recursion on already owned lock
	  lock->count++;
	} else {
	  // Wait for lock to become free
	  lock->waiters++;
	  pthread_cond_wait(&lock->cond, &global_mutex);
	  lock->waiters--;
	}
      } while (lock->owner != currentThread);
      pthread_mutex_unlock(&global_mutex);
    }
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(ljrt_monitor_enter)

GC_PROC_BEGIN(ljrt_monitor_leave,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_monitor_t *lock;
    ljrt_thread_t *thread;

//    printf("%s\n", __FUNCTION__);
    thread = (ljrt_thread_t*)pthread_getspecific(ljrt_tsd_key);
/* 	printf("Leave: %x %x %d\n",thread,lock->owner,lock->count); */
    GC_GET(lock, this, lock);
    pthread_mutex_lock(&global_mutex);
    assert(lock->owner == thread);
    assert(lock->count > 0);
    lock->count--;
    if (lock->count == 0) {
      // Unlock the nown unoccupied monitor
      pthread_mutex_unlock(&lock->mutex);
      if (lock->waiters) {
	// Free at least one waiting task
	pthread_cond_signal(&lock->cond);
      }
    }
    pthread_mutex_unlock(&global_mutex);
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(ljrt_monitor_leave)


GC_PROC_BEGIN(ljrt_monitor_leave_NOLOCK,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_monitor_t *lock;
    ljrt_thread_t *thread;

//    printf("%s\n", __FUNCTION__);
    thread = (ljrt_thread_t*)pthread_getspecific(ljrt_tsd_key);
/* 	printf("Leave: %x %x %d\n",thread,lock->owner,lock->count); */
    GC_GET(lock, this, lock);
    pthread_mutex_lock(&global_mutex);
    assert(lock->owner == thread);
    assert(lock->count > 0);
    lock->count--;
    if (lock->count == 0) {
      // Unlock the nown unoccupied monitor
      pthread_mutex_unlock(&lock->mutex);
      if (lock->waiters) {
	// Free at least one waiting task
	pthread_cond_signal(&lock->cond);
      }
    }
    pthread_mutex_unlock(&global_mutex);
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(ljrt_monitor_leave)

/*
 * java.lang.Object wait/notify methods
 */

int ljrt_wait_until(void **__eFlag__,ljrt_monitor_t *m, struct timespec *until) 
{
  int result, count, seq;
  ljrt_thread_t *currentThread;
  JBoolean interrupted;

  pthread_mutex_lock(&global_mutex);

  // Save monitor state
  count = m->count;
  seq = m->seq;
  currentThread = m->owner;
  m->count = 0;

  // Unlock monitor
  pthread_mutex_unlock(&m->mutex);
  if (m->waiters) {
    // Free at least one waiting task
    pthread_cond_signal(&m->cond);
  }
  // Wait for interrupt, timeout or notify
  while (1) {
    GC_GET(interrupted,currentThread->javathread,
	   _java_lang_Object_interruptedFlag);
    if (interrupted) { break; } // interrupt
    currentThread->state = thread_state_wait;
    currentThread->monitor = m;
    result = pthread_cond_timedwait(&m->cond, &global_mutex, until);
    currentThread->monitor = 0;
    currentThread->state = thread_state_run;
    if (result == ETIMEDOUT) { break; } // timeout
    if (m->seq != seq) { break; } // notify
  }
  if (interrupted) {
    GC_SET(currentThread->javathread,_java_lang_Object_interruptedFlag, 0);
    result = EINTR;
  }
  // Relock monitor
  do {
    if (m->count == 0) {
      // Reacquire free lock 
      m->count = count;
      m->owner = currentThread;
      pthread_mutex_lock(&m->mutex);
    } else {
      // Wait for lock to become free
      m->waiters++;
      pthread_cond_wait(&m->cond, &global_mutex);
      m->waiters--;
    }
  } while (m->owner != currentThread);  

 catch_0: //Is this correct? AndersN
  pthread_mutex_unlock(&global_mutex);  
  return result;
}  

GC_PROC_BEGIN(java_lang_Object_wait_long_int,
	      GC_PARAM(java_lang_Object,this),
	      JLong millis,JInt nanos) EXCEPTION_THROWS
{
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    int result;
    struct timeval now;
    struct timespec until;
    ljrt_monitor_t *monitor;
    
    if (millis == 0 && nanos == 0) {
      until.tv_sec = LONG_MAX;
      until.tv_nsec = 999999999;
    } else {
      gettimeofday(&now,0);
      until.tv_nsec = now.tv_usec * 1000 + ((millis%1000) * 1000000) + nanos;
      until.tv_sec  = now.tv_sec + (millis/1000) + until.tv_nsec / 1000000000;
      until.tv_nsec %= 1000000;
    }
    GC_GET(monitor,this,lock);
    result = ljrt_wait_until(__eFlag__,monitor, &until);
    if (result == EINTR) {
      EXCEPTION_THROW(0,JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX);
    }
  }  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
}
GC_PROC_END(java_lang_Object_wait_long_int)

GC_PROC_BEGIN(java_lang_Object_notify,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_monitor_t *lock;

    GC_GET(lock,this,lock);
    pthread_mutex_lock(&global_mutex);
    lock->seq++;
    pthread_cond_signal(&lock->cond);
    pthread_mutex_unlock(&global_mutex);  
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Object_notify)

GC_PROC_BEGIN(java_lang_Object_notifyAll,GC_PARAM(java_lang_Object,this))
  GC_PARAM_REF(java_lang_Object,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_monitor_t *lock;

    GC_GET(lock,this,lock);
    pthread_mutex_lock(&global_mutex);
    lock->seq++;
    pthread_cond_broadcast(&lock->cond);
    pthread_mutex_unlock(&global_mutex);  
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Object_notifyAll)

/*
 * java.lang.Thread methods
 */

int ljrt_sleep_until(void **__eFlag__,ljrt_thread_t *t, struct timespec *until) 
{
  int result;
  JBoolean interrupted;

  pthread_mutex_lock(&global_mutex);
  // Wait for interrupt or timeout
  while (1) {
    GC_GET(interrupted,t->javathread,_java_lang_Object_interruptedFlag);
    if (interrupted) { break; } // interrupt
    t->state = thread_state_sleep;
    result = pthread_cond_timedwait(&t->cond, &global_mutex, until);
    t->state = thread_state_run;
    if (result == ETIMEDOUT) { break; } // timeout
  }
  if (interrupted) { 
    GC_SET(t->javathread,_java_lang_Object_interruptedFlag, false);
    result = EINTR;
  }
 catch_0: //Is this correct? AndersN
  pthread_mutex_unlock(&global_mutex);  
  return result;
}
void ljrt_interrupt(void **__eFlag__,ljrt_thread_t *t) 
{
  pthread_mutex_lock(&global_mutex);
  GC_SET(t->javathread,_java_lang_Object_interruptedFlag, true);
  switch (t->state) {
    case thread_state_wait: {
      pthread_cond_broadcast(&t->monitor->cond);
    } break;
    case thread_state_sleep: {
      pthread_cond_signal(&t->cond);
    } break;
  }
 catch_0: //Is this correct? AndersN
  pthread_mutex_unlock(&global_mutex);  
}

/* Native implementation of java.lang.Thread.sleep(long millis,int nanos) */

GC_PROC_BEGIN(java_lang_Thread_sleep_long_int,JLong millis,int nanos)
  GC_FUNC_ENTER {
    int result;
    ljrt_thread_t *currentThread;
    struct timeval now;
    struct timespec until;
    ljrt_monitor_t *monitor;
    
    gettimeofday(&now,0);
    until.tv_nsec = now.tv_usec * 1000 + ((millis%1000) * 1000000) + nanos;
    until.tv_sec  = now.tv_sec + (millis/1000) + until.tv_nsec / 1000000000;
    until.tv_nsec %= 1000000000;
    currentThread = (ljrt_thread_t*)pthread_getspecific(ljrt_tsd_key);
    result = ljrt_sleep_until(__eFlag__,currentThread, &until);
/*     printf("%d %d %d\n",millis,now.tv_usec*1000,until.tv_sec); */
    if (result == EINTR) {
#ifdef JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX
	  GC_REF(java_lang_InterruptedException,_tmp_1);
	  GC_PUSH_ROOT(_tmp_1);
	  GC_REF_FUNC_CALL(_tmp_1,
					   java_lang_InterruptedException_InterruptedException,
					   &java_lang_InterruptedException_Class,
					   GC_PASS(java_lang_InterruptedException_Layout));CHECK(0);
	  EXCEPTION_STORE_OBJ(_tmp_1);
	  EXCEPTION_THROW(0,&java_lang_InterruptedException_Class)
		GC_POP_ROOT(_tmp_1);
#endif
    }
  } GC_FUNC_LEAVE
GC_PROC_END(java_lang_Thread_sleep_long_int)


/* Native implementation of java.lang.Thread.sleep(long millis) */

GC_PROC_BEGIN(java_lang_Thread_sleep_long, JLong millis) EXCEPTION_THROWS;
  GC_FUNC_ENTER
  GC_PROC_CALL(java_lang_Thread_sleep_long_int, millis, 0);	 
  GC_FUNC_LEAVE
GC_PROC_END(java_lang_Thread_sleep_long)

/* Native implementation of java.lang.Thread.interrupt() */
GC_PROC_BEGIN(java_lang_Thread_interrupt,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    ljrt_thread_t *thread;

    GC_GET(thread, this, _java_lang_Thread_target);
    ljrt_interrupt(__eFlag__,thread);
  }   
  GC_FUNC_LEAVE;  
  GC_POP_PARAM(this);	
GC_PROC_END(java_lang_Thread_interrupt)




GC_VAR_FUNC_BEGIN(JLong,java_lang_Thread_createThread,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER

  // We postpone pthread creation until start is called 
  GC_RETURN_VAR(0);

  GC_FUNC_LEAVE
  GC_POP_PARAM(this);     
GC_VAR_FUNC_END(JLong,java_lang_Thread_createThread)

/* Method to call when a new thread is created, initialises the TSD and call run method */

/* Native implementation of java.lang.Thread.setPriority() */

GC_PROC_BEGIN(java_lang_Thread_setPriority_int,GC_PARAM(java_lang_Thread,this),JInt priority)  EXCEPTION_THROWS
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER
  struct sched_param param;
  JInt policy;
  JLong tmp;
  JInt status;
  if (priority > 10 || priority < 1) {
    EXCEPTION_THROW(0,JAVA_LANG_ILLEGALARGUMENTEXCEPTION_INDEX)
  }
  else {
    GC_SET(this,_java_lang_Thread_priority,priority);
#if 0
    //FIXME: needs to be synchronized
    if(__thread_is_started){
	 GC_GET(tmp,this,_java_lang_Thread_target);
	 param.sched_priority = javaPr2schedPr(priority,_RT_SCHEDULING_POSIX);
	 if ((status = pthread_setschedparam_rt(tmp,_RT_SCHEDULING_POSIX,&param)) != 0 && status != ESRCH) {
	 	printf("pthread_setschedparam() %d[%d,%d]\n",param.sched_priority,sched_get_priority_min(_RT_SCHEDULING_POSIX),sched_get_priority_max(_RT_SCHEDULING_POSIX));
	}
    } else {
//      printf("setPriority: delaying priority %d\n", priority);
    }
#endif
  }
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_setPriority_int)

 


/* Transfer the priority of the currently running thread to this thread */
GC_PROC_BEGIN(java_lang_Thread_inheritPriority,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    struct sched_param param;
    JInt result;
    JInt policy;
    if (pthread_getschedparam(pthread_self(), &policy, &param)) {
      printf("Error pthread_getschedparam()\n");
    }
    /* Posix priority to Java priority */
    result = schedPr2javaPr(param.sched_priority,_RT_SCHEDULING_POSIX);
    GC_SET(this,_java_lang_Thread_priority,result);
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_inheritPriority)

typedef struct rundata_t {
  GC_REF(java_lang_Thread, jthread);
  GC_REF(java_lang_Object, arg);
  int state;
  void (*method)();
  pthread_t pthread
} rundata_t; 

void *ljrt_thread_start(void *arg)
{
  /* GC_THREAD_INIT declares gc_thread_t gc_thread */
  GC_THREAD_INIT {
    GC__GET_THREAD_ROOT; // This declares thread_root
    rundata_t *rundata = arg;
    int status;
    int priority;
    ljrt_thread_t ljrt_thread;
    void (*method)();
    void *eFlag = 0;
    void **__eFlag__ = &eFlag;
    GC_REF(java_lang_Object,arg);

    GC_PUSH_ROOT(ljrt_thread.javathread);   
    GC_PUSH_ROOT(arg);   
    
//    printf("Running\n");
    ljrt_thread.monitor = 0;
    ljrt_thread.state = thread_state_run;
    ljrt_thread.pthread = rundata->pthread;
    pthread_cond_init(&ljrt_thread.cond, 0);
    GC_ASSIGN(ljrt_thread.javathread, rundata->jthread);
    if ((status = pthread_setspecific(ljrt_tsd_key, &ljrt_thread)) != 0) {
      strerror(status);    	
    }
    method = rundata->method;
    GC_ASSIGN(arg, rundata->arg);
    GC_SET(ljrt_thread.javathread,_java_lang_Thread_target, &ljrt_thread);
    GC_SET(ljrt_thread.javathread,_java_lang_Thread_isAlive, true);
    inc_threads();

    // Now we have everything safely rooted and copied, and are running, so
    // let our creator continue
    pthread_mutex_lock(&global_mutex);
    rundata->state = 1;
    pthread_cond_broadcast(&global_cond);
    pthread_mutex_unlock(&global_mutex);
    
    // Run
#ifdef JAVA_LANG_THROWABLE
    EXCEPTION_TRY
      GC_GET(priority, ljrt_thread.javathread, _java_lang_Thread_priority);
      GC_PROC_CALL(java_lang_Thread_setPriority_int, 
		   GC_PASS(ljrt_thread.javathread), priority); 
      CHECK(1);
      GC_PROC_CALL(method,GC_PASS(arg));   	   
      CHECK(1);
    EXCEPTION_CATCH(1,&java_lang_Throwable_Class)
      // Check for uncaught exceptions
      GC_REF(java_lang_Throwable, t);
      void *class;
      GC_PUSH_ALIAS(t);
      EXCEPTION_GET_OBJ(t);
      GC_GET(class,t,class);
/*       printf("*__eFlag__: %x\n",*__eFlag__); */
      UNCAUGHT_EXCEPTION(class)
    EXCEPTION_AFTER_CATCH(1)
#else
      GC_GET(priority, ljrt_thread.javathread, _java_lang_Thread_priority);
      GC_PROC_CALL(java_lang_Thread_setPriority_int, 
		   GC_PASS(ljrt_thread.javathread), priority); 
      GC_PROC_CALL(method,GC_PASS(arg));   	   
#endif
      goto cont;
 catch_0: //Is this correct? AndersN
      printf("Thread caught exception on top level\n");
  cont:    // And terminate
    GC_SET(ljrt_thread.javathread,_java_lang_Thread_isAlive, false);
    dec_threads();

    GC_POP_ROOT(arg);
    GC_POP_ROOT(ljrt_thread.javathread); 
  } GC_THREAD_EXIT
}

GC_PROC_BEGIN(java_lang_Thread_start,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER {
    rundata_t rundata;
    int i;
    GC_REF(java_lang_Object,arg);
    GC_REF(java_lang_Runnable,run);

    GC_PUSH_ROOT(rundata.jthread);
    GC_PUSH_ROOT(rundata.arg);
    GC_PUSH_ROOT(arg);
    GC_PUSH_ROOT(run);

    // Pass on java thread 
    GC_ASSIGN(rundata.jthread, this);
    
    // Start the thread (see also java_lang_Thread_start())
    rundata.state = 0;
    GC_GET_REF(run,this,_java_lang_Thread_targetRun);
    if (!(GC_IS_NULL(run))) {      
      struct java_lang_Runnable_ClassStruct *run_class;
      GC_GET(run_class,run,class);
      rundata.method = run_class->ifazTbl->java_lang_Runnable_methodTbl->run;
      GC_ASSIGN(rundata.arg, run);
    } else {
      struct java_lang_Thread_ClassStruct *thread_class;  
      GC_GET(thread_class,this,class);
      rundata.method = thread_class->methodTbl->run;
      GC_ASSIGN(rundata.arg, this);
    }

    if (pthread_create(&rundata.pthread, 0, ljrt_thread_start, &rundata)) {
      printf("Error pthread_create()\n");	  
    }
    pthread_mutex_lock(&global_mutex);
    while (rundata.state == 0) {
      pthread_cond_wait(&global_cond, &global_mutex);
    }
    pthread_mutex_unlock(&global_mutex);   

    GC_POP_ROOT(run);
    GC_POP_ROOT(arg);
    GC_POP_ROOT(rundata.arg);
    GC_POP_ROOT(rundata.jthread);
  } GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_start)

/* Native implementation of java.lang.Thread.currentThread() */

GC_REF_FUNC_BEGIN(java_lang_Thread,java_lang_Thread_currentThread)
  GC_FUNC_ENTER
  GC_REF(java_lang_Thread,tmp);
  GC_PUSH_ROOT(tmp);
  {	
  	ljrt_thread_t *obj;
	obj = (ljrt_thread_t*) pthread_getspecific(ljrt_tsd_key);
	GC_ASSIGN(tmp,obj->javathread);
    GC_RETURN_REF(tmp);
  }  
  GC_POP_ROOT(tmp);
  GC_FUNC_LEAVE;
GC_REF_FUNC_END(java_lang_Thread,java_lang_Thread_currentThread)

/* Native implementation of java.lang.Thread.yield() */

GC_PROC_BEGIN(java_lang_Thread_yield)
  GC_FUNC_ENTER
  {	
    printf("java.lang.Thread.yield() has no implementation!\n");
  }  
  GC_FUNC_LEAVE;
GC_PROC_END(java_lang_Thread_yield)

/* Native implementation of java.lang.Thread.join() */

GC_PROC_BEGIN(java_lang_Thread_join,GC_PARAM(java_lang_Thread,this))
  GC_FUNC_ENTER
  {	
    printf("java.lang.Thread.join() has no implementation!\n");
  }  
  GC_FUNC_LEAVE;
GC_PROC_END(java_lang_Thread_join)

/* Native implementation of java.lang.Thread.join(long millis) */

GC_PROC_BEGIN(java_lang_Thread_join_long,GC_PARAM(java_lang_Thread,this),JLong millis)
  GC_FUNC_ENTER
  {	
    printf("java.lang.Thread.join(long millis) has no implementation!\n");
  }  
  GC_FUNC_LEAVE;
GC_PROC_END(java_lang_Thread_join_long)

/* Native implementation of java.lang.Thread.join(long millis, int nanos) */

GC_PROC_BEGIN(java_lang_Thread_join_long_int,GC_PARAM(java_lang_Thread,this),JLong millis, int nanos)
  GC_FUNC_ENTER
  {	
    printf("java.lang.Thread.join(long millis, int nanos) has no implementation!\n");
  }  
  GC_FUNC_LEAVE;
GC_PROC_END(java_lang_Thread_join_long_int)

GC_PROC_BEGIN(create_main_thread, int argc, char *argv[])
  GC_FUNC_ENTER {
    rundata_t rundata;
    int i;
    GC_REF(RefArray,arg);
    GC_REF(java_lang_String,s);

    GC_PUSH_ROOT(rundata.jthread);
    GC_PUSH_ROOT(rundata.arg);
    GC_PUSH_ROOT(arg);
    GC_PUSH_ROOT(s);
    // Create a new java thread
    GC_REF_FUNC_CALL(rundata.jthread, java_lang_Thread_Thread,
		     &java_lang_Thread_Class,
		     GC_PASS(java_lang_Thread_Layout)); 
    
    // Create the argument vector
    GC_REF_FUNC_CALL(arg,newRefArray,argc-1);
    for (i = 1 ; i < argc ; i++) {
      GC_REF_FUNC_CALL(s,java_lang_String_String_CCharArray,
		       &java_lang_String_Class,
		       GC_PASS(java_lang_String_Layout),
		       argv[i]);
      GC_SET_REF_ARRAY(arg,i-1,s);
    }

    // Start the thread (see also java_lang_Thread_start())
    rundata.state = 0;
    rundata.method = JAVAMAIN;
    GC_ASSIGN(rundata.arg, arg);
    if (pthread_create(&rundata.pthread, 0, ljrt_thread_start, &rundata)) {
      printf("Error pthread_create()\n");	  
    }
    pthread_mutex_lock(&global_mutex);
    while (rundata.state == 0) {
      pthread_cond_wait(&global_cond, &global_mutex);
    }
    pthread_mutex_unlock(&global_mutex);   

    GC_POP_ROOT(s);
    GC_POP_ROOT(arg);
    GC_POP_ROOT(rundata.arg);
    GC_POP_ROOT(rundata.jthread);
  } GC_FUNC_LEAVE;
GC_PROC_END(create_main_thread)

#ifdef VXWORKS
int ljrt_start(int argc, char argv[]) {
#else
int main(int argc, char argv[]) {
#endif
  GC_INIT(GC_HEAPSIZE);
  {
    void *eFlag = 0;
    void **__eFlag__ = &eFlag;
    thread_count = 0;
    if (pthread_key_create(&ljrt_tsd_key,NULL) != 0) {
      printf("Error pthread_key_create(&ljrt_tsd_key)\n");
    }
    javaClasses_Init(__eFlag__);
/*     Fields_Init(); */
    GC_PROC_CALL(create_main_thread, argc, argv);
    wait_for_all_threads_to_die(); 
  sleep(1);
  }
  GC_EXIT;
  printf("DONE\n");
  return 0;
}


#if 00
void thread_init(thread_t *t) 
{
  t->interrupted = 0;
  t->monitor = 0;
  t->state = thread_state_run;
  pthread_cond_init(&t->cond, 0);
}

typedef struct run_t {
  thread_t *thread;
  void (*run)(void*);
} run_t;

static void *runner(void *arg)
{
  run_t *r = (run_t*)arg;
  thread_t *thread = r->thread;
  void (*run)(void*) = r->run;
 
  pthread_mutex_lock(&global_mutex);
  thread->state = thread_state_run;
  pthread_cond_broadcast(&thread->cond);
  pthread_mutex_unlock(&global_mutex);  
  run(thread);
  pthread_mutex_lock(&global_mutex);
  thread->state = thread_state_terminate;
  pthread_cond_broadcast(&thread->cond);
  pthread_mutex_unlock(&global_mutex);  
}

void thread_start(thread_t *t, void (*run)(thread_t*))
{
  run_t r;
  
  r.thread = t;
  r.run = run;
  pthread_mutex_lock(&global_mutex);
  t->state = thread_state_start;
  pthread_create(&t->thread, 0, &runner, &r);
  while (t->state == thread_state_start) {
    pthread_cond_wait(&t->cond, &global_mutex);
  }
  pthread_mutex_unlock(&global_mutex);  
}

int monitor_wait(monitor_t *m) 
{
  static struct timespec forever = {
    .tv_sec = LONG_MAX,
    .tv_nsec = 999999999
  };
  return monitor_wait_until(m, &forever);
}
  
void monitor_notify(monitor_t *m) 
{
  pthread_mutex_lock(&global_mutex);
  m->seq++;
  pthread_cond_signal(&m->cond);
  pthread_mutex_unlock(&global_mutex);  
}

void monitor_notify_all(monitor_t *m) 
{
  pthread_mutex_lock(&global_mutex);
  m->seq++;
  pthread_cond_broadcast(&m->cond);
  pthread_mutex_unlock(&global_mutex);  
}



#if 0
void monitor_init(ObjectHead *this){
	if(this != 0 && this->lock == 0) {
		pthread_mutexattr_t attr;
		//printf("monitor init: %x\n", this);
		this->lock = malloc(sizeof(pthread_mutex_t));
		//printf("mallock lock...\n");
		//this->cond = malloc(sizeof(pthread_cond_t));
		//printf("mallock cond...\n");
		pthread_mutexattr_init_rt(&attr);
		pthread_mutexattr_settype_rt(&attr,PTHREAD_MUTEX_RECURSIVE);
		pthread_mutex_init_rt(this->lock,&attr);
		//printf("mutex init...\n");
		//pthread_mutex_init_rt((GC___PTR(((GC___REF(ObjectHead))this)))->lock,&attr);
		//pthread_cond_init_rt(this->cond,NULL);
		//printf("cond init...\n");
		pthread_mutexattr_destroy_rt(&attr);
	}

}

// void monitor_enter_static(Java_lang_Object *this){
//
// 	pthread_mutex_lock_rt(this.class_lock);
// }

void monitor_enter(ObjectHead *this){
//	if(isStatic){
//		pthread_mutex_lock_rt(((struct java_lang_Object_ClassStruct*)this)->class_lock);
//
//	}else;
  printf("monitor enter [%p]...\n", this->lock);
	pthread_mutex_lock_rt(this->lock);
}

void monitor_leave(ObjectHead *this){
	//printf("monitor leave...\n");
	pthread_mutex_unlock_rt(this->lock);
}

#endif
#endif

