/**
 * The T2 AVR Kernel is a fork of the TOBBE Kernel.
 * The TOBBE kernel was changed to simplify debugging.
 * Unit tests have been added for the priority and time queue
 * data structures.
 *
 * Several bugs were identified. These are some of the bugs:
 * - after rescheduling interrupts would be enabled in critical sections
 * - JAVAMAIN was used as a thread entry point
 * - gc_thread_exit was not protected with gc_lock/gc_unlock
 * - gc_thread_init was not protected with gc_lock/gc_unlock
 *
 * Some important changes:
 * - predecessor pointer in queue structures removed
 * - priority time queue simplified to time queue
 * - mutex semaphore has a thread queue for waiting threads
 *
 * Remaining problems:
 * - possible interaction problems between the T2 Kernel and GC macros
 */

/* include sfr definitions for IO registers */
#include <avr/io.h>
#include <compat/ina90.h>
#include <avr/interrupt.h>

#include <gc.h>
#include <config.h>

#include "AVR_Kernel.h"
#include "AVR_Memory.h"

volatile uint32_t	kernel_ticks = 0;
volatile kthread_t*	current_thread;
volatile queue_t	sleep_queue;
volatile queue_t	ready_queue;

#undef AVR_TEST_THREADS

#ifdef AVR_TEST_THREADS
#undef AVR_TEST_MUTEX

#ifdef AVR_TEST_MUTEX
static volatile mutex_t* __test_mutex;
#else
static volatile semaphore_t* __test_sem;
#endif

static void t1()
{
	int	i = 0;
	while (1) {
#ifdef AVR_TEST_MUTEX
		mutex_take(__test_mutex);
#else
		semaphore_take(__test_sem);
#endif
		LCD_moveCursor(0);
		LCD_printInt(i++);
		kernel_sleep(100);
#ifdef AVR_TEST_MUTEX
		mutex_give(__test_mutex);
#else
		semaphore_give(__test_sem);
#endif
		kernel_sleep(100);
	}
}

static void t2()
{
	while (1) {
#ifdef AVR_TEST_MUTEX
		mutex_take(__test_mutex);
#else
		semaphore_take(__test_sem);
#endif
		kernel_sleep(60);
#ifdef AVR_TEST_MUTEX
		mutex_give(__test_mutex);
#else
		semaphore_give(__test_sem);
#endif
		kernel_sleep(60);
	}
}

static void t3()
{
	while (1) {
#ifdef AVR_TEST_MUTEX
		mutex_take(__test_mutex);
#else
		semaphore_take(__test_sem);
#endif
		LCD_moveCursor(40);
		LCD_printInt32(kernel_get_ticks());
		kernel_sleep(75);
#ifdef AVR_TEST_MUTEX
		mutex_give(__test_mutex);
#else
		semaphore_give(__test_sem);
#endif
		kernel_sleep(75);
	}
}

#endif

static void idle()
{
#if 1
	/* idle */
	while (1) _NOP();
#else
	/* print tick counter on lower right part of lcd */
	while (1) {
		uint32_t ticks = kernel_get_ticks();
		cli();
		LCD_moveCursor(8);
		LCD_printInt32(ticks);
		sei();
	}
#endif
}

/* TODO: move this to autotools */
#define AVR_AUXIL_THREAD

/* Initialize the kernel and start the Java Main thread.
 */
void kernel_start()
{
	extern void java_main();

#ifdef AVR_AUXIL_THREAD
	extern void avr_auxil_thread();
#endif

	kthread_t*	javamain;
#ifdef AVR_AUXIL_THREAD
	kthread_t*	auxthread;
#endif

	/* allocate thread queues */
	queue_init(&sleep_queue);
	queue_init(&ready_queue);

#ifdef AVR_TEST_THREADS
	kthread_t*	test1;
	kthread_t*	test2;
	kthread_t*	test3;

#ifdef AVR_TEST_MUTEX
	__test_mutex = new_mutex(); 
#else
	__test_sem = new_semaphore(1); 
#endif
	test1 = new_kthread(&t1, PRIO_JAVA, JAVA_STACK_SIZE);
	pq_insert(&ready_queue, test1);
	test2 = new_kthread(&t2, PRIO_JAVA, JAVA_STACK_SIZE);
	pq_insert(&ready_queue, test2);
	test3 = new_kthread(&t3, PRIO_JAVA, JAVA_STACK_SIZE);
	pq_insert(&ready_queue, test3);
#else
	/* allocate threads... estimate some stack sizes
 	 */
	javamain = new_kthread(&java_main, PRIO_JAVA, 712);
	pq_insert(&ready_queue, javamain);
#ifdef AVR_AUXIL_THREAD
	auxthread = new_kthread(&avr_auxil_thread, PRIO_JAVA, 256);
	pq_insert(&ready_queue, auxthread);
#endif
#endif

	/* Set up the idle thread. Assuming we need no more
 	 * than 128 bytes of idle code stack space this gives
 	 * stack_end = 1100-80 = 1080. The stack pointer will
 	 * be initialized during the first context switch.
	 */
	current_thread = ealloc(sizeof(kthread_t));
	current_thread->succ = NULL;
	current_thread->last = current_thread;
	current_thread->prio_dyn = PRIO_IDLE;
	current_thread->prio_fix = PRIO_IDLE;
#ifdef CHECK_STACK
	current_thread->stack_end = IDLE_STACK_END;
#endif

	/* clear the "init" message */
	LCD_clear();

	/* set up timer interrupts */
#ifdef _AVR_EXT_CLOCK
	output(EICRB, input(EICRB) | 0x03);
	TIMSK = 0x00;
	EIMSK = 0x10;
#else
	output(TCCR2, 0x0C);
	output(OCR2, 0x9D);
	TIMSK = 0x80;
	EIMSK = 0x00;
#endif

	/* this is now the idle thread */
	idle();
}

#ifdef CHECK_STACK
/* Check current thread's stack pointer.
 * If the stack pointer has overflowed then print
 * an error message on the LCD and halt.
 */
static void kernel_check_stack()
{
	if (current_thread->sp < current_thread->stack_end) {
		/* Print error message and halt */
		LCD_clear();
		LCD_printStr("stack overflow");
		LCD_moveCursor(40);
		LCD_printInt(current_thread->sp);
		LCD_printChar('(');
		LCD_printInt(current_thread->stack_end);
		LCD_printChar(')');
		while (1) _NOP();
	}
}
#endif

void kernel_next_thread()
{
	kthread_t*	thread;

#ifdef CHECK_STACK
	kernel_check_stack();
#endif

	while (thread = tq_pop_one(&sleep_queue))
		pq_insert(&ready_queue, thread);

	current_thread = pq_pop_one(&ready_queue);

	if (!current_thread) {
		/* Print error message and halt */
		LCD_clear();
		LCD_printStr("thread failure");
		while (1) _NOP();
	}
}

void kernel_wait(uint32_t time)
{
	cli();

#ifdef CHECK_STACK
	kernel_check_stack();
#endif

	current_thread->time = time;
	tq_insert(&sleep_queue, current_thread);
	kernel_reschedule();
}

void kernel_sleep(uint32_t delay)
{
	cli();

#ifdef CHECK_STACK
	kernel_check_stack();
#endif

	current_thread->time = kernel_ticks + delay;
	tq_insert(&sleep_queue, current_thread);
	kernel_reschedule();
} 

uint32_t kernel_get_ticks()
{
	uint32_t val;
	cli();

#ifdef CHECK_STACK
	kernel_check_stack();
#endif

	val = kernel_ticks;
	sei();
	return val;
}

uint32_t kernel_get_tick_frac()
{
	uint32_t val;
	cli();

#ifdef CHECK_STACK
	kernel_check_stack();
#endif

	val = (uint32_t)OCR2;
	val *= ( 2<<16 / 0x9D );
	sei();
	return val;
}

