/**
 * Copyright (C) 2010  Jesper Öqvist <jesper.oqvist@cs.lth.se>
 *
 * This file is part of Ljrt.
 */

/**
 *
 * Tests for the priority queue and time queue
 * data structures in the T2 AVR Kernel.
 *
 * gcc -o test_queue AVR_Queue.c test_queue.c
 * ./test_queue
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "AVR_Queue.h"

#define NTHREADS 100
#define NTICKS 1000

volatile uint32_t	kernel_ticks = 0;

queue_t*		sleep_queue;
queue_t*		ready_queue;

kthread_t*		thread1;
kthread_t*		thread2;
kthread_t*		thread3;
kthread_t*		thread4;
kthread_t*		thread5;
kthread_t*		thread6;
kthread_t*		thread7;

void init_thread(kthread_t* thread)
{
	thread->succ = NULL;
	thread->last = thread;
	thread->time = 0;
	thread->prio_fix = thread->prio_dyn = 0;
}

void setup()
{
	sleep_queue = malloc(sizeof(queue_t));
	ready_queue = malloc(sizeof(queue_t));
	queue_init(sleep_queue);
	queue_init(ready_queue);

	thread1 = malloc(sizeof(kthread_t));
	thread2 = malloc(sizeof(kthread_t));
	thread3 = malloc(sizeof(kthread_t));
	thread4 = malloc(sizeof(kthread_t));
	thread5 = malloc(sizeof(kthread_t));
	thread6 = malloc(sizeof(kthread_t));
	thread7 = malloc(sizeof(kthread_t));
	init_thread(thread1);
	init_thread(thread2);
	init_thread(thread3);
	init_thread(thread4);
	init_thread(thread5);
	init_thread(thread6);
	init_thread(thread7);
}

void test_empty()
{
	assert(pq_pop_one(ready_queue) == NULL);
	assert(pq_pop_group(ready_queue) == NULL);
	assert(tq_pop_one(sleep_queue) == NULL);
}

void test_one()
{
	pq_insert(ready_queue, thread1);

	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == NULL);

	pq_insert(ready_queue, thread1);

	assert(pq_pop_group(ready_queue) == thread1);
	assert(pq_pop_group(ready_queue) == NULL);

}

void test_two()
{
	thread1->prio_dyn = 2;
	thread2->prio_dyn = 1;

	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread1);

	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == NULL);

	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread1);

	assert(pq_pop_group(ready_queue) == thread1);
	assert(pq_pop_group(ready_queue) == thread2);
	assert(pq_pop_group(ready_queue) == NULL);
}

void test_priority()
{
	thread1->prio_dyn = 2;
	thread2->prio_dyn = 1;
	thread3->prio_dyn = 2;
	thread4->prio_dyn = 1;
	thread5->prio_dyn = 2;
	thread6->prio_dyn = 2;
	thread7->prio_dyn = 2;

	/* ready_queue must be empty */
	assert(pq_pop_one(ready_queue) == NULL);

	pq_insert(ready_queue, thread1);
	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread3);
	pq_insert(ready_queue, thread4);

	assert(pq_pop_group(ready_queue) == thread1);
	assert(thread1->last == thread3);
	assert(thread1->succ == thread3);
	assert(pq_pop_group(ready_queue) == thread2);
	assert(thread2->last == thread4);
	assert(thread2->succ == thread4);
	assert(pq_pop_group(ready_queue) == NULL);

	pq_insert(ready_queue, thread1);
	pq_insert(ready_queue, thread2);
	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread3);
	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == thread4);
	assert(pq_pop_one(ready_queue) == NULL);
	assert(thread1->last == thread1);
	assert(thread1->succ == NULL);
	assert(thread2->last == thread2);
	assert(thread2->succ == NULL);
	assert(thread3->last == thread3);
	assert(thread3->succ == NULL);
	assert(thread4->last == thread4);
	assert(thread4->succ == NULL);

	/* larger group inserts */
	pq_insert(ready_queue, thread1);
	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread3);
	pq_insert(ready_queue, thread4);
	pq_insert(ready_queue, thread5);
	pq_insert(ready_queue, thread6);
	pq_insert(ready_queue, thread7);
	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread3);
	assert(pq_pop_one(ready_queue) == thread5);
	assert(pq_pop_one(ready_queue) == thread6);
	assert(pq_pop_one(ready_queue) == thread7);
	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == thread4);
	assert(pq_pop_one(ready_queue) == NULL);

	/* larger group inserts */
	pq_insert(ready_queue, thread4);
	pq_insert(ready_queue, thread7);
	pq_insert(ready_queue, thread6);
	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread5);
	pq_insert(ready_queue, thread3);
	pq_insert(ready_queue, thread1);
	assert(pq_pop_group(ready_queue) == thread7);
	assert(pq_pop_group(ready_queue) == thread4);
	assert(pq_pop_group(ready_queue) == NULL);
	assert(thread7->succ == thread6);
	assert(thread7->last == thread1);
	assert(thread6->succ == thread5);
	assert(thread5->succ == thread3);
	assert(thread3->succ == thread1);
	assert(thread1->succ == NULL);
	assert(thread4->succ == thread2);
	assert(thread4->last == thread2);
	assert(thread2->succ == NULL);

	pq_insert(ready_queue, thread7);
	pq_insert(ready_queue, thread4);
	assert(pq_pop_one(ready_queue) == thread7);
	assert(pq_pop_one(ready_queue) == thread6);
	assert(pq_pop_one(ready_queue) == thread5);
	assert(pq_pop_one(ready_queue) == thread3);
	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread4);
	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == NULL);
	assert(thread1->succ == NULL);
	assert(thread1->last == thread1);
	assert(thread2->succ == NULL);
	assert(thread2->last == thread2);
	assert(thread3->succ == NULL);
	assert(thread3->last == thread3);
	assert(thread4->succ == NULL);
	assert(thread4->last == thread4);
	assert(thread5->succ == NULL);
	assert(thread5->last == thread5);
	assert(thread6->succ == NULL);
	assert(thread6->last == thread6);
	assert(thread7->succ == NULL);
	assert(thread7->last == thread7);

	pq_insert(ready_queue, thread1);
	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread3);
	pq_insert(ready_queue, thread4);

	assert(pq_pop_group(ready_queue) == thread1);

	pq_insert(ready_queue, thread5);
	pq_insert(ready_queue, thread1);/* insert 1, 3 */
	pq_insert(ready_queue, thread6);
	pq_insert(ready_queue, thread7);

	assert(pq_pop_group(ready_queue) == thread5);
	assert(pq_pop_group(ready_queue) == thread2);
	assert(pq_pop_group(ready_queue) == NULL);

	assert(thread5->succ == thread1);
	assert(thread5->last == thread7);
	assert(thread1->succ == thread3);
	assert(thread3->succ == thread6);
	assert(thread6->succ == thread7);
	assert(thread7->succ == NULL);
	assert(thread2->succ == thread4);
	assert(thread2->last == thread4);
	assert(thread4->succ == NULL);

	pq_insert(ready_queue, thread5);
	pq_insert(ready_queue, thread2);
	assert(pq_pop_one(ready_queue) == thread5);
	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread3);
	assert(pq_pop_one(ready_queue) == thread6);
	assert(pq_pop_one(ready_queue) == thread7);
	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == thread4);
	assert(pq_pop_one(ready_queue) == NULL);
	assert(thread1->succ == NULL);
	assert(thread1->last == thread1);
	assert(thread2->succ == NULL);
	assert(thread2->last == thread2);
	assert(thread3->succ == NULL);
	assert(thread3->last == thread3);
	assert(thread4->succ == NULL);
	assert(thread4->last == thread4);
	assert(thread5->succ == NULL);
	assert(thread5->last == thread5);
	assert(thread6->succ == NULL);
	assert(thread6->last == thread6);
	assert(thread7->succ == NULL);
	assert(thread7->last == thread7);
}

void test_pq_last()
{
	thread1->prio_dyn = 0;
	thread2->prio_dyn = 0;
	thread3->prio_dyn = 0;
	thread4->prio_dyn = 0;
	thread5->prio_dyn = 0;
	thread6->prio_dyn = 0;
	thread7->prio_dyn = 0;

	/* ready_queue must be empty */
	assert(pq_pop_one(ready_queue) == NULL);

	pq_insert(ready_queue, thread1);
	pq_insert(ready_queue, thread2);
	pq_insert(ready_queue, thread3);

	assert(pq_pop_one(ready_queue) == thread1);
	assert(thread1->succ == NULL);
	assert(thread1->last == thread1);
	assert(thread2->succ == thread3);
	assert(thread2->last == thread3);

	pq_insert(ready_queue, thread1);
	assert(thread1->succ == NULL);
	assert(thread2->succ == thread3);
	assert(thread2->last == thread1);

	thread4->succ = thread5;
	thread4->last = thread7;
	thread5->succ = thread6;
	thread6->succ = thread7;
	thread7->succ = NULL;

	pq_insert(ready_queue, thread4);
	assert(thread1->succ = thread4);
	assert(thread2->succ == thread3);
	assert(thread2->last == thread7);

	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == thread3);
	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread4);
	assert(pq_pop_one(ready_queue) == thread5);
	assert(pq_pop_one(ready_queue) == thread6);
	assert(pq_pop_one(ready_queue) == thread7);
	assert(pq_pop_one(ready_queue) == NULL);
}

void test_time()
{
	kthread_t*	thread;

	thread1->prio_dyn = 2;
	thread2->prio_dyn = 2;
	thread3->prio_dyn = 1;
	thread4->prio_dyn = 1;

	thread1->time = 100;
	thread2->time = 1000;
	thread3->time = 500;
	thread4->time = 300;

	kernel_ticks = 2000;
	tq_insert(sleep_queue, thread1);
	tq_insert(sleep_queue, thread4);
	tq_insert(sleep_queue, thread2);
	tq_insert(sleep_queue, thread3);
	assert(tq_pop_one(sleep_queue) == thread1);
	assert(tq_pop_one(sleep_queue) == thread4);
	assert(tq_pop_one(sleep_queue) == thread3);
	assert(tq_pop_one(sleep_queue) == thread2);
	assert(tq_pop_one(sleep_queue) == NULL);

	tq_insert(sleep_queue, thread1);
	tq_insert(sleep_queue, thread2);
	tq_insert(sleep_queue, thread3);
	tq_insert(sleep_queue, thread4);

	kernel_ticks = 99;
	assert(tq_pop_one(sleep_queue) == NULL);

	/* test poping in chronological order from time queue */
	kernel_ticks = 300;
	assert(tq_pop_one(sleep_queue) == thread1);
	assert(tq_pop_one(sleep_queue) == thread4);
	assert(tq_pop_one(sleep_queue) == NULL);

	kernel_ticks = 600;
	assert(tq_pop_one(sleep_queue) == thread3);
	assert(tq_pop_one(sleep_queue) == NULL);

	kernel_ticks = 1000;
	assert(tq_pop_one(sleep_queue) == thread2);
	assert(tq_pop_one(sleep_queue) == NULL);

	tq_insert(sleep_queue, thread1);
	tq_insert(sleep_queue, thread2);
	tq_insert(sleep_queue, thread3);
	tq_insert(sleep_queue, thread4);

	while (thread = tq_pop_one(sleep_queue))
		pq_insert(ready_queue, thread);

	/* test time queue to ready queue transfer */
	assert(tq_pop_one(sleep_queue) == NULL);
	assert(pq_pop_one(ready_queue) == thread1);
	assert(pq_pop_one(ready_queue) == thread2);
	assert(pq_pop_one(ready_queue) == thread4);
	assert(pq_pop_one(ready_queue) == thread3);
	assert(pq_pop_one(ready_queue) == NULL);

}

int main(int argc, char** argv)
{
	setup();

	printf("running tests\n");
	test_empty();

	test_one();
	test_two();
	printf("simple tests passed\n");

	test_priority();
	test_pq_last();
	printf("priority queue passed\n");

	test_time();
	printf("time queue passed\n");

	printf("all tests passed!\n");
}
